#!/bin/bash

set -e
set -u

# Start time in seconds
let STARTTIME=`date +%s`

# Replace search path
export PATH="/bin:/usr/bin"

#Determine building operating system

BUILD_PLATFORM=""
declare -i NUMBER_OF_CPUS

if [ `uname -o | grep -i cygwin` ]; then
        BUILD_PLATFORM=CYGWIN
        NUMBER_OF_CPUS=$NUMBER_OF_PROCESSORS
fi

if [ `uname -o | grep -i linux` ]; then
        BUILD_PLATFORM=LINUX
        if [ -x /proc/acpi/processor ]; then
                NUMBER_OF_CPUS=`ls /proc/acpi/processor/ | wc -l`
        else
                NUMBER_OF_CPUS=1
        fi        
fi

if [ -z ${BUILD_PLATFORM} ]; then
        echo "  [ERROR] Unknown build platform."
        exit 1
fi

# Unset RUBYOPT to make it possible to run rails on windows without problems (???)
if [ ${BUILD_PLATFORM} = CYGWIN ]; then
        unset RUBYOPT
fi


#----- Search GT directory -----

#  GT_DIR is found by stripping /Make and anything behind it from current path 
#+ and convert to Windows syntax (but with / instead of \)

GT_DIR=""
ROOT_DIR=""

if [ ${BUILD_PLATFORM} = CYGWIN ]; then
        GT_DIR=$(cygpath -ml $(echo "`pwd`/$0"| sed 's?\(.*\)/Make.*?\1?'))
        ROOT_DIR=
	CXX="${ROOT_DIR}/opt/crosstool/gcc-4.1.1-glibc-2.3.6/i686-unknown-linux-gnu/bin/i686-unknown-linux-gnu-g++.exe"
	CC="${ROOT_DIR}/opt/crosstool/gcc-4.1.1-glibc-2.3.6/i686-unknown-linux-gnu/bin/i686-unknown-linux-gnu-gcc.exe"
	STRIP="${ROOT_DIR}/opt/crosstool/gcc-4.1.1-glibc-2.3.6/i686-unknown-linux-gnu/bin/i686-unknown-linux-gnu-strip.exe"
fi

if [ ${BUILD_PLATFORM} = LINUX ]; then
        GT_DIR=$(echo "`pwd`/$0" | sed 's?\(.*\)/Make.*?\1?')
        ROOT_DIR="/"
	CXX=g++
	CC=gcc
        STRIP=strip
fi

export BUILD_PLATFORM

#Check if the GT directory has been found
if [ -z "${GT_DIR}" ]; then
        echo "  [ERROR] Unable to determine GT directory"
        exit 1
else
        export GT_DIR
        export ROOT_DIR
fi  

# Check if the GT directory has a valid directory structure
if [ ! -x "${GT_DIR}/Src" -o ! -x "${GT_DIR}/Make" -o ! -x "${GT_DIR}/Config" ]; then
        echo "  [ERROR]  Unable to determine GT directory"
        exit 1
fi

cd "${GT_DIR}"


#----- Configuration -----

if [ $# -lt 1 -o $# -gt 2 ]; then
	echo "Usage: $0 NaoDebug/NaoOptimized/NaoRelease [clean]"
	echo "Example: $0 Debug"
	exit 1
fi

export CONFIGURATION=$1
OPTION=""
if [ $# -eq 2 ]; then
        OPTION=$2
fi
export OPTION
BUILD_DIR_BHUMAN=Build/Linux/$CONFIGURATION
BUILD_DIR_URC=Build/Linux/URC
BUILD_DIR_DEPEND=Build/Linux/depend

# Create variables if not existing (keeping values from environment)
export CFLAGS+=""
export CXXFLAGS+=""
export LDFLAGS+=""
export LIBS+=""
export CC
export CXX
export LD
export STRIP


# Global include path
CFLAGS+=" -ISrc"

# Optimise?
case $CONFIGURATION in
NaoRelease|NaoOptimized)
        CFLAGS+=" -O3 -mmmx -m3dnow"
        ;;
esac

# Debug
case $CONFIGURATION in
NaoRelease)
        CFLAGS+=" -DNDEBUG -DRELEASE"
        ;;
NaoDebug)
        CFLAGS+=" -D_DEBUG -DCONSOLE_OUTPUT -g"
        ;;
NaoOptimized)
        CFLAGS+=" -DNDEBUG -g"
        ;;
esac

#Needed to get right image byte order in robot code
CFLAGS+=" -DSWAP_IMAGE_BYTEORDER"

# Warnings, Language
CFLAGS+=" -D__STRICT_ANSI__"
CFLAGS+=" -Wall -Wno-non-virtual-dtor -Wno-deprecated -Wno-strict-aliasing"

case $CONFIGURATION in
NaoRelease)
        CFLAGS+=" -Wno-unused"
        ;;
esac

# robot system settings
CFLAGS+=" -march=k6-2 -DLINUX"

# Examine given configuration
case $CONFIGURATION in
# Compatibility behavior
NaoRelease|NaoDebug|NaoOptimized)
	CFLAGS+=" -DUSE_NAOQI"
	CFLAGS+=" -IUtil/alcommon/include -IUtil/alcommon/interface -IUtil/alcommon/soap"
        LDFLAGS+=" -LUtil/alcommon/lib/linux"
        LIBS+=" -lalcommon"
        TARGET_BHUMAN=library
	;;
*)
	echo "  [ERROR] Build.bash: Unsupported configuration: $CONFIGURATION"
	exit 1
	;;
esac

PHONY_TARGETS=""
case $OPTION in
clean)
	PHONY_TARGETS="clean"
	;;
'')
	;;
*)
	echo "  [ERROR] Build.bash: Unsupported option: $OPTION"
	exit 1
	;;
esac

# Summary
echo "  [INFO]    Configuration is $CONFIGURATION $OPTION"

# Create target directory
mkdir -p $BUILD_DIR_BHUMAN
mkdir -p $BUILD_DIR_URC
mkdir -p $BUILD_DIR_DEPEND

#Creating Config files
echo "  [CONFIG]  Creating .cfg from .h"
/bin/bash Make/GTBuild/h2cfg.bash

#----- Makefile call -----

# Create build directory (and parent directories if necessary)

echo "  [MAKE]    Rebuilding B-Human"
export BUILD_DIR=${BUILD_DIR_BHUMAN}
export TARGET=${TARGET_BHUMAN}


nice -n 5 make CFLAGS="$CFLAGS -DTARGET_ROBOT" -j$NUMBER_OF_CPUS -k -s -r -f Make/GTBuild/Makefile.bhuman $PHONY_TARGETS $TARGET 2>&1
MAKE_SUCCESS=${PIPESTATUS[0]}

# Exit if any errors occurred in the Makefile
if [ $MAKE_SUCCESS -ne 0 ]; then
	printf "  [ABORT]   Build was NOT successful\n\n"
	exit 1
fi


#----- Postprocessing -----

#----- Done -----

let "DURATION = `date +%s` - $STARTTIME"
echo "  [INFO]    $DURATION seconds elapsed"
