/** 
* @file ControllerQt/GroundTruth/GroundTruth.cpp
*
* Implementation of a class for image processing stuff for an external camera.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#include "GroundTruth.h"
#include "Tools/Debugging/Modify.h"
#include "../ConsoleRoboCupCtrl.h"
#include "Tools/Debugging/Stopwatch.h"

GroundTruth::GroundTruth(const std::string& name, int busID) : RobotConsole(theDebugReceiver,theDebugSender),  
  name(name), busID(busID), 
  camera(name, busID, image, cameraParameters),   transformation(name, cameraParameters, image), 
  rleCompression(name, image, rleImage), blobDetection(name, rleImage, yellowBlobs, orangeBlobs, blueBlobs),
  ballPerceptor(name, transformation, orangeBlobs, image.timeStamp, ballPercept), 
  robotPerceptor(name, image, transformation, yellowBlobs, blueBlobs, yellowRobotPercepts, blueRobotPercepts)
{
  mode = SystemCall::groundTruth;
}

GroundTruth::~GroundTruth()
{
  setGlobals();
  destructed = true;
}

void GroundTruth::announceStop()
{
  SYNC;
  camera.stop();

  Thread<GroundTruth>::announceStop();
}

void GroundTruth::init()
{
  SYNC;
  RobotConsole::init();
  camera.start();
  transformation.init();
}

void GroundTruth::run()
{
  while(isRunning())
    SystemCall::sleep(processMain());
}

int GroundTruth::main() 
{
  SYNC; // only for safe termination
  STOP_TIME_ON_REQUEST("CameraExecute", camera.execute(); );
  {
    SYNC_WITH(*(RobotConsole*) this); // in case RobotConsole also writes to theDebugSender
    
    OUTPUT(idProcessBegin, bin, 'g');
    // Do image processing only, if image data exists.
    // When the camera is stopped, image data becomes deleted but then
    // this code might still be called one more time :-(
    if(image.data)
    {
      STOP_TIME_ON_REQUEST("CameraConfigure", camera.configure(); );
      STOP_TIME_ON_REQUEST("TransformationConfigure", transformation.configure(); );
      STOP_TIME_ON_REQUEST("RLECompression", rleCompression.execute(); );
      STOP_TIME_ON_REQUEST("BlobDetection", blobDetection.execute(); );
      {
        SYNC_WITH(*ctrl); // ConsoleRoboCupCtrl will read percepts
        STOP_TIME_ON_REQUEST("BallPerceptor", ballPerceptor.execute(); );
        STOP_TIME_ON_REQUEST("RobotPerceptor", robotPerceptor.execute(); );
      }
    }
    DEBUG_RESPONSE("automated requests:DrawingManager", OUTPUT(idDrawingManager, bin, Global::getDrawingManager()));
    DEBUG_RESPONSE("automated requests:DrawingManager3D", OUTPUT(idDrawingManager3D, bin, Global::getDrawingManager3D()));
    DEBUG_RESPONSE("automated requests:StreamSpecification", OUTPUT(idStreamSpecification, bin, Global::getStreamHandler()));
    OUTPUT(idProcessFinished, bin, 'g');
    
    theDebugSender.moveAllMessages(theDebugReceiver);
  }
  return 1;
}

void GroundTruth::update()
{
  RobotConsole::update();
}
