/** 
* @file ControllerQt/GroundTruth/GroundTruth.h
*
* Declaration of a class for image processing stuff for an external camera.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef GroundTruth_h_
#define GroundTruth_h_

#include <string>
#include "../RobotConsole.h"
#include "../Platform/FireWireCamera.h"
#include "InternalData.h"
#include "Transformation.h"
#include "ImageProcessingData/blob.h"
#include "ImageProcessingData/color_class_rle_image.h"
#include "ImageProcessing/rle_compression_module.h"
#include "ImageProcessing/blob_detector_module.h"
#include "ImageProcessing/BallPerceptorGT.h"
#include "ImageProcessing/RobotPerceptorGT.h"


/**
* @class GroundTruth
*
* Main class for external image processing.
*/
class GroundTruth : public RobotConsole, private Thread<GroundTruth>
{
private:
  DECLARE_SYNC;
  std::string name;  /**< The name of the camera*/
  int busID;         /**< The firewire bus id of the camera*/

  /** The main loop of the process. */
  void run();

  /** Initializes stuff*/
  void init();

  /** The function is called from the framework once in every frame. */
  virtual int main();

  MessageQueue theDebugReceiver;
  MessageQueue theDebugSender;

public:
  /**
  * Constructor.
  * @param name The name of the camera.
  * @param busID The firewire bus id of the camera
  */
  GroundTruth(const std::string& name, int busID);

  /** Destructor */
  ~GroundTruth();

  /** The function starts the process. */
  void start() {Thread<GroundTruth>::start(this, &GroundTruth::run);}

  /** Is called before the process is stopped. */
  void announceStop();

  /** The function must be called to exchange data with SimRobot. */
  void update();

    /** Returns the name of the ground truth module */
  std::string getName() const
  { return name;}
 
  FireWireCamera camera;                       /**< A Firewire camera object*/
  Transformation transformation;               /**< Module for coordinate transformation*/
  RLE_Compression_Module rleCompression;       /**< First step in image processing*/
  Blob_Detector_Module blobDetection;          /**< Second step in image processing*/
  BallPerceptorGT ballPerceptor;               /**< Create ball percept */
  RobotPerceptorGT robotPerceptor;             /**< Creates robot percepts */
  FireWireImage image;                         /**< The image captured from the camera*/
  FireWireCameraParameters cameraParameters;   /**< Yes, as the name says...*/
  Color_Class_RLE_Image rleImage;              /**< A run-length encoded image */
  std::vector<Blob> blueBlobs;                 /**< The set of detected blue blobs */
  std::vector<Blob> yellowBlobs;               /**< The set of detected yellow blobs */
  std::vector<Blob> orangeBlobs;               /**< The set of detected orange blobs */
  BallPerceptGT ballPercept;                   /**< A perceived ball */
  RobotPerceptListGT yellowRobotPercepts;      /**< Perceived robots */
  RobotPerceptListGT blueRobotPercepts;        /**< More perceived robots */
};

#endif // GroundTruth_h_
