#include <vector>
#include <string>
#include <sstream>
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Settings.h"
#include "../InternalData.h"
#include "../ImageProcessingData/blob.h"
#include "../Transformation.h"
#include "RobotPerceptorGT.h"


RobotPerceptorGT::RobotPerceptorGT(const std::string& name, const FireWireImage& image, const Transformation& transformation,
                                   const std::vector<Blob>& yellowBlobs, const std::vector<Blob>& blueBlobs,
                                   RobotPerceptListGT& yellowRobots, RobotPerceptListGT& blueRobots) : 
  yellowRobots(yellowRobots),
  blueRobots(blueRobots),
  yellowBlobs(yellowBlobs),
  blueBlobs(blueBlobs),
  transformation(transformation),
  image(image),
  scan_radius(70),
  scan_points(36),
  securityDistance(25),
  name(name)
{
  loadConfig();
}


RobotPerceptorGT::~RobotPerceptorGT()
{
  for(unsigned int i=0; i<profiles.size(); i++)
    delete profiles[i];
  profiles.clear();
}


void RobotPerceptorGT::execute()
{
  DECLARE_DEBUG_DRAWING("ground truth:robot perceptor:robots", "drawingOnImage", "The perceived robots.");
  DECLARE_DEBUG_DRAWING("ground truth:robot perceptor:scan details", "drawingOnImage", "Some internal stuff.");
  create_robot_profiles();
  yellowRobots.clear();
  blueRobots.clear();

  std::vector<RobotProfile*>::iterator profile = profiles.begin();
  std::vector<RobotProfile*>::iterator profile_end = profiles.end();
  for (; profile != profile_end; ++profile)
  {
    RobotPerceptGT percept;
    percept.translation.x = (*profile)->get_x();
    percept.translation.y = (*profile)->get_y();        
    percept.rotation = (*profile)->get_rotation();
    percept.timeStamp = image.timeStamp;
    percept.id = (*profile)->get_id();

    Vector2<double> drawTrans(percept.translation);
    percept.translation = transformation.toField(percept.translation.x, percept.translation.y, ObjectHeights::ROBOT_HEIGHT);
    // Save the percept
    if ((*profile)->get_color() == TEAM_COLOR_BLUE)
    {  
      blueRobots.push_back(percept);
      // Draw:
      CIRCLE("ground truth:robot perceptor:robots", drawTrans.x, drawTrans.y, 20, 1,
        Drawings::ps_solid, ColorClasses::blue, Drawings::bs_solid, ColorRGBA(50,50,255,100));
    }
    else
    {
      yellowRobots.push_back(percept);
      // Draw:
      CIRCLE("ground truth:robot perceptor:robots", drawTrans.x, drawTrans.y, 20, 1,
        Drawings::ps_solid, ColorClasses::blue, Drawings::bs_solid, ColorRGBA(255,225,50,100));
    }
    // Tool tip
    std::stringstream s;
    s << "RobotPercept ID "<<percept.id<<": (" << percept.translation.x << ", " << percept.translation.y << ", " << percept.rotation << ")";
    TIP("ground truth:robot perceptor:robots", drawTrans.x, drawTrans.y, 20, s.str());
    // Rotation
  }
}


void RobotPerceptorGT::create_robot_profiles()
{
  for(unsigned int i=0; i<profiles.size(); i++)
    delete profiles[i];
  profiles.clear();

  std::vector<Blob>::const_iterator blob = blueBlobs.begin();
  std::vector<Blob>::const_iterator blob_end = blueBlobs.end();
  for (; blob != blob_end; ++blob)
  {
    if(distanceOK(int((*blob).center_x), int((*blob).center_y)))
    {
      RobotProfile* newProfile = new RobotProfile(*blob, image, transformation, scan_radius, patterns);
      profiles.push_back(newProfile);
    }
  }

  blob = yellowBlobs.begin();
  blob_end = yellowBlobs.end();
  for (; blob != blob_end; ++blob)
  {
    if(distanceOK(int((*blob).center_x), int((*blob).center_y)))
    {
      RobotProfile* newProfile = new RobotProfile(*blob, image, transformation, scan_radius, patterns);
      profiles.push_back(newProfile);
    }
  }
}


bool RobotPerceptorGT::distanceOK(int x, int y)
{
  return (x > securityDistance) && (x < int(image.width) - securityDistance) &&
    (y > securityDistance) && (y < int(image.height) - securityDistance);
}


void RobotPerceptorGT::loadConfig()
{
  std::string configFile("GroundTruth/patterns.cfg");
  InConfigFile stream(Global::getSettings().expandLocationFilename(configFile));
  if(stream.exists())
    for(int i = 0; i < RobotPatterns::NUMBER_OF_PATTERNS; i++)
      stream >> patterns.idOfPattern[i];
}
