#include <sstream>
#include <algorithm>
#include "../InternalData.h"
#include "../Transformation.h"
#include "../ImageProcessingData/blob.h"
#include "Platform/GTAssert.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Geometry.h"
#include "Tools/RingBufferWithSum.h"
#include "RobotProfile.h"


Vector2<int> vCast(const Vector2<double>& dv)
{
  return Vector2<int>(static_cast<int>(dv.x), static_cast<int>(dv.y));
}


RobotProfile::RobotProfile(const Blob& blob, const FireWireImage& image, const Transformation& trans,
                 int scan_radius, RobotPatterns& patterns)
    : rotation(0.0), image(image), scan_radius(scan_radius), patterns(patterns), trans(trans)
{
  minRadius = 52;
  maxRadius = 67;

  // Get the blobs color
  switch (blob.color)
  {
  case COLOR_CLASS_YELLOW:
    team_color = TEAM_COLOR_YELLOW;
    break;
  case COLOR_CLASS_BLUE:
    team_color = TEAM_COLOR_BLUE;
    break;
  default:
    team_color = TEAM_COLOR_NONE;
  }

  height = ObjectHeights::ROBOT_HEIGHT;

  // The initial profile center is the blobs center
  center.x = blob.center_x;
  center.y = blob.center_y;

  findInitialRotationThomas();
  findRealRotationAndId();

  // Draw stuff:
  Vector2<double> centerWorld(trans.toField(center.x, center.y, height));
  Vector2<double> endForDrawingLineWorld = centerWorld + Geometry::rotate(Vector2<double>(400.0, 0.0), rotation); 
  Vector2<double> endForDrawingLineWorldOpp = centerWorld + Geometry::rotate(Vector2<double>(400.0, 0.0), normalize(rotation+pi)); 
  Vector2<double> endForDrawingLineWorldRight = centerWorld + Geometry::rotate(Vector2<double>(400.0, 0.0), normalize(rotation-pi/2.0)); 
  Vector2<double> endForDrawingLineWorldLeft = centerWorld + Geometry::rotate(Vector2<double>(400.0, 0.0), normalize(rotation+pi/2.0)); 
  Vector2<double> endForDrawingLineImage = trans.toImage(endForDrawingLineWorld.x, endForDrawingLineWorld.y, height);
  Vector2<double> endForDrawingLineImageOpp = trans.toImage(endForDrawingLineWorldOpp.x, endForDrawingLineWorldOpp.y, height);
  Vector2<double> endForDrawingLineImageRight = trans.toImage(endForDrawingLineWorldRight.x, endForDrawingLineWorldRight.y, height);
  Vector2<double> endForDrawingLineImageLeft = trans.toImage(endForDrawingLineWorldLeft.x, endForDrawingLineWorldLeft.y, height);
  LINE("ground truth:robot perceptor:scan details", endForDrawingLineImageOpp.x, endForDrawingLineImageOpp.y,
    endForDrawingLineImage.x, endForDrawingLineImage.y, 1, Drawings::bs_solid, ColorClasses::white);
  LINE("ground truth:robot perceptor:scan details", endForDrawingLineImageLeft.x, endForDrawingLineImageLeft.y,
    endForDrawingLineImageRight.x, endForDrawingLineImageRight.y, 1, Drawings::bs_solid, ColorClasses::white);
}


Vector2<double> RobotProfile::t(const Vector2<double>& vec) const
{
    return trans.toImage(vec.x, vec.y, height);
}
  

void RobotProfile::findRealRotationAndId()
{   
  // Get average center color of all four segments:
  Vector2<double> centerInWorld(trans.toField(center.x, center.y, height));
  Vector2<double> segmentVector((minRadius+maxRadius)/2.0, 0);
  RGB_Pixel segmentColor[4];
  Vector2<int> segmentPos[4]; //kept for later drawing
  for(int i=0; i<4; i++)
  {
    Vector2<double> segWorldPos = centerInWorld + Geometry::rotate(segmentVector, rotation + pi/4.0 + (pi/2.0)*i);
    segmentPos[i] = vCast(t(segWorldPos));
    segmentColor[i] = getRGBPixelInterpolated(image, segmentPos[i].x, segmentPos[i].y);
  }

  // Find black segment:
  int blackMostIdx(0);
  unsigned char blackMostValue(segmentColor[0].toGrayScale());
  for(int i=1; i<4; i++)
  {
    unsigned char currentValue(segmentColor[i].toGrayScale());
    if(currentValue < blackMostValue)
    {
      blackMostValue = currentValue;
      blackMostIdx = i;
    }
  }

  // Set correct rotation:
  rotation += blackMostIdx*(pi/2.0);
  rotation = normalize(rotation);

  // Get id color sequence:
  RGB_Pixel colorSequence[3];
  Vector2<int> colorSequencePos[3]; //kept for later drawing
  for(int i=0; i<3; i++)
  {
    colorSequence[i] = segmentColor[(blackMostIdx + 1 + i) % 4];
    colorSequencePos[i] = segmentPos[(blackMostIdx + 1 + i) % 4];
  }

  // Find white segment:
  int whiteMostIdx(0);
  unsigned char whiteMostValue(colorSequence[0].toGrayScale());
  for(int i=1; i<3; i++)
  {
    unsigned char currentValue(colorSequence[i].toGrayScale());
    if(currentValue > whiteMostValue)
    {
      whiteMostValue = currentValue;
      whiteMostIdx = i;
    }
  }

  // Find magenta and green segment by checking r component of color
  int magentaIdx((whiteMostIdx+1)%3);
  int greenIdx((whiteMostIdx+2)%3);
  if(colorSequence[greenIdx].r > colorSequence[magentaIdx].r)
  {
    int temp = magentaIdx;
    magentaIdx = greenIdx;
    greenIdx = temp;
  }

  // Get id of robot
  id = patterns.getIdOfPattern(whiteMostIdx, greenIdx, magentaIdx);

  // Draw stuff:
  Vector2<double> rotationEndWorld = centerInWorld + Geometry::rotate(Vector2<double>(400.0, 0.0), rotation); 
  Vector2<int> rotationEndImage = vCast(trans.toImage(rotationEndWorld.x, rotationEndWorld.y, height));
  LINE("ground truth:robot perceptor:robots", center.x, center.y, rotationEndImage.x, rotationEndImage.y,
    3, Drawings::bs_solid, ColorRGBA(255,0,0,100));
  MID_DOT("ground truth:robot perceptor:scan details", segmentPos[blackMostIdx].x, segmentPos[blackMostIdx].y,
    ColorClasses::black, ColorClasses::black);
  MID_DOT("ground truth:robot perceptor:scan details", colorSequencePos[whiteMostIdx].x, colorSequencePos[whiteMostIdx].y,
    ColorClasses::white, ColorClasses::white);
  MID_DOT("ground truth:robot perceptor:scan details", colorSequencePos[greenIdx].x, colorSequencePos[greenIdx].y,
    ColorClasses::green, ColorClasses::green);
  MID_DOT("ground truth:robot perceptor:scan details", colorSequencePos[magentaIdx].x, colorSequencePos[magentaIdx].y,
    ColorRGBA(255,160,255), ColorRGBA(255,160,255));
}


RGB_Pixel RobotProfile::getRGBPixel(const FireWireImage& image, int x, int y) const
{
  RGB_Pixel result;
  unsigned char* pixel = &image.data[image.width*3*y + x*3];
  result.r = *pixel++;
  result.g = *pixel++;
  result.b = *pixel;
  return result;
}


RGB_Pixel RobotProfile::getRGBPixelInterpolated(const FireWireImage& image, int x, int y) const
{
  RGB_Pixel pix = getRGBPixel(image, x, y);
  int rSum(pix.r), gSum(pix.g), bSum(pix.b);
  pix = getRGBPixel(image, x+1, y);
  rSum += pix.r; gSum += pix.g; bSum += pix.b;
  pix = getRGBPixel(image, x-1, y);
  rSum += pix.r; gSum += pix.g; bSum += pix.b;
  pix = getRGBPixel(image, x, y+1);
  rSum += pix.r; gSum += pix.g; bSum += pix.b;
  pix = getRGBPixel(image, x, y-1);
  rSum += pix.r; gSum += pix.g; bSum += pix.b;
  return RGB_Pixel(rSum/5, gSum/5, bSum/5);
}


void RobotProfile::findInitialRotation()
{
  int maxSum = 0;
  double maxAngle = 0;
  double stepRadius = 3;
  Vector2<double> centerWorld = trans.toField(center.x, center.y, height);
  const double angleStep(pi/180.0);
  const double localOffset(fromDegrees(10));
  for(double angle = -pi/2.0; angle < pi/2.0; angle+=angleStep)
  {
    Vector2<double> angleVecWorld(1.0,0.0);
    Vector2<double> angleVecWorld0l = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + localOffset));
    Vector2<double> angleVecWorld0r = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) - localOffset));
    Vector2<double> angleVecWorld1l = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi/2.0 + localOffset));
    Vector2<double> angleVecWorld1r = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi/2.0 - localOffset));
    Vector2<double> angleVecWorld2l = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi + localOffset));
    Vector2<double> angleVecWorld2r = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi - localOffset));
    Vector2<double> angleVecWorld3l = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi*(3.0/2.0) + localOffset));
    Vector2<double> angleVecWorld3r = Geometry::rotate(angleVecWorld, normalize(static_cast<double>(angle) + pi*(3.0/2.0) - localOffset));
    int sum0l = 0, sum1l = 0, sum2l = 0, sum3l = 0;
    int sum0r = 0, sum1r = 0, sum2r = 0, sum3r = 0;
    for(double radius = minRadius; radius < maxRadius; radius += stepRadius)
    {
      Vector2<int> posImage0l = vCast(t(centerWorld + (angleVecWorld0l * radius)));
      Vector2<int> posImage0r = vCast(t(centerWorld + (angleVecWorld0r * radius)));
      Vector2<int> posImage1l = vCast(t(centerWorld + (angleVecWorld1l * radius)));
      Vector2<int> posImage1r = vCast(t(centerWorld + (angleVecWorld1r * radius)));
      Vector2<int> posImage2l = vCast(t(centerWorld + (angleVecWorld2l * radius)));
      Vector2<int> posImage2r = vCast(t(centerWorld + (angleVecWorld2r * radius)));
      Vector2<int> posImage3l = vCast(t(centerWorld + (angleVecWorld3l * radius)));
      Vector2<int> posImage3r = vCast(t(centerWorld + (angleVecWorld3r * radius)));
      sum0l += getRGBPixel(image, posImage0l.x, posImage0l.y).toGrayScale();
      sum1l += getRGBPixel(image, posImage1l.x, posImage1l.y).toGrayScale();
      sum2l += getRGBPixel(image, posImage2l.x, posImage2l.y).toGrayScale();
      sum3l += getRGBPixel(image, posImage3l.x, posImage3l.y).toGrayScale();
      sum0r += getRGBPixel(image, posImage0r.x, posImage0r.y).toGrayScale();
      sum1r += getRGBPixel(image, posImage1r.x, posImage1r.y).toGrayScale();
      sum2r += getRGBPixel(image, posImage2r.x, posImage2r.y).toGrayScale();
      sum3r += getRGBPixel(image, posImage3r.x, posImage3r.y).toGrayScale();
    }
    int sum = sqr(sum0l-sum0r) + sqr(sum1l-sum1r) + sqr(sum2l-sum2r) + sqr(sum3l-sum3r);
    if(sum > maxSum)
    {
      maxSum = sum;
      maxAngle = angle;
    }
  }
  rotation = maxAngle;
}


void RobotProfile::findInitialRotationThomas()
{
  RingBufferWithSum<int, 90> buffer0, buffer1, buffer2, buffer3;
  double maxSum = 0;
  int maxAngle = 0;
  double stepRadius = 3;
  Vector2<int> c = vCast(center);
  Vector2<double> cWorld = trans.toField(c.x, c.y, height);

  for(int angle = -90; angle < 90; ++angle)
  {
    int sum0 = 0, sum1 = 0, sum2 = 0, sum3 = 0;
    for(double radius = minRadius; radius < maxRadius; radius += stepRadius)
    {
      Vector2<double> v(radius, 0.0);
      Vector2<double> pWorld = cWorld + Geometry::rotate(v, fromDegrees(angle));
      Vector2<int> p = vCast(t(pWorld)) - c; 
      sum0 += getRGBPixel(image, c.x + p.x, c.y + p.y).toGrayScale();
      sum1 += getRGBPixel(image, c.x - p.y, c.y + p.x).toGrayScale();
      sum2 += getRGBPixel(image, c.x - p.x, c.y - p.y).toGrayScale();
      sum3 += getRGBPixel(image, c.x + p.y, c.y - p.x).toGrayScale();
    }
    buffer0.add(sum0);
    buffer1.add(sum1);
    buffer2.add(sum2);
    buffer3.add(sum3);
    double sum = sqr(double(buffer0.getSum() - buffer1.getSum())) +
                 sqr(double(buffer1.getSum() - buffer2.getSum())) +
                 sqr(double(buffer2.getSum() - buffer3.getSum())) +
                 sqr(double(buffer3.getSum() - buffer0.getSum()));
    if(sum > maxSum)
    {
      maxSum = sum;
      maxAngle = angle;
    }
  }
  rotation = fromDegrees(maxAngle);
}
