#include "blob_detector_module.h"
#include <sstream>
#include <string>
#include <cmath>
#include <limits>
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Common.h"
#include "Tools/Settings.h"
#include "Tools/Streams/InStreams.h"

typedef double Matrix2X2[2][2];

const int MAX_NUMBER_OF_REGIONS = 10000;
const int MAX_YELLOW_BLOBS = 200;
const int MAX_ORANGE_BLOBS = 200;
const int MAX_BLUE_BLOBS = 200;
const int MIN_DRAW_BLOB_SIZE = 5;

/** 
* Computes an eigenvalue decomposition of a symmetric 2x2 matrix in a way, 
* that is stable even for degenerated eigenvalues
* '(cos(phi), sin(phi))' will be an eigenvector with 
* eigenvalue 'l0' and '(-sin(phi),cos(phi))' will be 
* the orthogonal eigenvector with eigenvalue 'l1'. 'l0>=l1'.
*/
void eigenDecompositionSymmetric (const Matrix2X2& a, 
                                  double& phi, double& l0, double& l1)
{
  double ll0, ll1;
  phi = atan2(-2*a[1][0], a[1][1]-a[0][0])/2;
  double c = cos(phi);
  double s = sin(phi);
  double c2 = c*c, s2=s*s, cs = c*s;

  ll0 = c2*a[0][0]+2*cs*a[1][0]+s2*a[1][1];
  ll1 = s2*a[0][0]-2*cs*a[1][0]+c2*a[1][1];
  if (ll0<ll1) 
  {
    if (phi>0) phi-=pi/2;
    else phi+=pi/2;
    l0 = ll1;
    l1 = ll0;
  }
  else 
  {
    l0 = ll0;
    l1 = ll1;
  }
}


Blob_Detector_Module::Blob_Detector_Module(const std::string& name,
                                           Color_Class_RLE_Image& rleImage,
                                           std::vector<Blob>& yellowBlobs, 
                                           std::vector<Blob>& orangeBlobs, 
                                           std::vector<Blob>& blueBlobs): 
name(name),
rleImage(rleImage),
yellow_blobs(yellowBlobs),
blue_blobs(blueBlobs),
orange_blobs(orangeBlobs)
{
  potential_blobs.reserve(512);
  loadConfig();
}


void Blob_Detector_Module::execute()
{
  yellow_blobs.clear();
  blue_blobs.clear();
  orange_blobs.clear();
  if(rleImage.data.size())
  {
    group_regions();
    potential_blobs.clear();
    find_potential_blobs();
    check_and_convert_blobs();
  }
  configure();
}


void Blob_Detector_Module::group_regions()
{
  Run* begin = &*(rleImage.data.begin());
  Run* end = &(rleImage.data.back());
  ++end;
  Run* run1;

  // Init selfpointer:
  for (run1 = begin; run1 != end; ++run1)
    run1->parent = run1;

  // Go through the union-find loop:
  Run* run2 = run1 = begin;
  while (run1 != end)
  {
    while (run2->y < run1->y-1 || (run2->y == run1->y-1 && run2->x_end < run1->x_start))
      run2++;
    if (run1->color == run2->color &&
      run2->y == run1->y-1 && run2->x_start <= run1->x_end)
    {
      unite_regions(run1, run2);
    }
    run2->y >= run1->y || run2->x_end > run1->x_end ? run1++ : run2++;
  }

  // Assign numbers to different regions:
  int region_number = 0;
  for (run1 = begin; run1 != end; run1++) 
  {
    run1->parent == run1 ? run1->region = region_number++ : run1->region = run1->parent->region;
  }  
}


inline void Blob_Detector_Module::unite_regions(Run* region1, Run* region2)
{
  compress_path(region1);
  compress_path(region2);
  if (region1->parent < region2->parent) 
  {
    region2->parent->parent = region1->parent;
  }
  else 
  {
    region1->parent->parent = region2->parent;    
  }
}


inline void Blob_Detector_Module::compress_path(Run* run)
{
  Run* root = run;  
  while (root->parent != root)
  {
    root = root->parent;
  }
  Run* buffer;  
  while (run != root) 
  {
    buffer = run->parent;
    run->parent = root;
    run = buffer;
  }    
}


void Blob_Detector_Module::find_potential_blobs()
{
  try
  {
    Run* begin = &*(rleImage.data.begin());
    Run* end = &(rleImage.data.back());
    ++end;
    Run* run;

    // Suchen der max Anzahl der Regionen
    // Kann man vielleicht auch mit max_element(....) machen
    unsigned int number_of_regions(0);
    for (run = begin; run != end; ++run) 
    {
      if (run->region > number_of_regions)
        number_of_regions = run->region;
    }
    number_of_regions++;

    // Clip, if too big
    number_of_regions = number_of_regions > (unsigned int)MAX_NUMBER_OF_REGIONS ? (unsigned int)MAX_NUMBER_OF_REGIONS : number_of_regions;
    potential_blobs.resize(number_of_regions);  

    for (run = begin; run != end; run++) 
    {
      if (run->region < (unsigned int)MAX_NUMBER_OF_REGIONS)
      {
        Blob* blob = &potential_blobs[run->region];
        if(run->x_start < blob->x)
          blob->x = run->x_start;
        if(run->x_end > blob->w + blob->x)
          blob->w = run->x_end - blob->x;
        if(run->y < blob->y)
          blob->y = run->y;
        if(run->y > blob->h + blob->y)
          blob->h = run->y - blob->y;
        blob->size += run->x_end - run->x_start +1; // This is a bit redundant blob.I is the same
        blob->color = run->color;
        blob->I += run->x_end - run->x_start + 1;
        blob->I_x += (run->x_end * (run->x_end + 1) - run->x_start *(run->x_start - 1 ))/2;
        blob->I_y += (run->x_end - run->x_start +1 ) * run->y;
        blob->I_xx += (pow(run->x_end + 0.5, 3) - pow(run->x_start - 0.5, 3)) / 3;
        blob->I_yy += (run->x_end - run->x_start +1 ) * ( run->y * run->y + 1/12 );
        blob->I_xy += ((run->x_end * ( run->x_end +1) - run->x_start*(run->x_start-1))*run->y) / 2;
      }
    }

    for ( unsigned int i = 0; i < potential_blobs.size(); ++i)
    {
      Blob& blob = potential_blobs[i];
      blob.center_x = blob.I_x / blob.I;
      blob.center_y = blob.I_y / blob.I;
      // Haupttraegheitsachsen
      Matrix2X2 m;
      m[0][0] = (blob.I_xx / blob.I) - (blob.center_x*blob.center_x);
      m[0][1] = m[1][0] = (blob.I_xy / blob.I) - (blob.center_x * blob.center_y);
      m[1][1] = (blob.I_yy / blob.I) - (blob.center_y*blob.center_y);
      eigenDecompositionSymmetric (m, blob.main_axis, blob.large_length, blob.small_length );
      blob.small_length = 2.0 * sqrt(fabs(blob.small_length));
      blob.large_length = 2.0 * sqrt(fabs(blob.large_length));
    }
  }
  catch(...)
  {
    ;
  }
}


void Blob_Detector_Module::check_and_convert_blobs()
{
  int num_yellow_blobs(0);
  int num_blue_blobs(0);
  int num_orange_blobs(0);

  for(unsigned int i = 0; i < potential_blobs.size(); i++)
  {
    Blob& b = potential_blobs[i];
    switch(b.color)
    {
    case COLOR_CLASS_YELLOW: 
      if (num_yellow_blobs <= MAX_YELLOW_BLOBS &&
        blob_constraints.check_constraints(b, Blob_Constraints::TEAM))
      {
        yellow_blobs.push_back(b);
        num_yellow_blobs++;
      }
      break;

    case COLOR_CLASS_BLUE: 
      if (num_blue_blobs <= MAX_BLUE_BLOBS &&
        blob_constraints.check_constraints(b, Blob_Constraints::TEAM))
      {
        blue_blobs.push_back(b);
        num_blue_blobs++;
      }
      break;

    case COLOR_CLASS_ORANGE: 
      if (num_orange_blobs <= MAX_ORANGE_BLOBS &&
        blob_constraints.check_constraints(b, Blob_Constraints::BALL))
      {
        orange_blobs.push_back(b);
        num_orange_blobs++;
      }
      break;

    default:
      continue;
      break;
    };
  }    
}


void Blob_Detector_Module::loadConfig()
{
  // Load dimensions of area covered by transformation:
  std::string blobFile("GroundTruth/");
  blobFile += (name + "/BlobDetector.cfg");
  InConfigFile blobStream(Global::getSettings().expandLocationFilename(blobFile));
  if(blobStream.exists())
  {
    blobStream >> blob_constraints.constraint[Blob_Constraints::TEAM].min_size
      >> blob_constraints.constraint[Blob_Constraints::TEAM].max_size 
      >> blob_constraints.constraint[Blob_Constraints::TEAM].min_width 
      >> blob_constraints.constraint[Blob_Constraints::TEAM].xy_ratio 
      >> blob_constraints.constraint[Blob_Constraints::BALL].min_size
      >> blob_constraints.constraint[Blob_Constraints::BALL].max_size
      >> blob_constraints.constraint[Blob_Constraints::BALL].min_width
      >> blob_constraints.constraint[Blob_Constraints::BALL].xy_ratio;
  }
  else // Print a warning
  {
    OUTPUT(idText,text,"Warning! "<<blobFile.c_str()<<" does not exist, using default values.");
  }
}


void Blob_Detector_Module::saveConfig()
{/*
 std::string config_file(config_path + "blob_detection.cfg");
 BSmart::Configuration conf(config_file);
 const std::string section("Constraints");

 if (conf.is_open())
 {
 conf.add_param(section, "team_xy_ratio",
 blob_constraints.constraint[Blob_Constraints::TEAM].xy_ratio);
 conf.add_param(section, "team_min_size",
 blob_constraints.constraint[Blob_Constraints::TEAM].min_size);
 conf.add_param(section, "team_max_size",
 blob_constraints.constraint[Blob_Constraints::TEAM].max_size);
 conf.add_param(section, "team_min_width",
 blob_constraints.constraint[Blob_Constraints::TEAM].min_width);

 conf.add_param(section, "ball_xy_ratio",
 blob_constraints.constraint[Blob_Constraints::BALL].xy_ratio);
 conf.add_param(section, "ball_min_size",
 blob_constraints.constraint[Blob_Constraints::BALL].min_size);
 conf.add_param(section, "ball_max_size",
 blob_constraints.constraint[Blob_Constraints::BALL].max_size);
 conf.add_param(section, "ball_min_width",
 blob_constraints.constraint[Blob_Constraints::BALL].min_width);
 conf.add_param(section, "blob_radius",
 data.blob_radius);
 }*/
}


void Blob_Detector_Module::configure()
{
  // Draw potential blobs:
  DECLARE_DEBUG_DRAWING("ground truth:blob detection:potential blobs", "drawingOnImage", "Marks the centers of potential blobs.");
  DECLARE_DEBUG_DRAWING("ground truth:blob detection:found blobs", "drawingOnImage", "Marks the centers of detected blobs.");
  for(unsigned int i = 0; i < potential_blobs.size(); i++)
  {
    Blob& b = potential_blobs[i];
    if(b.size > MIN_DRAW_BLOB_SIZE)
    {
      std::stringstream s;
      s << "Blob   Size: " << b.size << "  Ratio: " << b.small_length / b.large_length;
      MID_DOT("ground truth:blob detection:potential blobs", b.center_x, b.center_y, ColorRGBA(255,255,255), ColorRGBA(255,255,255));
      TIP("ground truth:blob detection:potential blobs", b.center_x, b.center_y, 10, s.str());
    }
  }
  
  // Draw chosen blobs:
  for(unsigned int i = 0; i < yellow_blobs.size(); i++)
  {
    Blob& b = yellow_blobs[i];
    MID_DOT("ground truth:blob detection:found blobs", b.center_x, b.center_y, ColorRGBA(0,0,0), ColorRGBA(255,225,0));
  }
  for(unsigned int i = 0; i < orange_blobs.size(); i++)
  {
    Blob& b = orange_blobs[i];
    MID_DOT("ground truth:blob detection:found blobs", b.center_x, b.center_y, ColorRGBA(0,0,0), ColorRGBA(255,128,0));
  }
  for(unsigned int i = 0; i < blue_blobs.size(); i++)
  {
    Blob& b = blue_blobs[i];
    MID_DOT("ground truth:blob detection:found blobs", b.center_x, b.center_y, ColorRGBA(255,255,255), ColorRGBA(0,0,255));
  }
}
