
#include "rle_compression_module.h"
#include "Tools/Debugging/Debugging.h"
#include "Tools/Debugging/Modify.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Settings.h"


void RGBLUT64::update(const RGB_Pixel_Settings& s)
{
  // Clear old values:
  memset(colorClass, COLOR_CLASS_NONE, 64*64*64);
  // Set new values:
  for(int r=s.yellow_min.r; r <= s.yellow_max.r; r++) 
    for(int g=s.yellow_min.g; g <= s.yellow_max.g; g++)
      for(int b=s.yellow_min.b; b <= s.yellow_max.b; b++)
        colorClass[r >> 2][g >> 2][b >> 2] = COLOR_CLASS_YELLOW;
  for(int r=s.orange_min.r; r <= s.orange_max.r; r++) 
    for(int g=s.orange_min.g; g <= s.orange_max.g; g++)
      for(int b=s.orange_min.b; b <= s.orange_max.b; b++)
        colorClass[r >> 2][g >> 2][b >> 2] = COLOR_CLASS_ORANGE;
  for(int r=s.blue_min.r; r <= s.blue_max.r; r++) 
    for(int g=s.blue_min.g; g <= s.blue_max.g; g++)
      for(int b=s.blue_min.b; b <= s.blue_max.b; b++)
        colorClass[r >> 2][g >> 2][b >> 2] = COLOR_CLASS_BLUE;
}


RLE_Compression_Module::RLE_Compression_Module(const std::string& name,
                                               FireWireImage& image,
                                               Color_Class_RLE_Image& rleImage)
                                               : image(image), rleImage(rleImage), name(name)
{
  rgbLUT = new RGBLUT64;
  loadConfig();
}


RLE_Compression_Module::~RLE_Compression_Module()
{
  delete rgbLUT;
  saveConfig();
}


void RLE_Compression_Module::execute()
{
  rleImage.reset();
  image.lut = 0;
  Run run;
  run.region = 0;
  run.parent = 0;
  const unsigned int HEIGHT(image.height);
  const unsigned int WIDTH(image.width*3);

  for(unsigned int y = 0; y < HEIGHT; y++)
  {
    run.y = y;
    run.x_start = 0;
    unsigned char* start = &image.data[y * WIDTH];
    unsigned char* pixel = start;
    unsigned char* end = &image.data[(y + 1) * WIDTH];
    
    run.color = classify(pixel);
    Color_Class current_color = run.color;
    
    while (pixel != end)
    {        
      while (pixel != end && classify(pixel) == current_color)
      {
      }
      if (pixel != end)
      {
        int pos = ((unsigned int)pixel - (unsigned int)start) / 3;
        run.x_end = pos - 1;

        if(run.color != COLOR_CLASS_NONE)
        {
          rleImage.data.push_back(run);
        }
        run.x_start = pos;

        current_color = classify(pixel);
        run.color = current_color;
      }
      else
      {
        if (run.color != COLOR_CLASS_NONE)
        {
          run.x_end = WIDTH/3 - 1;
          rleImage.data.push_back(run);
        }
      }
    }
  }
  configure();
}


inline Color_Class RLE_Compression_Module::classify(unsigned char*& pixel) const
{
  unsigned char pixel_r = *pixel++;
  unsigned char pixel_g = *pixel++;
  unsigned char pixel_b = *pixel++;
  return Color_Class(rgbLUT->colorClass[pixel_r>>2][pixel_g>>2][pixel_b>>2]);
}


void RLE_Compression_Module::loadConfig()
{
  std::string configFile("GroundTruth/");
  configFile += (name + "/RLECompression.cfg");
  InConfigFile stream(Global::getSettings().expandLocationFilename(configFile));
  /** Be sure that the file exists. */
  if(stream.exists())
  {
    stream >> pixelSettings.yellow_min.r
           >> pixelSettings.yellow_max.r
           >> pixelSettings.yellow_min.g
           >> pixelSettings.yellow_max.g
           >> pixelSettings.yellow_min.b
           >> pixelSettings.yellow_max.b
           >> pixelSettings.orange_min.r
           >> pixelSettings.orange_max.r
           >> pixelSettings.orange_min.g
           >> pixelSettings.orange_max.g
           >> pixelSettings.orange_min.b
           >> pixelSettings.orange_max.b
           >> pixelSettings.blue_min.r
           >> pixelSettings.blue_max.r
           >> pixelSettings.blue_min.g
           >> pixelSettings.blue_max.g
           >> pixelSettings.blue_min.b
           >> pixelSettings.blue_max.b;
    rgbLUT->update(pixelSettings);
  }
  else
  {
    OUTPUT(idText, text, configFile.c_str()<<" does not exists. Using default parameters.");
  }
}


void RLE_Compression_Module::saveConfig()
{
  std::string configFile("GroundTruth/");
  configFile += (name + "/RLECompression.cfg");
  OutTextRawFile stream(Global::getSettings().expandLocationFilename(configFile));
  stream << "/**" << endl
         << "This file contains the maximum coordinates of three cubes inside the RGB color space." << endl
         << "The order is R (min, max), G (min, max), B (min, max)." <<endl
         << "One value per line*/"<<endl<<endl
         << "/**Yellow:*/" << endl
         << pixelSettings.yellow_min.r << endl
         << pixelSettings.yellow_max.r << endl
         << pixelSettings.yellow_min.g << endl
         << pixelSettings.yellow_max.g << endl
         << pixelSettings.yellow_min.b << endl
         << pixelSettings.yellow_max.b << endl << endl
         << "/**Orange:*/" << endl
         << pixelSettings.orange_min.r << endl
         << pixelSettings.orange_max.r << endl
         << pixelSettings.orange_min.g << endl
         << pixelSettings.orange_max.g << endl
         << pixelSettings.orange_min.b << endl
         << pixelSettings.orange_max.b << endl << endl
         << "/**Blue:*/" << endl
         << pixelSettings.blue_min.r << endl
         << pixelSettings.blue_max.r << endl
         << pixelSettings.blue_min.g << endl
         << pixelSettings.blue_max.g << endl
         << pixelSettings.blue_min.b << endl
         << pixelSettings.blue_max.b << endl << endl;
}


void RLE_Compression_Module::configure()
{
  MODIFY("ground truth:rle compression:color space", pixelSettings);
  DEBUG_RESPONSE("ground truth:rle compression:update LUT", rgbLUT->update(pixelSettings););
  DEBUG_RESPONSE("ground truth:rle compression:draw segmented image", {image.lut = rgbLUT;});
}


void RLE_Compression_Module::printPixelValue(int x, int y)
{
  unsigned char r = image.data[y*image.width*3 + x*3];
  unsigned char g = image.data[y*image.width*3 + x*3 + 1];
  unsigned char b = image.data[y*image.width*3 + x*3 + 2];
  OUTPUT(idText, text,"Pixel: ("<<x<<","<<y<<")  Color: ("<<r<<","<<g<<","<<b<<")");
}
