/** 
* @file ControllerQt/GroundTruth/InternalData.h
*
* Declaration of several internal data structures for gound truth computation
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef InternalData_h__
#define InternalData_h__

#include <string>
#include "Tools/Math/Vector2.h"
#include "Tools/Math/Pose2D.h"

class RGBLUT64;

/** Definition of different height classes*/
struct ObjectHeights
{
  /** Constructor, sets some default values*/
  ObjectHeights()
  {
    heights[FIELD_HEIGHT] = 0.0;
    heights[BALL_HEIGHT]  = 86.0;
    heights[ROBOT_HEIGHT] = 350.0;
  }

  enum HeightClass {FIELD_HEIGHT = 0, BALL_HEIGHT, ROBOT_HEIGHT, NUM_OF_HEIGHTS}; /**< The height classes */
  double heights[NUM_OF_HEIGHTS];                                                 /**< The heigths */
};


/**
* @class FireWireImage
* A simple structure for a captured image
* Data is stored in 24 bit RGB format
*/
struct FireWireImage
{
  /** Constructor */
  FireWireImage():data(0), lut(0)
  {}

  bool isInside(const Vector2<unsigned int>& p) const
  { return (p.x >= 0) && (p.x < width) && (p.y > 0) && (p.y < height);}

  bool isInside(const Vector2<double>& p) const
  { 
    Vector2<unsigned int> v(static_cast<unsigned int>(p.x), static_cast<unsigned int>(p.y));
    return isInside(v);
  }

  unsigned char* data;        /**< Buffer for the image*/
  unsigned long  width,       /**< The width of the image in pixels. */
                 height;      /**< The height of the image in pixels. */
  unsigned timeStamp;         /**< The time when the image was taken. */
  RGBLUT64* lut;              /**< Look-up table for segmented drawing. HACK! */
};


/**
* @class FireWireCameraParameters
* Parameters for the current camera (read from file camera.cfg):
*  - camera identifier
*  - intrisic parameters which are calculated externally
*  - video parameter
*/
struct FireWireCameraParameters
{
  std::string uniqueID;    /**< The unique identifier of the camera*/
  double focusLengthX;     /**< The focus length in x direction*/
  double focusLengthY;     /**< The focus length in y direction*/
  double principalPointX;  /**< The x coordinate of the lens' center*/
  double principalPointY;  /**< The y coordinate of the lens' center*/
  double distortion[4];    /**< The distortion coefficients*/
  int videoFormat;         /**< Format according to firewire standard*/
  int videoMode;           /**< Mode according to firewire standard (sets resolution)*/
  int videoFrameRate;      /**< Yes, the frame rate */
  unsigned short focus;          /**< The focus, may be configured to get a sharp image :-) */
  unsigned short shutter;        /**< Yes, the shutter speed */
  unsigned short whiteBalance1;  /**< One white balance configuration value */
  unsigned short whiteBalance2;  /**< The other white balance configuration value */
};


class BallPerceptGT: public Vector2<double>
{
public:
  BallPerceptGT() : timeStamp(0) {}

  BallPerceptGT(const Vector2<double>& other) : Vector2<double>(other), timeStamp(0) {}

  unsigned timeStamp;
};


class RobotPerceptGT: public Pose2D
{
public:
  RobotPerceptGT() : id(0), timeStamp(0) {}

  int id;
  unsigned timeStamp;
};


class RobotPatterns
{
public:
  RobotPatterns()
  {
    for(int i=0; i<NUMBER_OF_PATTERNS; i++)
      idOfPattern[i] = 42;
  }

  enum Pattern {
    MAGENTA_WHITE_GREEN = 0,
    MAGENTA_GREEN_WHITE = 1,
    WHITE_GREEN_MAGENTA = 2,
    WHITE_MAGENTA_GREEN = 3,
    GREEN_WHITE_MAGENTA = 4,
    GREEN_MAGENTA_WHITE = 5,
    NUMBER_OF_PATTERNS = 6
  };

  int idOfPattern[NUMBER_OF_PATTERNS];

  int getIdOfPattern(int white, int green, int magenta)
  {
    Pattern pattern;
    if(white == 0)
    {
      if(green == 1)
        pattern = WHITE_GREEN_MAGENTA;
      else
        pattern = WHITE_MAGENTA_GREEN;
    }
    else if(green == 0)
    {
      if(white == 1)
        pattern = GREEN_WHITE_MAGENTA;
      else
        pattern = GREEN_MAGENTA_WHITE;
    }
    else
    {
      if(white == 1)
        pattern = MAGENTA_WHITE_GREEN;
      else
        pattern = MAGENTA_GREEN_WHITE;
    }
    return idOfPattern[pattern];
  }
};


typedef std::vector<RobotPerceptGT> RobotPerceptListGT;

#endif // InternalData_h__
