/** 
* @file ControllerQt/GroundTruth/Transformation.cpp
*
* Implementation of a class for transforming coordinates between camera and world coordinates
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/


#include "Platform/GTAssert.h"
#include "Transformation.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Settings.h"
#include "Tools/Streams/InStreams.h"
#include <../Util/OpenCV/include/cv.h>

Transformation::Transformation(const std::string& name,
                               FireWireCameraParameters& cameraParameters,
                               const FireWireImage& image):
                               name(name),
                               cameraParameters(cameraParameters),
                               image(image),
                               lut_2D_to_3D(0), lutChanged(false)
{
  // Create CV matrices to hold the transformation parameters    
  intrinsic_params   = cvCreateMat(3, 3, CV_64F);
  distortion_coeffs  = cvCreateMat(1, 4, CV_64F);
  rotation_vector    = cvCreateMat(1, 3, CV_64F);
  translation_vector = cvCreateMat(1, 3, CV_64F);
  object_point = cvCreateMat(1, 1, CV_32FC3);
  image_point = cvCreateMat(1, 1, CV_32FC2);
}

Transformation::~Transformation()
{
  saveConfig();
  if(lut_2D_to_3D)
    delete lut_2D_to_3D;
}

void Transformation::init()
{
  lut_2D_to_3D = new unsigned int[image.width * image.height * ObjectHeights::NUM_OF_HEIGHTS];
  distortion_coeffs->data.db[0] = cameraParameters.distortion[0];
  distortion_coeffs->data.db[1] = cameraParameters.distortion[1];
  distortion_coeffs->data.db[2] = cameraParameters.distortion[2];
  distortion_coeffs->data.db[3] = cameraParameters.distortion[3];
  intrinsic_params->data.db[0] = cameraParameters.focusLengthX;
  intrinsic_params->data.db[1] = 0;
  intrinsic_params->data.db[2] = cameraParameters.principalPointX;
  intrinsic_params->data.db[3] = 0;
  intrinsic_params->data.db[4] = cameraParameters.focusLengthY;
  intrinsic_params->data.db[5] = cameraParameters.principalPointY;
  intrinsic_params->data.db[6] = 0;
  intrinsic_params->data.db[7] = 0;
  intrinsic_params->data.db[8] = 1;
  loadConfig();
}

void Transformation::loadConfig()
{
  // Load object heights:
  std::string configFile("GroundTruth/heights.cfg");
  InConfigFile stream(Global::getSettings().expandLocationFilename(configFile));
  if(stream.exists())
    for(int i = 0; i < ObjectHeights::NUM_OF_HEIGHTS; i++)
      stream >> objectHeights.heights[i];
  else
    OUTPUT(idText,text,"Warning! GroundTruth/heights.cfg does not exist, using default values.");

  // Load dimensions of area covered by transformation:
  std::string fieldFile("GroundTruth/");
  fieldFile += (name + "/field.cfg");
  InConfigFile fieldStream(Global::getSettings().expandLocationFilename(fieldFile));
  if(fieldStream.exists())
  {
    fieldStream >> maxFieldX
                >> maxFieldY;
  }
  else // Some default values and a warning
  {
    maxFieldX = 6000;
    maxFieldY = 6000;
    OUTPUT(idText,text,"Warning! "<<fieldFile.c_str()<<" does not exist, using default values.");
  }

  // Load points for extrinsic camera calibration:
  std::string pointsFile("GroundTruth/");
  pointsFile += (name + "/points.cfg");
  InConfigFile pointStream(Global::getSettings().expandLocationFilename(pointsFile));
  /** Be sure that the file exists. Continuing with default parameters
      does not make much sense*/
  ASSERT(pointStream.exists());
  while(!pointStream.eof())
  {
    ExtrinsicCameraCalibrationPoint p;
    std::string name;
    pointStream >> name;
    // Parse line for fancy drawing
    if(name == "connect")
    {
      CalibrationDrawingLine line;
      std::string p1, p2;
      pointStream >> p1 >> p2;
      bool p1Found(false), p2Found(false);
      std::vector<ExtrinsicCameraCalibrationPoint>::const_iterator it;
      for(it = calibrationPoints.begin(); it != calibrationPoints.end(); ++it)
      {
        if((!p1Found) && it->name == p1)
        {
          line.p1 = &*it;
          p1Found = true;
        }
        else if((!p2Found) && it->name == p2)
        {
          line.p2 = &*it;
          p2Found = true;
        }
        else if(p1Found && p2Found)
          break;
      }
      if(p1Found && p2Found)
        calibrationDrawingLines.push_back(line);
      else
        OUTPUT(idText, text, "Unknown point(s) for line from "<<p1<<" to "<<p2);
    }
    // Parse calibration point
    else
    {
      p.name = name;
      pointStream >> p.posWorld.x
        >> p.posWorld.y
        >> p.posImage.x
        >> p.posImage.y;
      calibrationPoints.push_back(p);
    }
  }

  // Read the previous lookup table from a file
  FILE* lut_file = 0;
  std::string lutFilePath("GroundTruth/");
  lutFilePath += (name + "/2D_to_3D.lut");
  lutFilePath = std::string(File::getGTDir())+"/Config/"+Global::getSettings().expandLocationFilename(lutFilePath);
  lut_file = fopen(lutFilePath.c_str(), "rb+" );
  if (lut_file)
  {
    // Read transformation parameters from the file
    fread(rotation_vector->data.ptr, 8, rotation_vector->cols, lut_file);
    fread(translation_vector->data.ptr, 8, translation_vector->cols, lut_file);
    fread(lut_2D_to_3D, sizeof(unsigned int), image.width * image.height *
      ObjectHeights::NUM_OF_HEIGHTS, lut_file);
    fclose(lut_file);
  }
  else
  {
    OUTPUT(idText, text, "No 2D to 3D lookup table found. Will be recomputed now.");
    computeTransformationParameters();
  }    
}

void Transformation::saveConfig()
{
  // Write transformation parameters and LUT to the file
  if(lutChanged)
  {
    FILE* lut_file = 0;
    std::string lutFilePath("GroundTruth/");
    lutFilePath += (name + "/2D_to_3D.lut");
    lutFilePath = std::string(File::getGTDir())+"/Config/"+Global::getSettings().expandLocationFilename(lutFilePath);
    lut_file = fopen(lutFilePath.c_str(), "wb+" );
    if(lut_file)
    {
      fwrite(rotation_vector->data.ptr, 8, rotation_vector->cols, lut_file);
      fwrite(translation_vector->data.ptr, 8, translation_vector->cols, lut_file);
      fwrite(lut_2D_to_3D, sizeof(unsigned int), image.width * image.height * ObjectHeights::NUM_OF_HEIGHTS, lut_file);
      fclose(lut_file);
    }
  }

  // Save points for extrinsic camera calibration:
  std::string pointsFile("GroundTruth/");
  pointsFile += (name + "/points.cfg");
  OutTextRawFile stream(Global::getSettings().expandLocationFilename(pointsFile));
  stream << "/**" << endl
         << "This file contains points for computing the extrinsic camera parameters. It should contain only points that are visible in the current image." << endl
         << "The format of one line is as follows:" << endl
         << "<Name of point>" << endl
         << "<x-coordinate in world>" << endl
         << "<y-coordinate in world>" << endl
         << "<x-coordinate in image>" << endl
         << "<y-coordinate in image>" << endl
         << "The two latter values should be set interactively." << endl
         << "Do not use spaces or any other separation characters in the name." << endl
         << "World coordinates are (0,0) at the center of the field." << endl
         << "The x-axis is in direction from the center of the yellow" << endl
         << "goal to the center of the blue goal." << endl
         << "*/" << endl
         << endl;

  std::vector<ExtrinsicCameraCalibrationPoint>::const_iterator it;
  for(it = calibrationPoints.begin(); it != calibrationPoints.end(); ++it)
    stream << it->name << " " 
           << it->posWorld.x << " "
           << it->posWorld.y << " "
           << it->posImage.x << " "
           << it->posImage.y << endl;

  stream << endl
         << "/*" << endl
         << "For a more comprehensible drawing of the points" << endl
         << "for interactive configuration, points can be connected" << endl
         << "via lines. Syntax:" << endl
         << "connect <name of point 1> <name of point 2>" << endl
         << "*/" << endl
         << endl;

  std::vector<CalibrationDrawingLine>::const_iterator itl;
  for(itl = calibrationDrawingLines.begin(); itl != calibrationDrawingLines.end(); ++itl)
    stream << "connect " << itl->p1->name << " " << itl->p2->name << endl;
}

inline bool Transformation::inImage(int x, int y) const
{
  return x >= 0 && x < int(image.width) && y >= 0 && y < int(image.height);
}

void Transformation::computeTransformationParameters()
{
  OUTPUT(idText, text, "Recomputing look-up table...");
  // Put all specific points on the field into a CV vector
  CvMat* object_points = cvCreateMat(1, calibrationPoints.size(), CV_32FC3);    
  CvPoint3D32f* op = (CvPoint3D32f*)object_points->data.fl;
  for (unsigned int i = 0; i < calibrationPoints.size(); i++)
  {
    *op++ = cvPoint3D32f(calibrationPoints[i].posWorld.x, calibrationPoints[i].posWorld.y, 0);
  }

  // Put all specific points in the image into a CV vector
  CvMat* image_points = cvCreateMat(1, calibrationPoints.size(), CV_32FC2);    
  CvPoint2D32f* ip = (CvPoint2D32f*)image_points->data.fl;
  for (unsigned int i = 0; i < calibrationPoints.size(); i++)
  {
    *ip++ = cvPoint2D32f(calibrationPoints[i].posImage.x, calibrationPoints[i].posImage.y);
  }

  // Let CV find the extrinsic parameters
  cvFindExtrinsicCameraParams2(object_points, image_points, intrinsic_params,
    distortion_coeffs, rotation_vector, translation_vector);

  // Create 2D to 3D lookup-table. The points are grouped according to their
  // y-coordinate. That is for every y-ccordinate one vector with all the
  // x-cooordinate and height combinations is passed to CV. Passing all points
  // in one go caused CV to run out of memory and the progress can not be
  // monitored.
  CvMat* mm_field = cvCreateMat(1, maxFieldX * ObjectHeights::NUM_OF_HEIGHTS, CV_32FC3);
  CvMat* mm_image = cvCreateMat(1, maxFieldX * ObjectHeights::NUM_OF_HEIGHTS, CV_32FC2);
  CvPoint3D32f* fp;

  int my = maxFieldY / 2;
  int mx = maxFieldX / 2;

  for (int y = -my; y < my; y += 1)
  {
    // Set all the field points for the current y coordinate
    fp = (CvPoint3D32f*)mm_field->data.fl;
    for (int x = -mx; x < mx; x += 1) 
    {
      for(int i = 0; i < ObjectHeights::NUM_OF_HEIGHTS; i++)
      {
        *fp++ = cvPoint3D32f(x, y, objectHeights.heights[i]);
      }
    }

    // Let CV translate the points to image coordinates
    cvProjectPoints2(mm_field, rotation_vector, translation_vector,
      intrinsic_params, distortion_coeffs, mm_image);

    // Put the entries into the lookup-table
    CvPoint2D32f* ip = (CvPoint2D32f*)mm_image->data.fl;
    for (int x = -mx; x < mx; x++)
    {
      unsigned int coordinates = ((x + mx) << 16) | (y + my);

      int u = static_cast<int>(ip->x);
      int v = static_cast<int>(ip->y);
      ip++;
      if (inImage(u, v))
      {
        lut_2D_to_3D[v * image.width * 3 + u * 3] = coordinates;
      }

      u = static_cast<int>(ip->x);
      v = static_cast<int>(ip->y);
      ip++;
      if (inImage(u, v))
      {         
        lut_2D_to_3D[v * image.width * 3 + u * 3 + 1] = coordinates;
      }

      u = static_cast<int>(ip->x);
      v = static_cast<int>(ip->y);
      ip++;
      if (inImage(u, v))
      {          
        lut_2D_to_3D[v * image.width * 3 + u * 3 + 2] = coordinates;
      }
    }
  }
  OUTPUT(idText, text, "...finished.");
  lutChanged = true;
}

Vector2<double> Transformation::toImage(double x, double y, ObjectHeights::HeightClass h) const
{
  // Determine the real height of the given height class
  double z = objectHeights.heights[h];
  
  // Let CV translate the 3D point into image coordinates
  *((CvPoint3D32f*)object_point->data.fl) = cvPoint3D32f(x, y, z);

  cvProjectPoints2(object_point, rotation_vector, translation_vector,
    intrinsic_params, distortion_coeffs, image_point);

  // Extract and return the translated coordinates                 
  CvPoint2D32f* ip = (CvPoint2D32f*)image_point->data.fl;
  return Vector2<double>(ip->x, ip->y);
}

Vector2<double> Transformation::toField(double u, double v, ObjectHeights::HeightClass h) const
{
  int my = maxFieldY / 2;
  int mx = maxFieldX / 2;

  // Round pixels. In this context 'down' is used as 'towards 0'. That is,
  //   numbers > 0 decrease and numbers < 0 increase when rounded down.
  int u_down = static_cast<int>(u);
  int u_up   = static_cast<int>(u < 0 ? u - 1 : u + 1);
  int v_down = static_cast<int>(v);
  int v_up   = static_cast<int>(v < 0 ? v - 1 : v + 1);

  // Build factors = how close is the rounded value to the original?
  double u_df = 1 - (u - u_down);
  double u_uf = 1 - u_df;
  double v_df = 1 - (v - v_down);
  double v_uf = 1 - v_df;

  // Look up the 4 adjacent pixels coordinates
  unsigned int coordinates = lut_2D_to_3D[v_down * image.width * 3 + u_down * 3 + h];
  int x1 = (coordinates >> 16) - mx;
  int y1 = (coordinates & 0xFFFF) - my;

  coordinates = lut_2D_to_3D[v_down * image.width * 3 + u_up * 3 + h];
  int x2 = (coordinates >> 16) - mx;
  int y2 = (coordinates & 0xFFFF) - my;

  coordinates = lut_2D_to_3D[v_up * image.width * 3 + u_down * 3 + h];
  int x3 = (coordinates >> 16) - mx;
  int y3 = (coordinates & 0xFFFF) - my;

  coordinates = lut_2D_to_3D[v_up * image.width * 3 + u_up * 3 + h];
  int x4 = (coordinates >> 16) - mx;
  int y4 = (coordinates & 0xFFFF) - my;

  // Combine the 4 pixels regarding their distance to the real pixel coordinates
  u = x1 * v_df * u_df + x2 * v_df * u_uf + x3 * v_uf * u_df + x4 * v_uf * u_uf;
  v = y1 * v_df * u_df + y2 * v_df * u_uf + y3 * v_uf * u_df + y4 * v_uf * u_uf;
  return Vector2<double>(u,v);
}

void Transformation::configure()
{
  DECLARE_DEBUG_DRAWING("ground truth:transformation:calibration points", "drawingOnImage", "Points for calibrating extrinsic camera parameters.");
  // Draw points:
  std::vector<ExtrinsicCameraCalibrationPoint>::const_iterator it;
  for(it = calibrationPoints.begin(); it != calibrationPoints.end(); ++it)
  {
    CIRCLE("ground truth:transformation:calibration points", it->posImage.x, it->posImage.y, 5, 1,
           Drawings::ps_solid, ColorClasses::orange, Drawings::bs_solid, ColorClasses::orange);
    TIP("ground truth:transformation:calibration points", it->posImage.x, it->posImage.y, 5, it->name.c_str());
  }
  // Draw lines:
  std::vector<CalibrationDrawingLine>::const_iterator itl;
  for(itl = calibrationDrawingLines.begin(); itl != calibrationDrawingLines.end(); ++itl)
  {
    LINE("ground truth:transformation:calibration points", itl->p1->posImage.x, itl->p1->posImage.y, 
                                            itl->p2->posImage.x, itl->p2->posImage.y, 
         3, Drawings::ps_solid, ColorClasses::orange);
  }

  // Recomputation of look-up table:
  DEBUG_RESPONSE("ground truth:transformation:recompute LUT", computeTransformationParameters(););

  // Reverse check of the transformation parameters
  DECLARE_DEBUG_DRAWING("ground truth:transformation:reverse check", "drawingOnImage", "Reverse check of the transformation parameters.");
  COMPLEX_DRAWING("ground truth:transformation:reverse check", drawReverseCheck(););
}

std::string Transformation::getPointAt(double x, double y) const
{
  std::vector<ExtrinsicCameraCalibrationPoint>::const_iterator it;
  for(it = calibrationPoints.begin(); it != calibrationPoints.end(); ++it)
    if(Vector2<double>(x - it->posImage.x, y - it->posImage.y).abs() <= 3)
      return it->name;
  return "";
}

void Transformation::setPointTo(const std::string& name, double x, double y)
{
  std::vector<ExtrinsicCameraCalibrationPoint>::iterator it;
  for(it = calibrationPoints.begin(); it != calibrationPoints.end(); ++it)
    if(it->name == name)
    {
      it->posImage.x = x;
      it->posImage.y = y;
      return;
    }
  ASSERT(false);
}

void Transformation::drawReverseCheck()
{
  // Revert all calibration drawing lines:
  for(unsigned int i=0; i<calibrationDrawingLines.size(); i++)
  {
    Vector2<double> p1 = calibrationDrawingLines[i].p1->posWorld;
    Vector2<double> p2 = calibrationDrawingLines[i].p2->posWorld;
    Vector2<double> vec = (p2 - p1)/10.0;
    for(int j=0; j<=10; j++)
    {
      Vector2<double> pos = p1 + (vec*j);
      pos = toImage(pos.x, pos.y, ObjectHeights::FIELD_HEIGHT);
      MID_DOT("ground truth:transformation:reverse check", pos.x, pos.y, ColorRGBA(255,0,255), ColorRGBA(255,0,255));
    }
  }
}
