/** 
* @file ControllerQt/GroundTruth/Transformation.h
*
* Declaration of a class for transforming coordinates between camera and world coordinates
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef Transformation_h_
#define Transformation_h_

#include "InternalData.h"
#include "Tools/Math/Vector2.h"

struct CvMat;

/**
* @class Transformation
*
* Class for coordinate transformation
*/
class Transformation
{
public:
  /**
  * Constructor.
  * @param name The name of the camera.
  * @param cameraParameters The camera's parameters
  * @param image The camera image.
  */
  Transformation(const std::string& name, 
                 FireWireCameraParameters& cameraParameters,
                 const FireWireImage& image);

  /** Destructor */
  ~Transformation();

  /**
  * Initializes the class
  * This function is needed since some information are not available
  * at construction time.
  */
  void init();

  /** Draws points on image and/or field, changes parameters */
  void configure();

  /** Draws the transformation from the field to the image to check the
      correctness of the parameters*/
  void drawReverseCheck();

  /**
  * Transforms a point <em>(x,y)</em> on the field to coordinates in the
  * corresponding camera image.
  *
  * @param x The x coordinate of the point
  * @param y The y coordinate of the point
  * @param h The height of the object the point belongs to
  * @return The image coordinates for this point
  */
  Vector2<double> toImage(double x, double y, ObjectHeights::HeightClass h) const;

  /**
  * Transforms a point <em>(u,v)</em> in an image to coordinates on the
  * field.
  *
  * @param  u  The u coordinate of the point
  * @param  v  The v coordinate of the point
  * @param  h  The height class of the object to which the point belongs
  * @return    The field coordinates for this point
  */
  Vector2<double> toField(double u, double v, ObjectHeights::HeightClass h) const;

  /** Saves the configuration*/
  void saveConfig();

  std::string getPointAt(double x, double y) const;

  void setPointTo(const std::string& name, double x, double y);

private:
  /** Data for a single calibration point*/
  struct ExtrinsicCameraCalibrationPoint
  {
    Vector2<double> posImage;  /**< Pixel coordinates*/
    Vector2<double> posWorld;  /**< World coordinates*/
    std::string name;          /**< Name (for drawing)*/
  };

  /** A simple line for drawing ines between calibration points*/
  struct CalibrationDrawingLine
  {
    const ExtrinsicCameraCalibrationPoint* p1; /**< First point of line */
    const ExtrinsicCameraCalibrationPoint* p2; /**< Second point of line */
    CalibrationDrawingLine() : p1(0), p2(0) {}
  };

  /** Loads the configuration*/
  void loadConfig();

  /**
  * Tests if the given coordinates are inside the image bounds.
  *
  * @param  x  the x coordinate of the point
  * @param  y  the y coordinate of the point
  * @return    <em>true</em>, if the point is inside the image
  */
  inline bool inImage(int x, int y) const;

  /**
  * Computes transformation parameters. In the first step, the extrinsic
  * camera parameters are calculated, based on the intrinsic parameters 
  * and the relation between image and field coordinates which is given by the 
  * set of specific points.
  * In the second step, a look-up-table is built for looking up 3D
  * coordinates to a given 2D image point. 
  */
  void computeTransformationParameters();

  const std::string name;       //!< The camera's name 
  FireWireCameraParameters& cameraParameters;  /**< The camera's parameters*/
  std::vector<ExtrinsicCameraCalibrationPoint> calibrationPoints; //!< Calibration points
  std::vector<CalibrationDrawingLine> calibrationDrawingLines;    //!< Lines between calibration points
  const FireWireImage& image;   //!< A reference to the image buffer
  CvMat* intrinsic_params;      //!< Camera specific parameters
  CvMat* distortion_coeffs;     //!< Camera's lens distortion
  CvMat* rotation_vector;       //!< Camera orientation in 3D Space
  CvMat* translation_vector;    //!< Camera location in 3D Space
  CvMat* object_point;          //!< Holds a single object point 
  CvMat* image_point;           //!< Holds a single image point 
  unsigned int* lut_2D_to_3D;   //!< Look-up-table from 2D to 3D points
  bool lutChanged;              //!< Flag. LUT will be saved, if true
  ObjectHeights objectHeights;  //!< Yes, the heights of different objects
  int maxFieldX;                //!< The resolution of the transformation in x direction
  int maxFieldY;                //!< The resolution of the transformation in y direction
};

#endif // GroundTruth_h_
