/**
* @file ControllerQt/Platform/Win32/FireWireCamera.cpp
* The file implements a class to communicate with a FireWire camera
* using the CMU1394 library
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:khuebner@informatik.uni-bremen.de">Kai Hbner</a>
*/

#include <windows.h>
#include <sstream>

#include "Platform/SystemCall.h"
#include "Platform/GTAssert.h"
#include <../Util/CMU1394Camera/include/1394Camera.h>
#include "Tools/Debugging/Debugging.h"
#include "Tools/Debugging/Modify.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Settings.h"
#include "FireWireCamera.h"
#include "ControllerQt/GroundTruth/InternalData.h"


FireWireCamera::FireWireCamera(const std::string& name, int busID, FireWireImage& image,
                               FireWireCameraParameters& cameraParameters)
                               : name(name), busID(busID), image(image),
                                 cameraParameters(cameraParameters), cameraConnected(false)
{
  theCamera = new C1394Camera;
  theFocusControl = 0;
  theShutterControl = 0;
  theWhiteControl = 0;
}

FireWireCamera::~FireWireCamera()
{
  delete theCamera;
  if(theFocusControl)
    delete theFocusControl;
  if(theShutterControl)
    delete theShutterControl;
  if(theWhiteControl)
    delete theWhiteControl;
}

void FireWireCamera::start()
{
  //Load camera configuration from file
  loadCameraParameters();

  if(theCamera->CheckLink() == CAM_SUCCESS && busID < theCamera->GetNumberCameras())
  {
    //General initialization of camera:
    ASSERT(busID < theCamera->GetNumberCameras());
    VERIFY(theCamera->SelectCamera(busID) == CAM_SUCCESS);
    VERIFY(theCamera->InitCamera() == CAM_SUCCESS);

    // Set camera parameters:
    VERIFY(theCamera->SetVideoFormat(cameraParameters.videoFormat) == CAM_SUCCESS);
    VERIFY(theCamera->SetVideoMode(cameraParameters.videoMode) == CAM_SUCCESS);
    VERIFY(theCamera->SetVideoFrameRate(cameraParameters.videoFrameRate) == CAM_SUCCESS);
    theCamera->GetVideoFrameDimensions(&image.width, &image.height);
    VERIFY(theCamera->StartImageAcquisition() == CAM_SUCCESS);
    VERIFY(theCamera->AcquireImage() == CAM_SUCCESS);
    image.data = new unsigned char[image.width * image.height * 3];
    memset(image.data, 0, image.width * image.height * 3);

    // Get some additional information from the camera:
    char buf[200];
    int bufLen(200);
    std::string cameraName, cameraVendor;
    theCamera->GetCameraName(&buf[0],bufLen);
    cameraName = buf;
    theCamera->GetCameraVendor(&buf[0],bufLen);
    cameraVendor = buf;
    LARGE_INTEGER uniqueID;
    theCamera->GetCameraUniqueID(&uniqueID);
    std::ostringstream s1;
    s1 << uniqueID.QuadPart;
    uniqueCameraIdentifier = uniqueID.QuadPart;
    OUTPUT(idText, text, "Found camera on bus "<<busID<<":");
    OUTPUT(idText, text, "Unique ID: "<<s1.str());
    OUTPUT(idText, text, "Name:        "<<cameraName);
    OUTPUT(idText, text, "Vendor:      "<<cameraVendor);
    cameraConnected = true;

    // Initialize features:
    if(theCamera->HasFeature(FEATURE_FOCUS))
    {
      theFocusControl = new C1394CameraControl(theCamera, FEATURE_FOCUS);
      unsigned short minF(42), maxF(42), current(42);
      theFocusControl->SetValue(cameraParameters.focus);
      theFocusControl->Inquire();
      theFocusControl->Status();
      theFocusControl->GetValue(&current);
      theFocusControl->GetRange(&minF, &maxF);
      OUTPUT(idText, text, "Camera has adjustable focus:");
      OUTPUT(idText, text, "Min:      "<<minF);
      OUTPUT(idText, text, "Max:      "<<maxF);
      OUTPUT(idText, text, "Current:  "<<current);
    }
    if(theCamera->HasFeature(FEATURE_SHUTTER))
    {
      theShutterControl = new C1394CameraControl(theCamera, FEATURE_SHUTTER);
      unsigned short minS(42), maxS(42), current(42);
      theShutterControl->SetValue(cameraParameters.shutter);
      theShutterControl->Inquire();
      theShutterControl->Status();
      theShutterControl->GetValue(&current);
      theShutterControl->GetRange(&minS, &maxS);
      OUTPUT(idText, text, "Camera has adjustable shutter:");
      OUTPUT(idText, text, "Min:      "<<minS);
      OUTPUT(idText, text, "Max:      "<<maxS);
      OUTPUT(idText, text, "Current:  "<<current);
    }
    if(theCamera->HasFeature(FEATURE_WHITE_BALANCE))
    {
      theWhiteControl = new C1394CameraControl(theCamera, FEATURE_WHITE_BALANCE);
      unsigned short minF(42), maxF(42), current1(42), current2(42);
      theWhiteControl->SetValue(cameraParameters.whiteBalance1, 
                                cameraParameters.whiteBalance2);
      theWhiteControl->Inquire();
      theWhiteControl->Status();
      theWhiteControl->GetValue(&current1, &current2);
      theWhiteControl->GetRange(&minF, &maxF);
      OUTPUT(idText, text, "Camera has adjustable white balance:");
      OUTPUT(idText, text, "Min:      "<<minF);
      OUTPUT(idText, text, "Max:      "<<maxF);
      OUTPUT(idText, text, "Current1:  "<<current1);
      OUTPUT(idText, text, "Current2:  "<<current2);
    }

    // Additional check and warning
    if(cameraParameters.uniqueID != s1.str())
    {
      OUTPUT(idText, text, " ");
      OUTPUT(idText, text, "Warning! Unique camera identifier in camera.cfg differs from the camera's identifier!");
      OUTPUT(idText, text, "Camera: "<<s1.str()<<"  camera.cfg: "<<cameraParameters.uniqueID);
    }
  }
  else
  {
    OUTPUT(idText, text, "There is no camera connected to this computer on bus "<<busID<<".");
    OUTPUT(idText, text, "Image processing uses a pacifying gray dummy image with some calibration lines now.");
    // Create the dummy image:
    image.width = 640;
    image.height = 480;
    image.data = new unsigned char[image.width * image.height * 3];
    memset(image.data, 128, image.width * image.height * 3);
    // Create light orange part
    for(int y=0; y < 20; y++)
    {
      int x=0;
      while(x<640)
      {
        image.data[y*480*3 + 3*x] = 255;
        image.data[y*480*3 + 3*x+1] = 160;
        image.data[y*480*3 + 3*x+2] = 80;
        ++x;
      }
    }
    // Create light blue part
    for(int y=20; y < 40; y++)
    {
      int x=0;
      while(x<640)
      {
        image.data[y*480*3 + 3*x] = 100;
        image.data[y*480*3 + 3*x+1] = 100;
        image.data[y*480*3 + 3*x+2] = 255;
        ++x;
      }
    }
    // Create light yellow part
    for(int y=40; y < 60; y++)
    {
      int x=0;
      while(x<640)
      {
        image.data[y*480*3 + 3*x] = 255;
        image.data[y*480*3 + 3*x+1] = 230;
        image.data[y*480*3 + 3*x+2] = 100;
        ++x;
      }
    }
  }
}

void FireWireCamera::stop()
{
  if(cameraConnected)
    VERIFY(theCamera->StopImageAcquisition() == CAM_SUCCESS);
  if(image.data)
    delete [] image.data;
  image.data = 0;
  cameraConnected = false;
}

bool FireWireCamera::execute()
{  
  if(cameraConnected && image.data && theCamera->AcquireImage() == CAM_SUCCESS)
  {
    theCamera->getRGB(image.data, image.width * image.height * 3);
    image.timeStamp = SystemCall::getCurrentSystemTime();
    return true;
  }
  else
  {
    Sleep(20);    //Pretend to do something ;-)
    return false;
  }
}

void FireWireCamera::configure()
{  
  int oldFocus = cameraParameters.focus;
  MODIFY("groundTruth:cameraFocus", cameraParameters.focus);
  if(cameraParameters.focus != oldFocus) //Check for change
  {
    if(theFocusControl)
    {
    }
    else
    {
      OUTPUT(idText,text,"Sorry, setting the focus is not possible for this camera");
    }
  }

  //Save and load image for testing
  DEBUG_RESPONSE("groundTruth:camera:save image", saveImage(););
  DEBUG_RESPONSE("groundTruth:camera:load image", loadImage(););
}

void FireWireCamera::loadCameraParameters()
{
  std::string configFile("GroundTruth/");
  configFile += (name + "/camera.cfg");
  InConfigFile stream(Global::getSettings().expandLocationFilename(configFile));
  /** Be sure that the file exists. Continuing with default parameters
      does not make much sense*/
  ASSERT(stream.exists());
  stream >> cameraParameters.uniqueID
         >> cameraParameters.focusLengthX
         >> cameraParameters.focusLengthY
         >> cameraParameters.principalPointX
         >> cameraParameters.principalPointY
         >> cameraParameters.distortion[0]
         >> cameraParameters.distortion[1]
         >> cameraParameters.distortion[2]
         >> cameraParameters.distortion[3]
         >> cameraParameters.videoFormat
         >> cameraParameters.videoMode
         >> cameraParameters.videoFrameRate
         >> cameraParameters.focus
         >> cameraParameters.shutter
         >> cameraParameters.whiteBalance1
         >> cameraParameters.whiteBalance2;
}

void FireWireCamera::saveImage()
{
  FILE* image_file = 0;
  std::string imageFilePath(std::string("GroundTruth/") + name + "/testimage.rgb");
  imageFilePath = std::string(File::getGTDir())+"/Config/"+Global::getSettings().expandLocationFilename(imageFilePath);
  image_file = fopen(imageFilePath.c_str(), "wb+" );
  if(image_file)
  {
    fwrite(image.data, sizeof(unsigned char), image.width * image.height * 3, image_file);
    fclose(image_file);
  }
}

void FireWireCamera::loadImage()
{
  FILE* image_file = 0;
  std::string imageFilePath(std::string("GroundTruth/") + name + "/testimage.rgb");
  imageFilePath = std::string(File::getGTDir())+"/Config/"+Global::getSettings().expandLocationFilename(imageFilePath);
  image_file = fopen(imageFilePath.c_str(), "rb+" );
  if(image_file)
  {
    fread(image.data, sizeof(unsigned char), image.width * image.height * 3, image_file);
    fclose(image_file);
  }
}
