/**
* @file ControllerQt/Platform/Win32/FireWireCamera.h
* The file declares a class to communicate with a FireWire camera.
* using the CMU1394 library
*
* @author <a href="mailto:khuebner@informatik.uni-bremen.de">Kai Hbner</a>
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef __FIREWIRECAMERA_H__
#define __FIREWIRECAMERA_H__

#include <string>

class C1394Camera;
class C1394CameraControl;
struct FireWireImage;
struct FireWireCameraParameters;

/**
* @class FireWireCamera
* The class communicates with a FireWire camera.
*/
class FireWireCamera
{
public:
  /** 
   * Constructor.
   * @param name The name of the camera.
   * @param busID The firewire bus id of the camera
   * @param image A reference to the image buffer
   * @param cameraParameters The parameters used for the camera.
   */
  FireWireCamera(const std::string& name, int busID, 
                 FireWireImage& image,
                 FireWireCameraParameters& cameraParameters);

  /** Destructor */
  ~FireWireCamera();

  /** Start recording. */
  void start();

  /** Stop recording. */
  void stop();

  /**
  * The function executes a further step in communication with 
  * the firewire camera. It has to be called quite regularily.
  * @return Was a new image received?
  */
  bool execute();

  /** Change parameters online, called every frame*/
  void configure();

  /** Returns the member
   * @return The unique camera identifier
   */
  long long getUniqueIdentifier() const
  { return uniqueCameraIdentifier;}

  /** Returns the member
   * @return The value of the cameraConnected flag
   */
  bool isCameraConnected() const
  { return cameraConnected;}

private:
   /** Loads the camera parameters from a file*/
  void loadCameraParameters();

  C1394Camera* theCamera;                       /**< A pointer to a CMU firewire camera object. */
  C1394CameraControl* theFocusControl;          /**< A pointer to control a camera's focus*/ 
  C1394CameraControl* theShutterControl;        /**< A pointer to control a camera's shutter*/ 
  C1394CameraControl* theWhiteControl;          /**< A pointer to control a camera's white balance*/ 
  std::string name;                             /**< The name of the camera. */
  int busID;                                    /**< The firewire bus id of the camera*/
  long long uniqueCameraIdentifier;             /**< The identifier of the currently connected camera. */
  FireWireImage& image;                         /**< A reference to the image */
  FireWireCameraParameters& cameraParameters;   /**< Yes, as the name says...*/
  bool cameraConnected;                         /**< Simple flag */

  /** Save the current image to a file for later testing*/
  void saveImage();

  /** Load an image from a file for testing*/
  void loadImage();
};

#endif //__FIREWIRECAMERA_H__
