/** 
* @file ControllerQt/TeamRobot.cpp
* Implementation of a class representing a process that communicates with a remote robot via team communication.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "TeamRobot.h"
#define Drawings ::Drawings // Base class also defines Drawings, but need the global class

int TeamRobot::main() 
{
  SYNC;
  OUTPUT(idProcessBegin, bin, 't');

  DECLARE_DEBUG_DRAWING("representation:RobotPose", "drawingOnField", "The robot pose");
  DECLARE_DEBUG_DRAWING("origin:RobotPose", "drawingOnField", "Set the origin to the robot's current position.");
  DECLARE_DEBUG_DRAWING("representation:BallModel", "drawingOnField", "drawing of the ball model");
  DECLARE_DEBUG_DRAWING("representation:GoalModel", "drawingOnField", "drawing of the goal model");
  DECLARE_DEBUG_DRAWING("representation:MotionRequest", "drawingOnField", "drawing of a request walk vector");
  fieldDimensions.declareDrawings();

  if(SystemCall::getTimeSince(robotPoseReceived) < 1000)
    robotPose.draw();
  if(SystemCall::getTimeSince(ballModelReceived) < 1000)
    ballModel.draw();
  if(SystemCall::getTimeSince(goalModelReceived) < 1000)
    goalModel.draw();
  if(SystemCall::getTimeSince(motionRequestReceived) < 1000)
    motionRequest.draw();

  int teamColor = 0,
      swapSides = 0;
  MODIFY("teamColor", teamColor);
  MODIFY("swapSides", swapSides);

  COMPLEX_DRAWING("field polygons",
    if(swapSides ^ teamColor)
    {
      ORIGIN("field polygons", 0, 0, pi2); // hack: swap sides!
    }
    fieldDimensions.drawPolygons(teamColor);
  );
  COMPLEX_DRAWING("field lines", 
    fieldDimensions.drawLines();
    fieldDimensions.drawCrosses();
  );

  DECLARE_DEBUG_DRAWING("robotState", "drawingOnField", "text decribing the state of the robot");
  int lineY = 2550;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "behavior: role: " << BehaviorData::getRoleName(behaviorData.role) << ", action: " << BehaviorData::getActionName(behaviorData.action));
  lineY -= 180;
  if(ballModel.timeWhenLastSeen)
  {
    DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "ballLastSeen: " << SystemCall::getRealTimeSince(ballModel.timeWhenLastSeen) << " ms");
  }
  else
  {
    DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "ballLastSeen: never");
  }
  lineY -= 180;
  if(goalModel.timeWhenLastSeen)
  {
    DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "goalLastSeen: " << SystemCall::getRealTimeSince(goalModel.timeWhenLastSeen) << " ms");
  }
  else
  {
    DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "goalLastSeen: never");
  }
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "voltage: " << robotHealth.voltage << " V");
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "max temperature: " << robotHealth.maxTemperature << " C");
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "cognition rate: " << roundNumberToInt(robotHealth.cognitionFrameRate) << " fps");
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "motion rate: " << roundNumberToInt(robotHealth.motionFrameRate) << " fps");
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "CM5 receive rate: " << roundNumberToInt(robotHealth.robotFramesPerSecond) << " fps");
  lineY -= 180;
  DRAWTEXT("robotState", -3600, lineY, 150, ColorClasses::white, "memory usage: " << static_cast<int>(robotHealth.memoryLoad) << " %");
  lineY -= 180;
  DRAWTEXT("robotState", -2600, 0, 900, ColorRGBA(255,255,255,50), robotHealth.robotName);

  DECLARE_DEBUG_DRAWING("robotOffline", "drawingOnField", "A huge X to display the online/offline state of the robot");
  if(SystemCall::getTimeSince(robotPoseReceived) > 500)
  {
    LINE("robotOffline", -3600, 2600, 3600, -2600, 50, Drawings::ps_solid, ColorRGBA(0xff, 0, 0));
    LINE("robotOffline", 3600, 2600, -3600, -2600, 50, Drawings::ps_solid, ColorRGBA(0xff, 0, 0));
  }    

  DEBUG_RESPONSE("automated requests:DrawingManager", OUTPUT(idDrawingManager, bin, Global::getDrawingManager()));
  DEBUG_RESPONSE("automated requests:DrawingManager3D", OUTPUT(idDrawingManager3D, bin, Global::getDrawingManager3D()));
  DEBUG_RESPONSE("automated requests:StreamSpecification", OUTPUT(idStreamSpecification, bin, Global::getStreamHandler()));

  OUTPUT(idProcessFinished, bin, 't');
  teamOut.moveAllMessages(teamIn);
  return 50;
}

TeamRobot::TeamRobot(const char* name, int number) :
RobotConsole(teamIn, teamOut), number(number)
{
  strcpy(this->name, name); 
  mode = SystemCall::teamRobot;
  fieldDimensions.load();
}
