/**
* @file ControllerQt/Views/FieldView.cpp
*
* Implementation of class FieldView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Colin Graf
*/

#include <QGLWidget>
#include <QPainter>
#include <QApplication>
#include <QMouseEvent>
#include <QResizeEvent>

#include <Representations/Configuration/FieldDimensions.h>
#include <Platform/Thread.h>

#include "FieldView.h"
#include "../RobotConsole.h"
#include "../Visualization/PaintMethods.h"

class FieldWidget : public QGLWidget
{
public:
  FieldWidget(RobotConsole& console, const std::string& name) : QGLWidget(QGLFormat(/*QGL::SampleBuffers*/)),
    console(console), name(name),
    lastDrawingsTimeStamp(0),
    widgetWidth(0), widgetHeight(0)
  {
    fieldDimensions.load();
  }

  void update()
  {
    if(needsRepaint())
      QGLWidget::update();
  }

private:

  void resizeGL(int width, int height)
  {
    widgetWidth = width;
    widgetHeight = height;

    int viewWidth = int(fieldDimensions.xPosOpponentGoal + fieldDimensions.flagRadius) * 2;
    int viewHeight = int(fieldDimensions.yPosLeftFlags + fieldDimensions.flagRadius) * 2;
    float xScale = float(widgetWidth)/viewWidth;
    float yScale = float(widgetHeight)/viewHeight;
    scale = xScale < yScale ? xScale : yScale;
  }

  void paintEvent(QPaintEvent *event)
  {     
    painter.begin(this);

    painter.setTransform(QTransform(scale, 0, 0, -scale, widgetWidth / 2, widgetHeight / 2));

    {
      SYNC_WITH(console);      
      paintDrawings();
    }

    painter.end();
  }

  void paintDrawings()
  {
    const QTransform baseTrans(painter.transform());
    const std::list<std::string>& drawings(console.fieldViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.fieldDrawings[*i]);
      PaintMethods::paintDebugDrawing(painter, debugDrawing, baseTrans);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        lastDrawingsTimeStamp = debugDrawing.timeStamp;
    }
    painter.setTransform(baseTrans);
  }

  bool needsRepaint() const
  {
    SYNC_WITH(console);
    const std::list<std::string>& drawings(console.fieldViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.fieldDrawings[*i]);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        return true;
    }
    return false;
  }

  void window2viewport(QPoint& point)
  {
    point = QPoint((point.x() - widgetWidth / 2) / scale, (point.y() - widgetHeight / 2) / scale);
  }

  void mousePressEvent(QMouseEvent* event)
  {
    if(event->button() == Qt::LeftButton || event->button() == Qt::MidButton)
      QApplication::beep();
  }

  void mouseMoveEvent(QMouseEvent* event)
  {
    {
      SYNC_WITH(console);
      if(!console.colorTableHandler.active)
      {
        QPoint pos(event->pos());
        window2viewport(pos);
        const char* text = 0;
        const std::list<std::string>& drawings(console.fieldViews[name]);
        for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
        {
          text = console.fieldDrawings[*i].getTip(pos.rx(), pos.ry());
          if(text)
            break;
        }
        if(text)
          setToolTip(QString(text));
        else
          setToolTip(QString());
      }
    }
  }

  RobotConsole& console;
  const std::string& name;
  unsigned int lastDrawingsTimeStamp;
  int widgetWidth;
  int widgetHeight;
  QPainter painter;
  FieldDimensions fieldDimensions; /**< The field dimensions. */
  float scale;
};



FieldView::FieldView(RobotConsole& console, const std::string& name) : 
  console(console), name(name) {}

QWidget* FieldView::createWidget()
{
  return new FieldWidget(console, name);
}

void FieldView::updateWidget(QWidget& widget)
{
  ((FieldWidget*)(&widget))->update();
}

