/**
* @file ControllerQt/Views/GroundTruthView.cpp
*
* Implementation of class GroundTruthView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Colin Graf
*/

#include <QGLWidget>
#include <QPainter>
#include <QApplication>
#include <QMouseEvent>

#include <Representations/Infrastructure/Image.h>
#include <Platform/Thread.h>
#include <GL/gl.h>

#include "GroundTruthView.h"
#include "../RobotConsole.h"
#include "../Visualization/PaintMethods.h"
#include "../GroundTruth/GroundTruth.h"

class GroundTruthWidget : public QGLWidget
{
public:
  GroundTruthWidget(RobotConsole& console, const std::string& name) : QGLWidget(QGLFormat(/*QGL::SampleBuffers*/)),
    console(console),
    name(name),
    imageData(0), imageWidth(0), imageHeight(0),
    lastImageTimeStamp(0), lastDrawingsTimeStamp(0),
    widgetWidth(0), widgetHeight(0) {}

  ~GroundTruthWidget()
  {
    if(imageData)
      delete[] imageData;
  }
  
  void update()
  {
    if(needsRepaint())
      QGLWidget::update();
  }

private:

  void resizeGL(int width, int height)
  {
    widgetWidth = width;
    widgetHeight = height;
  }

  void paintEvent(QPaintEvent *event)
  {
     
    painter.begin(this);
    
    {
      int width, height;

      SYNC_WITH(console);

      const unsigned char* image = ((GroundTruth&) console).image.data;
      if(image)
      {
        width = (int) ((GroundTruth&) console).image.width;
        height = (int) ((GroundTruth&) console).image.height;
      }
      else
      {
        width = cameraResolutionWidth; // @todo
        height = cameraResolutionHeight; // @todo
      }

      float scale, xOffset, yOffset;
      {
        float xScale = float(widgetWidth)/width;
        float yScale = float(widgetHeight)/height;
        scale = xScale < yScale ? xScale : yScale;
        xOffset = xScale > scale ? (widgetWidth - width * scale) / 2 : 0;
        yOffset = yScale > scale ? (widgetHeight - height * scale) / 2 : 0;
      }

      if(image)
        paintImage(image, xOffset, yOffset, scale);
      else
        lastImageTimeStamp = 0;

      painter.setTransform(QTransform(scale, 0, 0, scale, xOffset, yOffset));
      paintDrawings();
    }

    painter.end();
  }

  void paintDrawings()
  {
    const QTransform baseTrans(painter.transform());
    const std::list<std::string>& drawings(console.imageViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.imageDrawings[*i]);
      PaintMethods::paintDebugDrawing(painter, debugDrawing, baseTrans);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        lastDrawingsTimeStamp = debugDrawing.timeStamp;
    }
    painter.setTransform(baseTrans);
  }

  void copyImage(const unsigned char* srcImage)
  {
    int width = (int) ((GroundTruth&) console).image.width;
    int height = (int) ((GroundTruth&) console).image.height;
    char* p = imageData;
    for(int y = height - 1; y >= 0; y--) 
    {
      memcpy(p, srcImage + y * width * 3, width * 3);
      p += width * 3;
    }
  }

  void copyImageSegmented(const unsigned char* srcImage)
  {
    int width = (int) ((GroundTruth&) console).image.width;
    int height = (int) ((GroundTruth&) console).image.height;
    static const int colorClassColors[] = {
      0x7f7f7f, // unknown
      0x7f7f7f, // none
      0xffff00, // yellow
      0x0000ff, // blue
      0xff7f00, // orange
      0xffffff, // white
      0x00ff00, // green
      0xff00ff // magenta
    };
    char* p = imageData;
    RGBLUT64& lut = *((GroundTruth&) console).image.lut;
    for(int y = height - 1; y >= 0; y--) 
    {
      const unsigned char* cur = srcImage + y * width * 3;
      const unsigned char* end = cur + width * 3;
      for(; cur < end; cur += 3)
      {
        unsigned char r = cur[0];
        unsigned char g = cur[1];
        unsigned char b = cur[2];
        Color_Class clrclass = Color_Class(lut.colorClass[r >> 2][g >> 2][b >> 2]);
        if(clrclass == COLOR_CLASS_NONE || clrclass > 7)
        {
          *(p++) = r;
          *(p++) = g;
          *(p++) = b;
        }
        else
        {
          int color = colorClassColors[clrclass];
          *(p++) = color >> 16;
          *(p++) = color >> 8;
          *(p++) = color;
        }
      }
    }
  }

  void paintImage(const unsigned char* srcImage, float x, float y, float scale)
  {
    int width = (int) ((GroundTruth&) console).image.width;
    int height = (int) ((GroundTruth&) console).image.height;

    // make sure we have a buffer
    if(imageWidth != width || imageHeight != height)
    {
      if(imageData)
        delete[] imageData;
      imageWidth = width;
      imageHeight = height;
      imageData = new char[imageWidth * imageHeight * 3];
    }

    if(((GroundTruth&) console).image.timeStamp != lastImageTimeStamp)
    {
      if(((GroundTruth&) console).image.lut)
        copyImageSegmented(srcImage);
      else
        copyImage(srcImage);
      lastImageTimeStamp = ((GroundTruth&) console).image.timeStamp;
    }

    glRasterPos2f(x, widgetHeight - y - 1); 
    glPixelZoom(scale, scale);
    glDrawPixels(width, height, GL_RGB, GL_UNSIGNED_BYTE, imageData);
  }

  bool needsRepaint() const
  {
    SYNC_WITH(console);
    const std::list<std::string>& drawings(console.imageViews[name]);
    for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
    {
      const DebugDrawing& debugDrawing(console.imageDrawings[*i]);
      if(debugDrawing.timeStamp > lastDrawingsTimeStamp)
        return true;
    }
    return ((GroundTruth&) console).image.timeStamp != lastImageTimeStamp;
  }

  void window2viewport(QPoint& point)
  {
    float scale, xOffset, yOffset;
    {
      float xScale = float(widgetWidth)/imageWidth;
      float yScale = float(widgetHeight)/imageHeight;
      scale = xScale < yScale ? xScale : yScale;
      xOffset = xScale > scale ? (widgetWidth - imageWidth * scale) / 2 : 0;
      yOffset = yScale > scale ? (widgetHeight - imageHeight * scale) / 2 : 0;
    }
    point = QPoint((point.x() - xOffset) / scale, (point.y() - yOffset) / scale);
  }

  void mousePressEvent(QMouseEvent* event)
  {
    QGLWidget::mousePressEvent(event);
    
    if(((GroundTruth&) console).image.data &&
       (event->button() == Qt::LeftButton || event->button() == Qt::MidButton))
    {
      QPoint pos = event->pos();
      window2viewport(pos);
      selectedPoint = ((GroundTruth&) console).transformation.getPointAt(pos.x(), pos.y());
      QGLWidget::update();
    }
  }

  void mouseMoveEvent(QMouseEvent* event)
  {
    QGLWidget::mouseMoveEvent(event);
    QGLWidget::mousePressEvent(event);
    
    if(((GroundTruth&) console).image.data)
    {
      QPoint pos = event->pos();
      window2viewport(pos);
      if(selectedPoint != "")
      {
        {
          SYNC_WITH(console);
          ((GroundTruth&) console).transformation.setPointTo(selectedPoint, pos.x(), pos.y());
        }
        QGLWidget::update();
      }
      else
      {
        const char* text = 0;
        const std::list<std::string>& drawings(console.imageViews[name]);
        for(std::list<std::string>::const_iterator i = drawings.begin(), end = drawings.end(); i != end; ++i)
        {
          text = console.imageDrawings[*i].getTip(pos.rx(), pos.ry());
          if(text)
            break;
        }
        if(text)
          setToolTip(QString(text));
        else
          setToolTip(QString());
      }
    }
  }

  void mouseReleaseEvent(QMouseEvent* event)
  {
    QGLWidget::mouseReleaseEvent(event);
    
    if(((GroundTruth&) console).image.data && selectedPoint != "")
    {
      QPoint pos = event->pos();
      window2viewport(pos);
      {
        SYNC_WITH(console);
        if(selectedPoint != "")
          ((GroundTruth&) console).transformation.setPointTo(selectedPoint, pos.x(), pos.y());
        else
          ((GroundTruth&) console).rleCompression.printPixelValue(pos.x(), pos.y());
      }
      selectedPoint = "";
    }

    QGLWidget::update();
  }

  RobotConsole& console;
  const std::string& name;
  char* imageData;
  int imageWidth;
  int imageHeight;
  unsigned int lastImageTimeStamp;
  unsigned int lastDrawingsTimeStamp;
  int widgetWidth;
  int widgetHeight;
  QPainter painter;
  std::string selectedPoint;
};

GroundTruthView::GroundTruthView(RobotConsole& console, const std::string& name, bool) :
  console(console),
  name(name) {}

QWidget* GroundTruthView::createWidget()
{
  return new GroundTruthWidget(console, name);
}

void GroundTruthView::updateWidget(QWidget& widget)
{
  ((GroundTruthWidget*)(&widget))->update();
}
