/**
* @file ControllerQt/ModuleView.h
* Declaration of a class to represent a view displaying the module layout of the process.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef ModuleView_H
#define ModuleView_H

#include <SimRobotCore/Controller/DirectView.h>

class RobotConsole;
class QPixmap;

/**
* @class ModuleView 
* A class to represent a view displaying the module layout of the process.
*/
class ModuleView : public DirectView
{
public:
  /**
  * Constructor.
  * @param console The console object.
  * @param processIdentifier The identifier of the process the modules of which are displayed.
  * @param group Will the modules be grouped?
  */
  ModuleView(RobotConsole& console, char processIdentifier, bool group);

  /**
  * Destructor.
  */
  ~ModuleView();

private:
  RobotConsole& console; /**< A reference to the console object. */
  char processIdentifier; /**< The name of the view. */
  bool group; /**< Will the modules be grouped? */
  QPixmap* image; /**< The module graph image. */
  unsigned int lastModulInfoTimeStamp; /**< Module Info timestamp when the image was created. */

  /**
  * The method returns a new instance of a widget for this direct view.
  * The caller has to delete this instance. (Qt handles this)
  * @return The widget.
  */
  QWidget* createWidget();

  /**
  * The function is called to let the direct view update the widget. Do so if necessary.
  * @param widget The widget
  */
  void updateWidget(QWidget& widget);

  /**
  * The method creates the dot file of the module graph.
  * @param fileName The name of the file created.
  * @return Were any nodes generated?
  */
  bool generateDotFile(const char* fileName);

  /**
  * The method replaces all ' ' by '_'.
  * @param s The input string.
  * @return The string in which spaces were replaced.
  */
  std::string compress(const std::string& s) const;

  /**
  * The method creates the image of the module graph.
  */
  void generateImage();

  /**
  * The function is called to return a string that contains the configuration of the widget/view.
  * @param widget The widget
  * @return A text string containing all information to restore the internal state of the widget/view.
  */
  virtual std::string getConfiguration(QWidget& widget) const;

  /**
  * The function is called to restore the configuration of the widget/view.
  * @param widget The widget
  * @param A text string containing all information to restore the internal state of the widget/view.
  */
  virtual void setConfiguration(QWidget& widget, const std::string& configuration);

#if 0
private:
  RobotConsole& console; /**< A reference to the console object. */
  char processIdentifier; /**< The name of the view. */
  unsigned timeStamp; /**< The time stamp of the last module info that was displayed. */
  Vector2<int> origin, /**< The upper left corner of the image. */
               startPos; /**< The start position of dragging operations. */
  Gdiplus::Image* image; /**< The image showing the module graph. */
  char pngName[_MAX_PATH]; /**< The name of the png file used to create the graph. */
  bool group; /**< Will the modules be grouped? */




  /**
  * The function handles a mouse event.
  * @param event The event.
  * @param x The x coordinate of the cursor when the event occured.
  * @param y The y coordinate of the cursor when the event occured.
  * @param keys The set of keys pressed when the event occured.
  * @return Does the view need to be repainted?
  */
  virtual bool handleMouseEvent(MouseEvent event, int x, int y, int keys);


#endif
public:

#if 0
  /**
  * Destructor.
  */
  ~ModuleView();

  /**
  * The function is called to draw the view.
  * @param dc The device context to draw to.
  */
  void draw(CDC& dc);

  /** Needs the view to be repainted?
  * @return Yes or no? 
  */
  virtual bool needsRepaint() const {return timeStamp != console.moduleInfo.timeStamp;}

  /**
  * The function is called to return a string that contains the configuration of the view.
  * @return A text string containing all information to restore the internal state of the view.
  */
  virtual std::string getConfiguration() const;

  /**
  * The function is called to restore the configuration of the view.
  * @param configuration A text string containing all information to restore the internal state of the view.
  */
  virtual void setConfiguration(const std::string& configuration);
#endif
};

#endif //ModuleView_H
