/**
* @file ControllerQt/Views/PlotView.cpp
*
* Implementation of class PlotView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Colin Graf
*/

#include <QGLWidget>
#include <QPainter>

#include <Tools/Boundary.h>

#include "PlotView.h"
#include "../RobotConsole.h"

class PlotWidget : public QGLWidget
{
public:
  PlotWidget(PlotView& plotView) : QGLWidget(QGLFormat(/*QGL::SampleBuffers*/)),
    plotView(plotView) 
  {
    setBackgroundRole(QPalette::Base);
  }

private:

  void paintEvent(QPaintEvent *event)
  {     
    painter.begin(this);
    plotView.paintAll(painter);
    painter.end();
  }

  QPainter painter; /**< The painter used for painting the plot. */
  PlotView& plotView;
};


PlotView::PlotView(RobotConsole& console, const std::string& name, unsigned int plotSize, double minValue, double maxValue) :
  console(console), name(name), plotSize(plotSize), minValue(minValue), maxValue(maxValue), lastTimeStamp(0)
{
  points = new QPoint[plotSize];
  for(unsigned int i = 0; i < plotSize; ++i)
    points[i].rx() = i;
}

PlotView::~PlotView()
{
  delete [] points;
}

QWidget* PlotView::createWidget()
{
  return new PlotWidget(*this);
}

void PlotView::updateWidget(QWidget& widget)
{
  if(needsRepaint())
    ((PlotWidget*)(&widget))->update();
}

void PlotView::paintAll(QPainter& painter)
{
  SYNC_WITH(console);
  Boundary<int> boundingBox(Range<int>(0, plotSize), Range<int>((int) minValue, (int) maxValue));
  if(!boundingBox.isEmpty())
  {
    const QSize& size(painter.window().size());
    QTransform transform;
    transform.translate(0, size.height());
    transform.scale(1.0, -1.0);
    painter.setTransform(transform);
    Boundary<int> originalBoundingBox(boundingBox);
    bool axes = calcDrawingParameters(painter, 1, boundingBox);

    if(axes)
      drawCoordinateAxes(painter, originalBoundingBox, 1, true, false);
    
    QTransform yay;
    double yScale = double(innerSize.height()) / originalBoundingBox.y.getSize();
    yay.translate(border.left(), border.bottom() + yScale * originalBoundingBox.y.max);
    yay.scale(double(innerSize.width()) / originalBoundingBox.x.getSize(), -yScale);
    painter.setTransform(yay);
    
    paintPlots(painter);
  }
}

bool PlotView::calcDrawingParameters(QPainter& painter, int divisor, Boundary<int>& boundingBox)
{
  QSize size(painter.window().size());
  char buf1[20], buf2[20], buf3[20], buf4[20];
  sprintf(buf1, "%d ", boundingBox.y.min / divisor);
  sprintf(buf2, "%d ", boundingBox.y.max / divisor);
  sprintf(buf3, "%d", boundingBox.x.min / divisor);
  sprintf(buf4, "%d", boundingBox.x.max / divisor);

  const QFontMetrics& fontMetrics(painter.fontMetrics());
  const QSize& buf1Size(fontMetrics.size(Qt::TextSingleLine, buf1));
  const QSize& buf2Size(fontMetrics.size(Qt::TextSingleLine, buf2));
  const QSize& buf3Size(fontMetrics.size(Qt::TextSingleLine, buf3));
  const QSize& buf4Size(fontMetrics.size(Qt::TextSingleLine, buf4));

  border.setCoords(std::max<>(std::max<>(buf1Size.width(), buf2Size.width()), buf3Size.width() / 2),
    buf1Size.height() + 5, buf4Size.width() / 2, buf1Size.height() / 2);
  numberSize.rwidth() = std::max<>(buf3Size.width(), buf4Size.width());
  numberSize.rheight() = buf1Size.height();

  innerSize.rwidth() = size.width() - border.left() - border.right();
  innerSize.rheight() = size.height() - border.top() - border.bottom();
  
  if(innerSize.width() < 20 || innerSize.height() < 20)
    return false; // view is too small

  usedSize = innerSize;

  offset.rx() = (innerSize.width() - usedSize.width()) / 2,
  offset.ry() = (innerSize.height() - usedSize.height()) / 2;

  updateBoundingBox(usedSize, border, boundingBox);

  return true;
}

void PlotView::drawCoordinateAxes(QPainter& painter, Boundary<int>& box, int divisor, bool grid, bool square)
{
  QRect outerBorder(offset, offset);
  updateBoundingBox(usedSize, outerBorder, box);

  int yNum = innerSize.height() / numberSize.height(),
      xNum = innerSize.width() / numberSize.width();
  if(!xNum || !yNum)
    return;
  double stepsY = box.y.getSize() / yNum;
  stepsY = std::max<>((double) divisor, pow(10.0, ceil(log10(stepsY))));
  double stepsX = box.x.getSize() / xNum;
  stepsX = std::max<>((double) divisor, pow(10.0, ceil(log10(stepsX))));

  double steps;
  if(square)
    steps = std::max<>(stepsX, stepsY);
  else
    steps = stepsY;
  int yMin = int(ceil(box.y.min / steps) * steps),
      yMax = int(floor(box.y.max / steps) * steps);
  for(double y = yMin; y <= yMax; y += steps)
    printY(painter, border.left(), 
               border.top() + int((y - box.y.min) * innerSize.height() / box.y.getSize()), 
               grid ? border.left() + innerSize.width() : 0, int(y / divisor));

  if(!square)
    steps = stepsX;
  int xMin = int(ceil(box.x.min / steps) * steps),
      xMax = int(floor(box.x.max / steps) * steps);
  for(double x = xMin; x <= xMax; x += steps)
    printX(painter, border.left() + int((x - box.x.min) * innerSize.width() / box.x.getSize()),
    border.top(), grid ? border.top() + innerSize.height() : 0, int(x / divisor));

  QPoint points[5] =
  {
    QPoint(border.left(), border.top()),
    QPoint(border.left(), border.top() + usedSize.height() + 2 * offset.y()),
    QPoint(border.left() + usedSize.width() + 2 * offset.x(), border.top() + usedSize.height() + 2 * offset.y()),
    QPoint(border.left() + usedSize.width() + 2 * offset.x(), border.top()),
    QPoint(border.left(), border.top()),
  };
  painter.drawPolyline(points, 5);
}

void PlotView::updateBoundingBox(const QSize& usedSize, const QRect& border, Boundary<int>& box)
{
  double xSize = box.x.getSize(),
         ySize = box.y.getSize();
  // note that xSize and ySize are doubles, so an intermediate overflow is avoided 
  box.x.min -= int(xSize * border.left() / usedSize.width());
  box.y.min -= int(ySize * border.top() / usedSize.height());
  box.x.max += int(xSize * border.right() / usedSize.width());
  box.y.max += int(ySize * border.bottom() / usedSize.height());
}

void PlotView::printY(QPainter& painter, int x, int y, int x2, int num)
{
  char buf[20];
  sprintf(buf, "%d ", num);
  const QFontMetrics& fontMetrics(painter.fontMetrics());
  const QSize& size(fontMetrics.size(Qt::TextSingleLine, buf));
  QTransform oldTransform(painter.transform());
  painter.resetTransform();
  painter.drawText(QPoint(x - size.width(), painter.window().size().height() + border.bottom() + 4 - (y + size.height() / 2)), buf);
  painter.setTransform(oldTransform);
  painter.drawLine(x - fontMetrics.width(' ') + 1, y, x, y);
  if(x2)
  {
    QPen oldPen(painter.pen());
    painter.setPen(0xc0c0c0);
    painter.drawLine(x, y, x2, y);
    painter.setPen(oldPen);
  }
}

void PlotView::printX(QPainter& painter, int x, int y, int y2, int num)
{
  char buf[20];
  sprintf(buf, "%d", num);
  const QFontMetrics& fontMetrics(painter.fontMetrics());
  const QSize& size(fontMetrics.size(Qt::TextSingleLine, buf));
  int dist = fontMetrics.width(' ');
  QTransform oldTransform(painter.transform());
  painter.resetTransform();
  painter.drawText(QPoint(x - size.width() / 2, painter.window().size().height() + border.bottom() + 4 - (y - dist)), buf);
  painter.setTransform(oldTransform);
  painter.drawLine(x, y - dist + 1, x, y);
  if(y2)
  {
    QPen oldPen(painter.pen());
    painter.setPen(0xc0c0c0);
    painter.drawLine(x, y, x, y2);
    painter.setPen(oldPen);
  }
}

void PlotView::paintPlots(QPainter& painter)
{
  for(std::list<RobotConsole::Layer>::const_iterator i = console.plotViews[name].begin(), end = console.plotViews[name].end(); 
      i != end; ++i)
  {
    const std::list<double>& list = console.plots[i->layer].points;
    int numOfPoints = std::min<>(list.size(), plotSize);
    if(numOfPoints > 1)
    {
      std::list<double>::const_reverse_iterator k = list.rbegin();
      for(int j = plotSize - 1, minj = int(plotSize - numOfPoints); j >= minj; --j)
        points[j].ry() = int(*k++);
      ColorRGBA color(i->color);
      painter.setPen(QColor(color.r, color.g, color.b));
      painter.drawPolyline(points + plotSize - numOfPoints, numOfPoints);
    }
    unsigned int timeStamp = console.plots[i->layer].timeStamp;
    if(timeStamp > lastTimeStamp)
      lastTimeStamp = timeStamp;
  }
}

bool PlotView::needsRepaint() const
{
  SYNC_WITH(console);
  for(std::list<RobotConsole::Layer>::const_iterator i = console.plotViews[name].begin(), end = console.plotViews[name].end(); 
      i != end; ++i)
    if(console.plots[i->layer].timeStamp > lastTimeStamp)
      return true;
  return false;
}

