/**
* @file ControllerQt/Views/PlotView.h
*
* Declaration of class PlotView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Colin Graf
*/

#ifndef PlotView_H
#define PlotView_H

#include <QSize>
#include <QRect>
#include <QPoint>
#include <QPainter>

#include <SimRobotCore/Controller/DirectView.h>

class RobotConsole;
class QWidget;
template<class T> class Boundary;

/**
* @class PlotView 
* A class to represent a map with drawings sent by the robot code.
*/
class PlotView : public DirectView
{
public:
  /**
  * Constructor.
  * @param console The console object.
  * @param name The name of the view.
  * @param plotSize The number of entries in a plot.
  * @param minValue The minimum value in a plot.
  * @param maxValue The maximum value in a plot.
  */
  PlotView(RobotConsole& console, const std::string& name, 
           unsigned int plotSize, double minValue, double maxValue);

  /**
  * Destructor.
  */
  ~PlotView();

  /**
  * Paints the current plot on a QPainter.
  * @param painter The painter to draw to.
  */
  void paintAll(QPainter& painter);

private:
  RobotConsole& console; /**< A reference to the console object. */
  std::string name; /**< The name of the view. */
  unsigned int plotSize; /**< The number of entries in a plot. */
  double minValue, /**< The minimum value in a plot. */
         maxValue; /**< The maximum value in a plot. */
  QSize innerSize, /**< The size of the coordinate box. */
        usedSize, /**< The size actually used by the drawings. */
        numberSize; /**< The maximum size of the numbers. */
  QRect border; /**< The borders required for filling in the numbers. */
  QPoint offset; /**< The offset at which the isotropic drawing begins. */
  QPoint* points; /**< A buffer for drawing points. */
  unsigned int lastTimeStamp; /**< Timestamp of the last plot drawing. */

  /**
  * The method returns a new instance of a widget for this direct view.
  * The caller has to delete this instance. (Qt handles this)
  * @return The widget.
  */
  QWidget* createWidget();

  /**
  * The function is called to let the direct view update the widget. Do so if necessary.
  * @param widget The widget
  */
  void updateWidget(QWidget& widget);

  /**
  * The function calculates several parameters for drawing the coordinate axes.
  * @param painter The painter to draw to.
  * @param divisor A divisor the numbers at the coordinate axes are divided by.
  * @param boundingBox The bounding box of the drawing area.
  * @return Returns false if there is not enough space to draw the axes.
  */
  bool calcDrawingParameters(QPainter& painter, int divisor, Boundary<int>& boundingBox);

  /**
  * The function draws the coordinate axes and a grid.
  * @param painter The device context to draw to.
  * @param box The original bounding box for the drawings.
  * @param divisor A divisor the numbers at the coordinate axes are divided by.
  * @param grid Decides whether to draw a grid.
  * @param square Grid cells are squares.
  */
  void drawCoordinateAxes(QPainter& painter, Boundary<int>& box, int divisor, bool grid, bool square);

  /**
  * The function updates the bounding box so that it includes the border.
  * @param usedSize Will return the size actually used by the drawings.
  * @param border The borders required for filling in the numbers.
  * @param box The initial bounding box in logical coordinates. It will be enlarged.
  */
  void updateBoundingBox(const QSize& usedSize, const QRect& border, Boundary<int>& box);

  /**
  * The function write a number at the y axis and optionally draws a single line of the grid.
  * @param painter The painter to draw to.
  * @param x The x coordinate to write at in device coordinates.
  * @param y The y coordinate to write at in device coordinates.
  * @param x2 The end of the grid line. If 0 no line is drawn.
  * @param num The number to write.
  */
  void printY(QPainter& painter, int x, int y, int x2, int num);

  /**
  * The function write a number at the x axis and optionally draws a single line of the grid.
  * @param painter The painter to draw to.
  * @param x The x coordinate to write at in device coordinates.
  * @param y The y coordinate to write at in device coordinates.
  * @param y2 The end of the grid line. If 0 no line is drawn.
  * @param num The number to write.
  */
  void printX(QPainter& painter, int x, int y, int y2, int num);

  /**
  * The methods paints all plots to a graphics context.
  * @param painter The painter to draw to.
  */
  void paintPlots(QPainter& painter);

  /** Needs the view to be repainted?
  * @return Yes or no? 
  */
  virtual bool needsRepaint() const;

#if 0
private:
  CFont font; /**< The font for drawing the coordinate axes. */
  CSize innerSize, /**< The size of the coordinate box. */
        usedSize, /**< The size actually used by the drawings. */
        numberSize; /**< The maximum size of the numbers. */
  CRect border; /**< The borders required for filling in the numbers. */
  CPoint offset; /**< The offset at which the isotropic drawing begins. */
  unsigned plotSize; /**< The number of entries in a plot. */
  double minValue, /**< The minimum value in a plot. */
         maxValue; /**< The maximum value in a plot. */
  POINT* points; /**< A buffer for drawing points. */

private:
  class TimeStamp
  {
  public:
    unsigned timeStamp;
    TimeStamp() : timeStamp(0) {}
  };

  stdext::hash_map<std::string, TimeStamp> lastPlots; /**< The time stamps of the last plot drawn. */
  
  /**
  * The methods paints all plots to a graphics context.
  * @param dc The graphics context.
  */
  void paintPlots(CDC& dc);

  /**
  * The function draws the coordinate axes and a grid.
  * @param dc The device context to draw to.
  * @param box The original bounding box for the drawings.
  * @param divisor A divisor the numbers at the coordinate axes are divided by.
  * @param grid Decides whether to draw a grid.
  * @param square Grid cells are squares.
  */
  void drawCoordinateAxes(CDC& dc, Boundary<int>& box, int divisor, bool grid, bool square);

  /**
  * The function calculates several parameters for drawing the coordinate axes.
  * @param dc The device context that will drawn to.
  * @param divisor A divisor the numbers at the coordinate axes are divided by.
  * @param boundingBox The bounding box of the drawing area.
  * @return Returns false if there is not enough space to draw the axes.
  */
  bool calcDrawingParameters(CDC& dc, int divisor, Boundary<int>& boundingBox);

  /**
  * The function updates the bounding box so that it includes the border.
  * @param usedSize Will return the size actually used by the drawings.
  * @param border The borders required for filling in the numbers.
  * @param box The initial bounding box in logical coordinates. It will be enlarged.
  */
  void updateBoundingBox(const CSize& usedSize, const CRect& border, Boundary<int>& box);

  /**
  * The function write a number at the y axis and optionally draws a single line of the grid.
  * @param dc The device context to draw to.
  * @param x The x coordinate to write at in device coordinates.
  * @param y The y coordinate to write at in device coordinates.
  * @param x2 The end of the grid line. If 0 no line is drawn.
  * @param num The number to write.
  */
  void printY(CDC& dc, int x, int y, int x2, int num);

  /**
  * The function write a number at the x axis and optionally draws a single line of the grid.
  * @param dc The device context to draw to.
  * @param x The x coordinate to write at in device coordinates.
  * @param y The y coordinate to write at in device coordinates.
  * @param y2 The end of the grid line. If 0 no line is drawn.
  * @param num The number to write.
  */
  void printX(CDC& dc, int x, int y, int y2, int num);
#endif
#if 0
  /**
  * Destructor.
  */
  ~PlotView();

  /**
  * The function is called to draw the view.
  * @param dc The device context to draw to.
  */
  void draw(CDC& dc);

  /** Needs the view to be repainted?
  * @return Yes or no? 
  */
  virtual bool needsRepaint() const;
#endif
};

#endif //PlotView_H
