/**
* @file ControllerQt/Views/SensorView.cpp
*
* Implementation of class SensorView
*
* @author of the original sensorview <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
* @author Jeff
*/

#include <QPainter>
#include <QGLWidget>
#include "SensorView.h"

class SensorWidget : public QGLWidget
{
public:
  SensorWidget(RobotConsole &c, const JointData &jd, const SensorData &sd, QWidget* parent=0) : 
    QGLWidget((QWidget*)parent),
    console(c),
    jointData(jd),
    sensorData(sd)
  {
    setBackgroundRole(QPalette::Base);
  }

  enum Type
  {
    angle,
    distance,
    pressure,
    button,
    temperature,
    acceleration,
    voltage,
    angularSpeed,
    ratio,
    unknown,
    end
  };
  struct ViewInfo
  {
    int sensor; /**< The id of the sensor. */
    Type type;  /**< The type of measurements the sensor takes. */
    bool endOfSection; /**< Is this entry the last in a group of sensors? */
  };



  void paintEvent(QPaintEvent *event)
  {
    static const ViewInfo viewInfoNao[] =  {
      {JointData::headPan, angle, false},
      {JointData::headTilt, angle, true},

      {JointData::armLeft0, angle, false},
      {JointData::armLeft1, angle, false},
      {JointData::armLeft2, angle, false},
      {JointData::armLeft3, angle, true},

      {JointData::armRight0, angle, false},
      {JointData::armRight1, angle, false},
      {JointData::armRight2, angle, false},
      {JointData::armRight3, angle, true},

      {JointData::legLeft0, angle, false},
      {JointData::legLeft1, angle, false},
      {JointData::legLeft2, angle, false},
      {JointData::legLeft3, angle, false},
      {JointData::legLeft4, angle, false},
      {JointData::legLeft5, angle, true},

      {JointData::legRight0, angle, false},
      {JointData::legRight1, angle, false},
      {JointData::legRight2, angle, false},
      {JointData::legRight3, angle, false},
      {JointData::legRight4, angle, false},
      {JointData::legRight5, angle, false},
     
      {SensorData::gyroX,    angularSpeed, false},
      {SensorData::gyroY,    angularSpeed, true},

      {SensorData::accX,    acceleration, false},
      {SensorData::accY,    acceleration, false},
      {SensorData::accZ,    acceleration, true},

      {SensorData::voltage, ratio, true},

      {SensorData::fsrLFL,  pressure, false},
      {SensorData::fsrLFR,  pressure, false},
      {SensorData::fsrLBL,  pressure, false},
      {SensorData::fsrLBR,  pressure, false},
      {SensorData::fsrRFL,  pressure, false},
      {SensorData::fsrRFR,  pressure, false},
      {SensorData::fsrRBL,  pressure, false},
      {SensorData::fsrRBR,  pressure, true},

      {SensorData::us,  distance, false},
      {0, end, true}
    };
    const ViewInfo viewInfoBioloid[] =
    {
      {JointData::headPan, angle, false},
      {JointData::headTilt, angle, true},

      {JointData::armLeft0, angle, false},
      {JointData::armLeft1, angle, true},

      {JointData::armRight0, angle, false},
      {JointData::armRight1, angle, true},

      {JointData::legLeft0, angle, false},
      {JointData::legLeft1, angle, false},
      {JointData::legLeft2, angle, false},
      {JointData::legLeft3, angle, false},
      {JointData::legLeft4, angle, false},
      {JointData::legLeft5, angle, true},

      {JointData::legRight0, angle, false},
      {JointData::legRight1, angle, false},
      {JointData::legRight2, angle, false},
      {JointData::legRight3, angle, false},
      {JointData::legRight4, angle, false},
      {JointData::legRight5, angle, false},
     
      {SensorData::gyroX,    angularSpeed, false},
      {SensorData::gyroY,    angularSpeed, false},
      {SensorData::gyroZ,    angularSpeed, true},

      {SensorData::accX,    acceleration, false},
      {SensorData::accY,    acceleration, false},
      {SensorData::accZ,    acceleration, true},

      {SensorData::compassX,    unknown, false},
      {SensorData::compassY,    unknown, true},

      {SensorData::voltage, voltage, false},
      {0, end, true}
    };


    static const float colpos[] = {0.02f, 0.26f, 0.41f, 0.55f, 0.7f, 0.88f};
    static const int rowheight = 15;
    static const int spacerheight = 4;
    float valuespace = 0.14f;
    float widgetwidth = width();
    bool isBioloid = (jointData.angles[JointData::armLeft2] == JointData::off && sensorData.data[SensorData::fsrLBL] == SensorData::doesNotExist);
    int lowestLine = 0;
    int lowestSensorLine = 0;

    painter.begin(this);
    
    //draw header
    painter.drawText(colpos[0]*widgetwidth, rowheight , "Joint");
    painter.drawText(colpos[1]*widgetwidth, rowheight , "Value");
    painter.drawText(colpos[2]*widgetwidth, rowheight , "Load");
    painter.drawText(colpos[3]*widgetwidth, rowheight , "Temp");
    painter.drawText(colpos[4]*widgetwidth, rowheight , "Sensor");
    painter.drawText(colpos[5]*widgetwidth, rowheight , "Value");
    
    //draw spacer line below header
    painter.drawLine(0, rowheight + spacerheight, (colpos[5]+valuespace)*widgetwidth, rowheight + spacerheight);

    char val1[32], val2[32], val3[32];
    int y;
    int rowsensors=0, rowjoints=0;
    bool end = false;
    for(int i = 0; !end ; i++)
    {
      const ViewInfo &vi = isBioloid ? viewInfoBioloid[i] : viewInfoNao[i];
      switch(vi.type)
      {
        case SensorWidget::end:
          end = true;
          break;
        case SensorWidget::angle:
          if(jointData.angles[vi.sensor] != JointData::off)
          {
            sprintf(val1, "%.1f\xc2\xb0", toDegrees(jointData.angles[vi.sensor]));
            sprintf(val2, "%d", sensorData.loads[vi.sensor]);
            sprintf(val3, "%d", sensorData.temperatures[vi.sensor]);
          }
          else
          {
            sprintf(val1, "?");
            sprintf(val2, "?");
            sprintf(val3, "?");
            
          }
          break;
        case SensorWidget::distance:
          sprintf(val1, "%.1fmm", sensorData.data[vi.sensor]);
          break;
        case SensorWidget::pressure:
          if(sensorData.data[vi.sensor] == SensorData::doesNotExist)
            sprintf(val1, "?");
          else
            sprintf(val1, "%.1fg", sensorData.data[vi.sensor] * 1000);
          break;
        case SensorWidget::button:
          continue;
        case SensorWidget::temperature:
          continue;
        case SensorWidget::acceleration:
          sprintf(val1, "%.1fmg", sensorData.data[vi.sensor]*1000);
          break;
        case SensorWidget::voltage:
          sprintf(val1, "%.1fV", sensorData.data[vi.sensor]);
          break;
        case SensorWidget::angularSpeed:
          sprintf(val1, "%.1f\xc2\xb0/s", sensorData.data[vi.sensor]);
          break;
        case SensorWidget::ratio:
          sprintf(val1, "%.1f%%", sensorData.data[vi.sensor]*10);
          break;
        case SensorWidget::unknown:
          continue;
      }

     
      if(!end)
      {
        
        if(vi.type == SensorWidget::angle)
        {
          y = (rowjoints + 2) * rowheight + spacerheight;
          painter.drawText(colpos[0]*widgetwidth, y, JointData::getJointName((JointData::Joint)vi.sensor));
          painter.drawText(colpos[1]*widgetwidth, y, QString::fromUtf8(val1));
          painter.drawText(colpos[2]*widgetwidth, y, QString::fromUtf8(val2));
          painter.drawText(colpos[3]*widgetwidth, y, QString::fromUtf8(val3));
          if(vi.endOfSection)
            painter.drawLine(0, y+spacerheight, colpos[4]*widgetwidth-5, y+spacerheight);
          rowjoints++;

          if(y > lowestLine)
            lowestLine = y;
        }
        else
        {
          y = (rowsensors + 2) * rowheight + spacerheight;
          painter.drawText(colpos[4]*widgetwidth, y, SensorData::getSensorName((SensorData::Sensor)vi.sensor));
          painter.drawText(colpos[5]*widgetwidth, y, QString::fromUtf8(val1));
          if(vi.endOfSection)
            painter.drawLine(colpos[4]*widgetwidth-5, y+spacerheight, (colpos[5]+valuespace)*widgetwidth, y+spacerheight);
          rowsensors++;

          if(y > lowestLine)
            lowestLine = y;
          lowestSensorLine = y;
        }
      }
      
    }
    //draw vertical line between the two columns
    painter.drawLine(colpos[4]*widgetwidth-5, 0, colpos[4]*widgetwidth-5, lowestLine+spacerheight);
    painter.end();
  }

private:
  RobotConsole &console;
  const JointData &jointData;
  const SensorData &sensorData;
  QPainter painter;

};

QWidget* SensorView::createWidget()
{
  widget = new SensorWidget(console, jointData, sensorData);
  return widget;
}

void SensorView::updateWidget(QWidget& w)
{
  widget->update();
}

