/**
* @file ControllerQt/Views/XabslView.h
*
* Declaration of class XabslView
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef __XabslView_h_
#define __XabslView_h_

#include "SimRobotCore/Controller/DirectView.h"

class RobotConsole;
class XabslInfo;
class QGLWidget;

/**
* @class XabslView 
* 
* A class to represent a view with information about a Xabsl behavior.
*
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/
class XabslView : public DirectView
{
public:
  /**
  * Constructor.
  * @param console The console object.
  * @param info The Xabsl info object to be visualized.
  */
  XabslView(RobotConsole& console, const XabslInfo& info);

private:
  RobotConsole& console; /**< A reference to the console object. */
  const XabslInfo& info; /**< The Xabsl info structure. */
  QGLWidget* xabslWidget; /**< The Xabsl widget that displays the info. */
  unsigned int lastXabslInfoTimeStamp; /**< Timestamp of the last painted info. */

  /**
  * The method returns a new instance of a widget for this direct view.
  * The caller has to delete this instance. (Qt handles this)
  * @return The widget.
  */
  QWidget* createWidget();

  /**
  * The function is called to let the direct view update the widget. Do so if necessary.
  * @param widget The widget
  */
  void updateWidget(QWidget& widget);

  /**
  * The function is called to return a string that contains the configuration of the widget/view.
  * @param widget The widget
  * @return A text string containing all information to restore the internal state of the widget/view.
  */
  virtual std::string getConfiguration(QWidget& widget) const;

  /**
  * The function is called to restore the configuration of the widget/view.
  * @param widget The widget
  * @param A text string containing all information to restore the internal state of the widget/view.
  */
  virtual void setConfiguration(QWidget& widget, const std::string& configuration);


#if 0
  private:
    RobotConsole& console; /**< A reference to the console object. */
    const XabslInfo& info; /**< The Xabsl info structure. */
    static double leftOfColumnPercent[11]; /**< Ratios for the beginning of the columns. */
    CFont normalFont, /**< Normal font. */
          boldFont; /**< Bold font. */
    CDC* dc; /**< The current device context. */
    int line, /**< A line counter in pixels. */
        xpos, /**< Write x-position counter in pixels. */
        width, /**< The width of the view in pixels. */
        lineHeight; /**< The height of a line in pixels. */
    unsigned lastTimeStamp; /**< The time stamp of the last xabsl info displayed. */
  
    /**
    * The function prints a string to a certain column in the view.
    * @param s The string.
    * @param column A column index.
    * @param bold Should the string be printed in boldface?
    * @param offset An additional offset.
    * @param rightalign Adjustment it to the right side?.
    */    
    void print(const std::string& s, int column, bool bold = false, int offset = 0, bool rightalign = false);

    /**
    * The function sets the print position to the next line.
    * @param drawLine Should a line be drawn between the lines?
    */
    void newLine(bool drawLine = false);
#endif
  public:
#if 0
    /**
    * Destructor.
    */
    ~XabslView();

    /**
    * The function is called to draw the view.
    * @param dc The device context to draw to.
    */
    void draw(CDC& dc);

    /** Needs the view to be repainted?
    * @return Yes or no? 
    */
    virtual bool needsRepaint() const {return lastTimeStamp != info.timeStamp;}
#endif
};

#endif //__XabslView_h_
