/** 
 * @file Visualization/DebugDrawing3D.cpp
 * Implementation of class DebugDrawing3D.
 *
 * @author Philippe Schober
 */

#include "../RobotConsole.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include "DebugDrawing3D.h"
#include "Platform/GTAssert.h"
#include "Platform/SystemCall.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Configuration/RobotDimensions.h"


DebugDrawing3D::DebugDrawing3D()
{
  timeStamp = SystemCall::getCurrentSystemTime();
  drawn = false;
  scaleX  = scaleY  = scaleZ  = 1.0;
  rotateX = rotateY = rotateZ = 0.0;
  transX  = transY  = transZ  = 0.0;
}

const DebugDrawing3D& DebugDrawing3D::operator=(const DebugDrawing3D& other)
{
  reset();

  timeStamp = other.timeStamp;
  drawn = other.drawn;

  scaleX = other.scaleX;
  scaleY = other.scaleY;
  scaleZ = other.scaleZ;

  rotateX = other.rotateX;
  rotateY = other.rotateY;
  rotateZ = other.rotateZ;

  transX = other.transX;
  transY = other.transY;
  transZ = other.transZ;

  processIdentifier = other.processIdentifier;
  robotConsole  = other.robotConsole;

  return *this += other;
}

const DebugDrawing3D& DebugDrawing3D::operator+=(const DebugDrawing3D& other)
{
  lines.insert(lines.end(), other.lines.begin(), other.lines.end());
  dots.insert(dots.end(), other.dots.begin(), other.dots.end());
  polygons.insert(polygons.end(), other.polygons.begin(), other.polygons.end());
  quads.insert(quads.end(), other.quads.begin(), other.quads.end());
  return *this;
}

DebugDrawing3D::DebugDrawing3D(const DebugDrawing3D& other)
{
  *this = other;
}

DebugDrawing3D::DebugDrawing3D(const DebugDrawing3D* pDebugDrawing3D)
{
  *this = *pDebugDrawing3D;
}

DebugDrawing3D::~DebugDrawing3D()
{
}

void DebugDrawing3D::draw()
{
  SYNC_WITH(*robotConsole);

  // Convert mm to m.
  glScaled(0.001, 0.001, 0.001);

  // Custom scaling.
  glScaled(scaleX, scaleY, scaleZ);

  // Custom rotation.
  if (rotateX != 0) glRotated(toDegrees(rotateX), 1, 0, 0);
  if (rotateY != 0) glRotated(toDegrees(rotateY), 0, 1, 0);
  if (rotateZ != 0) glRotated(toDegrees(rotateZ), 0, 0, 1);

  // Custom translation.
  glTranslated(transX, transY, transZ);

  // Draw all quads.
  std::vector<Quad>::iterator q;
  for(q = quads.begin(); q != quads.end(); ++q)
  {
    glColor4d(q->color.red, q->color.green, q->color.blue, q->color.alpha);
    glBegin(GL_QUADS);
      glVertex3d(q->points[0].x, q->points[0].y, q->points[0].z);
      glVertex3d(q->points[1].x, q->points[1].y, q->points[1].z);
      glVertex3d(q->points[2].x, q->points[2].y, q->points[2].z);      
      glVertex3d(q->points[3].x, q->points[3].y, q->points[3].z);      
    glEnd();
  }

  // Draw all polygons/triangles.
  std::vector<Polygon>::iterator p;
  for(p = polygons.begin(); p != polygons.end(); ++p)
  {
    glColor4d(p->color.red, p->color.green, p->color.blue, p->color.alpha);
    glBegin(GL_TRIANGLES);
      glVertex3d(p->points[0].x, p->points[0].y, p->points[0].z);
      glVertex3d(p->points[1].x, p->points[1].y, p->points[1].z);
      glVertex3d(p->points[2].x, p->points[2].y, p->points[2].z);      
    glEnd();
  }

  // Draw all lines.
  // Get line width so we can restore it later on
  float lw[1];
  glGetFloatv(GL_LINE_WIDTH, lw);

  std::vector<Line>::iterator l;
  for(l = lines.begin(); l != lines.end(); ++l)
  {
    glLineWidth(l->width);
    glColor4d(l->color.red, l->color.green, l->color.blue, l->color.alpha);
    glBegin(GL_LINES);
      glVertex3d(l->points[0].x, l->points[0].y, l->points[0].z);
      glVertex3d(l->points[1].x, l->points[1].y, l->points[1].z);
    glEnd();    
  }
  // Restore line width
  glLineWidth(lw[0]);

  // Draw all points.
  // Get point size so we can restore it later on
  float ps[1];
  glGetFloatv(GL_POINT_SIZE, ps);

  std::vector<Dot>::iterator d;
  for(d = dots.begin(); d != dots.end(); ++d)
  {
    // Since each point may have a different size we can't handle all
    // points in a single glBegin(GL_POINTS). 
    // ( glPointSize is not allowed in a glBegin(...). )
    glPointSize(d->width);
    glColor4d(d->color.red, d->color.green, d->color.blue, d->color.alpha);
    glBegin(GL_POINTS);
      glVertex3d(d->point.x,d->point.y,d->point.z);
    glEnd();
  }
  // Restore point size
  glPointSize(ps[0]);  
}


void DebugDrawing3D::reset()
{
  timeStamp = SystemCall::getCurrentSystemTime();
  lines.clear();
  dots.clear();
  polygons.clear();
  quads.clear();
}

void DebugDrawing3D::quad(const Vector3<double>* points, float width, Color color)
{
  Quad element;
  element.points[0] = points[0];
  element.points[1] = points[1];
  element.points[2] = points[2];
  element.points[3] = points[3];
  element.color = color;
  element.width = width;

  quads.push_back(element);
}


void DebugDrawing3D::line
(
 double xStart, double yStart, double zStart,
 double xEnd,   double yEnd,   double zEnd,
 float width,   Color color
 )
{
  Line element;
  element.points[0] = Vector3<double>(xStart, yStart, zStart);
  element.points[1] = Vector3<double>(xEnd,   yEnd,   zEnd  );
  element.color = color;
  element.width = width;

  lines.push_back(element);
}

void DebugDrawing3D::line(Vector3<double> *points, float width, Color color)
{
  Line element;
  element.points[0] = points[0];
  element.points[1] = points[1];
  element.color = color;
  element.width = width;

  lines.push_back(element);
}

void DebugDrawing3D::line
(
 double xStart, 
 double yStart, 
 double zStart,
 double xEnd,
 double yEnd,
 double zEnd
 )
{
  line(xStart, yStart, zStart, xEnd, yEnd, zEnd, 1, Color(0.0, 0.0, 0.0, 1.0));
}

void DebugDrawing3D::polygon
(
 const Vector3<double>* points,
 float width,
 Color color
 )
{
  Polygon element;
  for(int i = 0; i < 3; i++)
    element.points[i] = points[i];
  element.width = width;
  element.color = color;

  polygons.push_back(element);
}

void DebugDrawing3D::dot
(
 Vector3<double> v,
 float  w,
 Color color
 )
{
  Dot element;
  element.point = v;
  element.color = color;
  element.width = w;

  dots.push_back(element);
}

bool DebugDrawing3D::addShapeFromQueue
(
 InMessage& message,
 Drawings3D::ShapeType shapeType,
 char identifier
 )
{          
  processIdentifier = identifier;

  switch ((Drawings3D::ShapeType)shapeType)
  {
  case Drawings3D::translate:
    {
      message.bin >> transX;
      message.bin >> transY;
      message.bin >> transZ;
    }
    break;
  case Drawings3D::scale:
    {
      message.bin >> scaleX;
      message.bin >> scaleY;
      message.bin >> scaleZ;
    }
    break;
  case Drawings3D::rotate:
    {
      message.bin >> rotateX;
      message.bin >> rotateY;
      message.bin >> rotateZ;
    }
    break;
  case Drawings3D::coordinates:
    {
      float width;
      double length;
      message.bin >> length;
      message.bin >> width;
      this->line(0,0,0, length,0,0, width, Color(1,0,0,1));
      this->line(0,0,0, 0,length,0, width, Color(0,1,0,1));
      this->line(0,0,0, 0,0,length, width, Color(0,0,1,1));
    }
    break;
  case Drawings3D::quad:
    {
      Vector3<double> points[4];
      Color c;
      message.bin >> points[0];
      message.bin >> points[1];
      message.bin >> points[2];
      message.bin >> points[3];
      message.bin >> c;
      this->quad(points, 1.0, c);
    }
    break;
  case Drawings3D::polygon:
    {
      Vector3<double> points[3];
      Color c;
      message.bin >> points[0];
      message.bin >> points[1];
      message.bin >> points[2];
      message.bin >> c;
      this->polygon(points, 1.0, c);
    }
    break;
  case Drawings3D::line:
    {
      Vector3<double> points[2];
      float width;
      Color c;
      message.bin >> points[0];
      message.bin >> points[1];
      message.bin >> width;
      message.bin >> c;
      this->line(points, width, c);
    }
    break;
  case Drawings3D::dot:
    {
      Vector3<double> v;
      float w;
      Color c;
      bool withLines;
      message.bin >> v;
      message.bin >> w;
      message.bin >> c;
      message.bin >> withLines;
      this->dot(v, w, c);

      if (withLines)
      {
        this->line(  0,  0,  0,   v.x,  0,  0, 1.0, c);
        this->line(  0,  0,  0,     0,v.y,  0, 1.0, c);
        this->line(  0,  0,  0,     0,  0,v.z, 1.0, c);
        this->line(v.x,  0,  0,   v.x,v.y,  0, 1.0, c);
        this->line(v.x,  0,  0,   v.x,  0,v.z, 1.0, c);
        this->line(  0,v.y,  0,     0,v.y,v.z, 1.0, c);
        this->line(v.x,v.y,  0,     0,v.y,  0, 1.0, c);
        this->line(v.x,v.y,  0,   v.x,v.y,v.z, 1.0, c);
        this->line(v.x,  0,v.z,   v.x,v.y,v.z, 1.0, c);
        this->line(v.x,  0,v.z,     0,  0,v.z, 1.0, c);
        this->line(  0,v.y,v.z,     0,  0,v.z, 1.0, c);
        this->line(  0,v.y,v.z,   v.x,v.y,v.z, 1.0, c);
      }

    }
    break;
  }
  return true;
}
