/** 
* @file BH2009BallSymbols.cpp
*
* Implementation of class BallSymbols.
*
* @author Max Risler
* \author Colin Graf
*/

#include "BH2009BallSymbols.h"
#include "BH2009TeamSymbols.h"
#include "Tools/Debugging/Modify.h"


PROCESS_WIDE_STORAGE BH2009BallSymbols* BH2009BallSymbols::theInstance = 0;

void BH2009BallSymbols::registerSymbols(xabsl::Engine& engine)
{
  engine.registerDecimalInputSymbol("ball.position.field.x", &ballPositionField.x);
  engine.registerDecimalInputSymbol("ball.position.field.y", &ballPositionField.y);
  engine.registerDecimalInputSymbol("ball.position.robot.x", &ballPositionRel.x);
  engine.registerDecimalInputSymbol("ball.position.robot.y", &ballPositionRel.y);
  engine.registerDecimalInputSymbol("ball.x", &ballPositionRel.x);
  engine.registerDecimalInputSymbol("ball.y", &ballPositionRel.y);
  engine.registerDecimalInputSymbol("ball.distance", &staticGetBallDistance);
  engine.registerDecimalInputSymbol("ball.angle", &staticGetBallAngle);
  engine.registerBooleanInputSymbol("ball.was_seen", &ballWasSeen);
  engine.registerDecimalInputSymbol("ball.time_since_last_seen", &timeSinceBallWasSeen);

  engine.registerDecimalInputSymbol("ball.speed.field.x", &ballSpeedField.x);
  engine.registerDecimalInputSymbol("ball.speed.field.y", &ballSpeedField.y);
  engine.registerDecimalInputSymbol("ball.speed.robot.x", &ballSpeedRel.x);
  engine.registerDecimalInputSymbol("ball.speed.robot.y", &ballSpeedRel.y);

  engine.registerDecimalInputSymbol("ball.time_when_own_y_axis_reached", &staticGetTimeWhenBallReachesOwnYAxis);
  engine.registerDecimalInputSymbol("ball.position_when_ball_reaches_own_y_axis.y", &staticGetYPosWhenBallReachesOwnYAxis);

  engine.registerBooleanInputSymbol("ball.was_seen.by_team_mate", &staticGetBallWasSeenByTeamMate);
  engine.registerBooleanInputSymbolEnumeratedParameter("ball.was_seen.by_team_mate","ball.was_seen.by_team_mate.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.time_since_last_seen.by_team_mate", &staticGetBallTimeSinceLastSeenByTeamMate);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.time_since_last_seen.by_team_mate","ball.time_since_last_seen.by_team_mate.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.position.field.by_team_mate.x", &staticGetBallPositionFieldByTeamMateX);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.position.field.by_team_mate.x","ball.position.field.by_team_mate.x.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.position.field.by_team_mate.y", &staticGetBallPositionFieldByTeamMateY);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.position.field.by_team_mate.y","ball.position.field.by_team_mate.y.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.position.robot.by_team_mate.x", &staticGetBallPositionRobotByTeamMateX);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.position.robot.by_team_mate.x","ball.position.robot.by_team_mate.x.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.position.robot.by_team_mate.y", &staticGetBallPositionRobotByTeamMateY);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.position.robot.by_team_mate.y","ball.position.robot.by_team_mate.y.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.distance.by_team_mate", &staticGetBallDistanceByTeamMate);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.distance.by_team_mate","ball.distance.by_team_mate.player", "team.mate", &player);
  engine.registerDecimalInputSymbol("ball.angle.by_team_mate", &staticGetBallAngleByTeamMate);
  engine.registerDecimalInputSymbolEnumeratedParameter("ball.angle.by_team_mate","ball.angle.by_team_mate.player", "team.mate", &player);
}

void BH2009BallSymbols::update()
{
  timeSinceBallWasSeen = frameInfo.getTimeSince(theInstance->ballModel.timeWhenLastSeen);
  ballWasSeen = timeSinceBallWasSeen < 500; 

  const BallState& lastSeenEstimate = ballModel.lastSeenEstimate;
  ballPositionRel = lastSeenEstimate.getPositionInRelativeCoordinates();
  ballPositionField = lastSeenEstimate.getPositionInFieldCoordinates(robotPose);
  ballSpeedRel = lastSeenEstimate.getVelocityInRelativeCoordinates();
  ballSpeedField = lastSeenEstimate.getVelocityInFieldCoordinates(robotPose);
}


double BH2009BallSymbols::getBallFieldRobotX()
{
  return ballPositionField.x;
}

double BH2009BallSymbols::getBallFieldRobotY()
{
  return ballPositionField.y;
}

double BH2009BallSymbols::getBallPositionRobotX()
{
  return ballPositionRel.x;
}

double BH2009BallSymbols::getBallPositionRobotY()
{
  return ballPositionRel.y;
}

double BH2009BallSymbols::getBallAngle()
{
  return toDegrees(ballModel.lastSeenEstimate.getAngle());
}

double BH2009BallSymbols::getBallDistance()
{
  return ballModel.lastSeenEstimate.getDistance();
}

double BH2009BallSymbols::getTimeWhenBallReachesOwnYAxis()
{
  double result = 1000000;
  if(ballSpeedRel.x < 0 )
  {
    result = fabs(ballPositionRel.x) / fabs(ballSpeedRel.x);
  }
  
  return result;
}

double BH2009BallSymbols::getYPosWhenBallReachesOwnYAxis()
{  
  double timeWhenAxisIsReached = getTimeWhenBallReachesOwnYAxis();
  
  if(timeWhenAxisIsReached == 1000000)
  {
    return 0.0;
  }
  
  Vector2<double> finalBallPos = ballPositionRel + (ballSpeedRel * timeWhenAxisIsReached);
  
  MODIFY("behavior symbols:ball:finalBallPos", finalBallPos);
  
  return finalBallPos.y;
}

bool BH2009BallSymbols::getBallWasSeenByTeamMate()
{
  if(player == theInstance->robotInfo.number)
    return ballWasSeen;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
    return frameInfo.getTimeSince(teamMateData.ballModels[player].timeWhenLastSeen) < 500;
  return false;
}

double BH2009BallSymbols::getBallTimeSinceLastSeenByTeamMate()
{
  if(player == robotInfo.number)
    return timeSinceBallWasSeen;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
    return frameInfo.getTimeSince(teamMateData.ballModels[player].timeWhenLastSeen);
  return 0;
}

double BH2009BallSymbols::getBallPositionFieldByTeamMateX()
{
  if(player == robotInfo.number)
    return ballPositionField.x;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionField = teamBallPositionsField[player];
    if(teamBallPositionsFieldTime[player] < frameInfo.time)
    {
      teamBallPositionsFieldTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionField = mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose);
    }
    return teamBallPositionField.x;
  }
  return 0;
}

double BH2009BallSymbols::getBallPositionFieldByTeamMateY()
{
  if(player == robotInfo.number)
    return ballPositionField.y;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionField = teamBallPositionsField[player];
    if(teamBallPositionsFieldTime[player] < frameInfo.time)
    {
      teamBallPositionsFieldTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionField = mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose);
    }
    return teamBallPositionField.y;
  }
  return 0;
}

double BH2009BallSymbols::getBallPositionRobotByTeamMateX()
{
  if(player == robotInfo.number)
    return ballPositionRel.x;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionRel = teamBallPositionsRel[player];
    if(teamBallPositionsRelTime[player] < frameInfo.time)
    {
      teamBallPositionsRelTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionRel = Geometry::fieldCoord2Relative(robotPose, mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose));
    }
    return teamBallPositionRel.x;
  }
  return 0;
}

double BH2009BallSymbols::getBallPositionRobotByTeamMateY()
{
  if(player == robotInfo.number)
    return ballPositionRel.y;
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionRel = teamBallPositionsRel[player];
    if(teamBallPositionsRelTime[player] < frameInfo.time)
    {
      teamBallPositionsRelTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionRel = Geometry::fieldCoord2Relative(robotPose, mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose));
    }
    return teamBallPositionRel.y;
  }
  return 0;
}

double BH2009BallSymbols::getBallDistanceByTeamMate()
{
  if(player == robotInfo.number)
    return getBallDistance();
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionRel = teamBallPositionsRel[player];
    if(teamBallPositionsRelTime[player] < frameInfo.time)
    {
      teamBallPositionsRelTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionRel = Geometry::fieldCoord2Relative(robotPose, mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose));
    }
    return teamBallPositionRel.abs();
  }
  return 0;
}

double BH2009BallSymbols::getBallAngleByTeamMate()
{
  if(player == robotInfo.number)
    return getBallAngle();
  if(player >= TeamMateData::firstPlayer && player < TeamMateData::numOfPlayers)
  {
    Vector2<double>& teamBallPositionRel = teamBallPositionsRel[player];
    if(teamBallPositionsRelTime[player] < frameInfo.time)
    {
      teamBallPositionsRelTime[player] = frameInfo.time;
      const RobotPose& matePose = teamMateData.robotPoses[player];
      const BallModel& mateBallModel = teamMateData.ballModels[player];
      teamBallPositionRel = Geometry::fieldCoord2Relative(robotPose, mateBallModel.lastSeenEstimate.getPositionInFieldCoordinates(matePose));
    }
    return toDegrees(teamBallPositionRel.angle());
  }
  return 0;
}
