/** 
* \file BH2009SoccerSymbols.cpp
* Implementation of symbols for our 4 roles.
* \author Colin Graf
* \author <a href="mailto:aschreck@informatik.uni-bremen.de">Andr&eacute; Schreck</a>
*/

#include <limits.h>
#include "Tools/Debugging/Modify.h"
#include "BH2009SoccerSymbols.h"
#include "BH2009TeamSymbols.h"
#include "BH2009LocatorSymbols.h"
#include "BH2009BallSymbols.h"
#include "Tools/Team.h"


PROCESS_WIDE_STORAGE BH2009SoccerSymbols* BH2009SoccerSymbols::theInstance = 0;

void BH2009SoccerSymbols::registerSymbols(xabsl::Engine& engine)
{
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.none", none); 
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.keeper", keeper);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.defender", defender);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.supporter", supporter);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.striker", striker);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.dribble", dribble);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.relative", relative);
  engine.registerEnumElement("soccer.walk_mode", "soccer.walk_mode.field", field);

  engine.registerEnumeratedOutputSymbol("soccer.walk_mode", "soccer.walk_mode", (int*)&walkMode);

  engine.registerDecimalOutputSymbol("soccer.walk_target.x", staticSetWalkTargetX, staticGetWalkTargetX);
  engine.registerDecimalOutputSymbol("soccer.walk_target.y", staticSetWalkTargetY, staticGetWalkTargetY);
  engine.registerDecimalOutputSymbol("soccer.walk_target.rot", staticSetWalkTargetRotation, staticGetWalkTargetRotation);

  engine.registerDecimalInputSymbol("soccer.opponent_goal.angle", &staticGetOpponentGoalAngle);

  engine.registerDecimalInputSymbol("soccer.keeper.block", &staticGetKeeperBlock);

  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.unknown", BehaviorData::unknown);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.dribble", BehaviorData::dribble);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.go_to_ball", BehaviorData::goToBall);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.go_to_target", BehaviorData::goToTarget);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.prepare_kick", BehaviorData::prepareKick);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.kick", BehaviorData::kick);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.block", BehaviorData::block);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.hold", BehaviorData::hold);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.stand_up", BehaviorData::standUp);
  engine.registerEnumElement("soccer.behavior_action", "soccer.behavior_action.patrol", BehaviorData::patrol);
  engine.registerEnumeratedOutputSymbol("soccer.behavior_action", "soccer.behavior_action", (int*)&behaviorData.action);

  engine.registerEnumeratedInputSymbol("soccer.nearest_to_ball_player", "team.mate", &staticGetNearestToBallPlayer);
  engine.registerEnumeratedInputSymbol("soccer.second_nearest_to_ball_player", "team.mate", &staticGetSecondNearestToBallPlayer);
  engine.registerEnumeratedInputSymbol("soccer.last_ball_seen_player", "team.mate", &staticGetLastBallSeenPlayer);
}

void BH2009SoccerSymbols::init()
{
}

void BH2009SoccerSymbols::update()
{
  MODIFY("parameters:BehaviorControl:SoccerSymbols", param);
  MODIFY_ENUM("module:BehaviorControl:SoccerSymbols:walkMode", walkMode, numOfWalkModes, getWalkModeName);

  updateWalkRequest();

  DECLARE_DEBUG_DRAWING("module:BehaviorControl:SoccerSymbols:walkTarget", "drawingOnField", "Draws the striker's target walk point on the field");
  COMPLEX_DRAWING("module:BehaviorControl:SoccerSymbols:walkTarget",
    ORIGIN("module:BehaviorControl:SoccerSymbols:walkTarget", robotPose.translation.x, robotPose.translation.y, robotPose.rotation);
    CROSS("module:BehaviorControl:SoccerSymbols:walkTarget", walkTarget.x, walkTarget.y, 80, 2, Drawings::ps_solid, ColorRGBA(0, 0, 128));
    CROSS("module:BehaviorControl:SoccerSymbols:walkTarget", walkTarget2.x, walkTarget2.y, 80, 2, Drawings::ps_solid, ColorRGBA(128, 0, 128, 128));
    CROSS("module:BehaviorControl:SoccerSymbols:walkTarget", walkSupport1.x, walkSupport1.y, 30, 2, Drawings::ps_solid, ColorRGBA(0, 0, 128));
    CROSS("module:BehaviorControl:SoccerSymbols:walkTarget", walkSupport2.x, walkSupport2.y, 30, 2, Drawings::ps_solid, ColorRGBA(0, 0, 128));
    Vector2<double> angle(1000.0, 0.0);
    angle.rotate(walkTargetRotation);
    ARROW("module:BehaviorControl:SoccerSymbols:walkTarget", 0.0, 0.0, angle.x, angle.y, 2, Drawings::ps_solid, ColorClasses::blue);
  );

  // update own ball hold situation
  if(behaviorData.role == BehaviorData::striker)
  {
    int timeSinceLastSeen(frameInfo.getTimeSince(ballModel.timeWhenLastSeen));
    if(timeSinceLastSeen < 500)
    {
      const BallState& lastSeenEstimate(ballModel.lastSeenEstimate);
      behaviorData.ballHold = lastSeenEstimate.getDistance() < 300.0 && 
        lastSeenEstimate.getPositionInRelativeCoordinates().x > 0.0 &&
        getOpponentGoalAngle() < 30;
    }
    else if(timeSinceLastSeen >= param.strikerBallHoldTimeout)
      behaviorData.ballHold = false;
  }
  else
    behaviorData.ballHold = false;
}

double BH2009SoccerSymbols::getOpponentGoalAngle()
{
  if(opponentGoalAngleTimeStamp == frameInfo.time)
    return toDegrees(opponentGoalAngle);
  if(frameInfo.getTimeSince(goalModel.timeWhenLastSeen) > param.strikerGoalModelTimeout)
   // use self locator for target calculation
    opponentGoalAngle = Geometry::angleTo(robotPose, Vector2<double>(fieldDimensions.xPosOpponentGoal, 0.0));
  else
    opponentGoalAngle = goalModel.largestFreePart.getCenter();
  return toDegrees(opponentGoalAngle);
}

void BH2009SoccerSymbols::setWalkTargetX(double value)
{
  walkTargetIn.x = value;
}

double BH2009SoccerSymbols::getWalkTargetX()
{
  return walkTarget.x;
}

void BH2009SoccerSymbols::setWalkTargetY(double value)
{
  walkTargetIn.y = value;
}

double BH2009SoccerSymbols::getWalkTargetY()
{
  return walkTarget.y;
}

void BH2009SoccerSymbols::setWalkTargetRotation(double value)
{
  walkTargetRotationIn = fromDegrees(value);
}

double BH2009SoccerSymbols::getWalkTargetRotation()
{
  return toDegrees(walkTargetRotation);
}

double BH2009SoccerSymbols::getKeeperBlock()
{
  const Vector2<double> ballPos = ballModel.lastSeenEstimate.getPositionInRelativeCoordinates();
  if(ballPos.x < 30.0)
    return 0.0; // ball is behind us or too near (we are to late!)

  const Vector2<double> ballVel = ballModel.lastSeenEstimate.getVelocityInRelativeCoordinates();
  if(ballVel.x >= 0.0)
    return 0.0; // ball is moving away
  const double timeTillBallArrives = fabs(ballPos.x / ballVel.x);
  const double ballYTarget = ballPos.y + ballVel.y * timeTillBallArrives;
  const double absBallYTarget = fabs(ballYTarget);
  if(absBallYTarget > 60.0 && absBallYTarget < 2000.0 && timeTillBallArrives < 3.0)
    return ballYTarget > 0.0 ? 1 :  -1;

  return 0;
}


double BH2009SoccerSymbols::getBallAvgPositionFieldX()
{
  double x = BH2009BallSymbols::staticGetBallFieldRobotX();
  int count = 1;
  const BallModel* ballModels = teamMateData.ballModels;
  const RobotPose* robotPoses = teamMateData.robotPoses;
  for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
    if(teamMateData.timeStamps[i])
      if(frameInfo.getTimeSince(ballModels[i].timeWhenLastSeen) <= 5500)
      {
        count++;
        const Vector2<double>& ballEstimate = ballModels[i].lastSeenEstimate.getPositionInFieldCoordinates(robotPoses[i]);
        x += ballEstimate.x;
      }
  x /= count;
  return x;
}

bool BH2009SoccerSymbols::getBallInOpponentArea()
{
  return getBallAvgPositionFieldX() > 0;
}

bool BH2009SoccerSymbols::getNearestToBall()
{
  calcNearestAndSecondNearestToBallPlayer();
  return nearestToBallPlayer == robotInfo.number;
}

bool BH2009SoccerSymbols::getSecondNearestToBall()
{
  calcNearestAndSecondNearestToBallPlayer();
  return secondNearestToBallPlayer == robotInfo.number;
}

int BH2009SoccerSymbols::getNearestToBallPlayer()
{
  calcNearestAndSecondNearestToBallPlayer();
  return nearestToBallPlayer;
}

int BH2009SoccerSymbols::getSecondNearestToBallPlayer()
{
  calcNearestAndSecondNearestToBallPlayer();
  return secondNearestToBallPlayer;
}

int BH2009SoccerSymbols::getLastBallSeenPlayer()
{
  if(lastBallSeenPlayerTimeStamp == frameInfo.time)
    return lastBallSeenPlayer;
  lastBallSeenPlayerTimeStamp = frameInfo.time;
  lastBallSeenPlayer = robotInfo.number;
  unsigned int minSeen = frameInfo.getTimeSince(ballModel.timeWhenLastSeen);
  for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
    if(teamMateData.timeStamps[i])
    {
      unsigned int curSeen = frameInfo.getTimeSince(teamMateData.ballModels[i].timeWhenLastSeen);
      if(curSeen < minSeen)
      {
        minSeen = curSeen;
        lastBallSeenPlayer = i;
      }
    }
  return lastBallSeenPlayer;
}

bool BH2009SoccerSymbols::getBallHold()
{
  if(behaviorData.ballHold)
    return true; 
  for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
    if(teamMateData.timeStamps[i] && teamMateData.behaviorData[i].ballHold)
      return true;
  return false;
}

void BH2009SoccerSymbols::calcNearestAndSecondNearestToBallPlayer()
{
  if(nearestToBallPlayerTimeStamp == frameInfo.time)
    return;
  nearestToBallPlayerTimeStamp = frameInfo.time;

  int timeToReachBall; // estimated in ms
  int timeToReachBallIncludingSnap; 
  int newNearestToBallPlayer = robotInfo.number;
  int newSecondNearestToBallPlayer = robotInfo.number;
  int minTimeToReachBall = INT_MAX; // infinite
  int minTimeToReachBallWithoutSnap = INT_MAX; // infinite
  int secondMinTimeToReachBall = INT_MAX; // infinite

  // get minimum timeToReachBall of whole team
  for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
    if(teamMateData.timeStamps[i] || i == robotInfo.number)
    {
      const BallModel& ballModel(i == robotInfo.number ? this->ballModel : teamMateData.ballModels[i]);
      const RobotPose& robotPose(i == robotInfo.number ? this->robotPose : teamMateData.robotPoses[i]);
      int timeSinceLastSeen = frameInfo.getTimeSince(ballModel.timeWhenLastSeen);

      if(ballModel.timeWhenLastSeen && timeSinceLastSeen < 12000)
      {
        const Vector2<double> ballPos = ballModel.lastSeenEstimate.getPositionInRelativeCoordinates();
        const Vector2<double> goalPos = Geometry::fieldCoord2Relative(robotPose, Vector2<double>(static_cast<double>(theInstance->fieldDimensions.xPosOpponentGoal), 0.0));

        // estimate time to reach ball
        timeToReachBall  = static_cast<int>(fabs(ballPos.abs()) * 1000 / param.averageWalkSpeed);
        timeToReachBall += static_cast<int>(fabs(ballPos.angle()) * 1000 / param.averageRotateSpeed);
        timeToReachBall += static_cast<int>(fabs(normalize(ballPos.angle() - goalPos.angle())) * 1000 / param.averageRotateSpeed);
        if(timeSinceLastSeen > 5500)
          timeToReachBall += (timeSinceLastSeen - 5500) * 2;

        // add "role snapping" bonus
        timeToReachBallIncludingSnap = timeToReachBall;
        if(nearestToBallPlayer == i)
          timeToReachBallIncludingSnap -= param.nearestToBallSnap;

        if(timeToReachBallIncludingSnap < minTimeToReachBall)
        {
          int lastNearestToBallPlayer = newNearestToBallPlayer;
          int lastMinTimeToReachBallWithoutSnap = minTimeToReachBallWithoutSnap;

          minTimeToReachBall = timeToReachBallIncludingSnap;
          newNearestToBallPlayer = i;
          minTimeToReachBallWithoutSnap = timeToReachBall;

          timeToReachBallIncludingSnap = lastMinTimeToReachBallWithoutSnap;
          if(secondNearestToBallPlayer == lastNearestToBallPlayer)
            timeToReachBallIncludingSnap -= param.nearestToBallSnap;
          if(timeToReachBallIncludingSnap < secondMinTimeToReachBall)
          {
            secondMinTimeToReachBall = timeToReachBallIncludingSnap;
            newSecondNearestToBallPlayer = lastNearestToBallPlayer;
          }
        }
        else
        {
          // add "role snapping" bonus
          timeToReachBallIncludingSnap = timeToReachBall;
          if(secondNearestToBallPlayer == i)
            timeToReachBallIncludingSnap -= param.nearestToBallSnap;

          if(timeToReachBallIncludingSnap < secondMinTimeToReachBall)
          {
            secondMinTimeToReachBall = timeToReachBallIncludingSnap;
            newSecondNearestToBallPlayer = i;
          }
        }
      }
    }

  nearestToBallPlayer = newNearestToBallPlayer;
  secondNearestToBallPlayer = newSecondNearestToBallPlayer;
}

void BH2009SoccerSymbols::updateWalkRequest()
{
  switch(walkMode)
  {
  case keeper:
    updateKeeperWalkRequest();
    break;
  case defender:
    updateDefenderWalkRequest();
    break;
  case supporter:
    updateSupporterWalkRequest();
    break;
  case striker:
    updateStrikerWalkRequest();
    break;
  case dribble:
    updateStrikerWalkRequest(true);
    break;
  case relative:
  {
    // relative to field 
    Vector2<double> target(walkTargetIn);
    target.rotate(-robotPose.rotation);
    target += robotPose.translation;      

    double rot(normalize(walkTargetRotationIn));

    generateWalkRequest(target, target, 
      Vector2<double>(0.0, fieldDimensions.yPosLeftFieldBorder), Vector2<double>(0.0, fieldDimensions.yPosRightFieldBorder),
      rot);
    break;
  }  
  case field:
  {
    // field to relative
    double rot(normalize(walkTargetRotationIn - robotPose.rotation));

    generateWalkRequest(walkTargetIn, walkTargetIn, 
      Vector2<double>(0.0, fieldDimensions.yPosLeftFieldBorder), Vector2<double>(0.0, fieldDimensions.yPosRightFieldBorder),
      rot);
    break;
  }
  default:
    break;
  }
}

void BH2009SoccerSymbols::updateStrikerWalkRequest(bool dribble)
{
  // just walk in ball direction when we are out of evade range
  Vector2<double> ballPos = ballModel.lastSeenEstimate.getPositionInRelativeCoordinates();
  if(ballPos.abs() >= param.strikerBallEvadeRange)
  {
    double rot(ballPos.angle());
    ballPos = ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose);
    generateWalkRequest(ballPos, ballPos, 
      Vector2<double>(0.0, fieldDimensions.yPosLeftSideline), Vector2<double>(0.0, fieldDimensions.yPosRightSideline),
      rot, false, true, false, false);
    return;
  }

  // some ball position and target faking for dribblig
  int strikerBallDistance = param.strikerBallDistance;
  if(dribble)
  {
    if(ballPos.x < 100)
      ballPos.x = 100;
    strikerBallDistance = 0;
  }

  // determine our target point
  Vector2<double> target;
  Vector2<double> offsetDir;
  if(frameInfo.getTimeSince(goalModel.timeWhenLastSeen) > param.strikerGoalModelTimeout)
  { // use self locator for target calculation
    target = Geometry::relative2FieldCoord(robotPose, ballPos.x, ballPos.y);

    // move "goal" in center of the penalty area when we can't play directly
    const int xToGoal = (fabs(target.y) > fieldDimensions.yPosLeftGoal &&
      target.x > fieldDimensions.xPosOpponentPenaltyArea) ?
        (fieldDimensions.xPosOpponentPenaltyArea + fieldDimensions.xPosOpponentGroundline) / 2 : 
          fieldDimensions.xPosOpponentGoal;

    target.x -= xToGoal;
    opponentGoalAngle = Geometry::angleTo(robotPose, Vector2<double>(xToGoal, 0.0));
    target.normalize(target.abs() + strikerBallDistance);
    if(param.strikerOffset)
    {
      offsetDir = target;
      offsetDir.rotateLeft();
      offsetDir.normalize(param.strikerOffset);
      offsetDir.rotate(-robotPose.getAngle());
    }
    target.x += xToGoal;
  }
  else 
  { // use goal locator for target calculation
    opponentGoalAngle = goalModel.largestFreePart.getCenter();
    target = ballPos;
    target -= Vector2<double>(static_cast<double>(strikerBallDistance), 0.0).rotate(opponentGoalAngle);
    
    if(param.strikerOffset)
      offsetDir = Vector2<double>(0.0, static_cast<double>(param.strikerOffset)).rotate(opponentGoalAngle);

    target = Geometry::relative2FieldCoord(robotPose, target.x, target.y);
  }

  // keep the calculated opponent goal angle for this frame (e.g. for soccer.opponent_goal.angle)
  opponentGoalAngleTimeStamp = frameInfo.time;

  Vector2<double> target1(target);
  Vector2<double> target2(target);
  
  double ballAngle(ballPos.angle());  
  double ballGoalAngle(fabs(normalize(ballAngle - opponentGoalAngle)));
  
  if(ballGoalAngle > fromDegrees(55))
  {
    const Vector2<double> ballPos(ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose));
    const Vector2<double> ballPosToTarget(target - ballPos);

    if(ballGoalAngle > fromDegrees(85))
    {
      target1 = Vector2<double>(ballPosToTarget).rotate(fromDegrees(35)).normalize(param.strikerBallEvadeDistance) + ballPos;
      target2 = Vector2<double>(ballPosToTarget).rotate(-fromDegrees(35)).normalize(param.strikerBallEvadeDistance) + ballPos;
    }
  }
  else if(param.strikerOffset)
  {
    target1 += offsetDir;
    target2 -= offsetDir;
  }

  double rot(ballGoalAngle < fromDegrees(45) ? opponentGoalAngle : ballAngle);
  generateWalkRequest(target1, target2, 
    target1, target2,
    rot, false, false, false, /*dribble ? false :*/ true);
}

void BH2009SoccerSymbols::updateSupporterWalkRequest()
{
  const Vector2<double> ballPosOnField(ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose));
  
  // somewhere near the striker
  Vector2<double> target1(getNearestToBall() ? ballPosOnField : teamMateData.robotPoses[getNearestToBallPlayer()].translation);
  target1.x += param.supporterXOffset;

  // keep supporter out of opponent penalty area
  if(target1.x > fieldDimensions.xPosOpponentPenaltyArea-30)
    target1.x = fieldDimensions.xPosOpponentPenaltyArea-30; 

  Vector2<double> target2(target1);
  target1.y += param.supporterYOffset;
  target2.y -= param.supporterYOffset;

  double rot(ballModel.lastSeenEstimate.getAngle());
  generateWalkRequest(target1, target2, 
    Vector2<double>(0.0, fieldDimensions.yPosLeftSideline), Vector2<double>(0.0, fieldDimensions.yPosRightSideline),
    rot);
}


void BH2009SoccerSymbols::updateDefenderWalkRequest()
{
  const Vector2<double> ballPosOnField(ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose));

  // ball vector from own goal
  Vector2<double> target(ballPosOnField);
  target.x -= fieldDimensions.xPosOwnGoal;
  
  // calc defender distance vector to own goal
  double defenderToGoalDistance = target.abs() - param.defenderOptimalBallDistance;
  if(defenderToGoalDistance > param.defenderMaxGoalDistance)
    defenderToGoalDistance = param.defenderMaxGoalDistance;
  if(defenderToGoalDistance < 1000.0)
    defenderToGoalDistance += param.defenderOptimalBallDistance * ((1000.0 - defenderToGoalDistance) / 900.0);
  target.normalize(defenderToGoalDistance);

  // convert vector from own goal in field coordinates
  target.x += fieldDimensions.xPosOwnGoal;

  // keep defender out of penalty area
  if(target.x < fieldDimensions.xPosOwnPenaltyArea + 30)
    target.x = fieldDimensions.xPosOwnPenaltyArea + 30;

  double rot(ballModel.lastSeenEstimate.getAngle());
  generateWalkRequest(target, target, 
    ballPosOnField, Vector2<double>(fieldDimensions.xPosOwnGoal, 0.0),    
    rot);
}

void BH2009SoccerSymbols::updateKeeperWalkRequest()
{
  // ball vector from own goal
  Vector2<double> ballPos = ballModel.lastSeenEstimate.getPositionInRelativeCoordinates();
  Vector2<double> target(ballPos);
  if(frameInfo.getTimeSince(ballModel.timeWhenLastSeen) >= param.keeperBallTimeout)
  { // else assume ball is at (0,0) on the field
    target.y = 0.0;
    target.x = -fieldDimensions.xPosOwnGoal;
  }
  else
    target.x -= fieldDimensions.xPosOwnGoal;

  // calc keeper distance vector to own goal
  double keeperAngle = target.angle();
  if(keeperAngle > param.keeperMaxAngle)
  {
    walkTarget.rotate(-(keeperAngle - param.keeperMaxAngle));
    keeperAngle = param.keeperMaxAngle;
  }
  else if(keeperAngle < -param.keeperMaxAngle)
  {
    walkTarget.rotate(-(keeperAngle + param.keeperMaxAngle));
    keeperAngle = -param.keeperMaxAngle;
  }
  keeperAngle = fabs(keeperAngle);
  const double keeperRadius = sqrt(
    pow(cos(keeperAngle) * static_cast<double>(param.keeperGoalDistanceX), 2)
    + pow(sin(keeperAngle) * static_cast<double>(param.keeperGoalDistanceY), 2));
  target.normalize(keeperRadius);

  // convert vector from own goal in field coordinates
  target.x += fieldDimensions.xPosOwnGoal;

  double rot(ballModel.lastSeenEstimate.getAngle());
  generateWalkRequest(target, target, 
    Vector2<double>(0.0, fieldDimensions.yPosLeftSideline), Vector2<double>(0.0, fieldDimensions.yPosRightSideline),
    rot, false, false, false, true);
}

void BH2009SoccerSymbols::generateWalkRequest(const Vector2<double>& target1, const Vector2<double>& target2, 
  const Vector2<double>& support1, const Vector2<double>& support2, double& rotation, bool evadeTeamMates, bool evadeObstacles, bool evadeBall, bool evadeBallFromWrongSide)
{
  Vector2<double> robotPos(robotPose.translation);
  Vector2<double> add, walk;

  // handle target attraction
  int chosenTarget = 1;
  Vector2<double> target(target1);
  if((target2 - robotPos).squareAbs() < (target1 - robotPos).squareAbs())
  {
    chosenTarget = 2;
    target = target2;
  }
  add = target - robotPos;
  add.normalize(param.targetAttraction * std::min(int(add.abs()), param.targetMaxDistance) / param.targetMaxDistance);
  walk += add;

  // handle support point attraction
  add = support1 - robotPos;
  add.normalize(param.supportAttraction * std::min(int(add.abs()), param.supportMaxDistance) / param.supportMaxDistance);
  walk += add;
  add = support2 - robotPos;
  add.normalize(param.supportAttraction * std::min(int(add.abs()), param.supportMaxDistance) / param.supportMaxDistance);
  walk += add;

  if(frameInfo.getTimeSince(ballModel.timeWhenLastSeen) < param.ballTimeout)
  {
    // handle ball repulse
    if(evadeBall)
    {
      Vector2<double> oppGoalPos(fieldDimensions.xPosOpponentGoal, 0.0);
      Vector2<double> ballPos(ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose));
      int ballAndOpponentGoalMinDistance = int((oppGoalPos - ballPos).abs()) + param.ballMinDistance;
      double ballDis = (ballPos - robotPos).abs();
      double oppGoalDis = (oppGoalPos - robotPos).abs();
      double dis = ballDis + oppGoalDis;    
      if(ballDis < param.ballMinDistance)
      {
        add = (robotPos - ballPos);
        add.normalize(param.ballRepulse * (param.ballMinDistance - ballDis) / param.ballMinDistance);
        walk += add;
      }
      else if(dis < ballAndOpponentGoalMinDistance)
      {
        add = (robotPos - ballPos) + (robotPos - oppGoalPos);
        add.x = 0.0;
        add.normalize(param.ballRepulse * (ballAndOpponentGoalMinDistance - dis) / param.ballMinDistance);
        walk += add;
      }
    }

    // handle ball repulse when the robot arrives from wrong side
    else if(evadeBallFromWrongSide)
    {
      getOpponentGoalAngle(); // make sure opponentGoalAngle is up to date
      double ballGoalAngle = fabs(normalize(ballModel.lastSeenEstimate.getPositionInRelativeCoordinates().angle() - opponentGoalAngle));
      if(ballGoalAngle > pi_4)
      {
        Vector2<double> ballPos(ballModel.lastSeenEstimate.getPositionInFieldCoordinates(robotPose));
        double ballDis = (ballPos - robotPos).abs();
        int ballMinDistance = int((ballGoalAngle - pi_4) * param.ballFromWrongSideMinDistance / (pi - pi_4));
        if(ballDis < ballMinDistance )
        {
          add = (robotPos - ballPos);
          add.normalize((param.ballFromWrongSideRepulse * (ballMinDistance - ballDis) / ballMinDistance));
          walk += add;
        }
      }
    }
  }

  // handle team mates repulse
  if(evadeTeamMates)
    for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
      if(teamMateData.timeStamps[i])
      {
        const Vector2<double>& teamMatePos(teamMateData.robotPoses[i].translation);
        double teamMateDis = (teamMatePos - robotPos).abs();
        if(teamMateDis < param.teamMateMinDistance)
        {
          add = (robotPos - teamMatePos);
          add.normalize(param.teamMateRepulse * (param.teamMateMinDistance - teamMateDis) / param.teamMateMinDistance);
          walk += add;
        }        
      }

  // set current walk target
  ASSERT(chosenTarget == 1 || chosenTarget == 2);
  walkTarget = Geometry::fieldCoord2Relative(robotPose, chosenTarget == 2 ? target2 : target1);
  walkTargetRotation = rotation;
#ifndef RELEASE
  walkTarget2 = Geometry::fieldCoord2Relative(robotPose, chosenTarget == 2 ? target1 : target2);
  walkSupport1 = Geometry::fieldCoord2Relative(robotPose, support1);
  walkSupport2 = Geometry::fieldCoord2Relative(robotPose, support2);
#endif


  // generate walk request from point with highest utility gain
  walk += robotPos;
  walk = Geometry::fieldCoord2Relative(robotPose, walk);
  Pose2D& speed(motionRequest.walkRequest.speed);
  int walkSpeedX = (walk.x > 0 ? param.walkSpeedForward : param.walkSpeedForward);
  walk.y *= double(walkSpeedX) / param.walkSpeedSideways;
  if(walk.abs() < 500)
    walk.y *= 3;
  speed.translation.x = walkSpeedX * walk.x / walk.abs();
  speed.translation.y = param.walkSpeedSideways * walk.y / walk.abs();
  speed.rotation = rotation * 2;
  if(speed.rotation > pi_4/3)
    speed.rotation = pi_4/3;
  if(speed.rotation < -pi_4/3)
    speed.rotation = -pi_4/3;

  // reduce walk speed, when we are very to close to our target
  if(walkTarget.abs() < 200)
  {
    if(speed.translation.x > 0)
      speed.translation.x *= 0.5;
  }



}
