/** 
* \file BH2009SoccerSymbols.h
* Implementation of symbols for our 4 roles.
* \author Colin Graf
*/

#ifndef __BH2009SoccerSymbols_h_
#define __BH2009SoccerSymbols_h_

#include "../../Symbols.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Modeling/BallModel.h"
#include "Representations/Modeling/GoalModel.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/TeamMateData.h"
#include "Representations/MotionControl/MotionRequest.h"


/**
* \class BH2009SoccerSymbols
* The Xabsl symbols that are defined in "soccer_symbols.xabsl"
* \author Colin Graf
*/ 
class BH2009SoccerSymbols : public Symbols
{
public:
  /**
  * Constructor.
  */
  BH2009SoccerSymbols(MotionRequest& motionRequest, BehaviorData& behaviorData, 
    const FrameInfo& frameInfo,
    const RobotInfo& robotInfo,
    const RobotPose& robotPose, 
    const BallModel& ballModel,
    const GoalModel& goalModel,
    const FieldDimensions& fieldDimensions,
    const TeamMateData& teamMateData) :

      motionRequest(motionRequest),
      behaviorData(behaviorData),
      frameInfo(frameInfo),
      robotInfo(robotInfo),
      robotPose(robotPose),
      ballModel(ballModel),
      goalModel(goalModel),
      fieldDimensions(fieldDimensions),
      teamMateData(teamMateData),

      walkMode(none),
      walkTargetRotation(0.0),    
      opponentGoalAngle(0.0),
      opponentGoalAngleTimeStamp(0),
 
      lastBallSeenPlayer(0),
      lastBallSeenPlayerTimeStamp(0),
      nearestToBallPlayer(0),
      secondNearestToBallPlayer(0),
      nearestToBallPlayerTimeStamp(0)

  {
    theInstance = this;
  }

  PROCESS_WIDE_STORAGE_STATIC BH2009SoccerSymbols* theInstance;

  /** registers the symbols at an engine */
  void registerSymbols(xabsl::Engine& engine);

  /** updates the symbols */
  void update();

  /** initialize the symbols */
  void init();

private:

  /** 
  * \class Parameters
  * A collection of all the parameters for the striker xabsl symbols 
  */
  class Parameters : public Streamable
  {
  private:

    /**
    * The method makes the object streamable.
    * \param in The stream from which the object is read 
    * \param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {  
      STREAM_REGISTER_BEGIN();
      STREAM(strikerBallHoldTimeout);
      STREAM(strikerBallEvadeDistance);
      STREAM(strikerBallEvadeRange);
      STREAM(strikerBallDistance);
      STREAM(strikerGoalModelTimeout);
      STREAM(strikerOffset);
      STREAM(defenderOptimalBallDistance);
      STREAM(defenderMaxGoalDistance);
      STREAM(supporterXOffset);
      STREAM(supporterYOffset);
      STREAM(keeperGoalDistanceX);
      STREAM(keeperGoalDistanceY);
      STREAM(keeperMaxAngle);
      STREAM(keeperBallTimeout);
      STREAM(averageWalkSpeed);
      STREAM(averageRotateSpeed);
      STREAM(nearestToBallSnap);

      STREAM(targetAttraction);
      STREAM(targetMaxDistance);
      STREAM(supportAttraction);
      STREAM(supportMaxDistance);
      STREAM(teamMateRepulse);
      STREAM(teamMateMinDistance);
      STREAM(ballTimeout);
      STREAM(ballRepulse);
      STREAM(ballMinDistance);
      STREAM(ballFromWrongSideRepulse);
      STREAM(ballFromWrongSideMinDistance);

      STREAM(walkSpeedForward);
      STREAM(walkSpeedBackward);
      STREAM(walkSpeedSideways);
      STREAM_REGISTER_FINISH();
    }

  public:

    /** Constructor. */
    Parameters() :
      strikerBallHoldTimeout(4000),
      strikerBallEvadeDistance(500),
      strikerBallEvadeRange(800),
      strikerBallDistance(160),
      strikerGoalModelTimeout(7000),
      strikerOffset(0),
      defenderOptimalBallDistance(1000),
      defenderMaxGoalDistance(3300),
      supporterXOffset(100),
      supporterYOffset(500),
      keeperGoalDistanceX(650),
      keeperGoalDistanceY(1000),
      keeperMaxAngle(fromDegrees(50)),
      keeperBallTimeout(10000),
      averageWalkSpeed(80),
      averageRotateSpeed(0.32),
      nearestToBallSnap(2000),

      targetAttraction(200),
      targetMaxDistance(500),
      supportAttraction(100),
      supportMaxDistance(500),
      teamMateRepulse(500),
      teamMateMinDistance(500),
      ballTimeout(7000),
      ballRepulse(500),
      ballMinDistance(300),
      ballFromWrongSideRepulse(700),
      ballFromWrongSideMinDistance(1000),

      walkSpeedForward(150),
      walkSpeedBackward(50),
      walkSpeedSideways(70)
    {};

    int strikerBallHoldTimeout; /**< After this period without seeing the ball, we are definitely not holding the ball. */

    int strikerBallEvadeDistance; /**< The distance we use to walk around the ball. */
    int strikerBallEvadeRange; /**< In this range we stop walking directly in ball direction to avoid hitting it from wrong direction. */
    int strikerBallDistance; /**< The distance we use to stand behind the ball. */
    int strikerGoalModelTimeout; /**< Prefer self location to locate the goal after this period of time without seeing the goal. */
    int strikerOffset; /**< The offset to prevent standing in center to ball. */

    int defenderOptimalBallDistance; /**< The distance a defender tries to stand between ball and own goal. */
    int defenderMaxGoalDistance; /**< The maximum distance the defender departs from own goal. */

    int supporterXOffset; /**< The offset in x direction on the field the supporter stands away form the striker. */
    int supporterYOffset; /**< The offset in y direction on the field the supporter stands away form the striker. */

    int keeperGoalDistanceX; /**< The distance in x direction the keeper stands away from the goal center. */
    int keeperGoalDistanceY; /**< The distance in y direction the keeper stands away from the goal center. */
    double keeperMaxAngle; /**< A limit to the keeper's angle to keep him in front of the goal. */
    int keeperBallTimeout; /**< After this period without seeing the ball, we assume that the ball is at (0,0) */

    int averageWalkSpeed; /**< A average walk speed in mm/s for a crude "time to reach target" estimation. */
    double averageRotateSpeed; /**< A average rotation speed in rad/s for a crude "time to reach target" estimation. */

    int nearestToBallSnap; /**< Time bonus in ms for a player which is already the "nearest to ball". */

    int targetAttraction;
    int targetMaxDistance;
    int supportAttraction;
    int supportMaxDistance;
    int teamMateRepulse;
    int teamMateMinDistance;
    int ballTimeout;
    int ballRepulse;
    int ballMinDistance;
    int ballFromWrongSideRepulse;
    int ballFromWrongSideMinDistance;

    int walkSpeedForward; /**< Maximun walk speed in X direction used for \c WalkRequest generation. */
    int walkSpeedBackward; /** < Maximun walk speed in -X direction used for \c WalkRequest generation. */
    int walkSpeedSideways; /** < Maximun walk speed in Y direction used for \c WalkRequest generation. */
  };

  double getOpponentGoalAngle();
  void setWalkTargetX(double value);
  double getWalkTargetX();
  void setWalkTargetY(double value);
  double getWalkTargetY();
  void setWalkTargetRotation(double value);
  double getWalkTargetRotation();
  double getKeeperBlock();
  double getBallAvgPositionFieldX();
  bool getNearestToBall();
  bool getSecondNearestToBall();
  int getNearestToBallPlayer();
  int getSecondNearestToBallPlayer();
  int getLastBallSeenPlayer();
  bool getBallInOpponentArea();
  bool getBallHold();

  MotionRequest& motionRequest;
  BehaviorData& behaviorData;
  const FrameInfo& frameInfo;
  const RobotInfo& robotInfo;
  const RobotPose& robotPose;
  const BallModel& ballModel;
  const GoalModel& goalModel;
  const FieldDimensions& fieldDimensions;
  const TeamMateData& teamMateData;

  Parameters param; /**< The parameters for the soccer symbols. */

  enum WalkMode
  {
    none,
    keeper,
    defender,
    supporter,
    striker,
    dribble,
    relative,
    field,
    numOfWalkModes
  };

  static const char* getWalkModeName(WalkMode mode)
  {
    const char* name[] = {
      "none",
      "keeper",
      "defender",
      "supporter",
      "striker",
      "dribble",
      "relative",
      "field",
    };
    return name[mode < numOfWalkModes ? mode : 0];
  }

  WalkMode walkMode; /**< A selected walk mode. */
  Vector2<double> walkTarget; /**< The last calculated target point. */
  double walkTargetRotation; /**< The last calculated target rotation. (something like angle to ball, angle to goal or angle to target point) */

#ifndef RELEASE
  Vector2<double> walkTarget2;
  Vector2<double> walkSupport1;
  Vector2<double> walkSupport2;
#endif

  Vector2<double> walkTargetIn; /**< A given target point for \c WalkMode::relative and \c WalkMode::field. */
  double walkTargetRotationIn; /**< A given target rotation for \c WalkMode::relative and \c WalkMode::field. */

  void updateWalkRequest();
  void updateStrikerWalkRequest(bool dribble = false);
  void updateSupporterWalkRequest();
  void updateDefenderWalkRequest();
  void updateKeeperWalkRequest();

  double opponentGoalAngle; /**< The angle to the opponent goal based on self locator or goal locator (depends on goal model validity) */
  unsigned int opponentGoalAngleTimeStamp; /**< Time stamp for preventing duplicated calculations of \c opponentGoalAngle in a single frame. */
 
  int lastBallSeenPlayer; /**< The player which has seen the ball at last. */
  unsigned int lastBallSeenPlayerTimeStamp; /**< Time stamp for preventing duplicated calculations of \c lastBallSeenPlayer in a single frame. */
  
  int nearestToBallPlayer; /**< The player which is the "nearest" to the ball. */
  int secondNearestToBallPlayer; /**< The player which is the second "nearest" to the ball. */
  unsigned int nearestToBallPlayerTimeStamp; /**< Time stamp for preventing duplicated calculations of \c  nearestToBallPlayer and \c secondNearestToBallPlayer in a single frame. */
  void calcNearestAndSecondNearestToBallPlayer(); /**< Calculates \c nearestToBallPlayer and secondNearestToBallPlayer. */

  void generateWalkRequest(const Vector2<double>& target1, const Vector2<double>& target2, 
    const Vector2<double>& support1, const Vector2<double>& support2, double& rotation, 
    bool evadeTeamMates = true, bool evadeObstacles = true, bool evadeBall = true, bool evadeBallFromWrongSide = false);


  // static symbol provider functions for the xabsl engine
  static double staticGetOpponentGoalAngle() { return theInstance->getOpponentGoalAngle(); }
  static void staticSetWalkTargetX(double value) { theInstance->setWalkTargetX(value); }
  static double staticGetWalkTargetX() { return theInstance->getWalkTargetX(); }
  static void staticSetWalkTargetY(double value) { theInstance->setWalkTargetY(value); }
  static double staticGetWalkTargetY() { return theInstance->getWalkTargetY(); }
  static void staticSetWalkTargetRotation(double value) { theInstance->setWalkTargetRotation(value); };
  static double staticGetWalkTargetRotation() { return theInstance->getWalkTargetRotation(); };
  static double staticGetKeeperBlock() { return theInstance->getKeeperBlock(); };
  static double staticGetBallAvgPositionFieldX() { return theInstance->getBallAvgPositionFieldX(); };
  static bool staticGetBallInOpponentArea() { return theInstance->getBallInOpponentArea(); }
  static bool staticGetNearestToBall() { return theInstance->getNearestToBall(); };
  static bool staticGetSecondNearestToBall() { return theInstance->getSecondNearestToBall(); };
  static int staticGetNearestToBallPlayer() { return theInstance->getNearestToBallPlayer(); }
  static int staticGetLastBallSeenPlayer() { return theInstance->getLastBallSeenPlayer(); }
  static int staticGetSecondNearestToBallPlayer() { return theInstance->getSecondNearestToBallPlayer(); }
  static bool staticGetBallHold() { return theInstance->getBallHold(); }

  friend class BH2009RoleSymbols;
};


#endif // __BH2009SoccerSymbols_h_

