/** 
* @file CognitionLogDataProvider.cpp
* This file implements a module that provides data replayed from a log file.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "CognitionLogDataProvider.h"
#include "Representations/Perception/JPEGImage.h"

PROCESS_WIDE_STORAGE CognitionLogDataProvider* CognitionLogDataProvider::theInstance = 0;

CognitionLogDataProvider::CognitionLogDataProvider() :
frameDataComplete(false),
theImage(0),
theCameraInfo(0),
theFrameInfo(0),
thePointsPercept(0),
theBallPercept(0),
theBallModel(0),
theRobotPose(0),
theGroundTruthBallModel(0),
theGroundTruthRobotPose(0)
{
  theInstance = this;
}

CognitionLogDataProvider::~CognitionLogDataProvider()
{
  if(theImage)
    delete theImage;
  if(theCameraInfo)
    delete theCameraInfo;
  if(theFrameInfo)
    delete theFrameInfo;
  if(thePointsPercept)
    delete thePointsPercept;
  if(theBallPercept)
    delete theBallPercept;
  if(theBallModel)
    delete theBallModel;
  if(theRobotPose)
    delete theRobotPose;
  if(theGroundTruthBallModel)
    delete theGroundTruthBallModel;
  if(theGroundTruthRobotPose)
    delete theGroundTruthRobotPose;
  theInstance = 0;
}

void CognitionLogDataProvider::update(Image& image)
{
  if(theImage)
  {
    image = *theImage;
    delete theImage;
    theImage = 0;
  }
  DEBUG_RESPONSE("representation:JPEGImage", OUTPUT(idJPEGImage, bin, JPEGImage(image)));
}

void CognitionLogDataProvider::update(CameraInfo& cameraInfo)
{
  if(theCameraInfo)
  {
    cameraInfo = *theCameraInfo;
    delete theCameraInfo;
    theCameraInfo = 0;
  }
}

void CognitionLogDataProvider::update(FrameInfo& frameInfo)
{
  if(theFrameInfo)
  {
    frameInfo = *theFrameInfo;
    delete theFrameInfo;
    theFrameInfo = 0;
  }
}

void CognitionLogDataProvider::update(PointsPercept& pointsPercept)
{
  if(thePointsPercept)
  {
    pointsPercept = *thePointsPercept;
    delete thePointsPercept;
    thePointsPercept = 0;
  }
}

void CognitionLogDataProvider::update(BallPercept& ballPercept)
{
  if(theBallPercept)
  {
    ballPercept = *theBallPercept;
    delete theBallPercept;
    theBallPercept = 0;
  }
}

void CognitionLogDataProvider::update(BallModel& ballModel)
{
  if(theBallModel)
  {
    ballModel = *theBallModel;
    delete theBallModel;
    theBallModel = 0;
  }
}

void CognitionLogDataProvider::update(RobotPose& robotPose)
{
  if(theRobotPose)
  {
    robotPose = *theRobotPose;
    delete theRobotPose;
    theRobotPose = 0;
  }
}

void CognitionLogDataProvider::update(GroundTruthBallModel& groundTruthBallModel)
{
  if(theGroundTruthBallModel)
  {
    groundTruthBallModel = *theGroundTruthBallModel;
    delete theGroundTruthBallModel;
    theGroundTruthBallModel = 0;
  }
}

void CognitionLogDataProvider::update(GroundTruthRobotPose& groundTruthRobotPose)
{
  if(theGroundTruthRobotPose)
  {
    groundTruthRobotPose = *theGroundTruthRobotPose;
    delete theGroundTruthRobotPose;
    theGroundTruthRobotPose = 0;
  }
}

bool CognitionLogDataProvider::handleMessage(InMessage& message)
{
  return theInstance && theInstance->handleMessage2(message);
}

bool CognitionLogDataProvider::isFrameDataComplete()
{
  if(!theInstance)
    return true;
  else if(theInstance->frameDataComplete)
  {
    OUTPUT(idLogResponse, bin, '\0');
    theInstance->frameDataComplete = false;
    return true;
  }
  else
    return false;
}

bool CognitionLogDataProvider::handleMessage2(InMessage& message)
{
  switch(message.getMessageID())
  {
  case idProcessFinished:
    frameDataComplete = true;
    return true;

  case idImage:
    if(!theImage)
      theImage = new Image;
    if(!theCameraInfo)
      theCameraInfo = new CameraInfo;
    if(!theFrameInfo)
      theFrameInfo = new FrameInfo;
    message.bin >> *theImage;
    *theCameraInfo = theImage->cameraInfo;
    theFrameInfo->time = theImage->timeStamp;
    return true;

  case idJPEGImage:
    if(!theImage)
      theImage = new Image;
    if(!theCameraInfo)
      theCameraInfo = new CameraInfo;
    if(!theFrameInfo)
      theFrameInfo = new FrameInfo;
    {
      JPEGImage jpegImage;
      message.bin >> jpegImage;
      jpegImage.toImage(*theImage);
    }
    *theCameraInfo = theImage->cameraInfo;
    theFrameInfo->time = theImage->timeStamp;
    return true;

  case idFrameInfo:
    if(!theFrameInfo)
      theFrameInfo = new FrameInfo;
    message.bin >> *theFrameInfo;
    return true;

  case idPointsPercept:
    if(!thePointsPercept)
      thePointsPercept = new PointsPercept;
    message.bin >> *thePointsPercept;
    return true;

  case idBallPercept:
    if(!theBallPercept)
      theBallPercept = new BallPercept;
    message.bin >> *theBallPercept;
    return true;

  case idGroundTruthBallModel:
    if(!theGroundTruthBallModel)
      theGroundTruthBallModel = new GroundTruthBallModel;
    message.bin >> *theGroundTruthBallModel;
    return true;

  case idGroundTruthRobotPose:
    if(!theGroundTruthRobotPose)
      theGroundTruthRobotPose = new GroundTruthRobotPose;
    message.bin >> *theGroundTruthRobotPose;
    return true;

  default:
    return false; 
  }
}

MAKE_MODULE(CognitionLogDataProvider, Infrastructure)
