/** 
* @file MotionConfigurationDataProvider.cpp
* This file implements a module that provides data loaded from configuration files.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "MotionConfigurationDataProvider.h"
#include "Tools/Settings.h"


MotionConfigurationDataProvider::MotionConfigurationDataProvider() :
theJointCalibration(0),
theSensorCalibration(0),
theRobotDimensions(0),
theWalkingParameters(0),
theBalanceParameters(0)
{
  readIfRequired();
}

MotionConfigurationDataProvider::~MotionConfigurationDataProvider()
{
  if(theJointCalibration)
    delete theJointCalibration;
  if(theSensorCalibration)
    delete theSensorCalibration;
  if(theRobotDimensions)
    delete theRobotDimensions;
  if(theWalkingParameters)
    delete theWalkingParameters;
  if(theBalanceParameters)
    delete theBalanceParameters;
}

void MotionConfigurationDataProvider::update(RobotName& robotName)
{
  readIfRequired();
  robotName.name = this->robot;
}

void MotionConfigurationDataProvider::update(JointCalibration& jointCalibration)
{
  readIfRequired();
  ASSERT(loadedJointCalibration);
  if(theJointCalibration)
  {
    jointCalibration = *theJointCalibration;
    delete theJointCalibration;
    theJointCalibration = 0;
  }
}

void MotionConfigurationDataProvider::update(SensorCalibration& sensorCalibration)
{
  readIfRequired();
  ASSERT(loadedSensorCalibration);
  if(theSensorCalibration)
  {
    sensorCalibration = *theSensorCalibration;
    delete theSensorCalibration;
    theSensorCalibration = 0;
  }
}

void MotionConfigurationDataProvider::update(RobotDimensions& robotDimensions)
{
  readIfRequired();
  ASSERT(loadedRobotDimensions);
  if(theRobotDimensions)
  {
    robotDimensions = *theRobotDimensions;
    delete theRobotDimensions;
    theRobotDimensions = 0;
  }
}

void MotionConfigurationDataProvider::update(WalkingParameters& walkingParameters)
{
  readIfRequired();
  ASSERT(loadedWalkingParameters);
  if(theWalkingParameters)
  {
    walkingParameters = *theWalkingParameters;
    delete theWalkingParameters;
    theWalkingParameters = 0;
  }
}

void MotionConfigurationDataProvider::update(BalanceParameters& balanceParameters)
{
  readIfRequired();
  ASSERT(loadedBalanceParameters);
  if(theBalanceParameters)
  {
    balanceParameters = *theBalanceParameters;
    delete theBalanceParameters;
    theBalanceParameters = 0;
  }
}

void MotionConfigurationDataProvider::readIfRequired()
{
  if(robot != Global::getSettings().robot)
  {
    robot = Global::getSettings().robot;
    readJointCalibration();
    readSensorCalibration();
    readRobotDimensions();
    readWalkingParameters();
    readBalanceParameters();
  }
}

void MotionConfigurationDataProvider::readJointCalibration()
{
  loadedJointCalibration = false;
  InConfigFile stream(Global::getSettings().expandRobotFilename("jointCalibration.cfg"));
  if(stream.exists())
  {
    if(!theJointCalibration)
      theJointCalibration = new JointCalibration;
    stream >> *theJointCalibration;
    loadedJointCalibration = true;
  }
}

void MotionConfigurationDataProvider::readWalkingParameters()
{
  loadedWalkingParameters = false;
  InConfigFile* stream = new InConfigFile(Global::getSettings().expandRobotFilename("walking.cfg"), "default");
  if(!stream->exists())
  {
    delete stream;
    stream = new InConfigFile("walking.cfg", "default");
  }

  if(stream->exists())
  {
    if(!theWalkingParameters)
      theWalkingParameters = new WalkingParameters;
    *stream >> *theWalkingParameters;
    loadedWalkingParameters = true;
  }

  delete stream;
}

  void MotionConfigurationDataProvider::readBalanceParameters()
{
  loadedBalanceParameters = false;
  InConfigFile* stream = new InConfigFile(Global::getSettings().expandRobotFilename("balanceCalibration.cfg"), "default");
  if(!stream->exists())
  {
    delete stream;
    stream = new InConfigFile("balanceCalibration.cfg", "default");
  }

  if(stream->exists())
  {
    if(!theBalanceParameters)
      theBalanceParameters = new BalanceParameters;
    *stream >> *theBalanceParameters;
    loadedBalanceParameters = true;
  }

  delete stream;
}

void MotionConfigurationDataProvider::readSensorCalibration()
{
  loadedSensorCalibration = false;
  InConfigFile stream(Global::getSettings().expandRobotFilename("sensorCalibration.cfg"));
  if(stream.exists())
  {
    if(!theSensorCalibration)
      theSensorCalibration = new SensorCalibration;
    stream >> *theSensorCalibration;
    loadedSensorCalibration = true;
  }
}

void MotionConfigurationDataProvider::readRobotDimensions()
{
  loadedRobotDimensions = false;
  InConfigFile stream(Global::getSettings().expandRobotFilename("robotDimensions.cfg"));
  if(stream.exists())
  {
    if(!theRobotDimensions)
      theRobotDimensions = new RobotDimensions;
    stream >> *theRobotDimensions;
    loadedRobotDimensions = true;
  }
}

MAKE_MODULE(MotionConfigurationDataProvider, Infrastructure)
