/** 
* @file MotionLogDataProvider.cpp
* This file implements a module that provides data replayed from a log file.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "MotionLogDataProvider.h"

PROCESS_WIDE_STORAGE MotionLogDataProvider* MotionLogDataProvider::theInstance = 0;

MotionLogDataProvider::MotionLogDataProvider() :
frameDataComplete(false),
theJointData(0),
theSensorData(0),
theKeyStates(0),
theFrameInfo(0),
theOdometryData(0)
{
  theInstance = this;
}

MotionLogDataProvider::~MotionLogDataProvider()
{
  if(theJointData)
    delete theJointData;
  if(theSensorData)
    delete theSensorData;
  if(theKeyStates)
    delete theKeyStates;
  if(theFrameInfo)
    delete theFrameInfo;
  if(theOdometryData)
    delete theOdometryData;
  theInstance = 0;
}

void MotionLogDataProvider::update(JointData& jointData)
{
  if(theJointData)
  {
    jointData = *theJointData;
    delete theJointData;
    theJointData = 0;
  }
  JointDataDeg jointDataDeg(jointData);
  MODIFY("representation:JointDataDeg", jointDataDeg);
}

void MotionLogDataProvider::update(SafeJointData& safeJointData)
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
    if(MotionLogDataProviderBase::theJointData.angles[i] != JointData::off)
      safeJointData.angles[i] = MotionLogDataProviderBase::theJointData.angles[i];
    else if(safeJointData.angles[i] == JointData::off)
      safeJointData.angles[i] = 0;
  safeJointData.timeStamp = MotionLogDataProviderBase::theJointData.timeStamp;
}

void MotionLogDataProvider::update(SensorData& sensorData)
{
  if(theSensorData)
  {
    sensorData = *theSensorData;
    delete theSensorData;
    theSensorData = 0;
  }
}

void MotionLogDataProvider::update(SafeSensorData& safeSensorData)
{
  for(int i = 0; i < SensorData::numOfSensors; ++i)
    if(MotionLogDataProviderBase::theSensorData.data[i] != JointData::off)
      safeSensorData.data[i] = MotionLogDataProviderBase::theSensorData.data[i];
    else if(safeSensorData.data[i] == JointData::off)
      safeSensorData.data[i] = 0;

  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    safeSensorData.loads[i] = MotionLogDataProviderBase::theSensorData.loads[i];
    safeSensorData.temperatures[i] = MotionLogDataProviderBase::theSensorData.temperatures[i];
  }

  safeSensorData.timeStamp = MotionLogDataProviderBase::theSensorData.timeStamp;
}

void MotionLogDataProvider::update(KeyStates& keyStates)
{
  if(theKeyStates)
  {
    keyStates = *theKeyStates;
    delete theKeyStates;
    theKeyStates = 0;
  }
}

void MotionLogDataProvider::update(FrameInfo& frameInfo)
{
  if(theFrameInfo)
  {
    frameInfo = *theFrameInfo;
    delete theFrameInfo;
    theFrameInfo = 0;
  }
}

void MotionLogDataProvider::update(OdometryData& odometryData)
{
  if(theOdometryData)
  {
    odometryData = *theOdometryData;
    delete theOdometryData;
    theOdometryData = 0;
  }
}

bool MotionLogDataProvider::handleMessage(InMessage& message)
{
  return theInstance && theInstance->handleMessage2(message);
}

bool MotionLogDataProvider::isFrameDataComplete()
{
  if(!theInstance)
    return true;
  else if(theInstance->frameDataComplete)
  {
    OUTPUT(idLogResponse, bin, '\0');
    theInstance->frameDataComplete = false;
    return true;
  }
  else
    return false;
}

bool MotionLogDataProvider::handleMessage2(InMessage& message)
{
  switch(message.getMessageID())
  {
  case idProcessFinished:
    frameDataComplete = true;
    return true;

  case idJointData:
    if(!theJointData)
      theJointData = new JointData;
    message.bin >> *theJointData;
    if(!theFrameInfo)
      theFrameInfo = new FrameInfo;
    theFrameInfo->time = theJointData->timeStamp;
    return true;

  case idSensorData:
    if(!theSensorData)
      theSensorData = new SensorData;
    message.bin >> *theSensorData;
    return true;

  case idKeyStates:
    if(!theKeyStates)
      theKeyStates = new KeyStates;
    message.bin >> *theKeyStates;
    return true;

  case idOdometryData:
    if(!theOdometryData)
      theOdometryData = new OdometryData;
    message.bin >> *theOdometryData;
    return true;

  default:
    return false; 
  }
}

MAKE_MODULE(MotionLogDataProvider, Infrastructure)
