/** 
* @file MotionLogDataProvider.h
* This file declares a module that provides data replayed from a log file.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef __MotionLogDataProvider_h_
#define __MotionLogDataProvider_h_

#include "Tools/Module/Module.h"
#include "Representations/Infrastructure/JointData.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Tools/MessageQueue/InMessage.h"

MODULE(MotionLogDataProvider)
  REQUIRES(JointData)
  USES(SensorData); // is required for SafeSensorData
  PROVIDES_WITH_MODIFY_AND_OUTPUT(JointData)
  PROVIDES_WITH_MODIFY(SafeJointData)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(SensorData)
  PROVIDES_WITH_MODIFY(SafeSensorData)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(KeyStates)
  PROVIDES_WITH_MODIFY(FrameInfo)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(OdometryData)
END_MODULE

class MotionLogDataProvider : public MotionLogDataProviderBase
{
private:
  PROCESS_WIDE_STORAGE_STATIC MotionLogDataProvider* theInstance; /**< Points to the only instance of this class in this process or is 0 if there is none. */
  bool frameDataComplete; /**< Were all messages of the current frame received? */

  JointData* theJointData;
  SensorData* theSensorData;
  KeyStates* theKeyStates;
  FrameInfo* theFrameInfo;
  OdometryData* theOdometryData;

  void update(JointData& jointData);
  void update(SafeJointData& safeJointData);
  void update(SensorData& sensorData);
  void update(SafeSensorData& safeSensorData);
  void update(KeyStates& keyStates);
  void update(FrameInfo& frameInfo);
  void update(OdometryData& odometryData);

  /** 
  * The method is called for every incoming debug message by handleMessage.
  * @param message An interface to read the message from the queue.
  * @return Was the message handled?
  */
  bool handleMessage2(InMessage& message);

public:
  /**
  * Default constructor.
  */
  MotionLogDataProvider();

  /**
  * Destructor.
  */
  ~MotionLogDataProvider();

  /** 
  * The method is called for every incoming debug message.
  * @param message An interface to read the message from the queue.
  * @return Was the message handled?
  */
  static bool handleMessage(InMessage& message);

  /** 
  * The method returns whether idProcessFinished was received.
  * @return Were all messages of the current frame received?
  */
  static bool isFrameDataComplete();
};

#endif // __MotionLogDataProvider_h_
