/**
* @file Modules/Infrastructure/NaoProvider.cpp
* The file declares a module that provides information from the Nao via DCM.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "Tools/Streams/InStreams.h"
#include "NaoProvider.h"
#include "Platform/GTAssert.h"
#include "Platform/SystemCall.h"
#include "Tools/Settings.h"

PROCESS_WIDE_STORAGE NaoProvider* NaoProvider::theInstance = 0;

#ifdef USE_NAOQI

#include <alproxy.h>
#include <almemoryfastaccess.h>

static const char* jointNames[] =
{
  "HeadYaw",
  "HeadPitch",
  "LShoulderPitch",
  "LShoulderRoll",
  "LElbowRoll",
  "LElbowYaw",
  "RShoulderPitch",
  "RShoulderRoll",
  "RElbowRoll",
  "RElbowYaw",
  "LHipYawPitch",
  "LHipRoll",
  "LHipPitch",
  "LKneePitch",
  "LAnklePitch",
  "LAnkleRoll",
  "RHipRoll",
  "RHipPitch",
  "RKneePitch",
  "RAnklePitch",
  "RAnkleRoll"
};

static const char* sensorNames[] =
{
  "InertialSensor/GyrX",
  "InertialSensor/GyrY",
  "InertialSensor/GyrRef",
  "InertialSensor/AccX",
  "InertialSensor/AccY",
  "InertialSensor/AccZ",
  "Battery/Charge",
  "LFoot/FSR/FrontLeft",
  "LFoot/FSR/FrontRight",
  "LFoot/FSR/RearLeft",
  "LFoot/FSR/RearRight",
  "RFoot/FSR/FrontLeft",
  "RFoot/FSR/FrontRight",
  "RFoot/FSR/RearLeft",
  "RFoot/FSR/RearRight",
  "US",
  "RFoot/Bumper/Right",
  "RFoot/Bumper/Left",
  "LFoot/Bumper/Right",
  "LFoot/Bumper/Left",
  "ChestBoard/Button"
};

static const char* ledNames[] =
{
  "Face/Led/Red/Left/0Deg",
  "Face/Led/Red/Left/45Deg",
  "Face/Led/Red/Left/90Deg",
  "Face/Led/Red/Left/135Deg",
  "Face/Led/Red/Left/180Deg",
  "Face/Led/Red/Left/225Deg",
  "Face/Led/Red/Left/270Deg",
  "Face/Led/Red/Left/315Deg",
  "Face/Led/Green/Left/0Deg",
  "Face/Led/Green/Left/45Deg",
  "Face/Led/Green/Left/90Deg",
  "Face/Led/Green/Left/135Deg",
  "Face/Led/Green/Left/180Deg",
  "Face/Led/Green/Left/225Deg",
  "Face/Led/Green/Left/270Deg",
  "Face/Led/Green/Left/315Deg",
  "Face/Led/Blue/Left/0Deg",
  "Face/Led/Blue/Left/45Deg",
  "Face/Led/Blue/Left/90Deg",
  "Face/Led/Blue/Left/135Deg",
  "Face/Led/Blue/Left/180Deg",
  "Face/Led/Blue/Left/225Deg",
  "Face/Led/Blue/Left/270Deg",
  "Face/Led/Blue/Left/315Deg",
  "Face/Led/Red/Right/0Deg",
  "Face/Led/Red/Right/45Deg",
  "Face/Led/Red/Right/90Deg",
  "Face/Led/Red/Right/135Deg",
  "Face/Led/Red/Right/180Deg",
  "Face/Led/Red/Right/225Deg",
  "Face/Led/Red/Right/270Deg",
  "Face/Led/Red/Right/315Deg",
  "Face/Led/Green/Right/0Deg",
  "Face/Led/Green/Right/45Deg",
  "Face/Led/Green/Right/90Deg",
  "Face/Led/Green/Right/135Deg",
  "Face/Led/Green/Right/180Deg",
  "Face/Led/Green/Right/225Deg",
  "Face/Led/Green/Right/270Deg",
  "Face/Led/Green/Right/315Deg",
  "Face/Led/Blue/Right/0Deg",
  "Face/Led/Blue/Right/45Deg",
  "Face/Led/Blue/Right/90Deg",
  "Face/Led/Blue/Right/135Deg",
  "Face/Led/Blue/Right/180Deg",
  "Face/Led/Blue/Right/225Deg",
  "Face/Led/Blue/Right/270Deg",
  "Face/Led/Blue/Right/315Deg",
  "Ears/Led/Left/0Deg",
  "Ears/Led/Left/36Deg",
  "Ears/Led/Left/72Deg",
  "Ears/Led/Left/108Deg",
  "Ears/Led/Left/144Deg",
  "Ears/Led/Left/180Deg",
  "Ears/Led/Left/216Deg",
  "Ears/Led/Left/252Deg",
  "Ears/Led/Left/288Deg",
  "Ears/Led/Left/324Deg",
  "Ears/Led/Right/0Deg",
  "Ears/Led/Right/36Deg",
  "Ears/Led/Right/72Deg",
  "Ears/Led/Right/108Deg",
  "Ears/Led/Right/144Deg",
  "Ears/Led/Right/180Deg",
  "Ears/Led/Right/216Deg",
  "Ears/Led/Right/252Deg",
  "Ears/Led/Right/288Deg",
  "Ears/Led/Right/324Deg",
  "ChestBoard/Led/Red",
  "ChestBoard/Led/Green",
  "ChestBoard/Led/Blue",
  "LFoot/Led/Red",
  "LFoot/Led/Green",
  "LFoot/Led/Blue",
  "RFoot/Led/Red",
  "RFoot/Led/Green",
  "RFoot/Led/Blue"
};

static const int numOfJoints = sizeof(jointNames) / sizeof(jointNames[0]);
static const int numOfLeds = sizeof(ledNames) / sizeof(ledNames[0]);
static const int numOfSwitches = 0;
static const int numOfSensors = sizeof(sensorNames) / sizeof(sensorNames[0]);

NaoProvider::NaoProvider() :
idRobot(-1),
proxy(0),
request(0),
memory(0),
lastUSMeasurement(0),
startUnpressed(0),
shuttingDown(false)
{
  NaoProvider::theInstance = this;
  try
  {
    proxy = new AL::ALProxy("ALMemory");
    Global::getSettings().robot = (const std::string&) proxy->call<ALValue>("getData", std::string("Device/DeviceList/ChestBoard/BodyNickName"));
    delete proxy;
    OUTPUT(idText, text, "Hi, I am " << Global::getSettings().robot << ".");

    proxy = new AL::ALProxy("DCM");
	ALValue delta = 0;
    timeOffset = (int) proxy->genericCall("getTime", delta) - SystemCall::getCurrentSystemTime();
    AL::ALValue params;
    params.arraySetSize(1);
    params[0].arraySetSize(2);
    params[0][0] = "allActuators";
    params[0][1].arraySetSize(numOfJoints * 2 + numOfLeds);
    TStringArray sensors;
    int j = 0;
    for(int i = 0; i < numOfJoints; ++i)
    {
      params[0][1][j++] = std::string(jointNames[i]) + "/Position/Actuator/Value";
      params[0][1][j++] = std::string(jointNames[i]) + "/Hardness/Actuator/Value";
      sensors.push_back(std::string("Device/SubDeviceList/") + jointNames[i] + "/Position/Sensor/Value");
      sensors.push_back(std::string("Device/SubDeviceList/") + jointNames[i] + "/ElectricCurrent/Sensor/Value");
      sensors.push_back(std::string("Device/SubDeviceList/") + jointNames[i] + "/Temperature/Sensor/Value");
    }
    for(int i = 0; i < numOfLeds; ++i)
      params[0][1][j++] = std::string(ledNames[i]) + "/Actuator/Value";
    proxy->genericCall("createAlias", params);

    params[0][0] = "usRequest";
    params[0][1].arraySetSize(1);
    params[0][1][0] = "US/Actuator/Value";
    proxy->genericCall("createAlias", params);

    request = new AL::ALValue;
    request->arraySetSize(1);
    (*request)[0].arraySetSize(6);
    (*request)[0][0] = "allActuators";
    (*request)[0][1] = "Merge";
    (*request)[0][2] = "time-separate";
    (*request)[0][3] = 0;
    (*request)[0][4].arraySetSize(1);
    (*request)[0][5].arraySetSize(numOfJoints * 2 + numOfLeds + numOfSwitches);
    for(int i = 0; i < numOfJoints * 2 + numOfLeds; ++i)
      (*request)[0][5][i].arraySetSize(1);

    for(int i = 0; i < numOfSensors; ++i)
      sensors.push_back(std::string("Device/SubDeviceList/") + sensorNames[i] + "/Sensor/Value");

    memory = new ALMemoryFastAccess;
    memory->ConnectToVariables(sensors, "bhuman");
    sensorValues.reserve(sensors.size());
    sensorValues.resize(sensors.size());

    usRequest = new AL::ALValue;
    usRequest->arraySetSize(1);
    (*usRequest)[0].arraySetSize(6);
    (*usRequest)[0][0] = "usRequest";
    (*usRequest)[0][1] = "Merge";
    (*usRequest)[0][2] = "time-separate";
    (*usRequest)[0][3] = 0;
    (*usRequest)[0][4].arraySetSize(1);
    (*usRequest)[0][5].arraySetSize(1);
    (*usRequest)[0][5][0].arraySetSize(1);
    (*usRequest)[0][5][0][0] = 1.0f;
  }
  catch(AL::ALError& e)
  {
    std::cout << e.toString() << std::endl;
    ASSERT(false);
  }
}

NaoProvider::~NaoProvider()
{
  delete memory;
  delete request;
  delete usRequest;
  delete proxy;
  NaoProvider::theInstance = 0;
}

void NaoProvider::send()
{
  float hardness = 0.8f;
  MODIFY("hardness", hardness);
  int dcmDelay = 20;
  MODIFY("dcmDelay", dcmDelay);
  int usDelay = 40;
  MODIFY("usDelay", usDelay);
  try
  {
    (*request)[0][4][0] = (int) SystemCall::getCurrentSystemTime() + timeOffset + dcmDelay;

    int j = 0;
    for(int i = 0; i < JointData::numOfJoints; ++i)
    { 
      if(i == JointData::legRight0) // missing on Nao
        ++i;

      if(theJointRequest.angles[i] == JointData::off || shuttingDown)
      {
        (*request)[0][5][j++][0] = 0.0f;
        (*request)[0][5][j++][0] = 0.0f;
      }
      else
      {
        double d = theJointRequest.angles[i] + theJointCalibration.joints[i].offset;
        if(d > theJointCalibration.joints[i].maxAngle) 
          d = theJointCalibration.joints[i].maxAngle;
        else if(d < theJointCalibration.joints[i].minAngle) 
          d = theJointCalibration.joints[i].minAngle;
        (*request)[0][5][j++][0] = float(d * theJointCalibration.joints[i].sign);
        (*request)[0][5][j++][0] = hardness;
      }
    }

    bool on = (SystemCall::getCurrentSystemTime() / 50 & 8) != 0;
    bool fastOn = (SystemCall::getCurrentSystemTime() / 10 & 8) != 0;
    for(int i = 0; i < numOfLeds; ++i)
      (*request)[0][5][j++][0] = (theLEDRequest.ledStates[i] == LEDRequest::on || 
                                  (theLEDRequest.ledStates[i] == LEDRequest::blinking && on) ||
                                  (theLEDRequest.ledStates[i] == LEDRequest::fastBlinking && fastOn)) 
                                 ? 1.0f : 0.0f;

    proxy->genericCall("setAlias", *request);

    if(SystemCall::getTimeSince(lastUSMeasurement) >= usDelay)
    {
      lastUSMeasurement = SystemCall::getCurrentSystemTime();
      (*usRequest)[0][4][0] = (int) SystemCall::getCurrentSystemTime() + timeOffset + usDelay;
      proxy->genericCall("setAlias", *usRequest);
    }
  }
  catch(AL::ALError& e)
  {
    std::cout << e.toString() << std::endl;
    ASSERT(false);
  }
}

void NaoProvider::update(JointData& jointData)
{
  try
  {
    memory->GetValues(sensorValues);
    int j = 0;
    for(int i = 0; i < JointData::numOfJoints; ++i)
    {
      if(i == JointData::legRight0)
      {
        jointData.angles[i] = jointData.angles[JointData::legLeft0];
        sensorData.loads[i] = sensorData.loads[JointData::legLeft0];
        sensorData.temperatures[i] = sensorData.temperatures[JointData::legLeft0];
      }
      else
      {
        jointData.angles[i] = sensorValues[j++] * theJointCalibration.joints[i].sign - theJointCalibration.joints[i].offset;
        sensorData.loads[i] = short(1000 * sensorValues[j++]); // somewhat compatible to Bioloid values, but without sign
        sensorData.temperatures[i] = (unsigned char) sensorValues[j++];
      }
    }

    float gyroRef = 0.0f;
    for(int i = 0; i < SensorData::numOfSensors; ++i)
    {
      if(i == SensorData::compassX)
        i += 2;
      if(i == SensorData::gyroZ)
        gyroRef = sensorValues[j++];
      else
        sensorData.data[i] = sensorValues[j++];
    }
    
    sensorData.data[SensorData::gyroX] *= (theSensorCalibration.gyroXGain/1600);
    bodyTiltMeasuredGyroX.add(sensorData.data[SensorData::gyroX]);
    sensorData.data[SensorData::gyroX] -= bodyTiltMeasuredGyroX.getAverage();
    sensorData.data[SensorData::gyroY] *= (theSensorCalibration.gyroYGain/1600);
    bodyTiltMeasuredGyroY.add(sensorData.data[SensorData::gyroY]);
    sensorData.data[SensorData::gyroY] -= bodyTiltMeasuredGyroY.getAverage();

    sensorData.data[SensorData::accX] *= theSensorCalibration.accXGain;
    sensorData.data[SensorData::accX] += theSensorCalibration.accXOffset;
    sensorData.data[SensorData::accY] *= theSensorCalibration.accYGain;
    sensorData.data[SensorData::accY] += theSensorCalibration.accYOffset;
    sensorData.data[SensorData::accZ] *= theSensorCalibration.accZGain;
    sensorData.data[SensorData::accZ] += theSensorCalibration.accZOffset;
    sensorData.data[SensorData::fsrLFL] = theSensorCalibration.fsrLFLGain / sensorData.data[SensorData::fsrLFL];
    sensorData.data[SensorData::fsrLFL] += theSensorCalibration.fsrLFLOffset;
    sensorData.data[SensorData::fsrLFR] = theSensorCalibration.fsrLFRGain / sensorData.data[SensorData::fsrLFR];
    sensorData.data[SensorData::fsrLFR] += theSensorCalibration.fsrLFROffset;
    sensorData.data[SensorData::fsrLBL] = theSensorCalibration.fsrLBLGain / sensorData.data[SensorData::fsrLBL];
    sensorData.data[SensorData::fsrLBL] += theSensorCalibration.fsrLBLOffset;
    sensorData.data[SensorData::fsrLBR] = theSensorCalibration.fsrLBRGain / sensorData.data[SensorData::fsrLBR];
    sensorData.data[SensorData::fsrLBR] += theSensorCalibration.fsrLBROffset;
    sensorData.data[SensorData::fsrRFL] = theSensorCalibration.fsrRFLGain / sensorData.data[SensorData::fsrRFL];
    sensorData.data[SensorData::fsrRFL] += theSensorCalibration.fsrRFLOffset;
    sensorData.data[SensorData::fsrRFR] = theSensorCalibration.fsrRFRGain / sensorData.data[SensorData::fsrRFR];
    sensorData.data[SensorData::fsrRFR] += theSensorCalibration.fsrRFROffset;
    sensorData.data[SensorData::fsrRBL] = theSensorCalibration.fsrRBLGain / sensorData.data[SensorData::fsrRBL];
    sensorData.data[SensorData::fsrRBL] += theSensorCalibration.fsrRBLOffset;
    sensorData.data[SensorData::fsrRBR] = theSensorCalibration.fsrRBRGain / sensorData.data[SensorData::fsrRBR];
    sensorData.data[SensorData::fsrRBR] += theSensorCalibration.fsrRBROffset;
    sensorData.data[SensorData::us] *= 1000;
 
    for(int i = 0; i < KeyStates::numberOfKeys; ++i)
      keyStates.pressed[i] = sensorValues[j++] != 0;

    jointData.timeStamp = sensorData.timeStamp = SystemCall::getCurrentSystemTime();
  }
  catch(AL::ALError& e)
  {
    std::cout << e.toString() << std::endl;
    ASSERT(false);
  }
  if(!keyStates.pressed[KeyStates::start])
    startUnpressed = SystemCall::getCurrentSystemTime();
  else if(startUnpressed && !shuttingDown && SystemCall::getTimeSince(startUnpressed) > 3000)
  {
    shuttingDown = true;
    if(keyStates.pressed[KeyStates::left] || keyStates.pressed[KeyStates::right] ||
       keyStates.pressed[KeyStates::up] || keyStates.pressed[KeyStates::down])
      system("harakiri --deep && shutdown -h now");
    else
      system("shutdown -h now");
  }

#ifndef RELEASE
  JointDataDeg jointDataDeg(jointData);
#endif
  MODIFY("representation:JointDataDeg", jointDataDeg);
}

#else // WIN32/WINCE
NaoProvider::NaoProvider() {}
NaoProvider::~NaoProvider() {}
void NaoProvider::send() {}
void NaoProvider::update(JointData& jointData) {}
#endif

void NaoProvider::update(SafeJointData& safeJointData)
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
    if(theJointData.angles[i] != JointData::off)
      safeJointData.angles[i] = theJointData.angles[i];
    else if(safeJointData.angles[i] == JointData::off)
      safeJointData.angles[i] = 0;
  safeJointData.timeStamp = theJointData.timeStamp;
}

void NaoProvider::update(SafeSensorData& safeSensorData)
{
  for(int i = 0; i < SensorData::numOfSensors; ++i)
    if(sensorData.data[i] != JointData::off || (i == SensorData::us && sensorData.data[i] != 0))
      safeSensorData.data[i] = sensorData.data[i];
    else if(safeSensorData.data[i] == JointData::off)
      safeSensorData.data[i] = 0;

  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    safeSensorData.loads[i] = sensorData.loads[i];
    safeSensorData.temperatures[i] = sensorData.temperatures[i];
  }

  safeSensorData.timeStamp = sensorData.timeStamp;
}

void NaoProvider::finishFrame()
{
  if(theInstance)
    theInstance->send();
}

MAKE_MODULE(NaoProvider, Infrastructure)
