/**
* @file Modules/Infrastructure/NaoProvider.h
* The file declares a module that provides information from the Nao via DCM.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Röfer</a>
*/

#ifndef __NaoProvider_H__
#define __NaoProvider_H__

#include "Tools/Module/Module.h"
#include "Representations/Configuration/JointCalibration.h"
#include "Representations/Configuration/SensorCalibration.h"
#include "Representations/Infrastructure/JointData.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/LEDRequest.h"
#include "Tools/RingBufferWithSum.h"

MODULE(NaoProvider)
  REQUIRES(JointCalibration)
  REQUIRES(SensorCalibration)
  REQUIRES(JointData)
  REQUIRES(LEDRequest)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(JointData)
  PROVIDES_WITH_MODIFY(SafeJointData)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(SensorData)
  PROVIDES_WITH_MODIFY(SafeSensorData)
  PROVIDES_WITH_MODIFY_AND_OUTPUT(KeyStates)
  PROVIDES_WITH_MODIFY(FrameInfo)
  USES(JointRequest) // Will be accessed in send()
END_MODULE

namespace AL 
{
  class ALProxy;
  class ALValue;
};
class ALMemoryFastAccess;

/**
* @class NaoProvider
* A module that provides information from the Bioloid control board.
*/
class NaoProvider : public NaoProviderBase
{
private:
  PROCESS_WIDE_STORAGE_STATIC NaoProvider* theInstance; /**< The only instance of this module. */
  SensorData sensorData; /**< The last sensor data received. */
  KeyStates keyStates; /**< The last key states received. */
  int idRobot; /**< The id of the robot currently connected. */
  AL::ALProxy* proxy; /**< The proxy to the DCM. */
  AL::ALValue* request; /**< The request sent to the DCM. */
  AL::ALValue* usRequest; /**< The request sent to the DCM for sonar readings. */
  ALMemoryFastAccess* memory; /**< Access to memory of the DCM. */
  std::vector<float> sensorValues; /**< The sensor values received from memory fast access. */
  int timeOffset; /**< Time offset between our time and NaoQi time. */
  unsigned lastUSMeasurement; /**< Last time when sonar was fired. */
  unsigned startUnpressed; /**< Last time when start button was not pressed. */
  bool shuttingDown; /**< Flag that indicates that Nao is shutting down. */
  RingBufferWithSum<double, 1000> bodyTiltMeasuredGyroY;
  RingBufferWithSum<double, 1000> bodyTiltMeasuredGyroX;

  void update(JointData& jointData);
  void update(SafeJointData& safeJointData);
  void update(SensorData& sensorData) {sensorData = this->sensorData;}
  void update(SafeSensorData& safeSensorData);
  void update(KeyStates& keyStates) {keyStates = this->keyStates;}
  void update(FrameInfo& frameInfo) {frameInfo.time = theJointData.timeStamp;}

  /**
  * The function sends a command to the CM-5.
  */
  void send();

public:
  /**
  * Constructor.
  */
  NaoProvider();

  /**
  * Destructor.
  */
  ~NaoProvider();

  /**
  * The method is called by process Motion to send the requests to the CM-5.
  */
  static void finishFrame();
};

#endif  //__NaoProvider_h_
