/** 
* @file TeamDataProvider.cpp
* This file implements a module that provides the data received by team communication.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "TeamDataProvider.h"
#include "Tools/Settings.h"
#include "Tools/Team.h"
#include "Tools/Debugging/ReleaseOptions.h"

PROCESS_WIDE_STORAGE TeamDataProvider* TeamDataProvider::theInstance = 0;

TeamDataProvider::TeamDataProvider() :
timeStamp(0),
robotNumber(-1),
lastSentTimeStamp(0),
gameControlTimeStamp(0)
{
  theInstance = this;
  memset(&gameControlData, 0, sizeof(gameControlData));  
  gameControlData.teams[0].teamNumber = Global::getSettings().teamNumber;
  gameControlData.teams[0].teamColor = Global::getSettings().teamColor;
  ASSERT(idGameControlData == 15); // if idGameControlData is changed, adapt it in GameController and here!
}

void TeamDataProvider::init()
{
  gameControlTimeStamp = theFrameInfo.time;
}

TeamDataProvider::~TeamDataProvider()
{
  theInstance = 0;
}

void TeamDataProvider::update(TeamMateData& teamMateData)
{
  teamMateData = theTeamMateData;
  teamMateData.numOfConnectedPlayers = 0;
  for(int i = TeamMateData::firstPlayer; i < TeamMateData::numOfPlayers; ++i)
    if(teamMateData.timeStamps[i])
    {
      if(SystemCall::getTimeSince(teamMateData.timeStamps[i]) < 500)
        ++teamMateData.numOfConnectedPlayers;
      else
        teamMateData.timeStamps[i] = 0;
    }
  if(teamMateData.numOfConnectedPlayers > 0)
    teamMateData.wasConnected = theTeamMateData.wasConnected = true;
  teamMateData.sendThisFrame = (ntp.doSynchronization(theRobotInfo.number, theFrameInfo.time, Global::getTeamOut()) ||
                             theFrameInfo.getTimeSince(lastSentTimeStamp) >= 100) && 
                             SystemCall::getTimeSince((theSafeSensorData.timeStamp)) < 500 && theRobotInfo.penalty == PENALTY_NONE; 
  if(teamMateData.sendThisFrame)
    lastSentTimeStamp = theFrameInfo.time;
}

void TeamDataProvider::update(RobotInfo& robotInfo)
{
  const RoboCup::TeamInfo& t = Global::getSettings().teamNumber == (int)gameControlData.teams[0].teamNumber 
    ? gameControlData.teams[0] : gameControlData.teams[1];
  ASSERT(Global::getSettings().playerNumber >= 1 && Global::getSettings().playerNumber <= 4);
  (RoboCup::RobotInfo&) robotInfo = t.players[Global::getSettings().playerNumber - 1];
  robotInfo.number         = Global::getSettings().playerNumber;
  robotInfo.robotModel = RobotInfo::getRobotModelByRobotName(theRobotName.name.c_str());
}

void TeamDataProvider::update(OwnTeamInfo& ownTeamInfo)
{
  if (SystemCall::getTimeSince(gameControlTimeStamp) < 500 || !&theBehaviorControlOutput || Global::getSettings().teamNumber != (int)theBehaviorControlOutput.ownTeamInfo.teamNumber) 
    (RoboCup::TeamInfo&) ownTeamInfo = Global::getSettings().teamNumber == (int)gameControlData.teams[0].teamNumber 
    ? gameControlData.teams[0] : gameControlData.teams[1];
  else if(&theBehaviorControlOutput && Global::getSettings().teamNumber == (int)theBehaviorControlOutput.ownTeamInfo.teamNumber)
    ownTeamInfo = theBehaviorControlOutput.ownTeamInfo;
  MODIFY("representation:OwnTeamInfo", ownTeamInfo);
  COMPLEX_DRAWING("field polygons",
    theFieldDimensions.drawPolygons(ownTeamInfo.teamColor);
  );
}

void TeamDataProvider::update(OpponentTeamInfo& opponentTeamInfo)
{
  (RoboCup::TeamInfo&) opponentTeamInfo = Global::getSettings().teamNumber == (int)gameControlData.teams[0].teamNumber 
    ? gameControlData.teams[1] : gameControlData.teams[0];
}

void TeamDataProvider::update(GameInfo& gameInfo)
{
  if (SystemCall::getTimeSince(gameControlTimeStamp) < 10000 || !&theBehaviorControlOutput || Global::getSettings().teamNumber != (int)theBehaviorControlOutput.ownTeamInfo.teamNumber) 
    memcpy(&(RoboCup::RoboCupGameControlData&) gameInfo, &gameControlData, (char*) gameControlData.teams - (char*) &gameControlData);
  else if(&theBehaviorControlOutput && Global::getSettings().teamNumber == (int)theBehaviorControlOutput.ownTeamInfo.teamNumber)
    gameInfo = theBehaviorControlOutput.gameInfo;
}

void TeamDataProvider::update(RobotPose& robotPose)
{
  robotPose = theGroundTruthRobotPose;
}

void TeamDataProvider::update(GroundTruthRobotPose& groundTruthRobotPose)
{
  groundTruthRobotPose = theGroundTruthRobotPose;
}

void TeamDataProvider::handleMessages(MessageQueue& teamReceiver)
{
  if(theInstance)
  {
    teamReceiver.handleAllMessages(*theInstance);
    TEAM_OUTPUT(idRobot, bin, theInstance->theRobotInfo.number);
  }

  teamReceiver.clear();
}

bool TeamDataProvider::handleMessage(InMessage& message)
{
  switch(message.getMessageID())
  {
  case idReceiveTimeStamp:
    message.bin >> timeStamp;
    return true;

  case idRobot:
    message.bin >> robotNumber;
    ntp.setReceiveTimeStamp(robotNumber, timeStamp);
    if(robotNumber != theRobotInfo.number) 
      if(robotNumber >= TeamMateData::firstPlayer && robotNumber < TeamMateData::numOfPlayers)
        theTeamMateData.timeStamps[robotNumber] = timeStamp;
    return true;

  case idSendTimeStamp:
    {
      unsigned timeStamp;
      message.bin >> timeStamp;
      ntp.setSendTimeStamp(robotNumber, timeStamp);
      return true;
    }

  case idReleaseOptions:
    message.bin >> Global::getReleaseOptions();
    return true;

  case idGameControlData:
    message.bin.read(&gameControlData, sizeof(gameControlData));
    gameControlTimeStamp = theFrameInfo.time;
    return true;

  case idTeamMateBallModel:
    if(robotNumber != theRobotInfo.number) 
      if(robotNumber >= TeamMateData::firstPlayer && robotNumber < TeamMateData::numOfPlayers)
      {
        BallModel& ballModel = theTeamMateData.ballModels[robotNumber];
        message.bin >> ballModel;

        if(ballModel.timeWhenLastSeen)
          ballModel.timeWhenLastSeen = ntp.getTeamMateTimeInOwnTime(ballModel.timeWhenLastSeen, robotNumber);
      }
    return true;

  case idTeamMateRobotPose:
    if(robotNumber != theRobotInfo.number) 
      if(robotNumber >= TeamMateData::firstPlayer && robotNumber < TeamMateData::numOfPlayers)
        message.bin >> theTeamMateData.robotPoses[robotNumber];
    return true;

  case idTeamMateBehaviorData:
    if(robotNumber != theRobotInfo.number) 
      if(robotNumber >= TeamMateData::firstPlayer && robotNumber < TeamMateData::numOfPlayers)
        message.bin >> theTeamMateData.behaviorData[robotNumber];
    return true;

  case idGroundTruthBallModel:
    {
      Vector2<double> position;
      message.bin >> theGroundTruthBallModel.timeWhenLastSeen >> position;
      theGroundTruthBallModel.timeWhenLastSeen = ntp.getTeamMateTimeInOwnTime(theGroundTruthBallModel.timeWhenLastSeen, 0);
      if(theOwnTeamInfo.teamColor == TEAM_BLUE)
        position *= -1;
      theGroundTruthBallModel.lastPerception.setPositionAndVelocityInFieldCoordinates(position, Vector2<double>(), 
                                                                                      theGroundTruthRobotPose);
      theGroundTruthBallModel.estimate = theGroundTruthBallModel.lastPerception;
    }
    return true;

  case idGroundTruthRobotPose:
    {
      char teamColor,
           id;
      unsigned timeStamp;
      Pose2D robotPose;
      message.bin >> teamColor >> id >> timeStamp >> robotPose;
      if(teamColor == (int)theOwnTeamInfo.teamColor && id == theRobotInfo.number)
      {
        if(theOwnTeamInfo.teamColor == TEAM_BLUE)
          robotPose = Pose2D(pi) + robotPose;
        (Pose2D&) theGroundTruthRobotPose = robotPose;
      }
    }
    return true;

  default:
    return ntp.handleMessage(message); 
  }
}

MAKE_MODULE(TeamDataProvider, Infrastructure)
