/**
* @file DribbleModelProvider.cpp
*
* This file implements a module that provides information about the current state of the robot's environment.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
*/

#include "DribbleModelProvider.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Streams/InStreams.h"

DribbleModelProvider::DribbleModelProvider()
{
  InConfigFile stream(Global::getSettings().expandLocationFilename("dribbleModelProvider.cfg"), "default");
  ASSERT(stream.exists());
  stream >> parameters;
  lastOdometry = theOdometryData;
  p1 = parameters.initialP1;
  p2 = parameters.initialP2;
  p3 = parameters.initialP3;
  timeLastSeenPole1 = 0;
  timeLastSeenPole2 = 0;
  timeLastSeenPole3 = 0;
}

void DribbleModelProvider::preExecution()
{
  odometryOffset = theOdometryData - lastOdometry;

  // Reset pole positions to start position (challenge restart)
  if(theGameInfo.state == STATE_PLAYING && 
    theKeyStates.pressed[KeyStates::left] && theKeyStates.pressed[KeyStates::right] && 
    !(theKeyStates.pressed[KeyStates::up]) && !(theKeyStates.pressed[KeyStates::down]))
  {
    p1 = parameters.initialP1;
    p2 = parameters.initialP2;
    p3 = parameters.initialP3;
  }
}


void DribbleModelProvider::postExecution()
{
  lastOdometry = theOdometryData;
}


void DribbleModelProvider::update(DribbleModel& dribbleModel)
{
  DECLARE_DEBUG_DRAWING("module:DribbleModelProvider:stuff", "drawingOnField", "Draws the internal representations of the dribble model provider");
  DECLARE_DEBUG_DRAWING("module:DribbleModelProvider:image", "drawingOnImage", "Draws the perception of the dribble model provider");
  preExecution();
  motionUpdate();
  collectSegments();
  mergeSegments(dribbleModel);
  generateModel(dribbleModel);
  postExecution();
  COMPLEX_DRAWING("module:DribbleModelProvider:stuff", draw());
}


void DribbleModelProvider::motionUpdate()
{
  p1 -= odometryOffset.translation;
  p2 -= odometryOffset.translation;
  p3 -= odometryOffset.translation;
  p1.rotate(-odometryOffset.rotation);
  p2.rotate(-odometryOffset.rotation); 
  p3.rotate(-odometryOffset.rotation);
}


void DribbleModelProvider::collectSegments()
{
  DEBUG_RESPONSE("module:DribbleModelProvider:currentsquaredist",;);

  percepts.clear();

  if(!theCameraMatrix.isValid)
    return;

  for(int i = int(thePointsPercept.points.size()) - 1; i >= 0; i--)
  {
    if((  thePointsPercept.points[i].type == PointsPercept::cyanBeacon
      || thePointsPercept.points[i].type == PointsPercept::magentaBeacon)
      && thePointsPercept.points[i].isCloser == false)
    {
      int numberOfMiniSegments(0);
      int pId = i;
      PointsPercept::Point pMax(thePointsPercept.points[i]);
      int lengthSum(0);
      while(thePointsPercept.points[i].link != i)
      {
        i = thePointsPercept.points[i].link;
        numberOfMiniSegments++;
        if(thePointsPercept.points[i].segmentId >= 0)
        {
          if(theSegmentsPercept.segmentsVertical[thePointsPercept.points[i].segmentId].onHalfScanLine)
            lengthSum += theSegmentsPercept.segmentsVertical[thePointsPercept.points[i].segmentId].length + parameters.minPixelHeight;
          else
            lengthSum += theSegmentsPercept.segmentsVertical[thePointsPercept.points[i].segmentId].length;
        }
      }
      if((numberOfMiniSegments >= 1 && lengthSum / numberOfMiniSegments > parameters.minPixelHeight)
         || lengthSum > parameters.minPixelHeight) // TODO
        percepts.push_back(pId);
    }
  }
}

void DribbleModelProvider::mergeSegments(DribbleModel& dribbleModel)
{
  int bestSquareLengthMagenta(0);
  int bestSquareLengthCyan1(0);
  int bestSquareLengthCyan2(0);
  PointsPercept::Point bestMinMagenta, bestMaxMagenta, bestMinCyan1, bestMaxCyan1, bestMinCyan2, bestMaxCyan2;

  // start merging of segments

  for(unsigned int i=0; i<percepts.size(); i++)
  {
    PointsPercept::Point pMax(thePointsPercept.points[percepts[i]]);
    int k = percepts[i];
    while(thePointsPercept.points[k].link != k)
    {
      k = thePointsPercept.points[k].link;
    }
    PointsPercept::Point pMin(thePointsPercept.points[k]);

    unsigned int j = i + 1;
    while(j < percepts.size())
    {
      if(abs(thePointsPercept.points[k].inImage.x - thePointsPercept.points[percepts[j]].inImage.x) < parameters.xMergingOffset
      && abs(thePointsPercept.points[k].inImage.y - thePointsPercept.points[percepts[j]].inImage.y) < parameters.yMergingOffset)
      {
        int l = percepts[j] ;
        while(thePointsPercept.points[l].link != l)
        {
          l = thePointsPercept.points[l].link;
        }

        if(thePointsPercept.points[l].inImage.x <= pMin.inImage.x)
        { 
          pMin = thePointsPercept.points[l];
          k = l;
        }
        j++;
      }
      else
        break;
    }

    /*
    for(unsigned int j=0; j<percepts.size(); j++)
    {
      if(i==j) continue;
      if(thePointsPercept.points[percepts[j]].type == thePointsPercept.points[percepts[i]].type)
      {
        if(abs(thePointsPercept.points[k].inImage.x - thePointsPercept.points[percepts[j]].inImage.x) < parameters.xMergingOffset
        && abs(thePointsPercept.points[k].inImage.y - thePointsPercept.points[percepts[j]].inImage.y) < parameters.yMergingOffset)
        {

          int l = percepts[j] ;
          while(thePointsPercept.points[l].link != l)
          {
            l = thePointsPercept.points[l].link;
          }

          if(thePointsPercept.points[l].inImage.x < pMin.inImage.x)
          { 
            pMin = thePointsPercept.points[l];   
          }
        }
      }
    }
    */
    
    int currentSquareDist = (pMax.onField - pMin.onField).squareAbs();
    DEBUG_RESPONSE("module:DribbleModelProvider:currentsquaredist",
      OUTPUT(idText, text, currentSquareDist);
    );
    if(currentSquareDist > parameters.minSquareDistance * parameters.minSquareDistance)
    {
      // TODO: Close enough check?
      if(pMax.type == PointsPercept::cyanBeacon) 
      { 
        if(currentSquareDist > bestSquareLengthCyan1)
        {
          bestSquareLengthCyan1 = currentSquareDist;
          bestMinCyan1 = pMin;
          bestMaxCyan1 = pMax;
        }
        else
        {
          bestSquareLengthCyan2 = currentSquareDist;
          bestMinCyan2 = pMin;
          bestMaxCyan2 = pMax;              
        }
      }
      else 
      {
        bestSquareLengthMagenta = currentSquareDist;
        bestMinMagenta = pMin;
        bestMaxMagenta = pMax;
      }
    }
  }

  bool p1mapped = false;
  Vector2<int> from;
  Vector2<int> to;
  if(bestSquareLengthMagenta)
  {
    COMPLEX_DRAWING("module:DribbleModelProvider:image",
     CIRCLE("module:DribbleModelProvider:image", bestMaxMagenta.inImage.x, bestMaxMagenta.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::magenta, Drawings::bs_solid, ColorClasses::magenta);
     CIRCLE("module:DribbleModelProvider:image", bestMinMagenta.inImage.x, bestMinMagenta.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::magenta, Drawings::bs_solid, ColorClasses::black);
     LINE("module:DribbleModelProvider:image", bestMaxMagenta.inImage.x, bestMaxMagenta.inImage.y, bestMinMagenta.inImage.x, bestMinMagenta.inImage.y, 2, Drawings::ps_solid, ColorClasses::magenta);
    );

    Vector2<double> from2 = thePointsPercept.origin * Vector2<double>(bestMaxMagenta.onField.x, bestMaxMagenta.onField.y);
    Vector2<double> to2   = thePointsPercept.origin * Vector2<double>(bestMinMagenta.onField.x, bestMinMagenta.onField.y);

    Vector2<double> center = to2 + (from2-to2)/2;
    Vector2<double> ext(50,0);
    ext.rotate(p2.angle());
    center += ext;
    if(dribbleModel.timeSincePoleLastSeen2 < parameters.blurMaxTime)
      p2 += (center-p2)/parameters.blurFactor;
    else
      p2 = center;
    timeLastSeenPole2 = theFrameInfo.time; 
  }
  if(bestSquareLengthCyan1)
  {
    COMPLEX_DRAWING("module:DribbleModelProvider:image",
     CIRCLE("module:DribbleModelProvider:image", bestMaxCyan1.inImage.x, bestMaxCyan1.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::cyan, Drawings::bs_solid, ColorClasses::cyan);
     CIRCLE("module:DribbleModelProvider:image", bestMinCyan1.inImage.x, bestMinCyan1.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::cyan, Drawings::bs_solid, ColorClasses::black);
     LINE("module:DribbleModelProvider:image", bestMaxCyan1.inImage.x, bestMaxCyan1.inImage.y, bestMinCyan1.inImage.x, bestMinCyan1.inImage.y, 2, Drawings::ps_solid, ColorClasses::cyan);
    );

    Vector2<double> from2 = thePointsPercept.origin * Vector2<double>(bestMaxCyan1.onField.x, bestMaxCyan1.onField.y);
    Vector2<double> to2   = thePointsPercept.origin * Vector2<double>(bestMinCyan1.onField.x, bestMinCyan1.onField.y);

    Vector2<double> center = to2 + (from2-to2)/2;
    Vector2<double> ext(50,0);
    ext.rotate(center.angle());
    center += ext;
    double distToP1 = (center-p1).abs();
    double distToP3 = (center-p3).abs();
    if(distToP1 < distToP3) 
    {
      if(dribbleModel.timeSincePoleLastSeen1 < parameters.blurMaxTime)
        p1 += (center-p1)/parameters.blurFactor;
      else
        p1 = center;
      p1mapped = true;
      timeLastSeenPole1 = theFrameInfo.time; 
    }
    else
    {
      if(dribbleModel.timeSincePoleLastSeen3 < parameters.blurMaxTime)
        p3 += (center-p3)/parameters.blurFactor;
      else
        p3 = center;
      timeLastSeenPole3 = theFrameInfo.time; 
    }

    if(bestSquareLengthCyan2 && bestMaxCyan2.inImage.x != bestMaxCyan1.inImage.x && bestMinCyan1.inImage.x != bestMinCyan2.inImage.x)
    {
      COMPLEX_DRAWING("module:DribbleModelProvider:image",
       CIRCLE("module:DribbleModelProvider:image", bestMaxCyan2.inImage.x, bestMaxCyan2.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::cyan, Drawings::bs_solid, ColorClasses::cyan);
       CIRCLE("module:DribbleModelProvider:image", bestMinCyan2.inImage.x, bestMinCyan2.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::cyan, Drawings::bs_solid, ColorClasses::black);
       LINE("module:DribbleModelProvider:image", bestMaxCyan2.inImage.x, bestMaxCyan2.inImage.y, bestMinCyan2.inImage.x, bestMinCyan2.inImage.y, 2, Drawings::ps_solid, ColorClasses::black);
      );

      Vector2<double> from2 = thePointsPercept.origin * Vector2<double>(bestMaxCyan2.onField.x, bestMaxCyan2.onField.y);
      Vector2<double> to2   = thePointsPercept.origin * Vector2<double>(bestMinCyan2.onField.x, bestMinCyan2.onField.y);
      
      Vector2<double> center2 = to2 + (from2-to2)/2;
      Vector2<double> ext(50,0);
      ext.rotate(center2.angle());
      center2 += ext;

      // Do another check: p1 and p3 position should be far away from each other (> 1.2m)
      if(fabs(center.abs()-center2.abs()) > 1200.0)
      {
        if(!p1mapped) 
        {
          if(dribbleModel.timeSincePoleLastSeen1 < parameters.blurMaxTime)
            p1 += (center2-p1)/parameters.blurFactor;
          else
            p1 = center2;
          timeLastSeenPole3 = theFrameInfo.time; 
        }
        else
        {
          if(dribbleModel.timeSincePoleLastSeen3 < parameters.blurMaxTime)
            p3 += (center2-p3)/parameters.blurFactor;
          else
            p3 = center2;
          timeLastSeenPole3 = theFrameInfo.time; 
        }
      }
    }
  }
}

void DribbleModelProvider::generateModel(DribbleModel& dribbleModel)
{
  dribbleModel.timeSincePoleLastSeen1 = theFrameInfo.getTimeSince(timeLastSeenPole1);
  dribbleModel.timeSincePoleLastSeen2 = theFrameInfo.getTimeSince(timeLastSeenPole2);
  dribbleModel.timeSincePoleLastSeen3 = theFrameInfo.getTimeSince(timeLastSeenPole3);
  dribbleModel.angleToPole1 = toDegrees(p1.angle());
  dribbleModel.angleToPole2 = toDegrees(p2.angle());
  dribbleModel.angleToPole3 = toDegrees(p3.angle());
  dribbleModel.distanceToPole1 = p1.abs();
  dribbleModel.distanceToPole2 = p2.abs();
  dribbleModel.distanceToPole3 = p3.abs();
  Vector2<double> pass1 = p2 + (p1-p2)/2;
  Vector2<double> pass2 = p3 + (p2-p3)/2;
  dribbleModel.angleToPass1 = toDegrees(pass1.angle());
  dribbleModel.distanceToPass1 = pass1.abs();
  dribbleModel.angleToPass2 = toDegrees(pass2.angle());
  dribbleModel.distanceToPass2 = pass2.abs();
  if (fabs(dribbleModel.angleToPole1) < parameters.nearAngle && dribbleModel.distanceToPole1 < parameters.nearDist)
  {
    dribbleModel.poleIsInFront = true;
    if(dribbleModel.angleToPole1 < 0)
      dribbleModel.goLeftAroundPole = false;
  }
  else if (fabs(dribbleModel.angleToPole2) < parameters.nearAngle && dribbleModel.distanceToPole2 < parameters.nearDist)
  {
    dribbleModel.poleIsInFront = true;
    if(dribbleModel.angleToPole2 < 0)
      dribbleModel.goLeftAroundPole = false;
  }
  else if (fabs(dribbleModel.angleToPole3) < parameters.nearAngle && dribbleModel.distanceToPole3 < parameters.nearDist)
  {
    dribbleModel.poleIsInFront = true;
    if(dribbleModel.angleToPole3 < 0)
      dribbleModel.goLeftAroundPole = false;
  }
  else
    dribbleModel.poleIsInFront = false;
}


void DribbleModelProvider::draw()
{
  Vector2<double> onField1(p1);
  onField1.rotate(theGroundTruthRobotPose.rotation);
  onField1 = onField1 + theGroundTruthRobotPose.translation;
  Vector2<double> onField2(p2);
  onField2.rotate(theGroundTruthRobotPose.rotation);
  onField2 = onField2 + theGroundTruthRobotPose.translation;
  Vector2<double> onField3(p3);
  onField3.rotate(theGroundTruthRobotPose.rotation);
  onField3 = onField3 + theGroundTruthRobotPose.translation;
  CIRCLE("module:DribbleModelProvider:stuff", onField1.x, onField1.y, 70, 2, Drawings::ps_solid, ColorClasses::cyan, Drawings::bs_solid, ColorClasses::cyan);
  CIRCLE("module:DribbleModelProvider:stuff", onField2.x, onField2.y, 70, 2, Drawings::ps_solid, ColorClasses::magenta, Drawings::bs_solid, ColorClasses::magenta);
  CIRCLE("module:DribbleModelProvider:stuff", onField3.x, onField3.y, 70, 2, Drawings::ps_solid, ColorClasses::black, Drawings::bs_solid, ColorClasses::black);
  Vector2<double> pass1 = p2 + (p1-p2)/2;
  Vector2<double> pass2 = p3 + (p2-p3)/2; 
  pass1.rotate(theGroundTruthRobotPose.rotation);
  pass2.rotate(theGroundTruthRobotPose.rotation);
  pass1 += theGroundTruthRobotPose.translation;
  pass2 += theGroundTruthRobotPose.translation;
  LINE("module:DribbleModelProvider:stuff", theGroundTruthRobotPose.translation.x, theGroundTruthRobotPose.translation.y, pass1.x, pass1.y, 2, Drawings::ps_solid, ColorClasses::orange);
  LINE("module:DribbleModelProvider:stuff", theGroundTruthRobotPose.translation.x, theGroundTruthRobotPose.translation.y, pass2.x, pass2.y, 2, Drawings::ps_solid, ColorClasses::orange);
}

MAKE_MODULE(DribbleModelProvider, Modeling)
