/**
* @file DribbleModelProvider.h
*
* This file declares a module that provides information about the current state of the robot's environment.
*
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
*/

#ifndef __DribbleModelProvider_h_
#define __DribbleModelProvider_h_

#include "Representations/Perception/SegmentsPercept.h"
#include "Representations/Perception/PointsPercept.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/CameraInfo.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Modeling/DribbleModel.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Infrastructure/GameInfo.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Tools/Module/Module.h"
#include "Tools/Debugging/Debugging.h"
#include "Tools/Math/Common.h"


MODULE(DribbleModelProvider)
  REQUIRES(SegmentsPercept)
  REQUIRES(PointsPercept)
  REQUIRES(OdometryData)
  REQUIRES(RobotPose)
  REQUIRES(GroundTruthRobotPose)
  REQUIRES(FrameInfo)
  REQUIRES(CameraInfo)
  REQUIRES(ImageCoordinateSystem)
  REQUIRES(CameraMatrix)
  REQUIRES(KeyStates)
  REQUIRES(GameInfo)
  PROVIDES_WITH_MODIFY(DribbleModel)
END_MODULE


/**
* @class DribbleModelProvider
*
* A module for computing the current state of the robot's environment
*/
class DribbleModelProvider: public DribbleModelProviderBase
{

public:
  /** Standard constructor */
  DribbleModelProvider();

private:

  class Parameters : public Streamable
  {
  private:
    /**
    * The method makes the object streamable.
    * @param in The stream from which the object is read 
    * @param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {  
      STREAM_REGISTER_BEGIN();
      STREAM(minPixelHeight);
      STREAM(xMergingOffset);
      STREAM(yMergingOffset);
      STREAM(minSquareDistance);
      STREAM(maxSquareDistance);
      STREAM(blurFactor);
      STREAM(blurMaxTime);
      STREAM(initialP1);
      STREAM(initialP2);
      STREAM(initialP3);
      STREAM(nearAngle);
      STREAM(nearDist);
      STREAM_REGISTER_FINISH();
    }

  public:
    int minPixelHeight;
    int xMergingOffset;
    int yMergingOffset;
    int minSquareDistance;
    int maxSquareDistance;
    int blurFactor;
    int blurMaxTime;
    Vector2<double> initialP1;
    Vector2<double> initialP2;
    Vector2<double> initialP3;
    double nearAngle;
    double nearDist;
  };

  /** Executes this module
  * @param dribbleModel The data structure that is filled by this module
  */
  void update(DribbleModel& dribbleModel);

  /** Updates contents of model according to robot motion*/
  void motionUpdate();

  void collectSegments();

  /** Find positive and negative segments from PointsPercept
  * @param dribbleModel The model
  */
  void mergeSegments(DribbleModel& dribbleModel);

  /** Generates model content from internal representations
  * @param dribbleModel The model
  */
  void generateModel(DribbleModel& dribbleModel);

  /** Prepares execution, initializes some values*/
  void preExecution();

  /** Sets some values after everything else of this module was executed*/
  void postExecution();

  /** Draws the internal representation*/
  void draw();

  Pose2D lastOdometry; /** save last odometry state*/
  Pose2D odometryOffset; /** odometry offset since last execution*/

  Parameters parameters; /**< The parameters of this module. */

  Vector2<double> p1; //cyan1
  Vector2<double> p2; //magenta
  Vector2<double> p3; //cyan2

  int timeLastSeenPole1;
  int timeLastSeenPole2;
  int timeLastSeenPole3;
  std::vector<int> percepts;
};

#endif// __DribbleModelProvider_h_
