/**
* @file FreeSpaceModelProvider.cpp
*
* This file implements a module that provides information about the current state of the robot's environment.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
*/

#include "FreeSpaceModelProvider.h"
#include "Tools/Math/Geometry.h"


FreeSpaceModelProvider::FreeSpaceModelProvider()
{
  // init internal model
  for(int i=0; i<MAX_SEGMENTS; i++)
  {
    FreeSpaceSegment seg;
    seg.distance = DEFAULT_DISTANCE;
    seg.lastSeen = 0;
    seg.safe =true;
    segments[i] = seg;
    segmentsWorking[i] = seg;
  }

  // calculate angles
  segmentWidth = fromDegrees(COVERED_ANGLE / MAX_SEGMENTS);
  baseAngle    = fromDegrees(COVERED_ANGLE / 2);

  timeSinceFrontIsFree = theFrameInfo.time;

  for(int i = 0; i < MAX_SEGMENTS / 2; i++)
  {
    sinus[i] = int(sin((i + 0.5) * segmentWidth) * (1 << TRIGONOMETRY_POWER_FACTOR));
    cosinus[i] = int(cos((i + 0.5) * segmentWidth) * (1 << TRIGONOMETRY_POWER_FACTOR));
  }
}

void FreeSpaceModelProvider::preExecution()
{
  xOrigin = int(thePointsPercept.origin.translation.x);
  yOrigin = int(thePointsPercept.origin.translation.y);
  rotationCos = int(cos(thePointsPercept.origin.rotation) * (1 << 10));
  rotationSin = int(sin(thePointsPercept.origin.rotation) * (1 << 10));
  odometryOffset = theOdometryData - lastOdometry;
  interpolatedGoalAngleBuffer.add(int(toDegrees(normalize(theGoalModel.largestFreePart.getCenter()))));
  interpolatedGoalAngle = interpolatedGoalAngleBuffer.getAverage();
}


void FreeSpaceModelProvider::postExecution()
{
  lastOdometry = theOdometryData;
}


void FreeSpaceModelProvider::update(FreeSpaceModel& freeSpaceModel)
{
  preExecution();
  motionUpdate();
  collectSegments();
  generateModel(freeSpaceModel);
  postExecution();
  DECLARE_DEBUG_DRAWING("module:FreeSpaceModelProvider:stuff", "drawingOnField", "Draws the internal representations of the free space model provider");
  DECLARE_DEBUG_DRAWING("module:FreeSpaceModelProvider:stuff2", "drawingOnField", "Draws the internal representations of the free distances of some angles");
  COMPLEX_DRAWING("module:FreeSpaceModelProvider:stuff", draw());
  COMPLEX_DRAWING("module:FreeSpaceModelProvider:stuff2", draw2());
}


void FreeSpaceModelProvider::motionUpdate()
{
  // Update rotation
  odometryAngleSum += odometryOffset.rotation;

  // Shift right
  if (odometryAngleSum > segmentWidth)
  {
    for(int i = 0; i < MAX_SEGMENTS-1; ++i)
    {
      segments[i].lastSeen = segments[i+1].lastSeen;
      segments[i].distance = segments[i+1].distance;
      segments[i].safe = segments[i+1].safe;
    }
    segments[MAX_SEGMENTS-1].distance = DEFAULT_DISTANCE;
    segments[MAX_SEGMENTS-1].lastSeen = theFrameInfo.time;
    odometryAngleSum -= segmentWidth;
  }
  // Shift left
  else if (odometryAngleSum < -segmentWidth)
  {
    for(int i = MAX_SEGMENTS-1; i > 0; --i)
    {
      segments[i].lastSeen = segments[i-1].lastSeen;
      segments[i].distance = segments[i-1].distance;
      segments[i].safe = segments[i-1].safe;
    }
    segments[0].distance = DEFAULT_DISTANCE;
    segments[0].lastSeen = theFrameInfo.time;
    odometryAngleSum += segmentWidth;
  }

  // Update translation
  double currentAngle = -baseAngle;

  for(int i = 0; i < MAX_SEGMENTS; ++i)
  {
    if (segments[i].distance != DEFAULT_DISTANCE)
    {
      Vector2<double> m((double) segments[i].distance, 0.0);
      m = Geometry::rotate(m, currentAngle + segmentWidth / 2);
      m -= odometryOffset.translation;
      segments[i].distance = (int) m.abs();
    }
    currentAngle += segmentWidth;
  }
}


void FreeSpaceModelProvider::collectSegments()
{
  // handle "timeouts"
  for(unsigned int i = 0; i < MAX_SEGMENTS; ++i)
  {
    if (theFrameInfo.getTimeSince(segments[i].lastSeen) > MAX_TIME)
    {
      segments[i].distance = DEFAULT_DISTANCE;
      segments[i].lastSeen = theFrameInfo.time;
      segments[i].safe     = true;
    }
    segmentsWorking[i].distance = DEFAULT_DISTANCE;
    segmentsWorking[i].lastSeen = 0;
    segmentsWorking[i].safe     = true;
  }

  // if the current camera matrix is not valid
  // we cannot update the internal model
  if(theCameraMatrix.isValid)
  {
    for(unsigned int i = 0; i < thePointsPercept.points.size(); ++i)
    {
      PointsPercept::Point p = thePointsPercept.points.at(i);
      if(p.type == PointsPercept::robot)
      {
        Vector2<int> v (xOrigin + ((p.onField.x * rotationCos - p.onField.y * rotationSin) >> 10),
                        yOrigin + ((p.onField.x * rotationSin + p.onField.y * rotationCos) >> 10));
        // map angle to segmentIndex
        int segmentNumber = (int) (((v.angle() + baseAngle) / segmentWidth));

        // if the point lies in the covered area
        // which may not be the case when the point is behind the robot
        if(segmentNumber >= 0 && segmentNumber < MAX_SEGMENTS)
        {
          int vectorLength = v.abs();
          
          // find the minimum distance in every segment
          if(segmentsWorking[segmentNumber].distance > vectorLength
            || segmentsWorking[segmentNumber].distance == DEFAULT_DISTANCE)
          {
            segmentsWorking[segmentNumber].distance = vectorLength;
            // if the point lies at the bottom of the image we cannot say whether
            // it is the base point and whether the distance computation is correct
            if(p.isCloser)
              segmentsWorking[segmentNumber].safe = false;
            else
              segmentsWorking[segmentNumber].safe = true;
          }
          segmentsWorking[segmentNumber].lastSeen = theFrameInfo.time;
        }
      }
    }

    // this is not the finest method for computing the currently visible angles
    // but for the moment it will work
    double leftAngle  =  (theCameraInfo.openingAngleWidth / 2) + thePointsPercept.origin.rotation;
    double rightAngle = -(theCameraInfo.openingAngleWidth / 2) + thePointsPercept.origin.rotation;
    if(rightAngle < -baseAngle)
      rightAngle = -baseAngle;
    if(leftAngle > baseAngle)
      leftAngle = baseAngle;
    int leftBound  = (int) (((leftAngle + baseAngle) / segmentWidth));
    int rightBound = (int) (((rightAngle + baseAngle) / segmentWidth));

    // Calculate distance to horizon
    int cameraHeight = int(theCameraMatrix.translation.z + 0.5);
    double pan = theCameraMatrix.rotation.getZAngle();
    if(pan > 2)
      pan -= pi;
    else if(pan < -2)
      pan += pi;
    RotationMatrix r(theCameraMatrix.rotation.invert());
    r.rotateZ(pan);
    double tilt = r.getYAngle();
    r.rotateY(-tilt);
    double roll = r.getXAngle();
    double factor = theCameraInfo.focalLengthInv * (1 << 20);
    int rotCos = int(cos(-roll) * factor + 0.5);
    int rotSin = int(sin(-roll) * factor + 0.5);
    int tiltCos2 = int(cos(-tilt) * (1 << 20) + 0.5);
    int tiltSin2 = int(sin(-tilt) * (1 << 20) + 0.5);
    int tiltCos = tiltCos2 >> 10;
    int tiltSin = tiltSin2 >> 10;

    // if we are looking high enough and the obstacle is not visible
    int yHorizon = theImageCoordinateSystem.fromHorizonBased(Vector2<double>()).y,
        yHorizonClipped = std::min(theCameraInfo.resolutionHeight - 1, std::max(0, yHorizon)),
        yHorizonCorrected = (int) theImageCoordinateSystem.toCorrected(Vector2<int>(0, yHorizonClipped)).y;

    yHorizon += yHorizonClipped - yHorizonCorrected;
    yHorizon = std::max(0, yHorizon);


    Vector2<int> p;
    Vector2<int> v = theImageCoordinateSystem.toCorrectedCenteredNeg((int) theCameraInfo.resolutionWidth / 2, yHorizon);
    int b0 = (v.x * rotSin + v.y * rotCos) >> 10,
      b1 = tiltCos2 + b0 * tiltSin,
      b2 = v.x * rotCos - v.y * rotSin,
      b3 = tiltSin2 - b0 * tiltCos;
    if (b3 == 0)
      b3 = 1;
    p.x = b1 * cameraHeight / b3;
    p.y = b2 * cameraHeight / b3;
    Vector2<int> p2 (xOrigin + ((p.x * rotationCos - p.y * rotationSin) >> 10),
                     yOrigin + ((p.x * rotationSin + p.y * rotationCos) >> 10));
    int distanceToHorizonCenter = (int) p2.abs();    

    // update internal model
    for(int i = 0; i < MAX_SEGMENTS; ++i)
    {
      // update the segment only if we see something in this segment in this frame
      if(segmentsWorking[i].distance != DEFAULT_DISTANCE)
      {
        if(segments[i].distance != DEFAULT_DISTANCE)
        {
          if(segmentsWorking[i].safe)
          {
            int increaseValue = MAX_INCREASE;
            int increaseMax   = abs(segmentsWorking[i].distance - segments[i].distance);
            if(increaseValue > increaseMax)
              increaseValue = increaseMax;

            if(segmentsWorking[i].distance > segments[i].distance)
              segments[i].distance += increaseValue;
            else
              segments[i].distance -= increaseValue;

            segments[i].lastSeen = segmentsWorking[i].lastSeen;
            segments[i].safe = segmentsWorking[i].safe;
          }
        }
        else
        {
          segments[i].distance = segmentsWorking[i].distance;
          segments[i].lastSeen = segmentsWorking[i].lastSeen;
          segments[i].safe = segmentsWorking[i].safe;
        }
      }
      // if we do not see anything in this segment test if we are looking high enough
      else if(i <= (leftBound + 1) && i >= (rightBound)   // TODO
              && segments[i].distance != DEFAULT_DISTANCE)
      {
        double currentAngle   = ((i+0.5) * segmentWidth) - baseAngle;
        int distanceToHorizon = (int) (distanceToHorizonCenter * cos(currentAngle));

        if(distanceToHorizon > segments[i].distance + 50)
        {
          segments[i].lastSeen = theFrameInfo.time;
          segments[i].distance = DEFAULT_DISTANCE;
          segments[i].safe     = true;
        }
      }
    }
  } 
}

void FreeSpaceModelProvider::generateModel(FreeSpaceModel& freeSpaceModel)
{
  // freeDistanceInFrontOfMe2 (experimental)
  // calculates the free distance of a number of angles within a specific width (with averaging)
  int maximalWeighting = 0;

  const double safetyDist = 0.34906585039886590;

  for(int i = 0; i < MAX_VERTICAL_SPACE_SEGMENTS; i++)
  {
    int minimalVerticalFreeSpacePerSegment = 0;
    for(int j = 0; j < MAX_SEGMENTS_PER_VERTICAL_SPACE; j++)
    {
      int segment = (MAX_SEGMENTS / 2) - (MAX_VERTICAL_SPACE_SEGMENTS / 2) + i - (MAX_SEGMENTS_PER_VERTICAL_SPACE / 2) + j;
      int relativeSegment = j - (MAX_SEGMENTS_PER_VERTICAL_SPACE / 2);
      int horizontalDistance = abs((segments[segment].distance * sin2(relativeSegment)) >> TRIGONOMETRY_POWER_FACTOR);
      int verticalFreeSpace = 0;
      if (segments[segment].safe) 
        verticalFreeSpace = (horizontalDistance < MAX_HORIZONTAL_DIFFERENCE) ? (segments[segment].distance * cos2(relativeSegment)) >> TRIGONOMETRY_POWER_FACTOR : DEFAULT_DISTANCE;
      else
        verticalFreeSpace = (horizontalDistance < MAX_HORIZONTAL_DIFFERENCE) ? (MIN_DISTANCE * cos2(relativeSegment)) >> TRIGONOMETRY_POWER_FACTOR : DEFAULT_DISTANCE;
      if(j == 0 || verticalFreeSpace < minimalVerticalFreeSpacePerSegment)
        minimalVerticalFreeSpacePerSegment = verticalFreeSpace;
    }
    verticalSpacePerSegment[i].add(minimalVerticalFreeSpacePerSegment);
    int averageVerticalSpace = verticalSpacePerSegment[i].getAverage();
    int weightingPerSegment = averageVerticalSpace - abs(i - MAX_VERTICAL_SPACE_SEGMENTS / 2) * 45 - abs((i - MAX_VERTICAL_SPACE_SEGMENTS / 2) * (COVERED_ANGLE / MAX_SEGMENTS) - interpolatedGoalAngle) * 35;
    if(i == 0 || weightingPerSegment > maximalWeighting)
    {
      maximalWeighting = weightingPerSegment;    
      bestSegment = i;
    }
  }

  if(theGoalModel.largestFreePart.max > safetyDist 
    && theGoalModel.largestFreePart.min < -safetyDist
    && std::min(verticalSpacePerSegment[MAX_VERTICAL_SPACE_SEGMENTS / 2].getAverage(), verticalSpacePerSegment[MAX_VERTICAL_SPACE_SEGMENTS / 2 - 1].getAverage()) > 1000
  )
  {
    bestSegment = MAX_VERTICAL_SPACE_SEGMENTS / 2;
    freeSpaceModel.freeGoalPartInFrontOfMe = true;
  }
  else
  {
     freeSpaceModel.freeGoalPartInFrontOfMe = false;
  }

  double bestAngle = (bestSegment - MAX_VERTICAL_SPACE_SEGMENTS / 2) * (COVERED_ANGLE / MAX_SEGMENTS);

  freeSpaceModel.timeSinceFrontIsFree = theFrameInfo.getTimeSince(timeSinceFrontIsFree);
  freeSpaceModel.freeDistanceInFrontOfMe2 = std::min(verticalSpacePerSegment[MAX_VERTICAL_SPACE_SEGMENTS / 2].getAverage(), verticalSpacePerSegment[MAX_VERTICAL_SPACE_SEGMENTS / 2 - 1].getAverage());
  if(freeSpaceModel.frontIsFree2 == false && freeSpaceModel.freeDistanceInFrontOfMe2 > MIN_DISTANCE)
    timeSinceFrontIsFree = theFrameInfo.time;
  else if(freeSpaceModel.frontIsFree2 == true && !(freeSpaceModel.freeDistanceInFrontOfMe2 > MIN_DISTANCE))
    freeSpaceModel.bestObstacleAvoidingAngle = bestAngle;
  freeSpaceModel.frontIsFree2 = freeSpaceModel.freeDistanceInFrontOfMe2 > MIN_DISTANCE;
  freeSpaceModel.bestAngle = bestAngle;
}


void FreeSpaceModelProvider::draw()
{
  double currentAngle(-baseAngle);
  ColorRGBA col(100,255,100);
  ColorRGBA col2(200,0,0);

  for(int i=0; i< MAX_SEGMENTS; i++)
  {
    ColorRGBA color = segments[i].safe ? col : col2;
    Vector2<double> l(segments[i].distance, 0.0);
    l = Geometry::rotate(l, currentAngle + theGroundTruthRobotPose.rotation);
    l += theGroundTruthRobotPose.translation;
    currentAngle += segmentWidth;
    Vector2<double> r(segments[i].distance, 0.0);
    r = Geometry::rotate(r, currentAngle + theGroundTruthRobotPose.rotation);
    r += theGroundTruthRobotPose.translation;
    LINE("module:FreeSpaceModelProvider:stuff", theGroundTruthRobotPose.translation.x, theGroundTruthRobotPose.translation.y,
      r.x, r.y,20,Drawings::ps_solid, color);
    LINE("module:FreeSpaceModelProvider:stuff", theGroundTruthRobotPose.translation.x, theGroundTruthRobotPose.translation.y,
      l.x, l.y,20,Drawings::ps_solid, color);
    LINE("module:FreeSpaceModelProvider:stuff", r.x, r.y,
      l.x, l.y,20,Drawings::ps_solid, color);
  }
}

void FreeSpaceModelProvider::draw2()
{
  ColorRGBA black(0, 0, 0);
  ColorRGBA red(255, 0, 0);

  for(int i = 0; i < MAX_VERTICAL_SPACE_SEGMENTS; i++)
  {
    double angle = (i - MAX_VERTICAL_SPACE_SEGMENTS / 2) * segmentWidth + theGroundTruthRobotPose.rotation;
    Vector2<double> freeSpace(verticalSpacePerSegment[i].getAverage(), 0.0);
    ColorRGBA tempColor = (i == bestSegment) ? red : black;
    freeSpace = Geometry::rotate(freeSpace, angle);
    freeSpace += theGroundTruthRobotPose.translation;
    Vector2<double> offset(MAX_HORIZONTAL_DIFFERENCE * sin(angle), MAX_HORIZONTAL_DIFFERENCE * cos(angle));
    LINE("module:FreeSpaceModelProvider:stuff2", freeSpace.x + offset.x, freeSpace.y - offset.y,
      freeSpace.x - offset.x, freeSpace.y + offset.y, 20, Drawings::ps_solid, tempColor);
  }
}


MAKE_MODULE(FreeSpaceModelProvider, Modeling)
