/**
* @file FreeSpaceModelProvider.h
*
* This file declares a module that provides information about the current state of the robot's environment.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
*/

#ifndef __FreeSpaceModelProvider_h_
#define __FreeSpaceModelProvider_h_

#include "Representations/Modeling/FreeSpaceModel.h"
#include "Representations/Perception/PointsPercept.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/CameraInfo.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Modeling/GoalModel.h"
#include "Tools/Module/Module.h"
#include "Tools/Math/Common.h"
#include "Tools/RingBufferWithSum.h"


MODULE(FreeSpaceModelProvider)
  REQUIRES(PointsPercept)
  REQUIRES(OdometryData)
  REQUIRES(RobotPose)
  REQUIRES(GroundTruthRobotPose)
  REQUIRES(FrameInfo)
  REQUIRES(CameraInfo)
  REQUIRES(GoalModel)
  REQUIRES(ImageCoordinateSystem)
  REQUIRES(CameraMatrix)
  PROVIDES_WITH_MODIFY(FreeSpaceModel)
END_MODULE


/**
* @class FreeSpaceModelProvider
*
* A module for computing the current state of the robot's environment
*/
class FreeSpaceModelProvider: public FreeSpaceModelProviderBase
{
public:
  /** Standard constructor */
  FreeSpaceModelProvider();

private:
  /** Executes this module
  * @param freeSpaceModel The data structure that is filled by this module
  */
  void update(FreeSpaceModel& freeSpaceModel);

  /** Updates contents of model according to robot motion*/
  void motionUpdate();

  /** Find positive and negative segments from PointsPercept*/
  void collectSegments();

  /** Generates model content from internal representations
  * @param freeSpaceModel The model
  */
  void generateModel(FreeSpaceModel& freeSpaceModel);

  /** Prepares execution, initializes some values*/
  void preExecution();

  /** Sets some values after everything else of this module was executed*/
  void postExecution();

  /** Draws the internal representation*/
  void draw();

  /** Draws the other internal representation*/
  void draw2();

  /** Returns the sinus of a specific segment shifted 
  * by TRIGONOMETRY_FACTOR (to avoid floating point operations)
  * @param segment The segment relative to the middle
  */
  int sin2(int segment) {return sinus[abs(segment + (segment < 0 ? 1 : 0))] * sgn(segment);}

  /** Returns the cosinus of a specific segment shifted 
  * by TRIGONOMETRY_FACTOR (to avoid floating point operations)
  * @param segment The segment relative to the middle
  */
  int cos2(int segment) {return cosinus[abs(segment + (segment < 0 ? 1 : 0))];}

  /** Some constant parameters*/
  enum {MAX_SEGMENTS = 30, DEFAULT_DISTANCE = 1500, COVERED_ANGLE = 240, MAX_TIME = 10000, MAX_INCREASE = 40, MAX_VERTICAL_SPACE_SEGMENTS = 10, MAX_SEGMENTS_PER_VERTICAL_SPACE = 10, MAX_HORIZONTAL_DIFFERENCE = 200, TRIGONOMETRY_POWER_FACTOR = 15, MIN_DISTANCE = 450, SAFE_DISTANCE = 550};

  struct FreeSpaceSegment
  {
    int distance;
    unsigned lastSeen;
    bool safe;
  };

  /** The segments which contain the distance information*/
  FreeSpaceSegment segments[MAX_SEGMENTS];
  FreeSpaceSegment segmentsWorking[MAX_SEGMENTS];

  Pose2D lastOdometry;
  Pose2D odometryOffset;
  double odometryAngleSum;
  double baseAngle;
  double segmentWidth;
  
  RingBufferWithSum<int, 15> verticalSpacePerSegment[MAX_VERTICAL_SPACE_SEGMENTS];
  RingBufferWithSum<int, 15> interpolatedGoalAngleBuffer;
  int bestSegment;
  int interpolatedGoalAngle;
  int sinus[MAX_SEGMENTS / 2];
  int cosinus[MAX_SEGMENTS / 2];

  int xOrigin;
  int yOrigin;
  int rotationCos;
  int rotationSin;

  int timeSinceFrontIsFree;
};

#endif// __FreeSpaceModelProvider_h_
