/**
* @file GoalLocatorLight.cpp
*
* Simple GoalLocator implementation.
* Computes free part only.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#include "GoalLocatorLight.h"
#include "Tools/Math/AngleTables.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Team.h"
#include <limits>

using namespace std;


GoalLocatorLight::GoalLocatorLight()
 : lastExecution(0), lastSeenGoalTime(0)
{
  hypotheses.reserve(MAX_HYPOTHESES);
  InConfigFile stream(Global::getSettings().expandLocationFilename("goalLocator.cfg"), "default");
  ASSERT(stream.exists());
  stream >> parameters;
}


void GoalLocatorLight::init()
{
  lastOdometry = theOdometryData;
}


void GoalLocatorLight::update(GoalModel& goalModel)
{
  DECLARE_DEBUG_DRAWING("module:GoalLocatorLight:stuff", "drawingOnField", "Draws the internal representations of the goal locator");
  DECLARE_DEBUG_DRAWING("module:GoalLocatorLight:image", "drawingOnImage", "Draws the internal representations of the goal locator");
  preExecution();
  clearOldPerceptsAndHypotheses();
  motionUpdate();
  addNewPercepts();
  mergeHypotheses();
  modelGeneration(goalModel);
  postExecution(goalModel);
  COMPLEX_DRAWING("module:GoalLocatorLight:stuff", draw());
  TEAM_OUTPUT(idTeamMateGoalModel, bin, goalModel);
}


void GoalLocatorLight::preExecution()
{
  oppColor = theOwnTeamInfo.teamColor == TEAM_YELLOW ? ColorClasses::blue : ColorClasses::yellow;
  ownColor = theOwnTeamInfo.teamColor == TEAM_YELLOW ? ColorClasses::yellow : ColorClasses::blue;
  odometryOffset = theOdometryData - lastOdometry;
}


bool GoalLocatorLight::getGoalLineSegments(Vector2<int>& from, Vector2<int>& to)
{
  segments.clear();

  if(!theCameraMatrix.isValid)
    return false;

  for(int i = int(thePointsPercept.points.size()) - 1; i >= 0; i--)
  {
    PointsPercept::PointType pType = theOwnTeamInfo.teamColor == TEAM_YELLOW ? PointsPercept::blueGoal : PointsPercept::yellowGoal;

    if(thePointsPercept.points[i].type == pType)
    {
      int numberOfMiniSegments(0);
      int pId = i;
      PointsPercept::Point pMax(thePointsPercept.points[i]);
      while(thePointsPercept.points[i].link != i)
      {
        i = thePointsPercept.points[i].link;
        numberOfMiniSegments++;
      }
      segments.push_back(pId);
    }
  }

  return mergeGoalLineSegments(from, to);
}

bool GoalLocatorLight::mergeGoalLineSegments(Vector2<int>& from, Vector2<int>& to)
{
  int bestSquareLength(0);
  PointsPercept::Point bestMin, bestMax;

  for(unsigned int i=0; i<segments.size(); i++)
  {
    PointsPercept::Point pMax(thePointsPercept.points[segments[i]]);

    int k = segments[i];
    while(thePointsPercept.points[k].link != k)
    {
      k = thePointsPercept.points[k].link;
    }
    PointsPercept::Point pMin(thePointsPercept.points[k]);

    unsigned int j = i + 1;
    while(j < segments.size())
    {
      if(abs(thePointsPercept.points[k].inImage.x - thePointsPercept.points[segments[j]].inImage.x) < xMergingOffset
      && abs(thePointsPercept.points[k].inImage.y - thePointsPercept.points[segments[j]].inImage.y) < yMergingOffset)
      {
        int l = segments[j] ;
        while(thePointsPercept.points[l].link != l)
        {
          l = thePointsPercept.points[l].link;
        }

        if(thePointsPercept.points[l].inImage.x <= pMin.inImage.x)
        { 
          pMin = thePointsPercept.points[l];
          k = l;
        }
        j++;
      }
      else
        break;
    }

    int currentSquareDist = (pMax.onField - pMin.onField).squareAbs();
    if((currentSquareDist > parameters.minSquareDistance * parameters.minSquareDistance) && (currentSquareDist > bestSquareLength))
    {
      bestSquareLength = currentSquareDist;
      bestMin = pMin;
      bestMax = pMax;
    }
  }
  if(bestSquareLength)
  {
    Vector2<double> from2 = thePointsPercept.origin * Vector2<double>(bestMax.onField.x, bestMax.onField.y);
    Vector2<double> to2   = thePointsPercept.origin * Vector2<double>(bestMin.onField.x, bestMin.onField.y);
    from.x = static_cast<int>(from2.x);
    from.y = static_cast<int>(from2.y);
    to.x = static_cast<int>(to2.x);
    to.y = static_cast<int>(to2.y);
    LINE("module:GoalLocatorLight:stuff",from.x,from.y,to.x,to.y,60,Drawings::ps_solid,ColorRGBA(255,0,0,200));

    COMPLEX_DRAWING("module:GoalLocatorLight:image",
     CIRCLE("module:GoalLocatorLight:image", bestMin.inImage.x, bestMin.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::white, Drawings::bs_solid, ColorClasses::white);
     CIRCLE("module:GoalLocatorLight:image", bestMax.inImage.x, bestMax.inImage.y, 3, 2, Drawings::ps_solid, ColorClasses::white, Drawings::bs_solid, ColorClasses::white);
     LINE("module:GoalLocatorLight:image", bestMin.inImage.x, bestMin.inImage.y, bestMax.inImage.x, bestMax.inImage.y, 2, Drawings::ps_solid, ColorClasses::white);
    );

    return true;
  }
  else
    return false;
}


void GoalLocatorLight::motionUpdate()
{
  // Odometry in integer:
  const int cosor = static_cast<int>(cos(odometryOffset.rotation) * 1024);
  const int sinor = static_cast<int>(sin(odometryOffset.rotation) * 1024);
  const Matrix2x2<int> R(Vector2<int>(cosor,-sinor), Vector2<int>(sinor,cosor));
  const Vector2<int> odoTrans(static_cast<int>(odometryOffset.translation.x), static_cast<int>(odometryOffset.translation.y));
  
  // Update information in buffers:
  std::vector<FreeGoalHypothesis>::iterator iter; 
  for(iter = hypotheses.begin(); iter != hypotheses.end(); ++iter)
  {
    (*iter).range.max = -pi2;
    (*iter).range.min = pi2;
    RingBuffer<FreeGoalSegmentPercept, MAX_PERCEPTS_PER_HYPOTHESIS>& p = (*iter).percepts;
    for(int i=p.getNumberOfEntries()-1; i >= 0; i--)
    {
      p[i].leftEnd = (R*p[i].leftEnd) / 1024;
      p[i].leftEnd -= odoTrans;
      p[i].rightEnd = (R*p[i].rightEnd) / 1024;
      p[i].rightEnd -= odoTrans;
      p[i].range.max = Atan2Approximation::atan2(p[i].leftEnd.y, p[i].leftEnd.x);
      p[i].range.min = Atan2Approximation::atan2(p[i].rightEnd.y, p[i].rightEnd.x);
      if(p[i].range.max > (*iter).range.max)
        (*iter).range.max = p[i].range.max;
      if(p[i].range.min < (*iter).range.min)
        (*iter).range.min = p[i].range.min;
    }
  }
}


void GoalLocatorLight::modelGeneration(GoalModel& goalModel)
{
  goalModel.goalColor = oppColor;
  if(newPerception)
    goalModel.timeWhenLastSeen = theFrameInfo.time;
  unsigned int indexOfBestHypothesis = getBestHypothesisIndex();
  if(indexOfBestHypothesis != MAX_HYPOTHESES)
  {
    goalModel.largestFreePart.min = hypotheses[indexOfBestHypothesis].range.min;
    goalModel.largestFreePart.max = hypotheses[indexOfBestHypothesis].range.max;
    // Compute min/max distance:
    goalModel.distancesToFreePart.min = numeric_limits<int>::max();
    goalModel.distancesToFreePart.max = numeric_limits<int>::min();
    RingBuffer<FreeGoalSegmentPercept, MAX_PERCEPTS_PER_HYPOTHESIS>& p = hypotheses[indexOfBestHypothesis].percepts;
    for(int i=p.getNumberOfEntries()-1; i >= 0; i--)
    {
      int currentDist = p[i].leftEnd * p[i].leftEnd;
      if(currentDist > goalModel.distancesToFreePart.max)
        goalModel.distancesToFreePart.max = currentDist;
      if(currentDist < goalModel.distancesToFreePart.min)
        goalModel.distancesToFreePart.min = currentDist;
      currentDist = p[i].rightEnd * p[i].rightEnd;
      if(currentDist > goalModel.distancesToFreePart.max)
        goalModel.distancesToFreePart.max = currentDist;
      if(currentDist < goalModel.distancesToFreePart.min)
        goalModel.distancesToFreePart.min = currentDist;
    }
    goalModel.distancesToFreePart.min = int(sqrt(static_cast<float>(goalModel.distancesToFreePart.min)));
    goalModel.distancesToFreePart.max = int(sqrt(static_cast<float>(goalModel.distancesToFreePart.max)));
  }
}


void GoalLocatorLight::mergeHypotheses()
{
  std::vector<FreeGoalHypothesis>::iterator i; 
  for(i = hypotheses.begin(); i != hypotheses.end(); ++i)
  {
    std::vector<FreeGoalHypothesis>::iterator i2 = i;
    ++i2;
    while(i2 != hypotheses.end())
    {
      if(rangesIntersect(i->range, i2->range))
      {
        mergeHypotheses(*i, *i2);
        i2 = hypotheses.erase(i2);
      }
      else
        ++i2;
    }
  }
}


void GoalLocatorLight::mergeHypotheses(FreeGoalHypothesis& h1, const FreeGoalHypothesis& h2)
{
  // Collect the newest perceptions from both buffers (higher index means newer percept):
  int numOfPercepts(0);
  int currentH1Index(h1.percepts.getNumberOfEntries()-1);
  int currentH2Index(h2.percepts.getNumberOfEntries()-1);
  FreeGoalSegmentPercept mergedPercepts[MAX_PERCEPTS_PER_HYPOTHESIS];
  while(numOfPercepts < MAX_PERCEPTS_PER_HYPOTHESIS)
  {
    //Both buffers contain elements, take one:
    if(currentH1Index >= 0 && currentH2Index >= 0)
    {
      if(h1.percepts[currentH1Index].timeStamp >= h2.percepts[currentH2Index].timeStamp)
        mergedPercepts[numOfPercepts] = h1.percepts[currentH1Index--];
      else
        mergedPercepts[numOfPercepts] = h2.percepts[currentH2Index--];
    }
    //Only the one contain elements
    else if(currentH1Index >= 0)
    {
      mergedPercepts[numOfPercepts] = h1.percepts[currentH1Index--];
    }
    //Only the other contain elements
    else if(currentH2Index >= 0)
    {
      mergedPercepts[numOfPercepts] = h2.percepts[currentH2Index--];
    }
    //No elements left
    else
      break;
    //If this point is reached, one more percept was selected
    ++numOfPercepts;
  }

  // Fill buffer of h1:
  h1.percepts.init();
  while(numOfPercepts > 0)
  {
    h1.percepts.add(mergedPercepts[numOfPercepts]);
    --numOfPercepts;
  }
}


unsigned int GoalLocatorLight::getBestHypothesisIndex() const
{
  if(hypotheses.size())
  {
    // Try to find a hypothesis in front of the robot:
    for(unsigned int i=0; i<hypotheses.size(); i++)
      if(hypotheses[i].range.isInside(0.0))
        return i;
    // Take the closest one, otherwise:
    double bestDist(pi2);
    unsigned int bestDistIndex(0);
    for(unsigned int i=0; i<hypotheses.size(); i++)
    {
      //Angular distance to range:
      double currentDist = std::min(std::fabs(hypotheses[i].range.min), std::fabs(hypotheses[i].range.max));
      if(currentDist < bestDist)
      {
        bestDist = currentDist;
        bestDistIndex = i;
      }
    }
    return bestDistIndex;
  }
  else
    return MAX_HYPOTHESES;
}


void GoalLocatorLight::postExecution(GoalModel& ballModel)
{
  lastExecution = theFrameInfo.time;
  lastOdometry = theOdometryData;
}


void GoalLocatorLight::draw()
{
  std::vector<FreeGoalHypothesis>::iterator iter = hypotheses.begin();
  while(iter != hypotheses.end())
  {
    RingBuffer<FreeGoalSegmentPercept, MAX_PERCEPTS_PER_HYPOTHESIS>& p = (*iter).percepts;
    Vector2<int> center;
    for(int i=p.getNumberOfEntries()-1; i >= 0; i--)
    {
      unsigned age(theFrameInfo.getTimeSince(p[i].timeStamp));
      ColorRGBA col(ownColor);
      col.a = static_cast<unsigned char>(100.0 + 155.0 * (double(maxPerceptAge - age) / maxPerceptAge));
      LINE("module:GoalLocatorLight:stuff",p[i].leftEnd.x, p[i].leftEnd.y,
        p[i].rightEnd.x, p[i].rightEnd.y,20,Drawings::ps_solid, col);
      center += p[i].rightEnd;
      center += p[i].leftEnd;
    }
    ++iter;
    //Draw circle around hypothesis:
    center /= (p.getNumberOfEntries()*2);
    int radius(0);
    for(int i=p.getNumberOfEntries()-1; i >= 0; i--)
    {
      int d1 = (p[i].rightEnd - center).abs();
      int d2 = (p[i].leftEnd - center).abs();
      if(d1 > radius)
        radius = d1;
      if(d2 > radius)
        radius = d2;
    }
    CIRCLE("module:GoalLocatorLight:stuff", center.x, center.y, radius, 20, Drawings::ps_solid, 
      ColorClasses::white, Drawings::bs_null, ColorClasses::white);
  }
}


void GoalLocatorLight::clearOldPerceptsAndHypotheses()
{
  // Delete old perceptions:
  std::vector<FreeGoalHypothesis>::iterator iter = hypotheses.begin();
  while(iter != hypotheses.end())
  {
    RingBuffer<FreeGoalSegmentPercept, MAX_PERCEPTS_PER_HYPOTHESIS>& p = (*iter).percepts;
    for(int i=p.getNumberOfEntries()-1; i >= 0; i--)
    {
      if(theFrameInfo.getTimeSince(p[i].timeStamp) > maxPerceptAge)
        p.removeFirst();
      else
        break;
    }
    ++iter;
  }
  // Delete empty hypotheses:
  iter = hypotheses.begin();
  while(iter != hypotheses.end())
  {
    if((*iter).percepts.getNumberOfEntries() == 0)
      iter = hypotheses.erase(iter);
    else
      ++iter;
  }
}


void GoalLocatorLight::addNewPercepts()
{
  FreeGoalSegmentPercept p;
  newPerception = getGoalLineSegments(p.rightEnd,p.leftEnd);
  if(newPerception)
  {
    p.timeStamp = theFrameInfo.time;
    p.range.min = Atan2Approximation::atan2(p.rightEnd.y, p.rightEnd.x);
    p.range.max = Atan2Approximation::atan2(p.leftEnd.y, p.leftEnd.x);
    if(!mergeNewGoalSegment(p))
      createNewHypothesis(p);
  }
}


bool GoalLocatorLight::mergeNewGoalSegment(const FreeGoalSegmentPercept& p)
{
  std::vector<FreeGoalHypothesis>::iterator iter; 
  for(iter = hypotheses.begin(); iter != hypotheses.end(); ++iter)
  {
    FreeGoalHypothesis& h = (*iter);
    // The new percept is completely inside the hypothesis:
    if(h.range.contains(p.range))
    {
      h.percepts.add(p);
      h.lastUpdate = theFrameInfo.time;
      return true;
    }
    // The hypothesis is completely inside the new percept:
    if(p.range.contains(h.range))
    {
      h.percepts.add(p);
      h.range = p.range;
      h.lastUpdate = theFrameInfo.time;
      return true;
    }
    // Hypothesis and percept overlap one the max side:
    if((p.range.max >= h.range.max) && (p.range.min <= h.range.max))
    {
      h.percepts.add(p);
      h.range.max = p.range.max;
      h.lastUpdate = theFrameInfo.time;
      return true;
    }
    // Hypothesis and percept overlap one the min side:
    if((p.range.min <= h.range.min) && (p.range.max >= h.range.min))
    {
      h.percepts.add(p);
      h.range.min = p.range.min;
      h.lastUpdate = theFrameInfo.time;
      return true;
    }
  }
  return false;
}


void GoalLocatorLight::createNewHypothesis(const FreeGoalSegmentPercept& p)
{
  // Check, if there is a place left:
  if(hypotheses.size() < MAX_HYPOTHESES)
  {
    FreeGoalHypothesis h;
    h.percepts.add(p);
    h.lastUpdate = theFrameInfo.time;
    h.range = p.range;
    hypotheses.push_back(h);
  }
  // Otherwise, replace oldest hypothesis:
  else
  {
    // Find oldest hypothesis:
    unsigned oldestIndex(0);
    unsigned oldestTime(theFrameInfo.getTimeSince(hypotheses[0].lastUpdate));
    for(int i=1; i < MAX_HYPOTHESES; i++)
    {
      unsigned hTime = theFrameInfo.getTimeSince(hypotheses[i].lastUpdate);
      if(hTime > oldestTime)
      {
        oldestTime = hTime;
        oldestIndex = i;
      }
    }
    hypotheses[oldestIndex].lastUpdate = theFrameInfo.time;
    hypotheses[oldestIndex].percepts.init();
    hypotheses[oldestIndex].percepts.add(p);
    hypotheses[oldestIndex].range = p.range;
  }
}


bool GoalLocatorLight::rangesIntersect(const Range<double>& r1, const Range<double>& r2) const
{
  return r1.contains(r2) || r2.contains(r1) || ((r1.max >= r2.max) && (r1.min <= r2.max)) ||
    ((r1.min <= r2.min) && (r1.max >= r2.min));
}


MAKE_MODULE(GoalLocatorLight, Modeling)
