/**
* @file GoalLocatorLight.h
*
* Simple GoalLocator implementation.
* Computes free part only.
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef Goal_Locator_Light_h_
#define Goal_Locator_Light_h_

#include "Tools/Module/Module.h"
#include "Tools/RingBuffer.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/TeamInfo.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Perception/PointsPercept.h"
#include "Representations/Modeling/GoalModel.h"
#include "Representations/MotionControl/OdometryData.h"


MODULE(GoalLocatorLight)
  REQUIRES(PointsPercept)
  REQUIRES(CameraMatrix)
  REQUIRES(OdometryData)
  REQUIRES(RobotPose)
  REQUIRES(FrameInfo)
  REQUIRES(OwnTeamInfo)
  PROVIDES_WITH_MODIFY_AND_DRAW(GoalModel)
END_MODULE


/**
* @class GoalLocatorLight
*
* A GoalLocator using no probabilistics at all
*/
class GoalLocatorLight : public GoalLocatorLightBase
{
public:
  /** Constructor */
  GoalLocatorLight();

  /** Executes the GoalLocatorLight module 
  * @param goalModel The goal representation which is updated by this module
  */
  void update(GoalModel& goalModel);

  /** Description of a set of perceived adjacent goal points */
  struct FreeGoalSegmentPercept
  {
    /** leftmost point*/
    Vector2<int> leftEnd;
    /** rightmost point*/
    Vector2<int> rightEnd;
    /** angles to these two points*/
    Range<double> range;
    /** point of time of perception*/
    unsigned timeStamp;
  };

private:  

  /**
  * A collection of all parameters of the module.
  */
  class Parameters : public Streamable
  {
  private:
    /**
    * The method makes the object streamable.
    * @param in The stream from which the object is read 
    * @param out The stream to which the object is written 
    */
    virtual void serialize(In* in, Out* out)
    {  
      STREAM_REGISTER_BEGIN();
      STREAM(minSquareDistance);
      STREAM_REGISTER_FINISH();
    }

  public:
    int minSquareDistance;
  };

  /** Some constants*/
  enum {MAX_PERCEPTS_PER_HYPOTHESIS=10,
        xMergingOffset = 60,
        yMergingOffset = 50,
        maxPerceptAge = 9000,
        MAX_HYPOTHESES=3};

  Parameters parameters; /**< The parameters of this module. */

  /** A cluster of overlapping perceptions*/
  struct FreeGoalHypothesis
  {
    /** Tee total angular range of all perceptions*/
    Range<double> range;
    /** The time stamp of the newest perception*/
    unsigned lastUpdate;
    /** A buffer for recent perceptions*/
    RingBuffer<FreeGoalSegmentPercept, MAX_PERCEPTS_PER_HYPOTHESIS> percepts;
  };

  /** prepares execution, initializes some values*/
  void preExecution();

  /** motion update step*/
  void motionUpdate();

  /** Sets the member values of the GoalModel
  * @param goalModel A reference to the GoalModel*/
  void modelGeneration(GoalModel& goalModel);

   /** Sets some values after everything else of this module was executed
  * @param goalModel A reference to the GoalModel*/
  void postExecution(GoalModel& goalModel);

  /** Extracts a segment from the PointsPercept
  * @param from Resulting rightmost point
  * @param to Resulting leftmost point
  * @return true, if an adequate segment has been found
  */
  bool getGoalLineSegments(Vector2<int>& from, Vector2<int>& to);

  bool mergeGoalLineSegments(Vector2<int>& from, Vector2<int>& to);

  /** Deletes old buffered information*/
  void clearOldPerceptsAndHypotheses();

  /** Organizes the buffering of new information*/
  void addNewPercepts();

  /** Tries to add the current perception to an existing hypothesis
  * @param p The perception
  * @return true, if p has been merged
  */
  bool mergeNewGoalSegment(const FreeGoalSegmentPercept& p);

  /** Merges hypotheses that became overlapping by the time*/
  void mergeHypotheses();

  /** Merges hypotheses that became overlapping by the time*/
  void mergeHypotheses(FreeGoalHypothesis& h1, const FreeGoalHypothesis& h2);

  /** Creates a new hypothesis based on a perception.
  *   If there is no place for a new hypothesis, the one with the oldest update
  * @param p The perception
  */
  void createNewHypothesis(const FreeGoalSegmentPercept& p);

  /** Determines the best hypothesis for generating a result
  * @return The index of that hypothesis
  */
  unsigned int getBestHypothesisIndex() const;

  /** Draws some stuff*/
  void draw();

  /** Init function, called after construction of all currently active modules*/
  void init();

  /** Checks, if two ranges have some common interval
  * @param r1 first range
  * @param r2 second range
  * @return true, if check was positive
  */
  bool rangesIntersect(const Range<double>& r1, const Range<double>& r2) const;

  unsigned lastExecution;                                    /** time stamp*/
  unsigned lastSeenGoalTime;                                 /** time stamp*/
  Pose2D lastOdometry;                                       /** save last odometry state*/
  Pose2D odometryOffset;                                     /** odometry offset since last execution*/
  ColorClasses::Color oppColor;                              /** color for drawing*/
  ColorClasses::Color ownColor;                              /** color for drawing*/
  bool newPerception;                                        /** flag, indicates that a goal segment was seen*/
  std::vector<FreeGoalHypothesis> hypotheses;                /** The set of hypotheses*/
  std::vector<int> segments;
};

#endif
