/**
* @file ParticleFilterBallLocator.cpp
* Contains a BallLocator implementation using a particle filter
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#include "ParticleFilterBallLocator.h"
#include "Tools/Math/AngleTables.h"
#include "Tools/Math/Matrix2x2.h"
#include "Tools/Team.h"


ParticleFilterBallLocator::ParticleFilterBallLocator()
 : slowWeighting(0.0), fastWeighting(0.0), lastExecution(0), lastSeenBallTime(0), lastCheatingTime(0), 
   calibrateUncertainty(false), currentCalibrationMeasurement(0), gaussian(5000, 100, 50, 400, 10),
   gaussianAngles(3200, 320, static_cast<int> (parameters.angleStandardDeviation),
   static_cast<int> (parameters.angleStandardDeviation), 1),newSamplesNumberForDrawing(0)
{
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    Pose2D pose(theFieldDimensions.randomPoseOnField());
    ballSamples[i].pos.x = static_cast<int>(pose.translation.x);
    ballSamples[i].pos.y = static_cast<int>(pose.translation.y);
    if(randomDouble() <= parameters.ballMotionProbability)
    {
      ballSamples[i].setRandomVelocity();
      ballSamples[i].state = BallSample::ROLLING;
    }
    else
    {
      ballSamples[i].state = BallSample::LYING;
    }
  }
}


void ParticleFilterBallLocator::init()
{
  lastOdometry = theOdometryData;
  numberOfGeneratedSamples = 0;
}


void ParticleFilterBallLocator::update(BallModel& ballModel)
{
  preExecution();
  motionUpdate();
  if(theBallPercept.ballWasSeen)
  {
    sensorUpdate();
    resampling();
  }
  modelGeneration(ballModel);
  postExecution(ballModel);

  //Send BallModel to TeamMate
  TEAM_OUTPUT(idTeamMateBallModel, bin, ballModel);	

  DECLARE_DEBUG_DRAWING("module:ball locator:ball samples", "drawingOnField", "Draws the internal representations of the ball locator");
  COMPLEX_DRAWING("module:ball locator:ball samples", drawSamples());
  DECLARE_DEBUG_DRAWING("module:ball locator:ball samples in image", "drawingOnImage", "Draws the internal representations of the ball locator to the image view");
  COMPLEX_DRAWING("module:ball locator:ball samples in image", drawSamplesToImage());
}

void ParticleFilterBallLocator::preExecution()
{
  //Remove old elements from percept buffer and try to add a new one (if it exists):
  while(ballPerceptBuffer.size() && theFrameInfo.getTimeSince(ballPerceptBuffer.front().timeStamp) > 1000)
  {
    ballPerceptBuffer.pop_front();
  }
  if(theBallPercept.ballWasSeen)
  {
    ballPerceptBuffer.push_back(BufferedPercept(theBallPercept, theFrameInfo.time, theOdometryData));
  }

  odometryOffset = theOdometryData - lastOdometry;
  MODIFY("module:ball locator:calibrate", calibrateUncertainty);
  DEBUG_RESPONSE("module:ball locator:reset calibration measurements", {currentCalibrationMeasurement = 0;});
  if(theBallPercept.ballWasSeen && calibrateUncertainty && 
     currentCalibrationMeasurement < NUM_OF_CALIBRATION_MEASUREMENTS)
  {
    uncertaintyCalibration();
  }
  MODIFY("parameters:ball locator", parameters);
}


void ParticleFilterBallLocator::motionUpdate()
{
  //Rotation matrix for odometry:
  const int cosor = static_cast<int>(cos(odometryOffset.rotation) * 1024);
  const int sinor = static_cast<int>(sin(odometryOffset.rotation) * 1024);
  const Matrix2x2<int> R(Vector2<int>(cosor,-sinor), Vector2<int>(sinor,cosor));
  //Some values:
  const Vector2<int> odoTrans(static_cast<int>(odometryOffset.translation.x), static_cast<int>(odometryOffset.translation.y));
  const int odometryLength = odoTrans.abs();
  const int t(theFrameInfo.getTimeSince(lastExecution));
  //Update samples:
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    const int velocity = ballSamples[i].velocity.abs();
    //Ball motion update:
    if(ballSamples[i].state == BallSample::ROLLING)
    {
      ballSamples[i].pos.x += (ballSamples[i].velocity.x + sampleTriangularDistribution(static_cast<int>(velocity*parameters.scaleFactorVelocityVariance)))* t / 1000;
      ballSamples[i].pos.y += (ballSamples[i].velocity.y + sampleTriangularDistribution(static_cast<int>(velocity*parameters.scaleFactorVelocityVariance)))* t / 1000;
    }
    //Robot odometry update:
    ballSamples[i].pos = (R*ballSamples[i].pos) / 1024;
    ballSamples[i].pos -= odoTrans;
    ballSamples[i].velocity = (R*ballSamples[i].velocity) / 1024;
    //Add uncertainty to position (depending on walked distance):
    ballSamples[i].pos.x += sampleTriangularDistribution(static_cast<int>(odometryLength*parameters.scaleFactorOdometryVariance));
    ballSamples[i].pos.y += sampleTriangularDistribution(static_cast<int>(odometryLength*parameters.scaleFactorOdometryVariance));
    //Add uncertainty to velocity:
    if(ballSamples[i].state == BallSample::ROLLING)
    {
      ballSamples[i].velocity.x += sampleTriangularDistribution(static_cast<int>(velocity*parameters.scaleFactorVelocityVariance))* t / 1000;
      ballSamples[i].velocity.y += sampleTriangularDistribution(static_cast<int>(velocity*parameters.scaleFactorVelocityVariance))* t / 1000;
    }
  }
}


void ParticleFilterBallLocator::sensorUpdate()
{
  averageWeighting = 0;
  const int observationDistance(static_cast<int>(theBallPercept.relativePositionOnField.abs()));
  const double observationAngle(atan2(theBallPercept.relativePositionOnField.y,theBallPercept.relativePositionOnField.x));
  double distStdDev;
  if(observationDistance <= parameters.closeThreshold)
    distStdDev = parameters.closeBallDistanceStandardDeviation;
  else if(observationDistance >= parameters.farThreshold)
    distStdDev = parameters.farBallDistanceStandardDeviation;
  else
    distStdDev = parameters.closeBallDistanceStandardDeviation + 
    ((observationDistance - parameters.closeThreshold)/(parameters.farThreshold - parameters.closeThreshold))*(parameters.farBallDistanceStandardDeviation - parameters.closeBallDistanceStandardDeviation);
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    const int modelDistance(ballSamples[i].pos.abs());
    const double modelAngle(Atan2Approximation::atan2(ballSamples[i].pos.y, ballSamples[i].pos.x));
    const double weighting = gaussian.value(abs(observationDistance - modelDistance), (int)distStdDev) *
      gaussianAngles.value(int(1000*fabs(observationAngle - modelAngle)), (int)parameters.angleStandardDeviation);
    ballSamples[i].weighting = weighting;
    averageWeighting += weighting;
  }
  averageWeighting /= MAX_SAMPLES;
  slowWeighting = slowWeighting + parameters.alphaSlow * (averageWeighting - slowWeighting);
  fastWeighting = fastWeighting + parameters.alphaFast * (averageWeighting - fastWeighting);
}


void ParticleFilterBallLocator::resampling()
{
  // swap sample arrays
  BallSample* oldSet = ballSamples.swap();

  const double weightingsSum(averageWeighting*MAX_SAMPLES);
  const double resamplingPercentage(std::max(0.0, 1.0 - fastWeighting / slowWeighting));
  const double RESAMPLED_SAMPLES = MAX_SAMPLES * (1.0 - resamplingPercentage);
  const double weightingBetweenTwoDrawnSamples(weightingsSum  / double(RESAMPLED_SAMPLES));
  double nextPos(randomDouble()*weightingBetweenTwoDrawnSamples); 
  double currentSum(0);

  // resample:
  int j(0);
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    currentSum += oldSet[i].weighting;
    while(currentSum > nextPos)
    {
      ballSamples[j++] = oldSet[i];
      nextPos += weightingBetweenTwoDrawnSamples;
      if(j == MAX_SAMPLES)  // This might be necessary because of floating point imprecision
        return;
    }
  }
  // Store number of generated samples
  numberOfGeneratedSamples = MAX_SAMPLES - j;
  // fill up remaining samples with new poses:
  for(; j < MAX_SAMPLES; ++j)
  {
    ballSamples[j] = generateNewSample();
  }
  newSamplesNumberForDrawing = numberOfGeneratedSamples;
}


BallSample ParticleFilterBallLocator::generateNewSample()
{
  BallSample newSample;
  // If there have been some percepts, use two of them to create a new sample
  if(ballPerceptBuffer.size() > 1)
  {
    int a = random(ballPerceptBuffer.size());
    int b = random(ballPerceptBuffer.size());
    // a and b must be different, of course:
    if(a == b)
    {
      if(a != 0)
        a--;
      else
        a++;
    }
    // a must be smaller than b:
    if(a > b)
    {
      int temp = b;
      b = a;
      a = temp;
    }
    Vector2<double> p1 = ballPerceptBuffer[a].getOdometryCorrectedPosition(theOdometryData);
    unsigned timeP1 = ballPerceptBuffer[a].timeStamp;
    Vector2<double> p2 = ballPerceptBuffer[b].getOdometryCorrectedPosition(theOdometryData);
    unsigned timeP2 = ballPerceptBuffer[b].timeStamp;
    double timeInSeconds((double(timeP2) - double(timeP1))/1000.0);
    Vector2<double> speed(p2-p1);
    speed /= timeInSeconds;
    newSample.pos.x = static_cast<int>(p2.x);
    newSample.pos.y = static_cast<int>(p2.y);
    newSample.velocity.x = static_cast<int>(speed.x);
    newSample.velocity.y = static_cast<int>(speed.y);
    // Compute current position of sample
    double timeToNow = theFrameInfo.getTimeSince(timeP2);
    timeToNow /= 1000;
    Vector2<double> offset(newSample.velocity.x, newSample.velocity.y);
    offset *= timeToNow;
    newSample.pos.x += static_cast<int>(offset.x);
    newSample.pos.y += static_cast<int>(offset.y);
    //If sample is close, decrease velocity.
    if(newSample.pos.abs() < parameters.templateSpeedDecreaseDistance)
    {
      newSample.velocity /= static_cast<int>(parameters.templateSpeedDecreaseFactor);
    }
  }
  // Use the current percept and a random speed
  else
  {
    newSample.pos.x = static_cast<int>(theBallPercept.relativePositionOnField.x);
    newSample.pos.y = static_cast<int>(theBallPercept.relativePositionOnField.y);
    newSample.setRandomVelocity();
    //If sample is close, decrease velocity.
    if(newSample.pos.abs() < parameters.templateSpeedDecreaseDistance)
    {
      newSample.velocity /= static_cast<int>(parameters.templateSpeedDecreaseFactor);
    }
  }
  //This must be integrated a bit better in the previous lines, but anyway:
  if(randomDouble() > parameters.ballMotionProbability)
    newSample.state = BallSample::LYING;
  else
    newSample.state = BallSample::ROLLING;
  return newSample;
}


void ParticleFilterBallLocator::modelGeneration(BallModel& ballModel)
{
  //The estimated ball position:
  Vector2<double> averagePosition;
  Vector2<double> averageVelocity;
  double sumOfAllWeights(0.0);
  // This is a bad case that might happen:
  if(numberOfGeneratedSamples == MAX_SAMPLES)
  {
    for(int i = 0; i < MAX_SAMPLES; ++i)
    {
      averagePosition.x += ballSamples[i].pos.x;
      averagePosition.y += ballSamples[i].pos.y;
      if(ballSamples[i].state == BallSample::ROLLING)
      {
        averageVelocity.x += ballSamples[i].velocity.x;
        averageVelocity.y += ballSamples[i].velocity.y;
      }
    }
    averagePosition /= MAX_SAMPLES;
    averageVelocity /= MAX_SAMPLES;
  }
  // This is the standard case:
  else
  {
    for(int i = 0; i < MAX_SAMPLES; ++i)
    {
      if(ballSamples[i].weighting != -1.0)
      {
        averagePosition.x += (ballSamples[i].pos.x * ballSamples[i].weighting);
        averagePosition.y += (ballSamples[i].pos.y * ballSamples[i].weighting);
        if(ballSamples[i].state == BallSample::ROLLING)
        {
          averageVelocity.x += (ballSamples[i].velocity.x * ballSamples[i].weighting);
          averageVelocity.y += (ballSamples[i].velocity.y * ballSamples[i].weighting);
        }
        sumOfAllWeights += ballSamples[i].weighting;
      }
    }
    averagePosition /= sumOfAllWeights;
    averageVelocity /= sumOfAllWeights;
  }
  ballModel.estimate.setPositionAndVelocityInRelativeCoordinates(averagePosition,averageVelocity);

  //Information about what has actually been seen:
  if(theBallPercept.ballWasSeen)
  {
    ballModel.timeWhenLastSeen = theFrameInfo.time;
    ballModel.lastPerception.setPositionAndVelocityInRelativeCoordinates(
      theBallPercept.relativePositionOnField);
    ballModel.lastSeenEstimate.setPositionAndVelocityInRelativeCoordinates(
      averagePosition, averageVelocity);
  }
  else
  {
    //Moved position of last ball perception according to the odometry
/*    Vector2<double> c1(cos(odometryOffset.rotation),-sin(odometryOffset.rotation));
    Vector2<double> c2(sin(odometryOffset.rotation),cos(odometryOffset.rotation));
    Matrix2x2<double> R(c1, c2);
    Vector2<double> seenBallPos(ballModel.lastPerception.getPositionInRelativeCoordinates());
    seenBallPos = R*seenBallPos;
    seenBallPos -= odometryOffset.translation;
    ballModel.lastPerception.setPositionAndVelocityInRelativeCoordinates(seenBallPos);
*/
    Vector2<double> lastPerception(ballModel.lastPerception.getPositionInRelativeCoordinates());
    lastPerception.rotate(-odometryOffset.rotation);
    lastPerception -= odometryOffset.translation;
    ballModel.lastPerception.setPositionAndVelocityInRelativeCoordinates(lastPerception);

    Vector2<double> lastSeenEstimate(ballModel.lastSeenEstimate.getPositionInRelativeCoordinates());
    lastSeenEstimate.rotate(-odometryOffset.rotation);
    lastSeenEstimate -= odometryOffset.translation;
    ballModel.lastSeenEstimate.setPositionAndVelocityInRelativeCoordinates(lastSeenEstimate);
  }
}


void ParticleFilterBallLocator::postExecution(BallModel& ballModel)
{
  PLOT("module:ParticleFilterBallLocator:numberOfGeneratedSamples", newSamplesNumberForDrawing);
  lastExecution = theFrameInfo.time;
  if(theBallPercept.ballWasSeen)
  {
    // There is currently one thing that cannot be handled properly:
    // Seeing a ball for the first time after a longer period. Then the model
    // does not adjust fast enough. Therefore we will deliver the percept for one
    // second ;-)

    //Currently cheating:
    if(theFrameInfo.getTimeSince(lastCheatingTime) < 1000)
    {
      ballModel.estimate.setPositionAndVelocityInRelativeCoordinates(theBallPercept.relativePositionOnField);
      ballModel.lastSeenEstimate.setPositionAndVelocityInRelativeCoordinates(theBallPercept.relativePositionOnField);
    }
    //Detect cheating situation:
    else if(theFrameInfo.getTimeSince(lastSeenBallTime) > 2000)
    {
      lastCheatingTime = theFrameInfo.time;
      ballModel.estimate.setPositionAndVelocityInRelativeCoordinates(theBallPercept.relativePositionOnField);
      ballModel.lastSeenEstimate.setPositionAndVelocityInRelativeCoordinates(theBallPercept.relativePositionOnField);
    }
    //Standard stuff:
    lastSeenBallTime = lastExecution;
  }
  lastOdometry = theOdometryData;
}


void ParticleFilterBallLocator::drawSamples()
{
  //Default drawing color
  ColorRGBA sampleColor(255,255,255);
  //Determine alpha offset for nicer drawing ;-)
  double alphaOffset(1.0);
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    const double offset(1.0 - ballSamples[i].weighting);
    if(offset < alphaOffset)
      alphaOffset = offset;
  }
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    Vector2<int> p(ballSamples[i].pos);
    Vector2<int> v(ballSamples[i].velocity);
    if(ballSamples[i].weighting != -1.0)
      sampleColor.a = (unsigned char)((ballSamples[i].weighting + alphaOffset)*255.0);
    else //Draw new samples in a different color:
      sampleColor = ColorRGBA(255,255,0);
    LARGE_DOT("module:ball locator:ball samples", p.x, p.y, sampleColor, sampleColor);
    if(ballSamples[i].state == BallSample::ROLLING)
    {
      ARROW("module:ball locator:ball samples", p.x, p.y, p.x+v.x, p.y+v.y,5,1,sampleColor);
    }
  }
}


void ParticleFilterBallLocator::drawSamplesToImage()
{
  ColorRGBA fillColor(255,128,64,100);
  ColorRGBA borderColor(255,128,64);
  for(int i = 0; i < MAX_SAMPLES; ++i)
  {
    Geometry::Circle c;
    Geometry::calculateBallInImage(Vector2<double>(ballSamples[i].pos.x, ballSamples[i].pos.y), theCameraMatrix, theCameraInfo, theFieldDimensions.ballRadius, c);
    CIRCLE("module:ball locator:ball samples in image", 
      c.center.x, 
      c.center.y, 
      c.radius, 
      1, // pen width 
      Drawings::ps_solid, 
      ColorClasses::orange,
      Drawings::bs_solid, 
      ColorRGBA(255,128,64,100));
  }
}


void ParticleFilterBallLocator::uncertaintyCalibration()
{
  if(currentCalibrationMeasurement < NUM_OF_CALIBRATION_MEASUREMENTS)
  {
    ballXMeasurements[currentCalibrationMeasurement] = theBallPercept.positionInImage.x;
    ballYMeasurements[currentCalibrationMeasurement] = theBallPercept.positionInImage.y;
    currentCalibrationMeasurement++;
  }
  if(currentCalibrationMeasurement == NUM_OF_CALIBRATION_MEASUREMENTS)
  {
    ballPosUncertainty.generateDistributionFromMeasurements(ballXMeasurements, NUM_OF_CALIBRATION_MEASUREMENTS,
      ballYMeasurements, NUM_OF_CALIBRATION_MEASUREMENTS);
    OUTPUT(idText, text, ballPosUncertainty.covariance[0][0] << " " << ballPosUncertainty.covariance[1][0] << " "
                      << ballPosUncertainty.covariance[0][1] << " " << ballPosUncertainty.covariance[1][1]);
  }
}


MAKE_MODULE(ParticleFilterBallLocator, Modeling)
