/**
* @file PointValidityChecker.h
*
* Checks, if a point may be used by self localization
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/

#ifndef _PointValidityChecker_H_
#define _PointValidityChecker_H_

#include "Representations/Perception/BallPercept.h"
#include "Representations/Perception/PointsPercept.h"


/**
* @class PointValidityChecker
* A class for checking points for localization
*/
class PointValidityChecker
{
private:
  const BallPercept& ballPercept;       /**< A reference to the ball percept*/
  const PointsPercept& pointsPercept;   /**< A reference to the points percepts*/
  bool performYellowInBallCheck;        /**< Flag, indicates that yellow points inside balls should be ignored*/
  int currentSquaredBallCheckRadius;    /**< What the long name says*/

public:
  /** Default constructor. */
  PointValidityChecker(const BallPercept& ballPercept, const PointsPercept& pointsPercept): ballPercept(ballPercept), 
    pointsPercept(pointsPercept)
  {}

  /** Select relevant points for localization and copy their indices to vector.
  *   The following point types will be sorted out:
  *   - Yellow points in orange ball
  *   - Team dependent points that are less frequent than others
  *   - Points that are not linked to others(!)
  * @param pointIndices The vector of all relevant indices
  */
  void getRelevantPointsForLocalization(std::vector<short>& pointIndices)
  {
    if((performYellowInBallCheck = ballPercept.ballWasSeen))
      currentSquaredBallCheckRadius = static_cast<int>(ballPercept.radiusInImage*ballPercept.radiusInImage);
    pointIndices.clear();
    PointsPercept::PointType preferredPointType = getPreferredTeamDependentPoints();
    for(unsigned i = 0; i < pointsPercept.points.size(); ++i)
    {
      const PointsPercept::Point& p = pointsPercept.points[i];
      // Check, if point belongs to field and is on the floor and is linked:
      if(p.type < PointsPercept::numberOfFieldPointTypes && !p.isCloser && (unsigned) p.link != i)
      {
        // Skip non-preferred point types:
        if(p.type < PointsPercept::numberOfTeamDependentPointTypes && p.type != preferredPointType)
          continue;
        // Special treatment of yellow goal points and yellow beacons (high unlikely but possible)
        if(performYellowInBallCheck && (p.type == PointsPercept::yellowGoal || p.type == PointsPercept::yellowBeacon))
        {
          if(!pointIsInBall(p))
            pointIndices.push_back(static_cast<short>(i));
        }
        // All other points can be used:
        else
          pointIndices.push_back(static_cast<short>(i));
      }
    }
  }

private:
  /** Checks, if a point is inside or close to (0.5*radius) a ball in the image
  * @param p The point
  * @return true, if the point is inside or close to the ball
  */
  bool pointIsInBall(const PointsPercept::Point& p)
  {
    Vector2<int> pToBall(static_cast<int>(ballPercept.positionInImage.x) - p.inImage.x, 
      static_cast<int>(ballPercept.positionInImage.y) - p.inImage.y);
    int pToBallSquaredDist = sqr(pToBall.x) + sqr(pToBall.y);
    return pToBallSquaredDist < (currentSquaredBallCheckRadius*3)/2;
  }

  /** Returns the most frequent point type (of linked points)
  * @return The type
  */
  PointsPercept::PointType getPreferredTeamDependentPoints()
  {
    const int maxPType(PointsPercept::numberOfTeamDependentPointTypes);
    int pointCount[maxPType];
    for(int i=0; i<maxPType; ++i)
      pointCount[i] = 0;
    for(unsigned i = 0; i < pointsPercept.points.size(); ++i)
    {
      const PointsPercept::Point& p = pointsPercept.points[i];
      if(p.type < maxPType && (unsigned) p.link != i)
        pointCount[p.type] += 1;
    }
    int maxCount(-1);
    int maxIdx(-1);
    for(int i=0; i<maxPType; ++i)
    {
      if(pointCount[i] > maxCount)
      {
        maxIdx = i;
        maxCount = pointCount[i];
      }
    }
    if(maxIdx > -1)
      return static_cast<PointsPercept::PointType>(maxIdx);
    else
      return PointsPercept::numberOfPointTypes;
  }
};


#endif  //_PointValidityChecker_H_
