/**
* @file SelfLocator.cpp
* Implements a class that performs self-localization using a particle filter.
*
* The difference to the ParticleFilterSelfLocator is as follows:
*
* - May choose different implementations for pose calculation
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include <limits>
#include "SelfLocator.h"
#include "PoseCalculator/PoseCalculator2DBinning.h"
#include "PoseCalculator/PoseCalculatorBestParticle.h"
#include "PoseCalculator/PoseCalculatorOverallAverage.h"
#include "Tools/Math/GaussianDistribution3D.h"
#include "Tools/Settings.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Streams/OutStreams.h"
#include "Tools/Team.h"


SelfLocator::SelfLocator() : 
slowWeighting(0.0), fastWeighting(0.0), updatedBySensors(false), 
pointSelector(theBallPercept, thePointsPercept), lastReenterTimeStamp(0)
{
  InBinaryFile stream(Global::getSettings().expandLocationFilename("closestPoints.tab"));
  if(stream.exists())
    for(int i = 0; i < numberOfTables; ++i)
      closestPointsTable[i].read(stream);
  pointIndices.reserve(1000);

  poseCalculator = new PoseCalculator2DBinning< Sample, SampleSet<Sample, numberOfSamples>, 10 > (samples, theFieldDimensions);
  poseCalculatorType = POSE_CALCULATOR_2D_BINNING;
}

SelfLocator::~SelfLocator()
{
  delete poseCalculator;
}

void SelfLocator::init()
{
  std::vector<Pose2D> poses;    
  std::vector<Pose2D> standardDeviations;
  if(theSelfLocatorParameter.knownStartPose)
  {
    poses.push_back(theSelfLocatorParameter.startPose);
    standardDeviations.push_back(theSelfLocatorParameter.startPoseStandardDeviation);
  }
  // If no start position is known, spread samples uniformly across state space
  initSamplesAtGivenPositions(poses, standardDeviations);
}

void SelfLocator::initSamplesAtGivenPositions(const std::vector<Pose2D>& poses,
    const std::vector< Pose2D >& standardDeviations)
{
  for(int i = 0; i < numberOfSamples; ++i)
  {
    Sample& sample = samples[i];
    if(poses.size())
    {
      // Select one of the poses from the list:
      unsigned int index = random(poses.size());
      Pose2D pose = poses[index];
      Pose2D stdDev = standardDeviations[index];
      // Create sample:
      sample.translation.x = static_cast<int>(pose.translation.x);
      sample.translation.x += sampleTriangularDistribution(static_cast<int>(stdDev.translation.x));
      sample.translation.y = static_cast<int>(pose.translation.y);
      sample.translation.y += sampleTriangularDistribution(static_cast<int>(stdDev.translation.y));
      double rotation = pose.rotation;
      rotation += sampleTriangularDistribution(stdDev.rotation);
      normalize(rotation);
      sample.angle = rotation;
      sample.rotation = Vector2<int>(int(cos(rotation) * 1024), int(sin(rotation) * 1024));
    }
    else //No given positions, spread uniformly:
    {
      Pose2D pose(theFieldDimensions.randomPoseOnField());
      sample.translation = Vector2<int>(int(pose.translation.x), int(pose.translation.y));
      sample.rotation = Vector2<int>(int(cos(pose.rotation) * 1024), int(sin(pose.rotation) * 1024));
      sample.angle = sample.rotation.angle();
    }
  }
  lastOdometry = theOdometryData;
  poseCalculator->init();
}

void SelfLocator::update(RobotPose& robotPose)
{
  preExecution();
  DEBUG_RESPONSE_NOT("module:ParticleFilterSelfLocator:just templates",
    DEBUG_RESPONSE_NOT("module:ParticleFilterSelfLocator:odometry only",
      motionUpdate(updatedBySensors);
      updatedBySensors = pointIndices.size() > 0 && theCameraMatrix.isValid;
      if(updatedBySensors)
      {
        sensorUpdate();
        resampling();
      }
      poseCalculator->calcPose(robotPose);
    );
  );

  DEBUG_RESPONSE("module:ParticleFilterSelfLocator:just templates",
    if(perceptBuffer.getNumberOfEntries())
      for(int i = 0; i < numberOfSamples; ++i)
        generateTemplate(samples[i]);
    poseCalculator->calcPose(robotPose);
  );

  DEBUG_RESPONSE("module:ParticleFilterSelfLocator:odometry only",
    robotPose += theOdometryData - lastOdometry;
    lastOdometry = theOdometryData;
  );

  // send own position to the team mates
  TEAM_OUTPUT(idTeamMateRobotPose, bin, robotPose);	

  // drawings
  DECLARE_DEBUG_DRAWING("module:ParticleFilterSelfLocator:poseCalculator", "drawingOnField", "Draws the internal representations for pose computation");
  COMPLEX_DRAWING("module:ParticleFilterSelfLocator:poseCalculator", poseCalculator->draw());

  DECLARE_DEBUG_DRAWING("module:ParticleFilterSelfLocator:samples", "drawingOnField", "Draws the internal representations of the goal locator");
  COMPLEX_DRAWING("module:ParticleFilterSelfLocator:samples", drawSamples());

  DECLARE_DEBUG_DRAWING("module:ParticleFilterSelfLocator:closestPoints", "drawingOnField", "Draws the closest points tables.");
  int index = 0;
  MODIFY("module:ParticleFilterSelfLocator:closestPoints", index);
  COMPLEX_DRAWING("module:ParticleFilterSelfLocator:closestPoints", closestPointsTable[index].draw(); );
}

void SelfLocator::preExecution()
{
  gaussian.computeTable(1000, 200, theSelfLocatorParameter.standardDeviation, theSelfLocatorParameter.standardDeviation, 1);

  DEBUG_RESPONSE("module:ParticleFilterSelfLocator:create tables",
    OutBinaryFile stream(Global::getSettings().expandLocationFilename("closestPoints.tab"));
    int index = 0;
    for(int i = 0; i < PointsPercept::numberOfBeaconPointTypes; ++i)
    {
      closestPointsTable[index].create(theFieldDimensions.table[i], 0, 1, 0);
      closestPointsTable[index++].write(stream);
    }
    for(int i = 0; i < numberOfLengthConstraints; ++i)
      for(int j = 0; j < numberOfOrientationConstraints; ++j)
        for(int k = 0; k < numberOfTableTypes; ++k)
        {
          double orientation = j == alongField ? pi_2 : 0; // normal of lines!
          int numberOfOrientations = j == noOrientationConstraint ? 1 : 2;
          double minLength = i == noLengthConstraint ? 0 : k == 2 ? theSelfLocatorParameter.maxCrossingLength : theSelfLocatorParameter.maxBeaconLength;
          closestPointsTable[index].create(theFieldDimensions.table[k + FieldDimensions::ownGoal], orientation, numberOfOrientations, minLength);
          closestPointsTable[index++].write(stream);
        }
  );

  // copy indices of relevant points from percept to local buffer
  pointSelector.getRelevantPointsForLocalization(pointIndices);

  // precompute several values as integers
  xOrigin = int(thePointsPercept.origin.translation.x);
  yOrigin = int(thePointsPercept.origin.translation.y);
  rotCos = int(cos(thePointsPercept.origin.rotation) * (1 << 10));
  rotSin = int(sin(thePointsPercept.origin.rotation) * (1 << 10));
  xOdoOrigin = int(theOdometryData.translation.x);
  yOdoOrigin = int(theOdometryData.translation.y);
  rotOdoCos = int(cos(theOdometryData.rotation) * (1 << 10));
  rotOdoSin = int(sin(theOdometryData.rotation) * (1 << 10));

  // update buffer from which sample templates can be constructed
  DEBUG_RESPONSE_NOT("module:ParticleFilterSelfLocator:freeze percept buffer",
    for(unsigned i = 0; i < pointIndices.size(); ++i)
    {
      const PointsPercept::Point& p = thePointsPercept.points[pointIndices[i]];
      if(p.type < PointsPercept::numberOfTeamDependentPointTypes && !p.isCloser)
        perceptBuffer.add(Percept(Vector2<int>(xOrigin + ((p.onField.x * rotCos - p.onField.y * rotSin) >> 10),
                                               yOrigin + ((p.onField.x * rotSin + p.onField.y * rotCos) >> 10)),
                                  p.type, p.length2 < sqr(theFieldDimensions.flagRadius * 2),
                                  xOdoOrigin, yOdoOrigin, rotOdoCos, rotOdoSin));
    }
  );

  // change module for pose calculation
  PoseCalculatorType newPoseCalculatorType(poseCalculatorType);
  MODIFY_ENUM("module:ParticleFilterSelfLocator:poseCalculatorType", 
    newPoseCalculatorType, NUMBER_OF_POSE_CALCULATORS, getPoseCalculatorTypeName);
  if(newPoseCalculatorType != poseCalculatorType)
  {
    delete poseCalculator;
    switch(newPoseCalculatorType)
    {
    case POSE_CALCULATOR_2D_BINNING: 
      poseCalculator = new PoseCalculator2DBinning< Sample, SampleSet<Sample, numberOfSamples>, 10 >
        (samples, theFieldDimensions);
      break;
    case POSE_CALCULATOR_BEST_PARTICLE: 
      poseCalculator = new PoseCalculatorBestParticle< Sample, SampleSet<Sample, numberOfSamples> >(samples);
      break;
    case POSE_CALCULATOR_OVERALL_AVERAGE:
      poseCalculator = new PoseCalculatorOverallAverage< Sample, SampleSet<Sample, numberOfSamples> >(samples);
      break;
    default: ASSERT(false);
    }
    poseCalculatorType = newPoseCalculatorType;
  }

  // Reset samples to re-enter positions:
  bool resetSamples = (theGameInfo.state == STATE_PLAYING) && 
    theKeyStates.pressed[KeyStates::left] && theKeyStates.pressed[KeyStates::right] && 
    !(theKeyStates.pressed[KeyStates::up]) && !(theKeyStates.pressed[KeyStates::down]);
  DEBUG_RESPONSE("module:ParticleFilterSelfLocator:resetSamplesToReenterPositions",
  {
    resetSamples = true;
  });
  if(resetSamples)
  {
    std::vector<Pose2D> poses;
    std::vector<Pose2D> stdDevs;
    poses.push_back(Pose2D(0.0, 0.0, theFieldDimensions.yPosLeftSideline));
    poses.push_back(Pose2D(pi/2.0, 0.0, theFieldDimensions.yPosRightSideline));
    stdDevs.push_back(Pose2D(0.2, 200, 200));
    stdDevs.push_back(Pose2D(0.2, 200, 200));
    initSamplesAtGivenPositions(poses, stdDevs);
  }

  // Re-init samples to start-pos
  bool reinitSamples = (theGameInfo.state == STATE_PLAYING) && 
    theKeyStates.pressed[KeyStates::up] && theKeyStates.pressed[KeyStates::down] && 
    !(theKeyStates.pressed[KeyStates::left]) && !(theKeyStates.pressed[KeyStates::right]);
  if(reinitSamples)
    init();

  // Initialize sample set again:
  DEBUG_RESPONSE("module:ParticleFilterSelfLocator:resetSamples", init(););

 /*Und hier gibt es Akkus:
  effektmodell
  justus-von liebig-strae 4
  31535 neustadt
  050 - 32 913675*/
}

void SelfLocator::motionUpdate(bool noise)
{
  Pose2D odometryOffset = theOdometryData - lastOdometry;

  const double transNoise = noise ? theSelfLocatorParameter.translationNoise : 0;
  const double rotNoise = noise ? theSelfLocatorParameter.rotationNoise : 0;
  const int transX = (int) odometryOffset.translation.x;
  const int transY = (int) odometryOffset.translation.y;
  const double dist = odometryOffset.translation.abs();
  const double angle = fabs(odometryOffset.rotation);

  lastOdometry += Pose2D(odometryOffset.rotation, transX, transY);

  // precalculate rotational error that has to be adapted to all samples
  const double rotError = std::max(rotNoise,
                             std::max(dist * theSelfLocatorParameter.movedDistWeight, 
                                      angle * theSelfLocatorParameter.movedAngleWeight));

  // precalculate translational error that has to be adapted to all samples
  const int transXError = (int) std::max(transNoise,
                                   std::max(fabs(transX * theSelfLocatorParameter.majorDirTransWeight),
                                            fabs(transY * theSelfLocatorParameter.minorDirTransWeight)));	
  const int transYError = (int) std::max(transNoise,
                                   std::max(fabs(transY * theSelfLocatorParameter.majorDirTransWeight),
                                            fabs(transX * theSelfLocatorParameter.minorDirTransWeight)));	
  for(int i = 0; i < numberOfSamples; i++)
  {
    Sample& s = samples[i];

    // the translational error vector
    const Vector2<int> transOffset(transX - transXError + (((transXError << 1) + 1) * (rand() & 0x3ff) >> 10),
                                   transY - transYError + (((transYError << 1) + 1) * (rand() & 0x3ff) >> 10));

    // update the sample
    s.translation += Vector2<int>((transOffset.x * s.rotation.x - transOffset.y * s.rotation.y) >> 10,
                                  (transOffset.x * s.rotation.y + transOffset.y * s.rotation.x) >> 10);
    s.angle += odometryOffset.rotation + (randomDouble() * 2 - 1) * rotError;
    s.angle = normalize(s.angle);
    s.rotation = Vector2<int>(int(cos(s.angle) * 1024), int(sin(s.angle) * 1024));
  }	
}

void SelfLocator::sensorUpdate()
{
  // select points for sensor update
  totalWeighting = 0;
  selectedIndices.resize(theSelfLocatorParameter.numberOfPoints);
  for(int j = 0; j < theSelfLocatorParameter.numberOfPoints; ++j)
    selectedIndices[j] = pointIndices[rand() % pointIndices.size()];

  for(int i = 0; i < numberOfSamples; ++i)
  {
    Sample& s = samples[i];

    const Vector2<int> origin(s.translation.x + ((xOrigin * s.rotation.x - yOrigin * s.rotation.y) >> 10),
                              s.translation.y + ((xOrigin * s.rotation.y + yOrigin * s.rotation.x) >> 10)),
                       rotation((rotCos * s.rotation.x - rotSin * s.rotation.y) >> 10,
                                (rotCos * s.rotation.y + rotSin * s.rotation.x) >> 10);

    s.weighting = 1.0;
    for(int j = 0; j < theSelfLocatorParameter.numberOfPoints; ++j)
    {
      const PointsPercept::Point& pp = thePointsPercept.points[selectedIndices[j]];
      const Vector2<int> pObs(origin.x + ((pp.onField.x * rotation.x - pp.onField.y * rotation.y) >> 10),
                              origin.y + ((pp.onField.x * rotation.y + pp.onField.y * rotation.x) >> 10));
      OrientationConstraints o = noOrientationConstraint;
      LengthConstraints l = noLengthConstraint;
      if(pp.type >= PointsPercept::numberOfBeaconPointTypes)
      {
        if(pp.link != i && !thePointsPercept.points[pp.link].isCloser)
        {
          const Vector2<int>& other = thePointsPercept.points[pp.link].onField;
          Vector2<int> pOther(origin.x + ((other.x * rotation.x - other.y * rotation.y) >> 10),
                              origin.y + ((other.x * rotation.y + other.y * rotation.x) >> 10)),
                       diff = pOther - pObs;
          if(std::abs(diff.x) > std::abs(diff.y))
            o = alongField;
          else
            o = acrossField;
        }
        if((pp.type == PointsPercept::fieldLine && pp.length2 > sqr(theSelfLocatorParameter.maxCrossingLength)) ||
           (pp.type != PointsPercept::fieldLine && pp.length2 > sqr(theSelfLocatorParameter.maxBeaconLength)))
          l = longSegment;
      }
      int type = pp.type;
      if(type < PointsPercept::numberOfTeamDependentPointTypes && theOwnTeamInfo.teamColor == TEAM_YELLOW)
        type ^= 1;
      Vector2<int> pModel = closestPointsTable[(l * numberOfOrientationConstraints + o) * numberOfTableTypes + type].getClosestPoint(pObs),
                         diffOnField = pModel - pObs,
                                       diffFromCamera((diffOnField.x * rotation.x + diffOnField.y * rotation.y) >> 10,
                                                      (diffOnField.y * rotation.x - diffOnField.x * rotation.y) >> 10);
      s.weighting *= gaussian.value(
        std::abs(1024 * diffFromCamera.x / (std::abs(pp.onField.x) + 450)),
        theSelfLocatorParameter.standardDeviation);
      s.weighting *= gaussian.value(
        pp.onField.x != 0 ? std::abs(1024 * diffFromCamera.y / pp.onField.x) : 0,
        theSelfLocatorParameter.standardDeviation);
    }
    totalWeighting += samples[i].weighting;
  }
  ASSERT(totalWeighting);
  const double averageWeighting = totalWeighting / numberOfSamples;
  if(slowWeighting)
  {
    slowWeighting = slowWeighting + theSelfLocatorParameter.alphaSlow * (averageWeighting - slowWeighting);
    fastWeighting = fastWeighting + theSelfLocatorParameter.alphaFast * (averageWeighting - fastWeighting);
  }
  else
  {
    slowWeighting = averageWeighting;
    if(theSelfLocatorParameter.knownStartPose)     
      fastWeighting = averageWeighting;   // Must be done to avoid complete re-init in first cycle
  }

  PLOT("module:ParticleFilterSelfLocator:averageWeighting", averageWeighting * 1e3);
  PLOT("module:ParticleFilterSelfLocator:slowWeighting", slowWeighting * 1e3);
  PLOT("module:ParticleFilterSelfLocator:fastWeighting", fastWeighting * 1e3);
}

void SelfLocator::resampling()
{
  // swap sample arrays
  Sample* oldSet = samples.swap();
  const double weightingsSum(totalWeighting);
  const double resamplingPercentage(perceptBuffer.getNumberOfEntries() ? std::max(0.0, 1.0 - fastWeighting / slowWeighting) : 0.0);
  const double numberOfResampledSamples = theSelfLocatorParameter.disableSensorResetting ? numberOfSamples : numberOfSamples * (1.0 - resamplingPercentage);
  const double threshold = theSelfLocatorParameter.resamplingThreshold * weightingsSum / numberOfSamples;
  const double weightingBetweenTwoDrawnSamples((weightingsSum + threshold * numberOfSamples) / numberOfResampledSamples);
  double nextPos(randomDouble() * weightingBetweenTwoDrawnSamples); 
  double currentSum(0);

  // resample:
  int j(0);
  for(int i = 0; i < numberOfSamples; ++i)
  {
    currentSum += oldSet[i].weighting + threshold;
    while(currentSum > nextPos && j < numberOfSamples)
    {
      samples[j++] = oldSet[i];
      nextPos += weightingBetweenTwoDrawnSamples;
    }
  }

  // fill up remaining samples with new poses:
  if(perceptBuffer.getNumberOfEntries())
    for(; j < numberOfSamples; ++j)
      generateTemplate(samples[j]);
  else if(j) // in rare cases, a sample is missing, so add one (or more...)
    for(; j < numberOfSamples; ++j)
      samples[j] = samples[rand() % j];
  else // resampling was not possible (for unknown reasons), so create a new sample set (fail safe)
#ifdef NDEBUG
  {
    for(int i = 0; i < numberOfSamples; ++i)
    {
      Sample& sample = samples[i];
      Pose2D pose(theFieldDimensions.randomPoseOnField());
      sample.translation = Vector2<int>(int(pose.translation.x), int(pose.translation.y));
      sample.rotation = Vector2<int>(int(cos(pose.rotation) * 1024), int(sin(pose.rotation) * 1024));
    }
    poseCalculator->init();
  }
#else
    ASSERT(false);
#endif
}

void SelfLocator::generateTemplate(Sample& sample)
{
  const Percept& p = perceptBuffer[rand() % perceptBuffer.getNumberOfEntries()]; // select point
  bool useBeacon = p.type < PointsPercept::numberOfBeaconPointTypes || (rand() % 100 <= theSelfLocatorParameter.beaconTemplateRatio && p.canBeBeacon); // goal or beacon?
  int length = p.ray.abs() + (useBeacon ? (int) theFieldDimensions.flagRadius : 0), // length from origin
      origin = useBeacon ? theFieldDimensions.xPosHalfWayLine
                         : theFieldDimensions.yPosLeftGoal + ((theFieldDimensions.yPosRightGoal - theFieldDimensions.yPosLeftGoal) * (rand() & 0x3ff) >> 10), // origin of construction (y-coordinate)
      minValue = std::max(origin - (15 * length >> 4), useBeacon ? theFieldDimensions.xPosOwnGroundline : theFieldDimensions.yPosRightFieldBorder), // minimum y-coordinate of constructed pose
      maxValue = std::min(origin + (15 * length >> 4), useBeacon ? theFieldDimensions.xPosOpponentGroundline : theFieldDimensions.yPosLeftFieldBorder), // maximum y-coordinate of constructed pose
      yPos = minValue + ((((maxValue - minValue) << 1) + 1) * (rand() & 0x3ff) >> 11), // y-coordinate of constructed pose
      xPos = (int) sqrt(float(sqr(length) - sqr(yPos - origin))); // since y is known, x can be computed

  Vector2<int> dir(xPos, yPos - origin); // direction from origin to constructed point
  
  // finish position of template
  if((static_cast<unsigned int>(p.type) & 1) == static_cast<unsigned int>(theOwnTeamInfo.teamColor))
  {
    dir.x = -dir.x;
    sample.translation = Vector2<int>((useBeacon ? theFieldDimensions.yPosRightFlags : theFieldDimensions.xPosOwnGoal) + xPos, yPos);
  }
  else
    sample.translation = Vector2<int>((useBeacon ? theFieldDimensions.yPosLeftFlags : theFieldDimensions.xPosOpponentGoal) - xPos, yPos);

  if(useBeacon)
  { // swap x and y
    sample.translation = Vector2<int>(sample.translation.y, sample.translation.x);
    dir = Vector2<int>(-dir.y, -dir.x);
  }

  // construct rotation of template
  Vector2<int> rotation(dir.x * p.ray.x - dir.y * p.ray.y, -dir.x * p.ray.y - dir.y * p.ray.x);
  length = rotation.abs();
  rotation = length > (1 << 20) ? rotation / (length >> 10) : rotation * 1024 / length;

  // so far the template is for the time when the landmark point was perceived.
  // add odometry offset since then to the pose

  // calc odometry offset
  Vector2<int> offset(xOdoOrigin - p.xOrigin, yOdoOrigin - p.yOrigin);
  offset = Vector2<int>((offset.x * p.rotCos + offset.y * p.rotSin) >> 10,
                        (offset.y * p.rotCos - offset.x * p.rotSin) >> 10);
  Vector2<int> rot((rotOdoCos * p.rotCos + rotOdoSin * p.rotSin) >> 10,
                   (rotOdoSin * p.rotCos - rotOdoCos * p.rotSin) >> 10);

  // add offset to sample pose
  sample.translation += Vector2<int>((offset.x * rotation.x - offset.y * rotation.y) >> 10,
                                     (offset.y * rotation.x + offset.x * rotation.y) >> 10);
  sample.rotation = Vector2<int>((rot.x * rotation.x - rot.y * rotation.y) >> 10,
                                 (rot.y * rotation.x + rot.x * rotation.y) >> 10);
  sample.angle = sample.rotation.angle();
  sample.weighting = 0;
  sample.cluster = poseCalculator->getNewClusterIndex();
}

void SelfLocator::drawSamples()
{
  const double maxWeighting = 2*totalWeighting / numberOfSamples;
  for(int i = 0; i < samples.size(); ++i)
  {
    const Pose2D pose(samples[i].angle, samples[i].translation.x, samples[i].translation.y);
    unsigned char weighting = (unsigned char) (samples[i].weighting / maxWeighting * 255);
    Vector2<double> bodyPoints[4] = {Vector2<double>(55,90),
                                     Vector2<double>(-55,90),
                                     Vector2<double>(-55,-90),
                                     Vector2<double>(55,-90)};
    for(int j = 0; j < 4; ++j)
    {
      bodyPoints[j] = Geometry::rotate(bodyPoints[j], pose.rotation);
      bodyPoints[j] += pose.translation;
    }
    Vector2<double> headPos(30,0);
    headPos = Geometry::rotate(headPos, pose.rotation);
    headPos += pose.translation;
    ColorRGBA color = samples[i].weighting ? ColorRGBA(weighting,weighting,weighting) : ColorRGBA(255,0,0);
    if(samples[i].cluster == poseCalculator->getIndexOfBestCluster())
      color = ColorRGBA(200,0,200);
    POLYGON("module:ParticleFilterSelfLocator:samples", 4, bodyPoints, 
        1, // pen width 
        Drawings::ps_solid, 
        ColorRGBA(180,180,180),
        Drawings::bs_solid, 
        color);
    CIRCLE("module:ParticleFilterSelfLocator:samples", 
        headPos.x, 
        headPos.y, 
        30, 
        1, // pen width 
        Drawings::ps_solid, 
        ColorRGBA(180,180,180),
        Drawings::bs_solid, 
        ColorRGBA(180,180,180));
  }
}

template<int xSize,int ySize, int cellSize> 
void SelfLocator::ClosestPointsTable<xSize, ySize, cellSize>::create(
  const FieldDimensions::LinesTable& table,
  double orientation,
  int numberOfOrientations,
  double minLength)
{
  const short conversionDifference = -std::numeric_limits<short>::min();
  for(int y = 0; y < ySize; ++y)
    for(int x = 0; x < xSize; ++x)
    {
      Vector2<double> px((x + 0.5 - xSize / 2) * cellSize, (y + 0.5 - ySize / 2) * cellSize);
      Vector2<double> p;
      VERIFY(table.getClosestPoint(p, Pose2D(orientation, px), numberOfOrientations, minLength));
      points[y][x][0] = static_cast<short>(p.x + conversionDifference) - conversionDifference;
      points[y][x][1] = static_cast<short>(p.y + conversionDifference) - conversionDifference;
    }
}

MAKE_MODULE(SelfLocator, Modeling)
