/**
* @file SelfLocator.h
* Declares a class that performs self-localization using a particle filter.
*
* The difference to the ParticleFilterSelfLocator is as follows:
*
* - May choose different implementations for pose calculation
*
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef _SelfLocator_H_
#define _SelfLocator_H_

#include "Tools/Module/Module.h"
#include "Tools/SampleSet.h"
#include "Representations/Configuration/FieldDimensions.h"
#include "Representations/Infrastructure/GameInfo.h"
#include "Representations/Infrastructure/KeyStates.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/TeamInfo.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/MotionControl/OdometryData.h"
#include "Representations/Modeling/RobotPose.h"
#include "Representations/Modeling/SelfLocatorParameter.h"
#include "Tools/RingBuffer.h"
#include "Tools/Math/GaussianTable.h"
#include "PointValidityChecker.h"
#include <vector>


template <typename Sample, typename SampleContainer>
class PoseCalculator;


MODULE(SelfLocator)
  REQUIRES(FieldDimensions)
  REQUIRES(GameInfo)
  REQUIRES(RobotInfo)
  REQUIRES(OwnTeamInfo)
  REQUIRES(KeyStates)
  REQUIRES(PointsPercept)
  REQUIRES(BallPercept)
  REQUIRES(OdometryData)
  REQUIRES(CameraMatrix)
  REQUIRES(SelfLocatorParameter)
  PROVIDES_WITH_MODIFY_AND_DRAW(RobotPose)
END_MODULE


/**
* @class SelfLocator
* A SelfLocator using a particle filter
*/
class SelfLocator : public SelfLocatorBase
{
private:
  /**
  * A sample/particle.
  */
  class Sample
  {
  public:
    Vector2<int> translation,   /**< The position in mm. */
                 rotation;      /**< Cosinus and sinus of the rotation multiplied by 1024. */
    double angle,               /**< The rotation in radians. */
           weighting;           /**< The weighting of a sample*/
    int cluster;                /**< The number of the cluster this sample belongs to*/
    Sample* next;               /**< The next robot sample when clustering samples using binning. */
    
    /** Default constructor*/
    Sample() : rotation(1024, 0), angle(0), weighting(0), cluster(0), next(0) {}
  };

  /**
  * An entry of a buffer that stores edge points from which sample templates can be constructed.
  */
  class Percept
  {
  public:
    Vector2<int> ray; /**< The ray from robot to edge point. Contains camera offset. */
    PointsPercept::PointType type; /**< The type of the edge point (yellowGoal/blueGoal). */
    bool canBeBeacon; /**< Whether the edge is short enough to be a beacon. */
    int xOrigin, /**< The x-coordinate of odometry when this edge point was perceived. */
        yOrigin, /**< The y-coordinate of odometry when this edge point was perceived. */
        rotCos, /**< The cosinus of the rotation of odometry when this edge point was perceived (10 bit fix point). */
        rotSin; /**< The sinus of the rotation of odometry when this edge point was perceived (10 bit fix point). */

    /**
    * Default constructor. Initializes nothing.
    */
    Percept() {}

    /**
    * Constructor.
    * @param ray The ray from robot to edge point. Contains camera offset.
    * @param type The type of the edge point (yellowGoal/blueGoal).
    * @param canBeBeacon Whether the edge is short enough to be a beacon.
    * @param xOrigin The x-coordinate of odometry when this edge point was perceived.
    * @param yOrigin The y-coordinate of odometry when this edge point was perceived.
    * @param rotCos The cosinus of the rotation of odometry when this edge point was perceived (10 bit fix point).
    * @param rotSin The sinus of the rotation of odometry when this edge point was perceived (10 bit fix point).
    */
    Percept(const Vector2<int>& ray, PointsPercept::PointType type, bool canBeBeacon,
            int xOrigin, int yOrigin, int rotCos, int rotSin) :
      ray(ray),
      type(type),
      canBeBeacon(canBeBeacon),
      xOrigin(xOrigin),
      yOrigin(yOrigin),
      rotCos(rotCos),
      rotSin(rotSin) {}
  };

  /**
  * The class realizes a table of closest points on the field model.
  */
  template<int xSize, int ySize, int cellSize> class ClosestPointsTable
  {
  private:
    short points[ySize][xSize][2]; /**< The table. For each [y][x] position a x- ([0]) and a y-coordinate ([1]). */

  public:
    /**
    * The method creates the table.
    * @param table The table of lines the closest points will be searched in.
    * @param orientation The orientation of the lines that are considered.
    * @param numberOfOrientations How many different orientations will be distinguished between?
    * @param minLength The minimum length of lines to be considered.
    */
    void create(const FieldDimensions::LinesTable& table,
                double orientation,
                int numberOfOrientations,
                double minLength);

    /**
    * The method returns the closest line point to the given point.
    * @param point The given point.
    * @return The line point closest to the given point.
    */
    Vector2<int> getClosestPoint(const Vector2<int>& point) const
    {
      int x = int((point.x  + (xSize / 2)*cellSize) / cellSize),
          y = int((point.y  + (ySize / 2)*cellSize) / cellSize);
      if(x < 0)
        x = 0;
      else if(x >= xSize)
        x = xSize - 1;
      if(y < 0)
        y = 0;
      else if(y >= ySize)
        y = ySize - 1;
      return Vector2<int>(points[y][x][0], points[y][x][1]);
    }

    /**
    * The method draws the table (by subsampling).
    */
    void draw()
    {
      for(int y = 0; y < ySize; y += 5)
        for(int x = 0; x < xSize; x += 5)
        {
          int x1 = (x - xSize / 2) * cellSize,
              y1 = (y - ySize / 2) * cellSize;
          Vector2<int> closest = getClosestPoint(Vector2<int>(x1, y1));
          LINE("module:ParticleFilterSelfLocator:closestPoints", x1, y1, closest.x, closest.y, 1, Drawings::ps_solid, ColorClasses::black);
        }
    }

    /**
    * The method writes the table to a stream.
    * @param stream The stream the table is written to.
    */
    void write(Out& stream) const {stream.write(points, sizeof(points));}

    /**
    * The method reads the table from a stream.
    * @param stream The stream the table is read from.
    */
    void read(In& stream) {stream.read(points, sizeof(points));}
  };

  enum LengthConstraints
  {
    noLengthConstraint,
    longSegment,
    numberOfLengthConstraints
  };

  enum OrientationConstraints
  {
    noOrientationConstraint,
    alongField,
    acrossField,
    numberOfOrientationConstraints
  };

  enum 
  {
    numberOfTableTypes = 3, /**< The number of point types in tables. */
    numberOfTables = PointsPercept::numberOfBeaconPointTypes + numberOfLengthConstraints * numberOfOrientationConstraints * numberOfTableTypes, /**< The number of tables to match points against. */
    numberOfSamples = 100 /**< The number of samples. */
  };

  enum PoseCalculatorType
  {
    POSE_CALCULATOR_2D_BINNING = 0,
    POSE_CALCULATOR_BEST_PARTICLE,
    POSE_CALCULATOR_OVERALL_AVERAGE,
    NUMBER_OF_POSE_CALCULATORS
  } poseCalculatorType;

  /** 
  * The function returns names for different pose calculator types
  * @param type The type
  * @return The corresponding name.
  */
  static const char* getPoseCalculatorTypeName(PoseCalculatorType type)
  {
    switch (type)
    {
    case POSE_CALCULATOR_2D_BINNING: return "PoseCalculator2DBinning";
    case POSE_CALCULATOR_BEST_PARTICLE: return "PoseCalculatorBestParticle";
    case POSE_CALCULATOR_OVERALL_AVERAGE: return "PoseCalculatorOverallAverage";
    default: return "none";
    }
  }

  ClosestPointsTable<310, 225, 24> closestPointsTable[numberOfTables]; /**< The tables for the three line types. */
  SampleSet<Sample, numberOfSamples> samples; /**< Container for all samples. */
  double totalWeighting; /**< The current weighting sum of all samples. */
  double slowWeighting; /**< This value follows the average weighting slowly. */
  double fastWeighting; /**< This value follows the average weighting more quickly. */
  Pose2D lastOdometry; /**< The last odometry state for calculating the offset since the last call. */
  std::vector<short> pointIndices, /**< The indices of all points in the points percept that might used by this module. */
                     selectedIndices; /**< The indices of the points in the points percept that were selected for the sensor update. */
  int xOrigin, /**< The x-coordinate of the relative origin of the points percept in integer. */
      yOrigin, /**< The y-coordinate of the relative origin of the points percept in integer. */
      rotCos, /**< The cosinus of the relative orientation of the points percept with a precision of 10 bit fractional digits. */
      rotSin, /**< The sinus of the relative orientation of the points percept with a precision of 10 bit fractional digits. */
      xOdoOrigin, /**< The x-coordinate of the odometry in integer. */
      yOdoOrigin, /**< The y-coordinate of the odometry in integer. */
      rotOdoCos, /**< The cosinus of the rotation of the odometry with a precision of 10 bit fractional digits. */
      rotOdoSin; /**< The sinus of the rotation of the odometry with a precision of 10 bit fractional digits. */
  bool updatedBySensors; /**< Was there an actual update during the previous observation update? */
  GaussianTable gaussian; /**< A table of Gaussians for assessing bearing measurements to points. */
  RingBuffer<Percept, 10> perceptBuffer; /**< A buffer of percepts from which templates can be generated. */
  PoseCalculator< Sample, SampleSet<Sample, numberOfSamples> >* poseCalculator; /**< External class for computing a pose from a set of samples */
  PointValidityChecker pointSelector;  /**< External class for selecting relevant points. */
  unsigned lastReenterTimeStamp; /**< Point of time of last re-enter event*/

  /** 
  * The method provides the robot pose.
  * @param robotPose The robot pose representation that is updated by this module.
  */
  void update(RobotPose& robotPose);

  /** 
  * The method prepares execution and initializes some values.
  */
  void preExecution();

  /** 
  * The method performs the motion update step of particle filter localization.
  * @param noise Add random noise to the particles?
  */
  void motionUpdate(bool noise);

  /** 
  * The method performs the sensor update step of particle filter localization.
  */
  void sensorUpdate();

  /** 
  * The method performs the resampling step of particle filter localization.
  * It might add new samples generated from templates.
  */
  void resampling();

  /** 
  * Computes a template pose.
  * @param sample The resulting template sample.
  */
  void generateTemplate(Sample& sample);

  /** 
  * Draws the particles.
  */
  void drawSamples();

  /** 
  * A method that contains all intialisations that require 
  * representations from the blackboard.
  */
  virtual void init();

  /** 
  * Initializes the sample set.
  * If the vectors are empty (i.e. no positions given) the samples will
  * represent a uniform probability distribution which covers the whole state space.
  * If one or more positions are given, the samples will be randomly (again uniform) assigned
  * to these positions. Each position is represented by a gaussian distribution. Within these
  * distributions, the samples are randomly placed.
  * @param poses A vector of preferred sample poses
  * @param standardDeviations The standard deviations of the distributions of the points
  */
  void initSamplesAtGivenPositions(const std::vector<Pose2D>& poses,
    const std::vector<Pose2D>& standardDeviations);

public:
  /** Default constructor. */
  SelfLocator();

  /** Destructor */
  ~SelfLocator();
};

#endif
