/** 
* @file Modules/MotionControl/MotionCombinator.cpp
* This file implements a module that combines the motions created by the different modules.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#include "MotionCombinator.h"

void MotionCombinator::update(JointRequest& jointRequest)
{
  headMotion(jointRequest);

  const JointRequest* jointRequests[MotionRequest::numOfMotions] = 
  {
    &theWalkingEngineOutput,
    &theSpecialActionsOutput,
    &theWalkingEngineStandOutput
  };

  int i;
  for(i = 0; i < MotionRequest::numOfMotions; ++i)
    if(theMotionSelection.ratios[i] == 1)
    {
      copy(*jointRequests[i], jointRequest);

      // default values
      motionInfo.executedMotionRequest.motion = MotionRequest::Motion(i);
      motionInfo.isMotionStable = true;
      motionInfo.positionInWalkCycle = 0;

      lastJointData = theSafeJointData;
  
      if(theMotionSelection.ratios[MotionRequest::walk] == 1)
      {
        odometryData += theWalkingEngineOutput.odometryOffset;
        motionInfo.executedMotionRequest.walkRequest.speed = theWalkingEngineOutput.speed;
        motionInfo.positionInWalkCycle = theWalkingEngineOutput.positionInWalkCycle;
      }
      else if(theMotionSelection.ratios[MotionRequest::specialAction] == 1)
      {
        odometryData += theSpecialActionsOutput.odometryOffset;
        motionInfo.executedMotionRequest.specialActionRequest = theSpecialActionsOutput.executedSpecialAction;
        motionInfo.isMotionStable = theSpecialActionsOutput.isMotionStable;
      }
      else if(theMotionSelection.ratios[MotionRequest::stand] == 1)
      {
        motionInfo.executedMotionRequest.motion = MotionRequest::specialAction;
        motionInfo.executedMotionRequest.specialActionRequest.specialAction = SpecialActionRequest::stand;
        motionInfo.executedMotionRequest.specialActionRequest.mirror = false;
      }
      break;
    }

  if(i == MotionRequest::numOfMotions)
  {
    for(i = 0; i < MotionRequest::numOfMotions; ++i)
      if(i != theMotionSelection.targetMotion && theMotionSelection.ratios[i])
      {
        interpolate(*jointRequests[i], *jointRequests[theMotionSelection.targetMotion], theMotionSelection.ratios[theMotionSelection.targetMotion], jointRequest);
        break;
      }
    ASSERT(i < MotionRequest::numOfMotions);
    
    if(theMotionSelection.targetMotion == MotionRequest::walk)
      motionInfo.positionInWalkCycle = theWalkingEngineOutput.positionInWalkCycle;
    else
      motionInfo.positionInWalkCycle = 0;
  }

#ifndef RELEASE
  JointDataDeg jointRequestDeg(jointRequest);
#endif
  MODIFY("representation:JointRequestDeg", jointRequestDeg);
}

void MotionCombinator::copy(const JointRequest& source, JointRequest& target) const
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
    if(source.angles[i] != JointData::ignore)
      target.angles[i] = source.angles[i];
}

void MotionCombinator::interpolate(const JointRequest& from, const JointRequest& to,
                                   double toRatio, JointRequest& target) const
{
  for(int i = 0; i < JointData::numOfJoints; ++i)
  {
    double f = from.angles[i],
      t = to.angles[i];
    if(t != JointData::ignore)
    {
      if(t == JointData::off)
      {
        target.angles[i] = JointData::off;
      }
      else
      {
        if(f == JointData::off || f == JointData::ignore)
          f = lastJointData.angles[i];
        target.angles[i] = toRatio * t + (1 - toRatio) * f;
      }
    }  
  }
}

void MotionCombinator::headMotion(JointRequest& jointRequest) const
{
  double speed = theHeadMotionRequest.speed * theRobotDimensions.motionCycleTime,
    lastPan = jointRequest.angles[JointData::headPan] == JointData::off ? theSafeJointData.angles[JointData::headPan] : jointRequest.angles[JointData::headPan],
    lastTilt = jointRequest.angles[JointData::headTilt] == JointData::off ? theSafeJointData.angles[JointData::headTilt] : jointRequest.angles[JointData::headTilt],
    diffPan = theHeadMotionRequest.pan == JointData::off ? 0 : theHeadMotionRequest.pan - lastPan,
    diffTilt = theHeadMotionRequest.tilt == JointData::off ? 0 : theHeadMotionRequest.tilt - lastTilt,
    diffMax = std::max(fabs(diffTilt), fabs(diffPan));

  if(diffMax > speed)
  {
    double factor =  speed / diffMax;
    diffPan *= factor;
    diffTilt *= factor;
  }

  jointRequest.angles[JointData::headPan] = theHeadMotionRequest.pan == JointData::off ? JointData::off : lastPan + diffPan;
  jointRequest.angles[JointData::headTilt] = theHeadMotionRequest.tilt == JointData::off ? JointData::off : lastTilt + diffTilt;
}

MAKE_MODULE(MotionCombinator, Motion Control)
