/** 
* @file Modules/MotionControl/MotionSelector.cpp
* This file implements a module that is responsible for controlling the motion.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#include "MotionSelector.h"

void MotionSelector::update(MotionSelection& motionSelection)
{
  static const int interpolationTimes[MotionRequest::numOfMotions] =
  {
    1000, // to walk
    200, // to specialAction
    200 // to stand
  };

  MotionRequest motionRequest = theMotionRequest;
  if(motionRequest.motion == MotionRequest::specialAction && 
     motionRequest.specialActionRequest.specialAction == SpecialActionRequest::stand)
    motionRequest.motion = MotionRequest::stand;

  // immediately switch to standup/block motions
  if(motionRequest.motion == MotionRequest::specialAction &&
     (motionRequest.specialActionRequest.specialAction == SpecialActionRequest::standUpBack ||
      motionRequest.specialActionRequest.specialAction == SpecialActionRequest::standUpFront))
  {
    if(lastMotion != motionRequest.motion)
      prevMotion = lastMotion;
    lastMotion = motionRequest.motion;
    switchTime = 0;
  }

  if(motionRequest.motion != lastMotion && theFrameInfo.time)
  {
    // default values
    motionSelection = MotionSelection();
    motionSelection.specialActionMode = MotionSelection::deactive;
    motionSelection.ratios[MotionRequest::specialAction] = 0;
    motionSelection.targetMotion = lastMotion;

    if((lastMotion == MotionRequest::walk && (!&theWalkingEngineOutput || theWalkingEngineOutput.isLeavingPossible)) ||
       lastMotion == MotionRequest::stand || // stand can always be left
       (lastMotion == MotionRequest::specialAction && (!&theSpecialActionsOutput || theSpecialActionsOutput.isLeavingPossible)))
    {
      switchTime = theFrameInfo.time;
      prevMotion = lastMotion;
      lastMotion = motionRequest.motion;
    }
    else if(lastMotion == MotionRequest::specialAction)
    {
      motionSelection.specialActionMode = MotionSelection::active;
      motionSelection.ratios[MotionRequest::specialAction] = 1;
      motionSelection.specialActionRequest.specialAction = SpecialActionRequest::numOfSpecialActions; // go to extern
    }
    else if(lastMotion < MotionRequest::numOfMotions)
      motionSelection.ratios[lastMotion] = 1;
    else // only called the first time
      ASSERT(false);
  }

  if(motionRequest.motion == lastMotion)
  {
    // default values
    motionSelection = MotionSelection();
    motionSelection.specialActionMode = MotionSelection::deactive;
    motionSelection.ratios[MotionRequest::specialAction] = 0;
    motionSelection.targetMotion = lastMotion;

    //no interpolation
    if(!switchTime || theFrameInfo.getTimeSince(switchTime) > std::max(interpolationTimes[motionRequest.motion], playDeadDelay))
    {
      motionSelection.ratios[motionRequest.motion] = 1;
      playDeadDelay = 0;
      if(motionRequest.motion == MotionRequest::specialAction)
      {
        motionSelection.specialActionMode = MotionSelection::active;
        motionSelection.specialActionRequest = motionRequest.specialActionRequest;
        motionSelection.specialActionRequest.specialize(theRobotInfo.robotModel);
        if(motionSelection.specialActionRequest.specialAction == SpecialActionRequest::playDead)
          playDeadDelay = 2000;
      }
      else if(motionRequest.motion == MotionRequest::walk)
        motionSelection.walkRequest = motionRequest.walkRequest;
      else
        ASSERT(motionRequest.motion == MotionRequest::stand);
    }
    //interpolation
    else
    {
      motionSelection.ratios[motionRequest.motion] = double(theFrameInfo.getTimeSince(switchTime)) / std::max(interpolationTimes[motionRequest.motion], playDeadDelay);
      motionSelection.ratios[prevMotion] = 1 - motionSelection.ratios[motionRequest.motion];
      if(motionRequest.motion == MotionRequest::specialAction)
      {
        motionSelection.specialActionMode = MotionSelection::first;
        motionSelection.specialActionRequest = motionRequest.specialActionRequest;
        motionSelection.specialActionRequest.specialize(theRobotInfo.robotModel);
      }
    }
  }
}

MAKE_MODULE(MotionSelector, Motion Control)
