/** 
* @file WalkingEngine.cpp
* This file implements a module that creates the walking motions.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
* @author <A href="mailto:Stefan.Czarnetzki@uni-dortmund.de">Stefan Czarnetzki</A>
* @author <a href="mailto:ciVic@web.de">Oliver Urbann</a> 
*/

#include "WalkingEngine.h"
#include "Platform/GTAssert.h"
#include "Tools/Debugging/Debugging.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Streams/InStreams.h"
#include "Tools/Math/Matrix.h"

WalkingEngine::WalkingEngine() :
phase(0),
wasActive(false),
bodyTiltGyroOffset(0),
bodyTiltLastGyroY(0),
lastBodyShift1(0),
lastBodyShift2(0),
bodyTiltOffset(0),
lastLeftRotation(0),
lastRightRotation(0)
{
  InConfigFile* stream = new InConfigFile(Global::getSettings().expandRobotFilename("walkingsprint.cfg"), "default");
  if(!stream->exists())
  {
    delete stream;
    stream = new InConfigFile("walkingsprint.cfg", "default");
  }

  delete stream;

  lastParameters = theWalkingParameters;
  resWalkingParameters = &theWalkingParameters;
  lastChangedTime = 0;
  wasInterpolating = false;
}

void WalkingEngine::init()
{
  footPositions[0] = theWalkingParameters.footPositions[0]->getPosition(0, 0, 0, 0);
  footPositions[1] = theWalkingParameters.footPositions[1]->getPosition(0.5, 0, 0, 0);
}

void WalkingEngine::checkCfg()
{
  const WalkingParameters& wp = theWalkingParameters;
  WalkingParameters& prev = lastParameters;

  int interpolationTime = 1000;
 
  if(!(wp == prev))
  {
    if(!wasInterpolating)
    {
      lastChangedTime = theFrameInfo.time;
      wasInterpolating = true;
    }
    
    if(theFrameInfo.getTimeSince(lastChangedTime) > interpolationTime && wasInterpolating)
    {
      wasInterpolating = false;
      prev = wp;
      resWalkingParameters = &wp;
      return;
    }

    double ratio = (theFrameInfo.getTimeSince(lastChangedTime))/((double)interpolationTime);

    PLOT("wpRatio", 100*ratio);

    res.stepDuration = (int)(wp.stepDuration * ratio + prev.stepDuration * (1-ratio));
    res.maxSpeedChange = wp.maxSpeedChange;
    res.odometryScale = wp.odometryScale * ratio + prev.odometryScale * (1-ratio);
    res.odometryRotationScaleCCW = wp.odometryRotationScaleCCW * ratio + prev.odometryRotationScaleCCW * (1-ratio);
    res.odometryRotationScaleCW = wp.odometryRotationScaleCW * ratio + prev.odometryRotationScaleCW * (1-ratio);
    res.maxForwardSpeed = wp.maxForwardSpeed * ratio + prev.maxForwardSpeed * (1-ratio);
    res.maxBackwardSpeed = wp.maxBackwardSpeed * ratio + prev.maxBackwardSpeed * (1-ratio);
    res.maxSidewaysSpeed = wp.maxSidewaysSpeed * ratio + prev.maxSidewaysSpeed * (1-ratio);
    res.maxRotationSpeed = wp.maxRotationSpeed * ratio + prev.maxRotationSpeed * (1-ratio);
    res.elbowAngle = wp.elbowAngle * ratio + prev.elbowAngle * (1-ratio);
    res.armDeactivationRatio = wp.armDeactivationRatio * ratio + prev.armDeactivationRatio * (1-ratio);
    res.bodyTiltOrigin = wp.bodyTiltOrigin * ratio + prev.bodyTiltOrigin * (1-ratio);
    res.bodyTiltStandOrigin = wp.bodyTiltStandOrigin * ratio + prev.bodyTiltStandOrigin * (1-ratio);
    res.bodyTiltSpeedFactor = wp.bodyTiltSpeedFactor * ratio + prev.bodyTiltSpeedFactor * (1-ratio);
    res.bodyTiltCorrection = wp.bodyTiltCorrection * ratio + prev.bodyTiltCorrection * (1-ratio);
    res.bodyShiftCorrection = wp.bodyShiftCorrection * ratio + prev.bodyShiftCorrection * (1-ratio);
    res.bodyShiftToAccel = wp.bodyShiftToAccel * ratio + prev.bodyShiftToAccel * (1-ratio);
    res.footPanCorrection = wp.footPanCorrection * ratio + prev.footPanCorrection * (1-ratio);
    res.footRollCorrection = wp.footRollCorrection * ratio + prev.footRollCorrection * (1-ratio);
    res.bodyShiftSidewaysOriginFactorLeft = wp.bodyShiftSidewaysOriginFactorLeft * ratio + prev.bodyShiftSidewaysOriginFactorLeft * (1-ratio);
    res.bodyShiftSidewaysOriginFactorRight = wp.bodyShiftSidewaysOriginFactorRight * ratio + prev.bodyShiftSidewaysOriginFactorRight * (1-ratio);
    res.stepHeight = wp.stepHeight * ratio + prev.stepHeight * (1-ratio);
    res.bodyShiftScale = wp.bodyShiftScale * ratio + prev.bodyShiftScale * (1-ratio);
    res.stepShape = wp.stepShape;
    res.stepOrigin = wp.stepOrigin * ratio + prev.stepOrigin * (1-ratio);
    res.stepTiltForward = wp.stepTiltForward * ratio + prev.stepTiltForward * (1-ratio);
    res.stepTiltBackward = wp.stepTiltBackward * ratio + prev.stepTiltBackward * (1-ratio);
    res.stepTiltSideways = wp.stepTiltSideways * ratio + prev.stepTiltSideways * (1-ratio);
    res.stepPhases[0] = wp.stepPhases[0] * ratio + prev.stepPhases[0] * (1-ratio);
    res.stepPhases[1] = wp.stepPhases[1] * ratio + prev.stepPhases[1] * (1-ratio);
    res.stepPhases[2] = wp.stepPhases[2] * ratio + prev.stepPhases[2] * (1-ratio);
    res.bodyShiftShape = wp.bodyShiftShape;
    res.bodyShiftOrigin = wp.bodyShiftOrigin * ratio + prev.bodyShiftOrigin * (1-ratio);
    res.bodyShiftScaleMode = wp.bodyShiftScaleMode;
    res.bodyShiftPhaseShift = wp.bodyShiftPhaseShift * ratio + prev.bodyShiftPhaseShift * (1-ratio);
    res.bodyShiftPhases[0] = wp.bodyShiftPhases[0] * ratio + prev.bodyShiftPhases[0] * (1-ratio);
    res.bodyShiftPhases[1] = wp.bodyShiftPhases[1] * ratio + prev.bodyShiftPhases[1] * (1-ratio);
    res.bodyShiftPhases[2] = wp.bodyShiftPhases[2] * ratio + prev.bodyShiftPhases[2] * (1-ratio);
    res.rotationShape = wp.rotationShape;
    res.rotationOrigin = wp.rotationOrigin * ratio + prev.rotationOrigin * (1-ratio);
    res.rotationScaleMode = wp.rotationScaleMode;
    res.rotationScale = wp.rotationScale * ratio + prev.rotationScale * (1-ratio);
    res.rotationPhaseShift = wp.rotationPhaseShift * ratio + prev.rotationPhaseShift * (1-ratio);
    res.rotationPhases[0] = wp.rotationPhases[0] * ratio + prev.rotationPhases[0] * (1-ratio);
    res.rotationPhases[1] = wp.rotationPhases[1] * ratio + prev.rotationPhases[1] * (1-ratio);
    res.rotationPhases[2] = wp.rotationPhases[2] * ratio + prev.rotationPhases[2] * (1-ratio);
    res.armShape = wp.armShape;
    res.armOrigin = wp.armOrigin * ratio + prev.armOrigin * (1-ratio);
    res.armScaleMode = wp.armScaleMode;
    res.armScale = wp.armScale * ratio + prev.armScale * (1-ratio);
    res.armPhaseShift = wp.armPhaseShift * ratio + prev.armPhaseShift * (1-ratio);
    res.armPhases[0] = wp.armPhases[0] * ratio + prev.armPhases[0] * (1-ratio);
    res.armPhases[1] = wp.armPhases[1] * ratio + prev.armPhases[1] * (1-ratio);
    res.armPhases[2] = wp.armPhases[2] * ratio + prev.armPhases[2] * (1-ratio);
    
    res.init();
    resWalkingParameters = &res;
  }
  else
    resWalkingParameters = &wp;
}

void WalkingEngine::update(WalkingEngineStandOutput& standOutput)
{
  DEBUG_RESPONSE("plot:angles", );
  DEBUG_RESPONSE("plot:gyroY", );
  DEBUG_RESPONSE("plot:accX", );
  
  bodyTiltGyroOffset = 0.99 * bodyTiltGyroOffset + theSafeSensorData.data[SensorData::gyroY];

  checkCfg();

  if(theMotionSelection.ratios[MotionRequest::stand])
  {
    double bodyTilt = resWalkingParameters->bodyTiltStandOrigin,
           bodyShift = resWalkingParameters->bodyShift->getPosition(0, 0);

    standOutput.angles[JointData::headPan] = JointData::ignore;
    standOutput.angles[JointData::headTilt] = JointData::ignore;

    standOutput.angles[JointData::legLeft0] = resWalkingParameters->rotation[0]->getPosition(0, 0);
    standOutput.angles[JointData::legRight0] = resWalkingParameters->rotation[1]->getPosition(0.5, 0);


    for(int side = 0; side < 2; ++side)
    {
      const Trajectory3D& t = *resWalkingParameters->footPositions[side];
      footPositions[side] = t.getPosition(side / 2.0, 0, 0, 0);
      calcLegJoints(footPositions[side], bodyTilt, bodyShift, 0, // no footRollCorrection when standing
        side ? JointData::legRight0 : JointData::legLeft0, standOutput);

      JointData::Joint arm = side ? JointData::armRight0 : JointData::armLeft0;
      standOutput.angles[arm] = resWalkingParameters->arm[side]->getPosition(0, 0);
      standOutput.angles[arm + 1] = theMotionSelection.ratios[MotionRequest::stand] == 1 && randomDouble() < resWalkingParameters->armDeactivationRatio ? JointData::off : resWalkingParameters->elbowAngle - pi_2;
      standOutput.angles[arm + 3] = theMotionSelection.ratios[MotionRequest::stand] == 1 && randomDouble() < resWalkingParameters->armDeactivationRatio ? JointData::off : -resWalkingParameters->elbowAngle;
      standOutput.angles[arm] +=calcGyroCorrection(theBalanceParameters.armPFactor, theBalanceParameters.armIFactor, theBalanceParameters.armDFactor, theBalanceParameters.armUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);

    }
  }
  PLOT("angles", toDegrees(standOutput.angles[JointData::legLeft2]));
  PLOT("gyroY", toDegrees(theSafeSensorData.data[SensorData::gyroY]));
  PLOT("accX", theSafeSensorData.data[SensorData::accX] * 100);
  bodyTiltLastGyroY = theSafeSensorData.data[SensorData::gyroY];
}

void WalkingEngine::update(WalkingEngineOutput& walkingEngineOutput)
{
  DEBUG_RESPONSE("plot:measuredBodyTilt", );
  DEBUG_RESPONSE("plot:bodyTiltError", );
  DEBUG_RESPONSE("plot:bodyTiltOffset", );

  checkCfg();

  DEBUG_RESPONSE("plot:wpbodyTilt",);
  DEBUG_RESPONSE("plot:wpRatio",);
  PLOT("wpbodyTilt", 100*resWalkingParameters->bodyTiltOrigin);

  if(theMotionSelection.ratios[MotionRequest::walk])
    update(walkingEngineOutput, *resWalkingParameters, wasActive);
  wasActive = theMotionSelection.ratios[MotionRequest::walk] != 0;
}

void WalkingEngine::update(WalkingEngineOutput& walkingEngineOutput, const WalkingParameters& theWalkingParameters, bool wasActive)
{
  Pose2D& currentSpeed = walkingEngineOutput.speed;
  bodyTiltGyroOffset = 0.999 * bodyTiltGyroOffset + theSafeSensorData.data[SensorData::gyroY];

  if(!wasActive)
  {
    lastIterationTime = theFrameInfo.time;
    //we start at with 0.25 to have both feet in the stance phase
    phase = 0.25;
    currentSpeed = Pose2D();
  }
  phase += (theFrameInfo.time - lastIterationTime) / (float)theWalkingParameters.stepDuration;
  if(phase > 1)
    phase -= 1;
  ASSERT(phase <= 1);
  lastIterationTime = theFrameInfo.time;

  Pose2D targetSpeed = theMotionSelection.walkRequest.speed;

  // invert odometry scale, so command executed hopefully result in desired speeds
  double odometryRotationScale = targetSpeed.rotation > 0
    ? theWalkingParameters.odometryRotationScaleCCW
    : theWalkingParameters.odometryRotationScaleCW;

  // clip maximum speeds
  //clip to ellipse
  #define USE_NEW_ROT_CLIPPING
  double f, f_rot;
  double maxSpeed;
  if(targetSpeed.translation.x >= 0)
  {
    //the factor we need to scale the y axis to get a circle
    f = (theWalkingParameters.maxForwardSpeed / theWalkingParameters.maxSidewaysSpeed);
    f_rot = theWalkingParameters.maxForwardSpeed / theWalkingParameters.maxRotationSpeed;
    maxSpeed = theWalkingParameters.maxForwardSpeed;  
  }
  else
  {
    f = (-theWalkingParameters.maxBackwardSpeed / theWalkingParameters.maxSidewaysSpeed);
    f_rot = theWalkingParameters.maxBackwardSpeed / theWalkingParameters.maxRotationSpeed;
    maxSpeed = -theWalkingParameters.maxBackwardSpeed;
  }

  Vector3<double> t;
  //transform the target speed ellipse to a circle
  t.x = targetSpeed.translation.x;
  t.y = targetSpeed.translation.y * f;
#ifdef USE_NEW_ROT_CLIPPING 
  t.z = targetSpeed.rotation * f_rot;
#else
  t.z = 0;
#endif

  double length = t.abs();
  //if speed is to fast
  if(length > maxSpeed) 
  {
    //scale to maxspeed
    t.normalize(maxSpeed);
    //translate back to ellipse
    t.y /= f;
    t.z /= f_rot;
    //set targetspeed
    targetSpeed.translation.x = t.x;
    targetSpeed.translation.y = t.y;
#ifdef USE_NEW_ROT_CLIPPING
    targetSpeed.rotation = t.z;
#endif
  }

#ifndef USE_NEW_ROT_CLIPPING
  if(targetSpeed.rotation < -theWalkingParameters.maxRotationSpeed)
    targetSpeed.rotation = -theWalkingParameters.maxRotationSpeed;
  else if(targetSpeed.rotation > theWalkingParameters.maxRotationSpeed)
    targetSpeed.rotation = theWalkingParameters.maxRotationSpeed;
#endif

  double xMax = theWalkingParameters.maxSpeedChange.translation.x * theRobotDimensions.motionCycleTime,
    yMax = theWalkingParameters.maxSpeedChange.translation.y * theRobotDimensions.motionCycleTime,
    rMax = theWalkingParameters.maxSpeedChange.rotation * theRobotDimensions.motionCycleTime;

  currentSpeed.translation.x += std::min(xMax, std::max(-xMax, (targetSpeed.translation.x - currentSpeed.translation.x)));
  currentSpeed.translation.y += std::min(yMax, std::max(-yMax, (targetSpeed.translation.y - currentSpeed.translation.y)));
  currentSpeed.rotation += std::min(rMax, std::max(-rMax, (targetSpeed.rotation - currentSpeed.rotation)));

  double speedRatio = theWalkingParameters.stepDuration / 4000.0, // 1000 ms, 1/2 step length, 2 steps per cycle
    xSize = currentSpeed.translation.x * speedRatio,
    ySize = currentSpeed.translation.y * speedRatio,
    rotation = currentSpeed.rotation * speedRatio * 4;

  double bodyTilt = theWalkingParameters.bodyTiltOrigin,
    bodyShift = theWalkingParameters.bodyShift->getPosition(phase, 1.0);

  //Sidewards walking bodyShift offset
  if(currentSpeed.translation.y > 0)
    bodyShift += currentSpeed.translation.y * theWalkingParameters.bodyShiftSidewaysOriginFactorLeft;
  else if(currentSpeed.translation.y < 0)
    bodyShift += currentSpeed.translation.y * theWalkingParameters.bodyShiftSidewaysOriginFactorRight;

  bodyTilt += currentSpeed.translation.x * theWalkingParameters.bodyTiltSpeedFactor;

  Vector2<double> rotationOffsets(-lastLeftRotation, lastRightRotation);
  lastLeftRotation = walkingEngineOutput.angles[JointData::legLeft0] = theWalkingParameters.rotation[0]->getPosition(phase, rotation / 4);
  lastRightRotation = walkingEngineOutput.angles[JointData::legRight0] = -theWalkingParameters.rotation[1]->getPosition(phase, rotation / 4);
  rotationOffsets += Vector2<double>(lastLeftRotation, -lastRightRotation);

  double shiftAccel = (bodyShift - 2 * lastBodyShift1 + lastBodyShift2) * theWalkingParameters.bodyShiftToAccel;
  lastBodyShift2 = lastBodyShift1;
  lastBodyShift1 = bodyShift;

  if(theSafeSensorData.data[SensorData::accX] != JointData::off)
  {
    double measuredBodyTilt = atan2(theSafeSensorData.data[SensorData::accX], -theSafeSensorData.data[SensorData::accZ]);
    PLOT("measuredBodyTilt", 10 * toDegrees(measuredBodyTilt));
    if(fabs(measuredBodyTilt) < 0.5)
    {
      bodyTiltError.add(measuredBodyTilt - bodyTilt);
      bodyTiltOffset += bodyTiltError.getAverage() * theWalkingParameters.bodyTiltCorrection;
      bodyTiltOffset = std::min(0.25, std::max(-0.25, bodyTiltOffset));
    }

    bodyShiftError.add(theSafeSensorData.data[SensorData::accY] - shiftAccel);
  }
  PLOT("bodyTiltError", 10 * toDegrees(bodyTiltError.getAverage()));
  PLOT("bodyTiltOffset", 10 * toDegrees(bodyTiltOffset));
  bodyTilt -= bodyTiltOffset;
  bodyShift -= bodyShiftError.getAverage() * theWalkingParameters.bodyShiftCorrection;

  Vector2<double> footOffsets[2];
  for(int side = 0; side < 2; ++side)
  {
    const Trajectory3D& t = *theWalkingParameters.footPositions[side];
    Vector3<double> lastFootPosition = footPositions[side];
    footPositions[side] = t.getPosition(phase, xSize, ySize, 0);
    footOffsets[side] = Vector2<double>(footPositions[side].x - lastFootPosition.x,
      footPositions[side].y - lastFootPosition.y);  
    calcLegJoints(footPositions[side], bodyTilt, bodyShift, theWalkingParameters.footRollCorrection,
      side ? JointData::legRight0 : JointData::legLeft0, walkingEngineOutput);

    JointData::Joint arm = side ? JointData::armRight0 : JointData::armLeft0;
    walkingEngineOutput.angles[arm] = theWalkingParameters.arm[side]->getPosition(phase, xSize);
    walkingEngineOutput.angles[arm] +=calcGyroCorrection(theBalanceParameters.armPFactor, theBalanceParameters.armIFactor, theBalanceParameters.armDFactor, theBalanceParameters.armUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);
    if(theMotionSelection.ratios[MotionRequest::walk] == 1)
    {
      walkingEngineOutput.angles[arm + 1] = randomDouble() < theWalkingParameters.armDeactivationRatio ? JointData::off : theWalkingParameters.elbowAngle - pi_2;
      walkingEngineOutput.angles[arm + 3] = randomDouble() < theWalkingParameters.armDeactivationRatio ? JointData::off : -theWalkingParameters.elbowAngle;
    }
    else
    {
      walkingEngineOutput.angles[arm + 1] = theWalkingParameters.elbowAngle - pi_2;
      walkingEngineOutput.angles[arm + 3] = -theWalkingParameters.elbowAngle;
    }
  }

  walkingEngineOutput.angles[JointData::headPan] = JointData::ignore;
  walkingEngineOutput.angles[JointData::headTilt] = JointData::ignore;

  if(phase < 0.25 || phase >= 0.75)
    walkingEngineOutput.odometryOffset =
    Pose2D(rotationOffsets[0] * odometryRotationScale,
    Vector2<double>(-footOffsets[0].x * theWalkingParameters.odometryScale.x,
    -footOffsets[0].y * theWalkingParameters.odometryScale.y));
  else
    walkingEngineOutput.odometryOffset =
    Pose2D(rotationOffsets[1] * odometryRotationScale,
    Vector2<double>(-footOffsets[1].x * theWalkingParameters.odometryScale.x,
    -footOffsets[1].y * theWalkingParameters.odometryScale.y));

  walkingEngineOutput.isLeavingPossible = fabs(currentSpeed.translation.x) < 5 &&
    fabs(currentSpeed.translation.y) < 5 &&
    fabs(currentSpeed.rotation) < 0.05 &&
    (fabs(phase - 0.25) < 0.05 || fabs(phase - 0.75) < 0.05);
  walkingEngineOutput.positionInWalkCycle = phase;

   bodyTiltLastGyroY = theSafeSensorData.data[SensorData::gyroY];
}

double WalkingEngine::calcGyroCorrection(double pFactor, double iFactor, double dFactor, double upperLimit, double lastGyro, double gyro) const
{
  double sign = gyro > 0 ? 1 : -1;
  double gyroScale = pow(fabs(gyro*2), 0.5)*sign;
 
  double gyroCorrection = pFactor * gyroScale +
    iFactor * bodyTiltGyroOffset +
    dFactor * (gyro - lastGyro);

  if(gyroCorrection > upperLimit) 
    gyroCorrection = upperLimit;
  else if(gyroCorrection < -upperLimit) 
    gyroCorrection = -upperLimit;
   return gyroCorrection;
}

void WalkingEngine::calcLegJoints(const Vector3<double>& position, double bodyTilt, double bodyShift,
                                  double footRollCorrection, JointData::Joint joint, JointRequest& jointRequest) const
{
  double sign = joint == JointData::legLeft0 ? 1 : -1;

  if(theRobotInfo.robotModel == RobotInfo::nao)
  {
    double gyroCorrection = calcGyroCorrection(theBalanceParameters.hipPFactor, theBalanceParameters.hipIFactor, theBalanceParameters.hipDFactor, theBalanceParameters.hipUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);
    Vector3<double> p = RotationMatrix().rotateY(gyroCorrection - bodyTilt) * position - Vector3<double>(0, bodyShift, 0);
    calcLegJoints(joint, p, gyroCorrection - bodyTilt, resWalkingParameters->footPanCorrection + jointRequest.angles[joint], jointRequest);
    return;
  }

  // foot is assumed to be flat on the ground, so we can remove offset between ground and joint 5
  Vector3<double> offset = position - Vector3<double>(0, bodyShift + sign * theRobotDimensions.lengthBetweenLegs / 2, -theRobotDimensions.heightLeg5Joint);
  offset.y *= sign;

  if(resWalkingParameters->footPanCorrection)
  {
    Vector2<double> v = (Pose2D(resWalkingParameters->footPanCorrection) + Pose2D(offset.x, offset.y)).translation;
    offset.x = v.x;
    offset.y = v.y;
  }

  // now we can calculate joints 1 and 5
  jointRequest.angles[joint + 1] = -atan2(offset.y, -offset.z);
  jointRequest.angles[joint + 5] = -jointRequest.angles[joint + 1] + footRollCorrection;
  jointRequest.angles[joint + 2] = -atan2(offset.x, Vector2<double>(offset.y, offset.z).abs());

  double diagonal = offset.abs();

  // upperLegLength, lowerLegLength, and diagonal form a triangle, use cosine theorem
  double a1 = (theRobotDimensions.upperLegLength * theRobotDimensions.upperLegLength - 
    theRobotDimensions.lowerLegLength * theRobotDimensions.lowerLegLength + diagonal * diagonal) / 
    (2 * theRobotDimensions.upperLegLength * diagonal);
  a1 = fabs(a1) > 1.0 ? 0 : acos(a1);

  double a2 = (theRobotDimensions.upperLegLength * theRobotDimensions.upperLegLength +
    theRobotDimensions.lowerLegLength * theRobotDimensions.lowerLegLength - diagonal * diagonal) / 
    (2 * theRobotDimensions.upperLegLength * theRobotDimensions.lowerLegLength);
  a2 = fabs(a2) > 1.0 ? pi : acos(a2);

  jointRequest.angles[joint + 2] -= a1;
  jointRequest.angles[joint + 3] = pi - a2;
  jointRequest.angles[joint + 4] = -jointRequest.angles[joint + 2] - jointRequest.angles[joint + 3];

  if(resWalkingParameters->footPanCorrection)
  {
    RotationMatrix r = RotationMatrix().rotateZ(-resWalkingParameters->footPanCorrection)
      .rotateX(jointRequest.angles[joint + 1])
      .rotateY(-bodyTilt);
    jointRequest.angles[joint + 1] = r.getXAngle();
    jointRequest.angles[joint + 2] += r.getYAngle();

    r = RotationMatrix().rotateZ(resWalkingParameters->footPanCorrection)
      .rotateX(jointRequest.angles[joint + 5]);
    jointRequest.angles[joint + 4] += r.getYAngle();
    jointRequest.angles[joint + 5] = r.getXAngle();

    jointRequest.angles[joint] += resWalkingParameters->footPanCorrection;
  }
  else
    jointRequest.angles[joint + 2] -= bodyTilt;

  //use gyroData for tilt balance

  jointRequest.angles[joint + 2] += calcGyroCorrection(theBalanceParameters.hipPFactor, theBalanceParameters.hipIFactor, theBalanceParameters.hipDFactor, theBalanceParameters.hipUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);
  jointRequest.angles[joint + 3] += calcGyroCorrection(theBalanceParameters.kneePFactor, theBalanceParameters.kneeIFactor, theBalanceParameters.kneeDFactor, theBalanceParameters.kneeUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);
  jointRequest.angles[joint + 4] += calcGyroCorrection(theBalanceParameters.footForwardPFactor, theBalanceParameters.footForwardIFactor, theBalanceParameters.footForwardDFactor, theBalanceParameters.footForwardUpperLimit,  bodyTiltLastGyroY, theSafeSensorData.data[SensorData::gyroY]);
}

void WalkingEngine::calcLegJoints(JointData::Joint whichSideJoint0, const Vector3<double>& position, double yAngle, double zAngle, JointRequest& jointRequest) const
{
  const double footHeight = theRobotDimensions.heightLeg5Joint;
  const double tibiaLength = theRobotDimensions.lowerLegLength;
  const double thighLength = theRobotDimensions.upperLegLength;
  const double hipOffsetY = 0.5 * theRobotDimensions.lengthBetweenLegs;

  double right = (whichSideJoint0 == JointData::legLeft0 ? -1 : 1);

  double tiltAngleOfFirstJointAxis =- pi / 2 + right * pi / 4;

  Matrix3x3<double> M1;
  M1[1][1] = cos(tiltAngleOfFirstJointAxis);
  M1[2][1] = -sin(tiltAngleOfFirstJointAxis);
  M1[1][2] = sin(tiltAngleOfFirstJointAxis);
  M1[2][2] = cos(tiltAngleOfFirstJointAxis);

  Matrix3x3<double> M2;
  M2[1][1] = cos(tiltAngleOfFirstJointAxis);
  M2[2][1] = sin(tiltAngleOfFirstJointAxis);
  M2[1][2] = -sin(tiltAngleOfFirstJointAxis);
  M2[2][2] = cos(tiltAngleOfFirstJointAxis);

  // BEGIN p = M1 * rotMatrixAroundZ(-zAngle)*M2*([x;y;z]+[0;0;46]+[0;right*50;85]);
  RotationMatrix rotZ, footRot;
  rotZ.rotateZ(-zAngle);

  Vector3<double> footHeightVec(0, 0, footHeight);
  footRot.rotateY(yAngle);
  footHeightVec = footRot * footHeightVec;

  Vector3<double> p(position.x, position.y + right * hipOffsetY, position.z);
  p = p + footHeightVec;
  p = M1 * rotZ * M2 * p;
  // END p = M1 * rotMatrixAroundZ(-zAngle)*M2*([x;y;z]+[0;0;46]+[0;right*50;85]);

  double ysign = -1*right;

  double r = p.abs();
  if (r > tibiaLength + thighLength)
  {
    OUTPUT(idText, text, "Leg too short!\r\n");
    return;
  }

  // @todo Check whether this assumes tibiaLength == thighLength
  double t3 = 2 * asin((r / 2) / tibiaLength);
  double t1 = atan(p[1]/p[2]);

  Vector3<double> p1(p);
  p1[0] = 0;
  double t2a = (pi-t3)/2;

  double soll;
  if (p[0]!=0)
    soll=atan(p[0]/p1.abs());
  else
    soll=0;

  double t2=t2a+soll;


  tiltAngleOfFirstJointAxis = - pi / 2 +  pi / 4;

  M1[1][1] = cos(tiltAngleOfFirstJointAxis);
  M1[2][1] = -sin(tiltAngleOfFirstJointAxis);
  M1[1][2] = sin(tiltAngleOfFirstJointAxis);
  M1[2][2] = cos(tiltAngleOfFirstJointAxis);

  M2[1][1] = cos(tiltAngleOfFirstJointAxis);
  M2[2][1] = sin(tiltAngleOfFirstJointAxis);
  M2[1][2] = -sin(tiltAngleOfFirstJointAxis);
  M2[2][2] = cos(tiltAngleOfFirstJointAxis);

  // BEGIN R=M1*rotMatrixAroundZ(zAngle)*M2; R=R*rotMatrixAroundX(ysign*t1)*rotMatrixAroundY(-t2)*rotMatrixAroundY(pi-t3);
  Matrix3x3<double> R;
  RotationMatrix rotZ2, rotY4, rotX, rotY5;
  rotZ2.rotateZ(zAngle);
  rotY4.rotateY(-t2);
  rotX.rotateX(ysign * t1);
  rotY5.rotateY(pi - t3);
  R = M1 * rotZ2 * M2 * rotX * rotY4 * rotY5;
  // END R=M1*rotMatrixAroundZ(zAngle)*M2; R=R*rotMatrixAroundX(ysign*t1)*rotMatrixAroundY(-t2)*rotMatrixAroundY(pi-t3);

  // BEGIN schnitt=cross(R*[0;1;0], [0;0;1]);
  Vector3<double> v3(0, 0, 1);
  Vector3<double> v4(0, 1, 0);
  v4 = R * v4;
  Vector3<double> schnitt(v4 ^ (footRot * v3));
  // END schnitt=cross(R*[0;1;0], [0;0;1]);


  Vector3<double> vek(R * v3);

  double t4 = -asin((schnitt * vek) / (schnitt.abs() * vek.abs()));

  // BEGIN R=R*rotMatrixAroundY(t4);
  RotationMatrix rotY6;
  rotY6.rotateY(t4);
  R = R * rotY6;
  // END R=R*rotMatrixAroundY(t4);

  // BEGIN schnitt=cross(R*[1;0;0], [0;0;1]);
  Vector3<double> v5(1, 0, 0);
  schnitt = (R * v5) ^ (footRot * v3);
  // END schnitt=cross(R*[1;0;0], [0;0;1]);

  vek = R * v3;

  // BEGIN t5=-asin((schnitt'*vek)/(norm(schnitt)*norm(vek)));
  double t5 = -asin((schnitt * vek) / (schnitt.abs() * vek.abs()));
  // END

  jointRequest.angles[whichSideJoint0 + 0] = zAngle; // [-90 ;  0 ] = [...
  jointRequest.angles[whichSideJoint0 + 1] = ysign * t1; // [-100; 25 ] = [...
  jointRequest.angles[whichSideJoint0 + 2] = -t2; // [-45 ; 25 ]
  jointRequest.angles[whichSideJoint0 + 3] = pi - t3; // [ 0  ; 130]
  jointRequest.angles[whichSideJoint0 + 4] = t4; // [-75 ; 45 ]
  jointRequest.angles[whichSideJoint0 + 5] = t5; // [-25 ; 45 ]	
}

MAKE_MODULE(WalkingEngine, Motion Control)
