/** 
* @file Modules/MotionControl/WalkingEngine.h
* This file declares a module that creates the walking motions.
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#ifndef __WalkingEngine_h_
#define __WalkingEngine_h_

#include "Tools/Module/Module.h"
#include "Representations/Configuration/RobotDimensions.h"
#include "Representations/Configuration/WalkingParameters.h"
#include "Representations/Configuration/BalanceParameters.h"
#include "Representations/Infrastructure/FrameInfo.h"
#include "Representations/Infrastructure/RobotInfo.h"
#include "Representations/Infrastructure/SensorData.h"
#include "Representations/MotionControl/MotionSelection.h"
#include "Representations/MotionControl/WalkingEngineStandOutput.h"
#include "Representations/MotionControl/WalkingEngineOutput.h"
#include "Tools/Math/Vector3.h"
#include "Tools/RingBufferWithSum.h"


MODULE(WalkingEngine)
  REQUIRES(RobotDimensions)
  REQUIRES(WalkingParameters)
  REQUIRES(BalanceParameters)
  REQUIRES(FrameInfo)
  REQUIRES(RobotInfo)
  REQUIRES(SafeSensorData)
  REQUIRES(MotionSelection)
  PROVIDES_WITH_MODIFY(WalkingEngineOutput)
  PROVIDES_WITH_MODIFY(WalkingEngineStandOutput)
END_MODULE

class WalkingEngine : public WalkingEngineBase
{
private:

  double phase; /**< The current position in walk phase. */
  unsigned lastIterationTime;
  Vector3<double> footPositions[2]; /**< The previous foot positions. */
  bool wasActive; /**< Was this module active in the previous frame? */
  unsigned startTime; /**< The time when the first motion cycle began after motion type was switched to walking. */
  RingBufferWithSum<double, 50> bodyTiltError; /**< A buffer for deviations from the desired body tilt. */
  double bodyTiltGyroOffset;
  double bodyTiltLastGyroY;
  double lastBodyShift1, /**< The body shift of the previous frame. */
         lastBodyShift2; /**< The body shift of the frame before the previous one. */
  RingBufferWithSum<double, 21> bodyShiftError; /**< A buffer for deviations from the desired body shift. */
  WalkingParameters lastParameters;
  WalkingParameters res;
  int lastChangedTime;
  bool wasInterpolating;
  const WalkingParameters *resWalkingParameters;
  double bodyTiltOffset, /**< Accumulated offset between desired body tilt and required body tilt. */
         lastLeftRotation, /**< The last rotation of the left leg. */
         lastRightRotation; /**< The last rotation of the right leg. */

  /**
  * The method calculates a correction value for the body tilt to prevent falling down.
  * @return A correction value in radians.
  */

double calcGyroCorrection(double pFactor, double iFactor, double dFactor, double upperLimit, double lastGyro, double gyro) const;

  /**
  * The method calculates the joint angles for a certain leg.
  * @param position The desired position of the foot sole.
  * @param bodyTilt The tilt angle of the body.
  * @param bodyShift The side shift of the body center in mm.
  * @param footRollCorrection An offset for the roll angle of the foot.
  * @param joint The index of the first joint of this leg in the class JointAngles.
  * @param jointRequest The joint angles to be calculated.
  */
  void calcLegJoints(const Vector3<double>& position, double bodyTilt, double bodyShift, 
                     double footRollCorrection, JointData::Joint joint, JointRequest& jointRequest) const;

  /**
  * The method calculates the joint angles for a certain Nao leg.
  * @param whichSideJoint0 The index of the first joint of this leg in the class JointAngles.
  * @param position The desired position of the foot sole.
  * @param yAngle The tilt angle of the foot sole.
  * @param zAngle The pan angle of the foot sole.
  * @param jointRequest The joint angles to be calculated.
  */
  void calcLegJoints(JointData::Joint whichSideJoint0, const Vector3<double>& position, 
                     double yAngle, double zAngle, JointRequest& jointRequest) const;

  void update(WalkingEngineOutput& walkingEngineOutput, const WalkingParameters& theWalkingParameters, bool wasActive);

  void update(WalkingEngineOutput& walkingEngineOutput);
  void update(WalkingEngineStandOutput& standOutput);

  void checkCfg();

public:
  /**
  * Default constructor.
  */
  WalkingEngine();

  /**
  * The method is called once before the first call of an update method.
  */
  void init();
};

#endif // __WalkingEngine_h_
