/**
* @file BallPerceptor.cpp
* This file declares a module that provides the ball percept based on the
* biggest orange segment found by the PointsPerceptor. It is based on the
* BallSpecialist in GT2005.
* @author <A href="mailto:juengel@informatik.hu-berlin.de">Matthias Juengel</A>
* @author Max Risler
* @author <A href="mailto:brunn@sim.tu-darmstadt.de">Ronnie Brunn</A>
* @author <A href="mailto:mkunz@sim.tu-darmstadt.de">Michael Kunz</A>
* @author <a href="mailto:walter.nistico@uni-dortmund.de">Walter Nistico</a>
* @author <a href="mailto:judy@informatik.uni-bremen.de">Judith Mller</a>
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "BallPerceptor.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Math/Matrix_nxn.h"
#include "Tools/ImageProcessing/BresenhamLineScan.h"

void BallPerceptor::update(BallPercept& ballPercept)
{
  DECLARE_DEBUG_DRAWING("module:BallPerceptor:ball","drawingOnImage","");
  ballPercept.ballWasSeen = false;

  if(theBallSpot.found && theCameraMatrix.isValid)
    searchBall(theBallSpot.position, ballPercept);
}

void BallPerceptor::searchBall(const Vector2<int>& ballStart, BallPercept& ballPercept)
{
  BallPointList ballPoints;
  Vector2<int> center;
  double radius;
  int countOrange = 0;
  int countAmbiguous = 0;	
  int maxOrangePerLine = 0;
  int countPixel = 0;

  scanForBallPoints(ballStart, ballPoints, countAmbiguous, countOrange, maxOrangePerLine, countPixel);

  //ignore ball if not enough points are orange
  int numberOfSoftEdgePoints = 0;
  int numberOfPointsInYellow = 0;

  for(int i = 0; i < ballPoints.number; i++)
  {
    if (ballPoints[i].yellowIsClose && !ballPoints[i].atBorder) numberOfPointsInYellow++;
    if (!ballPoints[i].hardEdge && !ballPoints[i].atBorder) numberOfSoftEdgePoints++;
  }

  if ((countOrange > countAmbiguous) && 
    (countOrange * 6 > countPixel) &&                          // >  16.66% orange
    (numberOfPointsInYellow * 4 <= ballPoints.number * 3)      // <= 75% with Yellow close
    )
  {
    considerBallPoints(ballPoints);

    //try only points near green with hard edge first
    BallPointList testPoints;
    for(int i = 0; i < ballPoints.number; i++)
    {
      if (ballPoints[i].greenIsClose && ballPoints[i].hardEdge) 
        testPoints.add(ballPoints[i]);
    }
    if (
      testPoints.number * 2 >= ballPoints.number &&
      computeBallInImageLevenbergMarquardt(testPoints, center, radius) &&
      checkIfPointsAreInsideBall(ballPoints, center, radius))
    {
      checkCurrentBall(center, radius, ballPercept);
    }
    else
    {
      //now all points if not at border with hard edge
      testPoints.number = 0;
      for(int i = 0; i < ballPoints.number; i++)
      {
        if (!ballPoints[i].atBorder && ballPoints[i].hardEdge) 
          testPoints.add(ballPoints[i]);
      }
      if (computeBallInImageLevenbergMarquardt(testPoints, center, radius) &&
          checkIfPointsAreInsideBall(ballPoints, center, radius))
      {
        checkCurrentBall(center, radius, ballPercept);
      }
      else
      {
        //now all points if not at border
        testPoints.number = 0;
        for(int i = 0; i < ballPoints.number; i++)
        {
          if (!ballPoints[i].atBorder) 
            testPoints.add(ballPoints[i]);
        }
        if (computeBallInImageLevenbergMarquardt(testPoints, center, radius) &&
            checkIfPointsAreInsideBall(ballPoints, center, radius))
        {
          checkCurrentBall(center, radius, ballPercept);
        }
        else
        {
          //take all points if nothing else works
          if (computeBallInImageLevenbergMarquardt(ballPoints, center, radius))
          {
            calculateDeviationOfBallPoints(ballPoints, center, radius);
            checkCurrentBall(center, radius, ballPercept);
          }
        }
      }
    }
  }
}


void BallPerceptor::BallPointList::add(const BallPoint& ballPoint)
{
  if(number < maxNumberOfPoints)
  {
    ballPoints[number++] = ballPoint;
    DOT("module:BallPerceptor:ball", ballPoint.x, ballPoint.y,
      (ballPoint.hardEdge) ? ColorClasses::blue : ColorClasses::orange, 
      (ballPoint.atBorder) ? ColorClasses::black :
      (ballPoint.greenIsClose) ? ColorClasses::green :
      (ballPoint.yellowIsClose) ? ColorClasses::yellow :
      ColorClasses::white
      );
  }
}


void BallPerceptor::scanForBallPoints
(
 const Vector2<int>& ballStart,
 BallPointList& ballPoints,
 int& countAmbiguous, 
 int& countOrange,
 int& maxOrangePerLine,
 int& countPixel
 )
{
  // search for ball variables
  BallPoint north;
  BallPoint east;
  BallPoint south;
  BallPoint west;

  BallPoint start;
  start = ballStart;
  Vector2<int>step;
  BallPoint destination;
  BallPoint start2;

  DOT("module:BallPerceptor:ball", start.x, start.y, ColorClasses::black, ColorClasses::white);
  //find north
  step.x = 0; step.y = -1;

  findEndOfBall(start, step, north, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
  if(north.atBorder)
  {
    start2 = north - step;
    //find east
    step.x = 1; step.y = 0;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
    //find west
    step.x = -1; step.y = 0;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
  }
  else
  {
    ballPoints.add(north);
  }

  //find east
  step.x = 1; step.y = 0;
  findEndOfBall(start, step, east, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
  if(east.atBorder)
  {
    start2 = east - step;
    //find north
    step.x = 0; step.y = -1;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
    //find south
    step.x = 0; step.y = 1;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
  }
  else
  {
    ballPoints.add(east);
  }

  //find south
  step.x = 0; step.y = 1;
  findEndOfBall(start, step, south, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
  if(south.atBorder)
  {
    start2 = south - step;
    //find east
    step.x = 1; step.y = 0;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
    //find west
    step.x = -1; step.y = 0;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
  }
  else
  {
    ballPoints.add(south);
  }

  //find west
  step.x = -1; step.y = 0;
  findEndOfBall(start, step, west, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
  if(west.atBorder)
  {
    start2 = west - step;
    //find north
    step.x = 0; step.y = -1;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
    //find south
    step.x = 0; step.y = 1;
    if(findEndOfBall(start2, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel))
    {
      ballPoints.add(destination);
    }
  }
  else
  {
    ballPoints.add(west);
  }

  if( (south.y - north.y) > (east.x - west.x) )
  {
    if ((north.y + south.y) / 2 != start.y)
    {
      start.y = (north.y + south.y) / 2;
      //find east
      step.x = 1; step.y = 0;
      findEndOfBall(start, step, east, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
      if(!east.atBorder)
      {
        ballPoints.add(east);
      }
      //find west
      step.x = -1; step.y = 0;
      findEndOfBall(start, step, west, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
      if (!west.atBorder)
      {
        ballPoints.add(west);
      }
      start.x = (west.x + east.x) / 2;
    }
  }
  else
  {
    if ((west.x + east.x) / 2 != start.x)
    {
      start.x = (west.x + east.x) / 2;
      //find north
      step.x = 0; step.y = -1;
      findEndOfBall(start, step, north, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
      if(!north.atBorder)
      {
        ballPoints.add(north);
      }
      //find south
      step.x = 0; step.y = 1;
      findEndOfBall(start, step, south, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
      if(!south.atBorder)
      {
        ballPoints.add(south);
      }
      start.y = (north.y + south.y) / 2;
    }
  }

  // find in diagonal
  for (step.x = -1; step.x <= 1; step.x += 2)
  {
    for (step.y = -1; step.y <= 1; step.y += 2)
    {
      findEndOfBall(start, step, destination, countAmbiguous, countOrange, maxOrangePerLine, countPixel);
      if (!destination.atBorder)
      {
        ballPoints.add(destination);
      }
    }
  }
}


bool BallPerceptor::findEndOfBall
(
 const BallPoint& start,
 const Vector2<int>& step,
 BallPoint& destination,
 int& countAmbiguous,
 int& countOrange,
 int& maxOrangePerLine,
 int& countPixel
 )
{
  /*
  stopColors are colors indicating the end of the ball:
  green, yellow, skyblue
  more than 3 pixels of a stopColor indicate the end of the ball
  */
  int stopColorCounter = 0;
  int currentOrange = 0;
  int currentAmbiguous = 0;
  int len = 0;
  int stopLen = 0;

  ColorClasses::Color currentColorClass;

  Vector2<int> firstStopColor = start;
  unsigned char currentOrangeSim = 0;
  Vector2<int> lastPoint = start;
  destination = start;
  destination.greenIsClose = false;
  destination.yellowIsClose = false;
  destination.atBorder = false;
  destination.hardEdge = true;

  bool isOrange = false;

  bool goOn = true;

  while(goOn)
  {
    lastPoint = destination;
    destination += step;
    if(destination.x < 0 || destination.x >= theImage.cameraInfo.resolutionWidth ||
       destination.y < 0 || destination.y >= theImage.cameraInfo.resolutionHeight) 
    {
      if (stopColorCounter == 0)
      {
        countPixel += len;
        destination.atBorder = true;
      }
      else
      {
        destination = firstStopColor;
        countPixel += stopLen;
        destination.atBorder = false;
      }
      goOn = false;
    }
    else
    {
      currentColorClass = (ColorClasses::Color) theColorTable64.colorClasses
        [theImage.image[destination.y][destination.x].y >> 2]
        [theImage.image[destination.y][destination.x].cb >> 2]
        [theImage.image[destination.y][destination.x].cr >> 2];

      // counting all orange pixels on the scanned horz./vert./diag. lines
      len++;
      if ( currentColorClass == ColorClasses::orange )
      {
        currentOrange++;
      }
      if (currentColorClass == ColorClasses::yellow)
      {
        currentAmbiguous++;
      }

      LINE("module:BallPerceptor:ball",destination.x,destination.y,destination.x + 1,destination.y,1,Drawings::ps_solid,ColorClasses::green);

      if (currentColorClass == ColorClasses::orange)
        isOrange = true;
      else
      {
        currentOrangeSim = getSimilarityToOrange(
          theImage.image[destination.y][destination.x].y,
          theImage.image[destination.y][destination.x].cb,
          theImage.image[destination.y][destination.x].cr
          );
        isOrange = currentOrangeSim > 30;
      }
      if(currentColorClass == ColorClasses::green || 
        currentColorClass == ColorClasses::yellow ||
        currentColorClass == ColorClasses::blue ||
        !isOrange
        )
      {
        if (stopColorCounter == 0)
        {
          firstStopColor = destination;
          stopLen = len;
        }
        if (currentColorClass == ColorClasses::green) destination.greenIsClose = true;
        if (currentColorClass == ColorClasses::yellow) destination.yellowIsClose = true;
        stopColorCounter++;

        LINE("module:BallPerceptor:ball",destination.x,destination.y,destination.x + 1,destination.y,1,Drawings::ps_solid,ColorRGBA(255,0,0));

        if(stopColorCounter > 8) 
        {
          destination = firstStopColor;
          countPixel += stopLen;
          goOn = false;
        }
      }
      else
      {
        destination.greenIsClose = false;
        destination.yellowIsClose = false;
        stopColorCounter = 0;
      }
    } 
  }
  destination -= step;

  // compute sum of all orange pixels and max-pixels-per-line
  countOrange += currentOrange;
  countAmbiguous += currentAmbiguous;
  maxOrangePerLine = std::max ( maxOrangePerLine, currentOrange);
  if (destination.greenIsClose) destination.yellowIsClose = false;

  if (!destination.atBorder)
  {
    unsigned char lastGoodColorOrangeSim = getSimilarityToOrange(
      theImage.image[destination.y][destination.x].y,
      theImage.image[destination.y][destination.x].cb,
      theImage.image[destination.y][destination.x].cr
      );
    destination.hardEdge = (2 * currentOrangeSim < lastGoodColorOrangeSim);
  }
  else 
    destination.hardEdge = false;

  if (!destination.atBorder)
  {
    if (destination.greenIsClose)
    {
      LINE("module:BallPerceptor:ball",destination.x,destination.y,destination.x + 1,destination.y,1,Drawings::ps_solid,ColorClasses::blue);
    }
    else
      if (destination.yellowIsClose)
      {
        LINE("module:BallPerceptor:ball",destination.x,destination.y,destination.x + 1,destination.y,1,Drawings::ps_solid,ColorRGBA(0,128,255));
      }
      else
      {
        LINE("module:BallPerceptor:ball",destination.x,destination.y,destination.x + 1,destination.y,1,Drawings::ps_solid,ColorRGBA(255,128,128));
      }
  }
  return true;
}

bool BallPerceptor::computeBallInImageLevenbergMarquardt( 
  const BallPointList& ballPoints,
  Vector2<int>& center,
  double& radius)
{
  if (ballPoints.number < 3)
    return false;
  int Mx(0), My(0), Mxx(0), Myy(0), Mxy(0), Mz(0), Mxz(0), Myz(0);
  for (int i = 0; i < ballPoints.number; ++i)
  {
    int x = ballPoints[i].x;
    int y = ballPoints[i].y;
    int xx = x*x;
    int yy = y*y;
    int z = xx + yy;
    Mx += x;
    My += y;
    Mxx += xx;
    Myy += yy;
    Mxy += x*y;
    Mz += z;
    Mxz += x*z;
    Myz += y*z;
  }

  // Construct and solve matrix (might fail and throw exception).
  // Result will be center and radius of ball in theImage.
  try
  {
    double matrix_values[9] = 
    {
      static_cast<double>(Mxx), static_cast<double>(Mxy), static_cast<double>(Mx),
      static_cast<double>(Mxy), static_cast<double>(Myy), static_cast<double>(My),
      static_cast<double>(Mx),  static_cast<double>(My),  static_cast<double>(ballPoints.number)
    };
    Matrix_nxn<double, 3> M(matrix_values);

    Vector_n<double, 3> v;
    v[0] = static_cast<double>(-Mxz);
    v[1] = static_cast<double>(-Myz);
    v[2] = static_cast<double>(-Mz);

    Vector_n<double, 3> BCD(M.solve(v));
    center.x = static_cast<int>(-BCD[0]/2.0);
    center.y = static_cast<int>(-BCD[1]/2.0);
    double radicand = BCD[0]*BCD[0]/4.0 + BCD[1]*BCD[1]/4.0 - BCD[2];
    if (radicand < 0.0)
      return false;
    radius = sqrt(radicand);
  }
  catch (...)
  {
    OUTPUT(idText, text, "Exception in BallPerceptor::computeBallInImageLevenbergMarquardt");
    return false;
  }
  return true;
}


bool BallPerceptor::checkIfPointsAreInsideBall(const BallPointList& ballPoints, Vector2<int>& center, double radius)
{
  double d;
  int numberOfNonBorderBallPoints = 0;
  double dev = 0;
  for(int i = 0; i < ballPoints.number; i++)
  {
    d = Geometry::distance(center, ballPoints[i]);
    if (d > radius * 1.1)
    {
      return false;
    }
    if (!ballPoints[i].atBorder)
    {
      numberOfNonBorderBallPoints++;
      dev += fabs(d-radius);
    }
  }
  deviationOfBallPoints = (numberOfNonBorderBallPoints ? dev / numberOfNonBorderBallPoints : 0);
  return true;
}


void BallPerceptor::calculateDeviationOfBallPoints(const BallPointList& ballPoints, Vector2<int>& center, double radius)
{
  double d;
  int numberOfNonBorderBallPoints = 0;
  double dev = 0;
  for(int i = 0; i < ballPoints.number; i++)
  {
    d = Geometry::distance(center, ballPoints[i]);
    if (!ballPoints[i].atBorder)
    {
      numberOfNonBorderBallPoints++;
      dev += fabs(d-radius);
    }
  }
  deviationOfBallPoints = (numberOfNonBorderBallPoints ? dev / numberOfNonBorderBallPoints : 0);
}


void BallPerceptor::considerBallPoints(const BallPointList& ballPoints)
{
  int numberOfNearGreenPoints = 0;
  int numberOfNearYellowPoints = 0;
  int numberOfHardEdgePoints = 0;
  int numberOfNonBorderPoints = 0;

  for (int i=0; i<ballPoints.number; i++)
  {
    if (!ballPoints[i].atBorder)
    {
      if (ballPoints[i].hardEdge)
      {
        numberOfHardEdgePoints++;
      }
      if (ballPoints[i].greenIsClose)
      {
        numberOfNearGreenPoints++;
      }
      if (ballPoints[i].yellowIsClose)
      {
        numberOfNearYellowPoints++;
      }
      numberOfNonBorderPoints++;
    }
  }
  if (numberOfNonBorderPoints>0)
  {
    durabilityOfBallPoints=1.0;
    if (!(numberOfHardEdgePoints>=3 && numberOfNearGreenPoints*2>numberOfNonBorderPoints)) 
      durabilityOfBallPoints *= 0.8;
    if (numberOfHardEdgePoints*3<numberOfNonBorderPoints*2) 	// less than 2/3 hard edge
      durabilityOfBallPoints *= 0.9;
    if (numberOfHardEdgePoints*3<numberOfNonBorderPoints) 	  // less than 1/3 hard edge
      durabilityOfBallPoints *= 0.9;
    if (numberOfNearGreenPoints*4<numberOfNonBorderPoints) 	  // less than 1/4 near green
      durabilityOfBallPoints *= 0.5;
    if (numberOfNearGreenPoints==0)                           // no points near green
      durabilityOfBallPoints *= 0.5;
    if (numberOfNearYellowPoints*2>numberOfNonBorderPoints) 	// more than 1/2 near yellow
      durabilityOfBallPoints *= 0.9;
    if (numberOfNonBorderPoints<4)                            // only 1-3 non border points 
      durabilityOfBallPoints *= 0.5;
  }
  else
  {// there are only border ball points, so the ball is right in front of the camera, and the durability can stay 1.0
    durabilityOfBallPoints=1.0;
  }
}


void BallPerceptor::checkCurrentBall
(
 const Vector2<int>& center,
 double radius,
 BallPercept& ballPercept
 )
{
  // test if ball is below horizon
  double factor = theImage.cameraInfo.focalLengthInv;
  Vector3<double> vectorToCenter(1, 
                                 (theImage.cameraInfo.opticalCenter.x - center.x) * factor, 
                                 (theImage.cameraInfo.opticalCenter.y - center.y) * factor);
  Vector3<double> vectorToCenterWorld = theCameraMatrix.rotation * vectorToCenter;

  //Is the point above the horizon ? - return
  if(vectorToCenterWorld.z < 1 * factor)
  {
    double percentOfOrange = 1.0; // big balls are already scanned enough. They are supposed to have nearly 100% orange inside
    if (radius <= 12.0)
    {//small ghost balls can easily appear on an orange fringe of a pink/yellow landmark
      // this because the scan follows along the fringe; this new scan starts from the 
      // hypothetical center of the "ghost" ball, which in such case lies inside the landmark
      // and hence produces a high number of "non orange" points; a true ball however,
      // should have enough orange inside
      Vector2<int> current(center), clippedCenter(center);
      int orangeCount = 0;
      int totalCount = 0;
      double scanAngle = 0;
      ColorClasses::Color currentColor;
      Geometry::clipPointInsideRectangle(Vector2<int>(0,0), Vector2<int>(theImage.cameraInfo.resolutionWidth, theImage.cameraInfo.resolutionHeight), clippedCenter);
      currentColor = (ColorClasses::Color) theColorTable64.colorClasses
        [theImage.image[clippedCenter.y][clippedCenter.x].y >> 2]
        [theImage.image[clippedCenter.y][clippedCenter.x].cb >> 2]
        [theImage.image[clippedCenter.y][clippedCenter.x].cr >> 2];
      if(currentColor == ColorClasses::orange)
        orangeCount++;
      totalCount++;
      for (int sector=0; sector<8; sector++)
      {
        BresenhamLineScan scan(clippedCenter, scanAngle, theImage.cameraInfo);
        scan.init();
        current = clippedCenter;
        int steps = (sector%2) ? (int)(radius/sqrt(2.0)) : (int)(radius);
        for (int i=0; i<steps && i<scan.numberOfPixels; i++)
        {
          scan.getNext(current);
          currentColor = (ColorClasses::Color) theColorTable64.colorClasses
            [theImage.image[current.y][current.x].y >> 2]
            [theImage.image[current.y][current.x].cb >> 2]
            [theImage.image[current.y][current.x].cr >> 2];
          if(currentColor == ColorClasses::orange)
            orangeCount++;
          totalCount++;
          DOT("module:BallPerceptor:ball", current.x, current.y, ColorRGBA(128, 128, 128), currentColor);
        }
        scanAngle += pi_4;
      }
      if (orangeCount*3 < totalCount) //not enough orange to be ball
        return;
      percentOfOrange = (double)orangeCount/totalCount;
    }

    Vector2<double> correctedCenter = theImageCoordinateSystem.toCorrected(center);
    Vector2<double>angles;
    Geometry::calculateAnglesForPoint(Vector2<int>(int(correctedCenter.x + 0.5), int(correctedCenter.y + 0.5)),
                                      theCameraMatrix, theImage.cameraInfo, angles);
    double radiusAsAngle = Geometry::pixelSizeToAngleSize(radius, theImage.cameraInfo);

    computeOffsetToBall(center,
      angles,
      radius,
      radiusAsAngle, 
      theCameraMatrix.translation, 
      theCameraMatrix.isValid,
      ballPercept);
  }
}


void BallPerceptor::computeOffsetToBall(
 const Vector2<int>& centerInPixel,
 const Vector2<double>& centerAsAngles,
 double radiusInPixel,
 double radiusAsAngle,
 const Vector3<double>& translationOfCamera,
 bool isCameraMatrixValid,
 BallPercept& ballPercept)
{
  if (!ballPercept.ballWasSeen)
  {
    double ballRadius = theFieldDimensions.ballRadius;

    // Bearing based ball offset:
    Vector2<double> bbo; // bearing based
    const double distanceBearingBased = centerAsAngles.y < 0 ? (translationOfCamera.z - ballRadius) / tan(-centerAsAngles.y) : 10000;
    bbo.x = translationOfCamera.x + distanceBearingBased * cos(centerAsAngles.x);
    bbo.y = translationOfCamera.y + distanceBearingBased * sin(centerAsAngles.x);

    // Size based ball offset:
    Vector2<double> sbo; 
    const double distanceFromCameraToBallCenter = Geometry::getBallDistanceByAngleSize(int(2.0 * ballRadius), 2.0 * radiusAsAngle);
    double distanceFromCameraToBallCenterOnGround;
    if(distanceFromCameraToBallCenter > fabs(translationOfCamera.z - ballRadius))
      distanceFromCameraToBallCenterOnGround = sqrt(sqr(distanceFromCameraToBallCenter) - sqr(translationOfCamera.z  - ballRadius));
    else
      distanceFromCameraToBallCenterOnGround = 0;
    sbo.x = translationOfCamera.x + distanceFromCameraToBallCenterOnGround * cos(centerAsAngles.x);
    sbo.y = translationOfCamera.y + distanceFromCameraToBallCenterOnGround * sin(centerAsAngles.x);

    Vector2<double> o; // offset
    const double sbDist(sbo.abs());
    const double SIZE_BASED_ONLY_DIST(600.0);
    const double BEARING_BASED_ONLY_DIST(300.0);
    //In some cases, use size based computation only:
    if(!isCameraMatrixValid || centerAsAngles.y >= 0 || sbDist > SIZE_BASED_ONLY_DIST) 
      o = sbo;
    //At medium distance, interpolate between both offsets:
    else if(sbDist >= BEARING_BASED_ONLY_DIST)
    {
      double factor = (SIZE_BASED_ONLY_DIST - sbDist) / (SIZE_BASED_ONLY_DIST - BEARING_BASED_ONLY_DIST);
      o = bbo * factor + sbo * (1.0 - factor);
    }
    //In all remainung cases (ball is near, no other problems), use bearing based distance only:
    else 
      o = bbo;
    // Finally, create the percept
    fillPercept(o, bbo, sbo, Vector2<double>(centerInPixel.x, centerInPixel.y), radiusInPixel, isCameraMatrixValid, ballPercept);
  }
}

void BallPerceptor::fillPercept
(
 const Vector2<double>& offsetOnField,
 const Vector2<double>& bearingBasedOffsetOnField,
 const Vector2<double>& sizeBasedOffsetOnField,
 const Vector2<double>& centerInImage,
 double radiusInImage,
 bool isCameraMatrixValid,
 BallPercept& ballPercept
 )
{
  ballPercept.relativePositionOnField = offsetOnField;
  ballPercept.positionInImage = centerInImage;
  ballPercept.radiusInImage = radiusInImage;
  ballPercept.ballWasSeen = radiusInImage < parameters.maxBallRadius;
}

MAKE_MODULE(BallPerceptor, Perception)
