/**
* @file ColorChangePerceptor.cpp
* This file declares a module that provides line points found in the image.
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
* @author <a href="mailto:felipe@informatik.uni-bremen.de">Felipe Lopez</a>
*/

#include "ColorChangePerceptor.h"
#include "Tools/Streams/InStreams.h"


ColorChangePerceptor::ColorChangePerceptor()
{
  InConfigFile stream("pointsColorChange.cfg", "default");
  ASSERT(stream.exists());
  stream >> parameters;
}

void ColorChangePerceptor::update(PointsPercept& pointsColorChange)
{
  INIT_DEBUG_IMAGE(ColorChangePerceptor, theImage);
  MODIFY("parameters:ColorChangePerceptor", parameters);

  pointsColorChange.points.clear();

  if(theCameraMatrix.isValid)
  {
    init(pointsColorChange.origin);

    //If there are vertical segments, the color changes in these segments will be recognized
    scanVertically(pointsColorChange);
    connectVertically(pointsColorChange);
    unsigned start = pointsColorChange.points.size();
    scanHorizontally(pointsColorChange);
    connectHorizontally(pointsColorChange, start);
    calcLengths(pointsColorChange);
  }
  SEND_DEBUG_IMAGE(ColorChangePerceptor);
}

void ColorChangePerceptor::init(Pose2D& origin)
{
  cameraHeight = int(theCameraMatrix.translation.z + 0.5);
  xCenter = int(theImage.cameraInfo.opticalCenter.x + 0.5);
  yCenter = int(theImage.cameraInfo.opticalCenter.y + 0.5);
  double pan = theCameraMatrix.rotation.getZAngle();
  if(pan > 2)
   pan -= pi;
  else if(pan < -2)
   pan += pi;
  RotationMatrix r(theCameraMatrix.rotation.invert());
  r.rotateZ(pan);
  double tilt = r.getYAngle();
  r.rotateY(-tilt);
  double roll = r.getXAngle();
  origin = Pose2D(pan, theCameraMatrix.translation.x, theCameraMatrix.translation.y);

  Vector3<int> v((int) toDegrees(roll), (int) toDegrees(tilt), (int) toDegrees(pan));
  MODIFY("LinePointPerceptor:angles", v);

  double factor = theImage.cameraInfo.focalLengthInv * (1 << 20);
  rotCos = int(cos(-roll) * factor + 0.5);
  rotSin = int(sin(-roll) * factor + 0.5);
  tiltCos2 = int(cos(-tilt) * (1 << 20) + 0.5);
  tiltSin2 = int(sin(-tilt) * (1 << 20) + 0.5);
  tiltCos = tiltCos2 >> 10;
  tiltSin = tiltSin2 >> 10;
  skipBody = tilt < -1.3;
}


void ColorChangePerceptor::pointToField(int x, int y, Vector2<int>& p) const
{
  Vector2<int> v = theImageCoordinateSystem.toCorrectedCenteredNeg(x, y);
  int b0 = (v.x * rotSin + v.y * rotCos) >> 10,
  b1 = tiltCos2 + b0 * tiltSin,
  b2 = v.x * rotCos - v.y * rotSin,
  b3 = tiltSin2 - b0 * tiltCos;
  if (b3 == 0) b3 = 1;
  p.x = b1 * cameraHeight / b3;
  p.y = b2 * cameraHeight / b3;
}

void ColorChangePerceptor::scanVertically(PointsPercept& pointsColorChange)
{
  unsigned fieldBorderSize = theSegmentsPercept.fieldBorders.size(); // SegmentsPerceptor takes care that size is >= 2
  unsigned fieldBorderIndex = fieldBorderSize-1;
  bool containsFieldColorSegment = false; 
  if (theSegmentsPercept.segmentsVertical.size() > 0 && (theSegmentsPercept.segmentsVertical.at(0).color == ColorClasses::orange || theSegmentsPercept.segmentsVertical.at(0).color == ColorClasses::blue
    || theSegmentsPercept.segmentsVertical.at(0).color == ColorClasses::yellow || theSegmentsPercept.segmentsVertical.at(0).color == ColorClasses::green || theSegmentsPercept.segmentsVertical.at(0).color == ColorClasses::white))
    containsFieldColorSegment = true;  

  for(unsigned int i = 1; (i+1) < theSegmentsPercept.segmentsVertical.size(); ++i)
  {
    SegmentsPercept::Segment currentSegment = theSegmentsPercept.segmentsVertical[i];
    
    int maxY = -1;

    if(fieldBorderIndex >= 1
        && currentSegment.x > theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x)
    {
      fieldBorderIndex--;
    }

    if(currentSegment.x < theSegmentsPercept.fieldBorders.at(fieldBorderSize-1).x) 
    {
      maxY = theSegmentsPercept.fieldBorders.at(fieldBorderSize-1).y;
    }
    else if(currentSegment.x > theSegmentsPercept.fieldBorders.at(0).x)
    {
      maxY = theSegmentsPercept.fieldBorders.at(0).y;
    }
    else if(currentSegment.x == theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x)
    {
      maxY = theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y;
    }
    else if(fieldBorderIndex >= 1 
             && currentSegment.x == theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x)
    {
      maxY = theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).y;
    }
    else if(fieldBorderIndex >= 1
             && currentSegment.x > theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x
             && currentSegment.x < theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x)
    {
      int xdiff = (theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x-theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x);
      int ydiff = (theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).y-theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y);
      double vh = (double)(theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x - currentSegment.x) / ((double) xdiff);
      maxY = (int) (theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y - (ydiff * vh));
    }

    maxY -= parameters.fieldBorderExceed;

    // if the segment lies inside the field border
    if(currentSegment.y > maxY)
    {
      // COLOR
      SegmentsPercept::Segment prevSegment = currentSegment;
      SegmentsPercept::Segment nextSegment = currentSegment;
      int prevDistance = -1;
      int nextDistance = -1;

      if(theSegmentsPercept.segmentsVertical[i-1].x == currentSegment.x)
      {
        prevSegment = theSegmentsPercept.segmentsVertical[i-1];
        prevDistance = (prevSegment.y - prevSegment.length) - currentSegment.y;
      }
      else
      {
        if(!containsFieldColorSegment && !theSegmentsPercept.segmentsVertical[i-1].onHalfScanLine)
        {
          PointsPercept::Point point;
          point.type = PointsPercept::robot;
          point.inImage.x = theSegmentsPercept.segmentsVertical[i-1].x;
          point.inImage.y = theImage.cameraInfo.resolutionHeight - 1;
          point.link = pointsColorChange.points.size();
          point.length2 = 0;
          point.isCloser = true;
          point.segmentId = i-1;
          pointToField(point.inImage.x,point.inImage.y,point.onField);
          pointsColorChange.points.push_back(point);
        }
        containsFieldColorSegment = false;
      }
      if(theSegmentsPercept.segmentsVertical[i+1].x == currentSegment.x)
      {
        nextSegment = theSegmentsPercept.segmentsVertical[i+1];
        nextDistance = (currentSegment.y - currentSegment.length) - nextSegment.y;
      }

      // YELLOW
      if(currentSegment.color == ColorClasses::yellow &&
         ((prevDistance == -1 && currentSegment.y >= theImage.cameraInfo.resolutionHeight - 1 - parameters.distanceFromGreen) ||
          (prevDistance != -1 && prevDistance < parameters.distanceFromGreen && prevSegment.color == ColorClasses::green)))
      {
        PointsPercept::Point point;
        point.inImage.x = currentSegment.x;
        point.inImage.y = currentSegment.y;
        point.link = pointsColorChange.points.size();
        point.length2 = 0;
        point.isCloser = prevDistance == -1;          
        point.segmentId = i;
        pointToField(point.inImage.x,point.inImage.y,point.onField);

        if(nextDistance != -1 && nextDistance < parameters.distanceFromGreen
            && nextSegment.color == ColorClasses::blue)
          point.type = PointsPercept::yellowBeacon;
        else
          point.type = PointsPercept::yellowGoal;

        pointsColorChange.points.push_back(point);
      }

      // BLUE
      else if(currentSegment.color == ColorClasses::blue &&
         ((prevDistance == -1 && currentSegment.y >= theImage.cameraInfo.resolutionHeight - 1 - parameters.distanceFromGreen) ||
          (prevDistance != -1 && prevDistance < parameters.distanceFromGreen && prevSegment.color == ColorClasses::green)))
      {
        PointsPercept::Point point;
        point.inImage.x = currentSegment.x;
        point.inImage.y = currentSegment.y;
        point.link = pointsColorChange.points.size();
        point.length2 = 0;
        point.isCloser = prevDistance == -1;
        point.segmentId = i;
        pointToField(point.inImage.x, point.inImage.y, point.onField);

        if(nextDistance != -1 && nextDistance < parameters.distanceFromGreen
            && nextSegment.color == ColorClasses::yellow)
          point.type = PointsPercept::blueBeacon;
        else
          point.type = PointsPercept::blueGoal;

        pointsColorChange.points.push_back(point);
      }     

      // WHITE
      else if(currentSegment.color == ColorClasses::white)
      {
        if(prevDistance != -1 && prevDistance < parameters.distanceFromGreen
            && nextDistance != -1 && nextDistance < parameters.distanceFromGreen
            && prevSegment.color == ColorClasses::green && nextSegment.color == ColorClasses::green) 
        {
         PointsPercept::Point point;
         point.type = PointsPercept::fieldLine;
         Vector2<int> lineStart, lineEnd;
         pointToField(currentSegment.x, currentSegment.y, lineStart);
         pointToField(currentSegment.x, currentSegment.y - currentSegment.length, lineEnd);
         int diff2 = (lineStart - lineEnd).squareAbs();
         if(diff2 > parameters.minWhiteLength * parameters.minWhiteLength 
            && diff2 < parameters.maxWhiteLength * parameters.maxWhiteLength)        
         {
           point.inImage.x = currentSegment.x;
           point.inImage.y = currentSegment.y-currentSegment.length/2;
           point.link = pointsColorChange.points.size();
           point.length2 = 0;
           point.isCloser = false;
           point.segmentId = i;
           pointToField(point.inImage.x,point.inImage.y,point.onField);
           pointsColorChange.points.push_back(point);  
          }   
        }
      }

      // BLACK
      else if(currentSegment.color == ColorClasses::black && (!skipBody || currentSegment.y < theImage.cameraInfo.resolutionHeight / 2) &&
         ((prevDistance == -1 && currentSegment.y >= theImage.cameraInfo.resolutionHeight - 1 - parameters.distanceFromGreen) ||
          (prevDistance != -1 && prevDistance < parameters.distanceFromGreen &&
          (prevSegment.color == ColorClasses::green || prevSegment.color == ColorClasses::white))))
      {
        PointsPercept::Point point;
        point.type = PointsPercept::robot;
        point.inImage.x = currentSegment.x;
        point.inImage.y = currentSegment.y;
        point.link = pointsColorChange.points.size();
        point.length2 = 0;
        point.isCloser = prevDistance == -1;
        point.segmentId = i;
        pointToField(point.inImage.x,point.inImage.y,point.onField);
        pointsColorChange.points.push_back(point);
      }

      // MAGENTA
      else if(currentSegment.color == ColorClasses::magenta &&
        ((prevDistance == -1 && currentSegment.y >= theImage.cameraInfo.resolutionHeight - 1 - parameters.distanceFromGreen) ||
        (prevDistance != -1 && prevDistance < parameters.distanceFromGreen &&
        (prevSegment.color == ColorClasses::green || prevSegment.color == ColorClasses::white))))
      {
        PointsPercept::Point point;
        point.type = PointsPercept::magentaBeacon;
        point.inImage.x = currentSegment.x;
        point.inImage.y = currentSegment.y;
        point.link = pointsColorChange.points.size();
        point.length2 = 0;
        point.isCloser = prevDistance == -1;
        point.segmentId = i;
        pointToField(point.inImage.x,point.inImage.y,point.onField);
        pointsColorChange.points.push_back(point);
      }

      // CYAN
      else if(currentSegment.color == ColorClasses::cyan &&
        ((prevDistance == -1 && currentSegment.y >= theImage.cameraInfo.resolutionHeight - 1 - parameters.distanceFromGreen) ||
        (prevDistance != -1 && prevDistance < parameters.distanceFromGreen &&
        (prevSegment.color == ColorClasses::green || prevSegment.color == ColorClasses::white))))
      {
        PointsPercept::Point point;
        point.type = PointsPercept::cyanBeacon;
        point.inImage.x = currentSegment.x;
        point.inImage.y = currentSegment.y;
        point.link = pointsColorChange.points.size();
        point.length2 = 0;
        point.isCloser = prevDistance == -1;
        point.segmentId = i;
        pointToField(point.inImage.x,point.inImage.y,point.onField);
        pointsColorChange.points.push_back(point);
      }

      if(!(currentSegment.color == ColorClasses::black) && !(currentSegment.color == ColorClasses::none))
        containsFieldColorSegment = true;
    }
  }
}

void ColorChangePerceptor::connectVertically(PointsPercept& pointsColorChange) const
{
  unsigned i = 0,
   j = 0,
   size = pointsColorChange.points.size();
  if(size)
  {
    PointsPercept::Point* points = &pointsColorChange.points[0];

    while(i < size)
    {
      if(points[j].type == points[i].type &&
       points[j].inImage.x == points[i].inImage.x - parameters.gridStepSize &&
       std::abs(points[j].inImage.y - points[i].inImage.y) < 2 * parameters.gridStepSize)
      {
        PointsPercept::Point& pj(points[j]),
         & pi(points[i]);
        Vector2<int> middle((pi.inImage.x + pj.inImage.x) >> 1, (pi.inImage.y + pj.inImage.y) >> 1);
        switch(pj.type)
        {
         case PointsPercept::fieldLine:
          {
            unsigned color1(SWAP(theImage.image[middle.y][middle.x].color));
            if(theColorTable64.colorClasses[color1 >> 10 & 0x3f][color1 >> 2 & 0x3f][color1 >> 18 & 0x3f] == ColorClasses::white)
             pi.link = j;
            break;
          }
          case PointsPercept::blueBeacon:
          case PointsPercept::blueGoal:
          {
            unsigned color(SWAP(theImage.image[middle.y][middle.x].color));
            if(theColorTable64.colorClasses[color >> 10 & 0x3f][color >> 2 & 0x3f][color >> 18 & 0x3f] == ColorClasses::blue)
             pi.link = j;
            break;
          }
          case PointsPercept::yellowBeacon:
          case PointsPercept::yellowGoal:
          {
            unsigned color(SWAP(theImage.image[middle.y][middle.x].color));
            if(theColorTable64.colorClasses[color >> 10 & 0x3f][color >> 2 & 0x3f][color >> 18 & 0x3f] == ColorClasses::yellow)
             pi.link = j;
            break;
          }
          case PointsPercept::robot:
          {
            unsigned color(SWAP(theImage.image[middle.y][middle.x].color));
            if(theColorTable64.colorClasses[color >> 10 & 0x3f][color >> 2 & 0x3f][color >> 18 & 0x3f] == ColorClasses::black)
             pi.link = j;
            break;
          }
          case PointsPercept::magentaBeacon:
          {
            unsigned color(theImage.image[middle.y][middle.x].color);
            if(theColorTable64.colorClasses[color >> 10 & 0x3f][color >> 2 & 0x3f][color >> 18 & 0x3f] == ColorClasses::magenta)
              pi.link = j;
            break;
          }
          case PointsPercept::cyanBeacon:
            {
              unsigned color(theImage.image[middle.y][middle.x].color);
              if(theColorTable64.colorClasses[color >> 10 & 0x3f][color >> 2 & 0x3f][color >> 18 & 0x3f] == ColorClasses::cyan)
                pi.link = j;
              break;
            }         
          case PointsPercept::numberOfPointTypes:
           break;
        }
      }
      if((j < size - 1 &&
       points[j].inImage.x < points[i].inImage.x - parameters.gridStepSize) ||
       (points[j].inImage.x == points[i].inImage.x - parameters.gridStepSize && 
       (points[j].inImage.y >= points[i].inImage.y || i == size - 1 || 
       points[j].inImage.x < points[i + 1].inImage.x - parameters.gridStepSize)))
        ++j;
      else
        ++i;
    }
  }
}

void ColorChangePerceptor::scanHorizontally(PointsPercept& pointsColorChange) 
{
  unsigned fieldBorderSize = theSegmentsPercept.fieldBorders.size();
  unsigned fieldBorderIndex = fieldBorderSize-1;
  int lastY = -1;

  for(unsigned int i = 1; (i+1) < theSegmentsPercept.segmentsHorizontal.size(); ++i)
  {
    SegmentsPercept::Segment currentSegment = theSegmentsPercept.segmentsHorizontal[i];

    bool insideFieldBorder = false;
    int mid = currentSegment.x + (currentSegment.length / 2);    

    // next scan line
    if(currentSegment.y != lastY)
    {
      lastY = currentSegment.y;
      fieldBorderIndex = fieldBorderSize-1;
    }

    while(fieldBorderIndex >= 1
     && mid > theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x)
    {
     fieldBorderIndex--;
    }

    if((  mid < theSegmentsPercept.fieldBorders.at(fieldBorderSize-1).x
        || mid > theSegmentsPercept.fieldBorders.at(0).x
        || mid == theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x
        || (fieldBorderIndex >= 1 && mid == theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x))
       && currentSegment.y > theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y) 
    {
     insideFieldBorder = true;
    }
    else if(fieldBorderIndex >= 1
     && mid > theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x
     && mid < theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x)
    {
      int xdiff = (theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).x-theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x);
      int ydiff = (theSegmentsPercept.fieldBorders.at(fieldBorderIndex-1).y-theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y);
      double vh = (double)(theSegmentsPercept.fieldBorders.at(fieldBorderIndex).x - mid) / ((double) xdiff);
      if((int) (theSegmentsPercept.fieldBorders.at(fieldBorderIndex).y - (ydiff * vh)) < currentSegment.y)
      {
        insideFieldBorder = true;
      }
    }

    // if the segment lies inside the field border
    if(insideFieldBorder)
    {
      // COLOR
      if(currentSegment.color != ColorClasses::none)
      {
        SegmentsPercept::Segment prevSegment = currentSegment;
        SegmentsPercept::Segment nextSegment = currentSegment;
        int prevDistance = -1;
        int nextDistance = -1;

        if(theSegmentsPercept.segmentsHorizontal[i-1].y == currentSegment.y)
        {
          prevSegment = theSegmentsPercept.segmentsHorizontal[i-1];
          prevDistance = currentSegment.x - (prevSegment.x + prevSegment.length);
        }
        if(theSegmentsPercept.segmentsHorizontal[i+1].y == currentSegment.y)
        {
          nextSegment = theSegmentsPercept.segmentsHorizontal[i+1];
          nextDistance = nextSegment.x - (currentSegment.x + currentSegment.length);
        }

        // WHITE
        if(currentSegment.color == ColorClasses::white)
        {
          if(prevDistance != -1 && prevDistance < parameters.distanceFromGreen
           && nextDistance != -1 && nextDistance < parameters.distanceFromGreen
           && prevSegment.color == ColorClasses::green && nextSegment.color == ColorClasses::green) 
          {
            Vector2<int> lineStart, lineEnd;
            pointToField(currentSegment.x, currentSegment.y, lineStart);
            pointToField(currentSegment.x + currentSegment.length, currentSegment.y, lineEnd);
            int diff2 = (lineStart - lineEnd).squareAbs();
            if(diff2 > parameters.minWhiteLength * parameters.minWhiteLength 
              && diff2 < parameters.maxWhiteLength * parameters.maxWhiteLength)        
            {
              PointsPercept::Point point;
              point.type = PointsPercept::fieldLine;
              point.inImage.x = currentSegment.x+currentSegment.length/2;
              point.inImage.y = currentSegment.y;
              point.link = pointsColorChange.points.size();
              point.length2 = 0;
              point.isCloser = false;
              point.segmentId = -1;
              pointToField(point.inImage.x,point.inImage.y,point.onField);
              pointsColorChange.points.push_back(point);
            }     
          }
        }
      }
    }
  }
}

void ColorChangePerceptor::connectHorizontally(PointsPercept& pointsColorChange, unsigned start) const
{
  unsigned i = start,
   j = start,
   size = pointsColorChange.points.size();
  if(size)
  {
    PointsPercept::Point* points = &pointsColorChange.points[0];
    while(i < size)
    {
      if(points[j].inImage.y == points[i].inImage.y + parameters.gridStepSize &&
       std::abs(points[j].inImage.x - points[i].inImage.x) < 2 * parameters.gridStepSize)
      {
        PointsPercept::Point& pj(points[j]),
         & pi(points[i]);
        Vector2<int> middle((pi.inImage.x + pj.inImage.x) >> 1, (pi.inImage.y + pj.inImage.y) >> 1);
        unsigned color1((theImage.image[middle.y][middle.x].color)),
         color2(SWAP(theImage.image[middle.y][middle.x + 1].color));
        if(theColorTable64.colorClasses[color1 >> 10 & 0x3f][color1 >> 2 & 0x3f][color1 >> 18 & 0x3f] == ColorClasses::white ||
         theColorTable64.colorClasses[color2 >> 10 & 0x3f][color2 >> 2 & 0x3f][color2 >> 18 & 0x3f] == ColorClasses::white)
         pi.link = j;
      }
      if((j < size - 1 &&
       points[j].inImage.y > points[i].inImage.y + parameters.gridStepSize) ||
       (points[j].inImage.y == points[i].inImage.y + parameters.gridStepSize && 
       ((points[j].inImage.x - points[i].inImage.x) <= 0 || i == size - 1 || 
       points[j].inImage.y > points[i + 1].inImage.y + parameters.gridStepSize)))
        ++j;
      else
        ++i;
    }
  }
}

void ColorChangePerceptor::calcLengths(PointsPercept& pointsPercept) const
{
  for(short i = short(pointsPercept.points.size() - 1); i >= 0; --i)
   if(!pointsPercept.points[i].length2 && pointsPercept.points[i].link != i)
    pointsPercept.points[i].length2 = calcLength2(pointsPercept.points[i].onField, pointsPercept, pointsPercept.points[i].link);
}

int ColorChangePerceptor::calcLength2(const Vector2<int>& end, PointsPercept& pointsPercept, short index) const
{
  if(pointsPercept.points[index].link == index)
  {
    Vector2<int> diff = pointsPercept.points[index].onField - end;
    return pointsPercept.points[index].length2 = diff * diff;
  }
  else
    return pointsPercept.points[index].length2 = calcLength2(end, pointsPercept, pointsPercept.points[index].link);
}

MAKE_MODULE(ColorChangePerceptor, Perception)
