/**
* @file ColorChangePerceptor.h
* This file declares a module that provides line points found in the image.
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
* @author <a href="mailto:felipe@informatik.uni-bremen.de">Felipe Lopez</a>
*/

#ifndef __ColorChangePerceptor_h_
#define __ColorChangePerceptor_h_

#include "Tools/Module/Module.h"
#include "Representations/Configuration/ColorTable64.h"
#include "Representations/Infrastructure/Image.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Perception/SegmentsPercept.h"
#include "Representations/Perception/PointsPercept.h"
#include "Representations/Perception/BallSpot.h"
#include "Tools/Debugging/DebugImages.h"

MODULE(ColorChangePerceptor)
REQUIRES(ColorTable64)
REQUIRES(Image)
REQUIRES(CameraMatrix)
REQUIRES(ImageCoordinateSystem)
REQUIRES(SegmentsPercept)
PROVIDES_WITH_MODIFY_AND_DRAW(PointsPercept)
END_MODULE

class ColorChangePerceptor: public ColorChangePerceptorBase
{
private:
 class Parameters : public Streamable
 {
 private:
  void serialize(In* in, Out* out)
  {
   STREAM_REGISTER_BEGIN();
   STREAM(gridStepSize);
   STREAM(distanceFromGreen);
   STREAM(fieldBorderExceed);
   STREAM(minWhiteLength);
   STREAM(maxWhiteLength);
   STREAM_REGISTER_FINISH();
  }

 public:
  int
   gridStepSize, /**< Threshold for recognizing green/blue edges. */
   distanceFromGreen, /**< Maximum distance of line/goal/beacon segments from green in pixels */
   fieldBorderExceed, /**< Number of pixels the scan will exceed the field borders */
   minWhiteLength, /**< The minimum length of a white segment to be accepted as field line (in mm). */
   maxWhiteLength; /**< The maximum length of a white segment to be accepted as field line (in mm). */
 };


 Parameters parameters; /**< The parameters of this module. */
 int cameraHeight, /**< The height of the camera in mm. */
  xCenter, /**< The horizontal center of the image in pixels. */
  yCenter, /**< The vertical center of the image in pixels. */
  rotCos, /**< The cosinus of camera roll multiplied by 1024. */
  rotSin, /**< The sinus of camera roll multiplied by 1024. */
  rotCos2, /**< The cosinus of camera roll multiplied by 1048576. */
  rotSin2, /**< The sinus of camera roll multiplied by 1048576. */
  tiltCos, /**< The cosinus of camera tilt multiplied by 1024. */
  tiltSin, /**< The sinus of camera tilt multiplied by 1024. */
  tiltCos2, /**< The cosinus of camera tilt multiplied by 1048576. */
  tiltSin2, /**< The sinus of camera tilt multiplied by 1048576. */
  skipBody; 
 DECLARE_DEBUG_IMAGE(ColorChangePerceptor); /**< Debug image for visualization. */

 void update(PointsPercept& pointsColorChange);

 /**
 * The method initializes all variables for quickly projecting points to the field.
 * @param origin Will be set to the relative offset of the camera projected to the field.
 */
 void init(Pose2D& origin);

 /**
 * The method scans vertically for edge points on the field.
 * @param pointsColorChange The points will be added to this structure.
 */
 void scanVertically(PointsPercept& pointsColorChange);

 void connectVertically(PointsPercept& pointsColorChange) const;

 void scanHorizontally(PointsPercept& pointsColorChange);

 void connectHorizontally(PointsPercept& pointsColorChange, unsigned start) const;

 /**
 * The method efficiently projects an image point to the field.
 * @param x The x coordinate of the point.
 * @param y The y coordinate of the point.
 * @param p Will be set to the point in relative field coordinates.
 */
 void pointToField(int x, int y, Vector2<int>& p) const;

 void calcLengths(PointsPercept& pointsPercept) const;

 int calcLength2(const Vector2<int>& end, PointsPercept& pointsPercept, short index) const;

public:
 /**
 * Default constructor.
 */
 ColorChangePerceptor();
};

#endif// __ColorChangePerceptor_h_
