/**
* @file NaoCameraMatrixProvider.cpp
* This file implements a class to calculate the position of the camera for the Nao.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#include "NaoCameraMatrixProvider.h"

void NaoCameraMatrixProvider::update(CameraMatrix& cameraMatrix)
{
  // calculate both foots position from actualAngles, origin point between first leg joints
  Pose3D leftFoot, 
         rightFoot;
  lowerFootPositionFromJoints(true, 
                              theSafeJointData.angles[JointData::legLeft0], 
                              theSafeJointData.angles[JointData::legLeft1], 
                              theSafeJointData.angles[JointData::legLeft2],
                              theSafeJointData.angles[JointData::legLeft3],
                              theSafeJointData.angles[JointData::legLeft4],
                              theSafeJointData.angles[JointData::legLeft5],
                              leftFoot);
  lowerFootPositionFromJoints(false, 
                              theSafeJointData.angles[JointData::legRight0], 
                              theSafeJointData.angles[JointData::legRight1], 
                              theSafeJointData.angles[JointData::legRight2],
                              theSafeJointData.angles[JointData::legRight3],
                              theSafeJointData.angles[JointData::legRight4],
                              theSafeJointData.angles[JointData::legRight5],
                              rightFoot);
  
  // decide base foot
  bool left = leftFoot.translation.z < rightFoot.translation.z;

  // transform origin to feet.
  Pose3D body(left ? leftFoot.invert() : rightFoot.invert());

  // The direction of the body is always 0
  body.rotateZ(-body.rotation.getZAngle());

  cameraMatrix = Pose3D(Vector3<double>(0,0,0))
                 .conc(Pose3D(Vector3<double>(0, (left ? -0.5 : 0.5) * (rightFoot.translation.y - leftFoot.translation.y), 0)))
                 .conc(body)
                 .conc(Pose3D(Vector3<double>(theRobotDimensions.xLegJoint1ToHeadPan, 0, theRobotDimensions.zLegJoint1ToHeadPan)))
                 .rotateY(theCameraCalibration.bodyTiltCorrection) 
                 .rotateZ(theSafeJointData.angles[JointData::headPan])
                 .conc(Pose3D(Vector3<double>(theRobotDimensions.xHeadPanToHeadTilt, 0, theRobotDimensions.zHeadPanToHeadTilt))) 
                 .rotateY(-theSafeJointData.angles[JointData::headTilt])
                 .conc(Pose3D(Vector3<double>(theRobotDimensions.xHeadTiltToCamera, 0, theRobotDimensions.zHeadTiltToCamera)))
                 .rotateY(theCameraCalibration.cameraTiltCorrection) 
                 .rotateX(theCameraCalibration.cameraRollCorrection); 

  cameraMatrix.isValid = theMotionInfo.isMotionStable && 
                         theFallDownState.state == FallDownState::upright &&
                         theFrameInfo.getTimeSince(theSafeJointData.timeStamp) < 500 &&
                         theRobotInfo.penalty == PENALTY_NONE;
}

void NaoCameraMatrixProvider::lowerFootPositionFromJoints(bool isLeft, double j0, double j1, double j2, double j3, 
                                                          double j4, double j5, Pose3D& lowerFootPosition) const
{
  lowerFootPosition = Pose3D(0.0, theRobotDimensions.lengthBetweenLegs / 2.0, 0.0).
                      rotateX(-pi_4).
                      rotateY(j0).
                      rotateX(pi_4 - j1).
                      translate(0.0, 0.0, -theRobotDimensions.lengthHipToLegJoint2).
                      rotateY(j2).
                      translate(0.0, 0.0, -theRobotDimensions.upperLegLength).
                      rotateY(j3).
                      translate(0.0, 0.0, -theRobotDimensions.lowerLegLength).
                      rotateY(j4).
                      translate(0.0, 0.0, -theRobotDimensions.lengthFootToLegJoint4).
                      rotateX(-j5).
                      translate(0.0, 0.0, -theRobotDimensions.heightLeg5Joint);

  // mirror y for left leg 
  if(isLeft)
    lowerFootPosition.translation.y = -lowerFootPosition.translation.y;
}

MAKE_MODULE(NaoCameraMatrixProvider, Perception);
