/**
* @file SegmentsPerceptor.cpp
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
* @author <a href="mailto:felipe@informatik.uni-bremen.de">Felipe Lopez</a>
*/

#include "SegmentsPerceptor.h"
#include "Tools/Streams/InStreams.h"

SegmentsPerceptor::SegmentsPerceptor()
{
  InConfigFile stream("pointsSegments.cfg", "default");
  ASSERT(stream.exists());
  stream >> parameters;
}

void SegmentsPerceptor::update(SegmentsPercept& segmentsPercept)
{
  MODIFY("parameters:SegmentsPerceptor", parameters);
  ballSpot.found = false;
  maxBallSegmentLength = 5;
  segmentsPercept.segmentsVertical.clear();
  segmentsPercept.segmentsHorizontal.clear();
  segmentsPercept.fieldBorders.clear();
  int yEnd = scanVertically(segmentsPercept);
  scanHorizontally(yEnd, segmentsPercept);
  calculateConvexHull(segmentsPercept);
}

int SegmentsPerceptor::scanVertically(SegmentsPercept& segmentsPercept)
{
  // imagewidth - 2*minborder is the area which has to be scanned
  int xStart = parameters.minBorder + ((theImage.cameraInfo.resolutionWidth - 2 * parameters.minBorder - 1) % parameters.gridStepSize) / 2,
    xEnd = theImage.cameraInfo.resolutionWidth - xStart,
    yStart = theImage.cameraInfo.resolutionHeight - 1,
    yHorizon = theImageCoordinateSystem.fromHorizonBased(Vector2<double>()).y,
    // prevent horizon coordinates from exeding image-coordinates
    yHorizonClipped = std::min(theImage.cameraInfo.resolutionHeight - 1, std::max(0, yHorizon)),
    yHorizonCorrected = (int) theImageCoordinateSystem.toCorrected(Vector2<int>(0, yHorizonClipped)).y,
    whitePixelCount = 0;

  // The horizon is given in image coordinates. However, the points found will be corrected
  // to compensate for image distortion due to camera motion. So they may be moved on or
  // above the horizon. To prevent this, we adjust the scan grid in advance.
  // We would need fromCorrected() here, to determine which y-coordinate would be projected 
  // on the horizon. However, moving the horizon by its own projection offset should also 
  // work, since the grid will start lower anyway.
  yHorizon += yHorizonClipped - yHorizonCorrected;

  // yEnd is the y-coordinate to which the vertical scan has to be done
  int yEnd = std::max(0, yHorizon + parameters.pixelsBelowHorizon);

  //Every second vertical line has to be half scanned
  bool halfScan = false;

  for(int x = xStart; x < xEnd; x += parameters.gridStepSize/2)
  {
    //If the line has to be half scanned, the start of the y coordinate will be incremented in pixelsUnderHorizon
     int yScanStart = halfScan && yEnd + parameters.pixelsUnderHorizon < yStart ? yEnd + parameters.pixelsUnderHorizon : yStart;

    // actual color we are actually interested in
    unsigned lastColor(SWAP(theImage.image[yScanStart][x].color));
    ColorClasses::Color lastColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[lastColor >> 10 & 0x3f][lastColor >> 2 & 0x3f][lastColor >> 18 & 0x3f];
 
    int currentStart = yScanStart;

    //The next line has/ has not to be scanned
    halfScan = !halfScan;

    for(int y = yScanStart; y > yEnd; --y)
    {     
      // color of current scanned pixel
      unsigned currentColor(SWAP(theImage.image[y][x].color));
      ColorClasses::Color currentColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[currentColor >> 10 & 0x3f][currentColor >> 2 & 0x3f][currentColor >> 18 & 0x3f];

      // If the current color does not equal the color we are interested in
      // we try to scan the following (skipOffset) pixels for the interesting color.
      if(currentColorConverted != lastColorConverted)
      {
        bool skip = false;
        int i=1;
        whitePixelCount = 0;

        if (currentColorConverted == ColorClasses::white)
           whitePixelCount++;

        for(; i<parameters.skipOffset && y-i>yEnd; i++)
        {
          unsigned tmpColor(SWAP(theImage.image[y-i][x].color));
          ColorClasses::Color tmpColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[tmpColor >> 10 & 0x3f][tmpColor >> 2 & 0x3f][tmpColor >> 18 & 0x3f];

          // if the color repeats we treat the current pixel as 'Rauschen'
          // but skip small white lines
          if(lastColorConverted != ColorClasses::white)
          {
            if(tmpColorConverted == ColorClasses::white)
             whitePixelCount++;
            else
             whitePixelCount=0;

            if(whitePixelCount >= parameters.skipOffsetWhite)
             break;
          }
          // skip pixels
          if(tmpColorConverted == lastColorConverted)
          {
           skip = true;
           break;
          }
        }

        // Skip the pixels
        if(skip)
        {
          y = y-i;
        }
        // otherwise insert new segment
        // and change the color we are looking for
        else
        {
          if((currentStart - y >= parameters.minSegmentLength || (lastColorConverted == ColorClasses::white && currentStart - y >= parameters.minSegmentLengthWhite)) && lastColorConverted != ColorClasses::none )
          {
            SegmentsPercept::Segment seg;
            seg.color = lastColorConverted;
            seg.x = x;
            seg.y = currentStart;
            seg.length = currentStart - y;
            seg.onHalfScanLine = !halfScan;
            segmentsPercept.segmentsVertical.push_back(seg);

            // Search for Ballspot
            if(lastColorConverted == ColorClasses::orange && currentStart-y>maxBallSegmentLength)
            {
              maxBallSegmentLength = currentStart - y;
              ballSpot.found = true;
              ballSpot.position.x = x;
              ballSpot.position.y = (currentStart + y) >> 1;
            }
          }
          currentStart = y;
          lastColorConverted = currentColorConverted;
        }
      }
    }

    // when the line is finished insert a new segment
    if((currentStart - yEnd >= parameters.minSegmentLength  || (lastColorConverted == ColorClasses::white && currentStart - yEnd >= parameters.minSegmentLengthWhite)) && lastColorConverted != ColorClasses::none )
    {    
      SegmentsPercept::Segment seg;
      seg.color = lastColorConverted;
      seg.x = x;
      seg.y = currentStart;
      seg.length = currentStart - yEnd;
      seg.onHalfScanLine = !halfScan;
      segmentsPercept.segmentsVertical.push_back(seg);

      // Search for Ballspot
      if(lastColorConverted == ColorClasses::orange && currentStart-yEnd>maxBallSegmentLength)
      {
       maxBallSegmentLength = currentStart - yEnd;
       ballSpot.found = true;
       ballSpot.position.x = x;
       ballSpot.position.y = (currentStart + yEnd) >> 1;
      }
    }
  }
  return yEnd;
}

void SegmentsPerceptor::scanHorizontally(int yEnd, SegmentsPercept& segmentsPercept)
{
  // imagewidth - 2*minborder is the area which has to be scanned
  int xStart = parameters.minBorder + ((theImage.cameraInfo.resolutionHeight - 2 * parameters.minBorder - 1) % parameters.gridStepSize) / 2,
    xEnd = theImage.cameraInfo.resolutionWidth - xStart,
    yStart = theImage.cameraInfo.resolutionHeight-1,
    stepSize = parameters.gridStepSize,
    whitePixelCount = 0;
    
  

  for(int y = yStart; y > yEnd; y -= stepSize)
  {
    // actual color we are actualy interested in
    unsigned lastColor(SWAP(theImage.image[y][xStart].color));
    ColorClasses::Color lastColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[lastColor >> 10 & 0x3f][lastColor >> 2 & 0x3f][lastColor >> 18 & 0x3f];
    int currentStart = xStart;
    if(y<yEnd+parameters.pixelsUnderHorizon)    
      stepSize = parameters.gridStepSize / 2;

    for(int x = xStart; x < xEnd; x++)
    {
      // color of current scanned pixel
      unsigned currentColor(SWAP(theImage.image[y][x].color));
      ColorClasses::Color currentColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[currentColor >> 10 & 0x3f][currentColor >> 2 & 0x3f][currentColor >> 18 & 0x3f];

      // If the current color does not equal the color we are interested in
      // we try to scan the following (skipOffset) pixels for the interesting color.
      if(currentColorConverted != lastColorConverted)
      {
        bool skip = false;
        int i=1;
        whitePixelCount = 0;

        if(currentColorConverted == ColorClasses::white)
         whitePixelCount++;

        for(; i<parameters.skipOffset && x+i<=xEnd; i++)
        {
          unsigned tmpColor(SWAP(theImage.image[y][x+i].color));
          ColorClasses::Color tmpColorConverted = (ColorClasses::Color) theColorTable64.colorClasses[tmpColor >> 10 & 0x3f][tmpColor >> 2 & 0x3f][tmpColor >> 18 & 0x3f];
          
          // if the color repeats we treat the current pixel as 'Rauschen'
          // but skip small white lines
          if(lastColorConverted != ColorClasses::white)
          {
           if (tmpColorConverted == ColorClasses::white)
            whitePixelCount++;
           else
            whitePixelCount=0;

           if(whitePixelCount >= parameters.skipOffsetWhite)
            break;
          }
          // skip pixels
          if(tmpColorConverted == lastColorConverted)
          {
            skip = true;
            break;
          }
        }

        // Skip the pixels
        if(skip)
        {
          x = x+i;
        }
        // otherwise insert new segment
        // and change the color we are looking for
        else
        {
          if((x-currentStart >= parameters.minSegmentLength  || (lastColorConverted == ColorClasses::white && x-currentStart >= parameters.minSegmentLengthWhite)) && lastColorConverted != ColorClasses::none )
          {
            SegmentsPercept::Segment seg;
            seg.color = lastColorConverted;
            seg.x = currentStart;
            seg.y = y;
            seg.length = x - currentStart;
            seg.onHalfScanLine = false;
            segmentsPercept.segmentsHorizontal.push_back(seg);
          }
          currentStart = x;
          lastColorConverted = currentColorConverted;
        }
      }
    }

    // when the line is finished insert a new segment
    if((xEnd-currentStart >= parameters.minSegmentLength || (lastColorConverted == ColorClasses::white && xEnd-currentStart >= parameters.minSegmentLengthWhite)) && lastColorConverted != ColorClasses::none )
    {    
      SegmentsPercept::Segment seg;
      seg.color = lastColorConverted;
      seg.x = currentStart;
      seg.y = y;
      seg.length = xEnd - currentStart;
      seg.onHalfScanLine = false;
      segmentsPercept.segmentsHorizontal.push_back(seg);
    }
  }
 }

 void SegmentsPerceptor::calculateConvexHull(SegmentsPercept& segmentsPercept)
 {
  // If segments were found
  if(segmentsPercept.segmentsVertical.size())
  {
   SegmentsPercept::Segment * firstGreenSegment = NULL;
   int currentX = segmentsPercept.segmentsVertical.at(segmentsPercept.segmentsVertical.size()-1).x;
   bool foundOnCurrentScanLine = false;

   for(int i = (int) segmentsPercept.segmentsVertical.size()-1; i >= 0; --i)
   {
    SegmentsPercept::Segment currentSegment = segmentsPercept.segmentsVertical[i];

    if(currentX != currentSegment.x)
    {
      if(firstGreenSegment != NULL && !foundOnCurrentScanLine)
      {
       Vector2<int> b (firstGreenSegment->x, firstGreenSegment->y - firstGreenSegment->length);
       segmentsPercept.fieldBorders.push_back(b);      
      }
      foundOnCurrentScanLine = false;
      currentX = currentSegment.x;
      firstGreenSegment = NULL;
    }

    if(!foundOnCurrentScanLine)
    {
     if(currentSegment.color == ColorClasses::green && firstGreenSegment == NULL)
     {
       firstGreenSegment = &segmentsPercept.segmentsVertical[i];
     }
     // TODO: Handle Distance
     else if(currentSegment.color != ColorClasses::green && currentSegment.color != ColorClasses::white)
     {
       firstGreenSegment = NULL;
     }
     else if(currentSegment.color == ColorClasses::white && firstGreenSegment != NULL) // TODO
     {
       foundOnCurrentScanLine = true;
       Vector2<int> b (firstGreenSegment->x, firstGreenSegment->y - firstGreenSegment->length);
       segmentsPercept.fieldBorders.push_back(b);
     }
    }
   }

   if(firstGreenSegment != NULL && !foundOnCurrentScanLine)
   {
     Vector2<int> b (firstGreenSegment->x, firstGreenSegment->y - firstGreenSegment->length);
     segmentsPercept.fieldBorders.push_back(b);      
   }

   // if convex hull points exist
   if(segmentsPercept.fieldBorders.size() > 0)
   {
    std::vector<Vector2<int> > ch = jarvisMarch(segmentsPercept, 0, segmentsPercept.fieldBorders.size()-1);
    segmentsPercept.fieldBorders = ch;
   }
  }
  if(segmentsPercept.fieldBorders.size() < 2)
  {
    segmentsPercept.fieldBorders.clear();
    Vector2<int> pr (theImage.cameraInfo.resolutionWidth-1, 0);
    Vector2<int> pl (0, 0);
    segmentsPercept.fieldBorders.push_back(pr);
    segmentsPercept.fieldBorders.push_back(pl);
  }
 }

std::vector<Vector2<int> > SegmentsPerceptor::jarvisMarch(SegmentsPercept& segmentsPercept, int n, int h)
{
 std::vector<Vector2<int> > chp;
 unsigned int i=n;
 while(i < segmentsPercept.fieldBorders.size())
 {
  chp.push_back(segmentsPercept.fieldBorders.at(i));
  i=indexOfRightmostPointFrom(segmentsPercept, i);
 }
 if((int)i != n && i < segmentsPercept.fieldBorders.size())
   chp.push_back(segmentsPercept.fieldBorders.at(i));
 return chp;
}

int SegmentsPerceptor::indexOfRightmostPointFrom(SegmentsPercept& segmentsPercept, int q)
{
 unsigned int b = q+1;
 for(unsigned int j = q+1; j < segmentsPercept.fieldBorders.size(); ++j)
 {
  // If j is better
  double gkj = (double) segmentsPercept.fieldBorders.at(q).y-segmentsPercept.fieldBorders.at(j).y;
  double akj = (double) segmentsPercept.fieldBorders.at(q).x-segmentsPercept.fieldBorders.at(j).x;
  double gkb = (double) segmentsPercept.fieldBorders.at(q).y-segmentsPercept.fieldBorders.at(b).y;
  double akb = (double) segmentsPercept.fieldBorders.at(q).x-segmentsPercept.fieldBorders.at(b).x;
  
  double rj = gkj / akj;
  double rb = gkb / akb;

  if(rj >= rb)
    b = j;
 }
 return b;
}

MAKE_MODULE(SegmentsPerceptor, Perception)
