/**
* @file SegmentsPerceptor.h
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
* @author <a href="mailto:felipe@informatik.uni-bremen.de">Felipe Lopez</a>
*/

#ifndef __SegmentsPerceptor_h_
#define __SegmentsPerceptor_h_

#include "Tools/Module/Module.h"
#include "Representations/Configuration/ColorTable64.h"
#include "Representations/Infrastructure/Image.h"
#include "Representations/Perception/CameraMatrix.h"
#include "Representations/Perception/ImageCoordinateSystem.h"
#include "Representations/Perception/SegmentsPercept.h"
#include "Representations/Perception/BallSpot.h"
#include "Tools/Debugging/DebugImages.h"

MODULE(SegmentsPerceptor)
REQUIRES(ColorTable64)
REQUIRES(Image)
REQUIRES(ImageCoordinateSystem)
REQUIRES(SegmentsPercept)
PROVIDES_WITH_MODIFY_AND_DRAW(SegmentsPercept)
PROVIDES_WITH_MODIFY(BallSpot)
END_MODULE

class SegmentsPerceptor: public SegmentsPerceptorBase
{
private:
  class Parameters : public Streamable
  {
  private:
    void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM(gridStepSize);
      STREAM(minBorder);
      STREAM(skipOffset);
      STREAM(skipOffsetWhite);
      STREAM(pixelsBelowHorizon);
      STREAM(minSegmentLength);
      STREAM(minSegmentLengthWhite);
      STREAM(pixelsUnderHorizon);
      STREAM_REGISTER_FINISH();
    }

  public:
    int gridStepSize, /**< The pixels distance between neighboring scan lines. */
      minBorder, /**< The minimum distance between scan line and image border. */
      skipOffset, /**< The maximum number of pixels to skip when grouping pixels to segments. */
      skipOffsetWhite, /**< The minimum number of white pixels to avoid segments grouping. */
      pixelsBelowHorizon, /**< The number of pixels scanning stops below the horizon. */ 
      minSegmentLength, /**<The minimal length of a segment. */
      minSegmentLengthWhite, /**<The minimal length of a white segment. */
      pixelsUnderHorizon; /**< The number of pixels under horizon to scan with more frequency. */
  };

  Parameters parameters; /**< The parameters of this module. */
  BallSpot ballSpot; /**< The best ball spot found (if any). */
  int maxBallSegmentLength; /**< The maximum length of a ball segment found so far. */

  void update(SegmentsPercept& segmentsPercept);
  void update(BallSpot& ballSpot) {ballSpot = this->ballSpot;}

  /**
  * The method scans vertically for segments in the image.
  * @param segmentsPercept The segments will be added to this structure.
  * @return The top y coordinate reached in the middle of the image.
  */
  int scanVertically(SegmentsPercept& segmentsPercept);

  /**
  * The method scans horizontally for segments in the image.
  * @param yEnd The top y coordinate used for scanning.
  * @param segmentsPercept The points will be added to this structure.
  */
  void scanHorizontally(int yEnd, SegmentsPercept& segmentsPercept);

  void calculateConvexHull(SegmentsPercept& segmentsPercept);

  std::vector<Vector2<int> > jarvisMarch(SegmentsPercept& segmentsPercept, int n, int h);

  int indexOfRightmostPointFrom(SegmentsPercept& segmentsPercept, int q);

public:
  /**
  * Default constructor.
  */
  SegmentsPerceptor();
};

#endif// __SegmentsPerceptor_h_
