/** 
* @file Platform/SimRobotQt/LocalRobot.cpp
*
* Implementation of LocalRobot.
*
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
* @author <A href="mailto:kspiess@tzi.de">Kai Spiess</A>
*/

#include "LocalRobot.h"
#include <ControllerQt/ConsoleRoboCupCtrl.h>
#include "Platform/itoa.h"

LocalRobot::LocalRobot()
: RobotConsole(theDebugReceiver,theDebugSender),
theDebugReceiver(this,"Receiver.MessageQueue.O",false),
theDebugSender(this,"Sender.MessageQueue.S",false),
imageTimeStamp(0),
imageLastTimeStampSent(0)
{
  mode = ConsoleRoboCupCtrl::getMode();
  addViews();
  std::string robotName = RoboCupCtrl::getRobotFullName();

  if(mode == SystemCall::logfileReplay)
  {
    logFile = ConsoleRoboCupCtrl::getLogFile();
    logPlayer.open(logFile.c_str());
    logPlayer.play();
  }
  else if(mode == SystemCall::simulatedRobot)
  {
    VERIFY(objRobot = ctrl->getObjectReference(robotName));
    VERIFY(objBody = ctrl->getObjectReference(robotName));

    model = objRobot->getChildNode(0)->getName() == "CM5Box" ? Settings::bioloid : Settings::nao;

    // Get joint-ids
    std::vector<std::string> parts;
    parts.resize(2);
    parts[0] = ConsoleRoboCupCtrl::getRobotFullName();
    for(int i = 0; i < JointData::numOfJoints; ++i)
    {
      std::string angleName(JointData::getJointName(JointData::Joint(i)));
      parts[1] = angleName;
      spJoints[i] = ctrl->getSensorportIdByPartOfName(parts);
      parts[1] += ".motorAngle";
      apJoints[i] = ctrl->getActuatorportIdByPartOfName(parts);
      parts[1] = angleName + ".switch";
      psJoints[i] = ctrl->getActuatorportIdByPartOfName(parts);
    }

    // default: LED not available
    for(int i = 0; i < LEDRequest::numOfLEDs; ++i)
      apLeds[i] = -1;

    if(model == Settings::bioloid)
      for(int i = 0; i < LEDRequest::numOfBioloidLEDs; ++i)
      {
        std::string ledName(LEDRequest::getBioloidLEDName(LEDRequest::LED(i)));
        parts[1] = "LED" + ledName + ".state";
        apLeds[i] = ctrl->getActuatorportIdByPartOfName(parts);
      }

    parts[1] = "camera";
    spCamera = ctrl->getSensorportIdByPartOfName(parts);
    parts[1] = "accelerometer";
    spAcc = ctrl->getSensorportIdByPartOfName(parts);
    parts[1] = "gyroscope";
    spGyro = ctrl->getSensorportIdByPartOfName(parts);

    InConfigFile stream(Global::getSettings().expandRobotFilename("jointCalibration.cfg"));
    ASSERT(stream.exists());
    stream >> jointCalibration;
  }
}

int LocalRobot::main() 
{
  // Only one thread can access *this now.
  SYNC;

  if(mode == SystemCall::simulatedRobot)
  {
    debugOut.out.bin << 'm';
    debugOut.out.finishMessage(idProcessBegin);
    debugOut.out.bin << jointData;
    debugOut.out.finishMessage(idJointData);
    debugOut.out.bin << sensorData;
    debugOut.out.finishMessage(idSensorData);
    debugOut.out.bin << 'm';
    debugOut.out.finishMessage(idProcessFinished);

    if(imageLastTimeStampSent != image.timeStamp)
    {
      debugOut.out.bin << 'c';
      debugOut.out.finishMessage(idProcessBegin);
      debugOut.out.bin << image;
      debugOut.out.finishMessage(idImage);
      debugOut.out.bin << 'c';
      debugOut.out.finishMessage(idProcessFinished);
      imageLastTimeStampSent = image.timeStamp;
    }
  }
  else if(mode == SystemCall::physicalRobot &&
          imageLastTimeStampSent != image.timeStamp)
  {
    debugOut.out.bin << 'c';
    debugOut.out.finishMessage(idProcessBegin);
    debugOut.out.bin << image;
    debugOut.out.finishMessage(idImage);
    debugOut.out.bin << 'c';
    debugOut.out.finishMessage(idProcessFinished);
    imageLastTimeStampSent = image.timeStamp;
  }

  theDebugSender.send();
  return -20;
}

void LocalRobot::update()
{
  RobotConsole::update();

  // Only one thread can access *this now.
  SYNC;

  if(mode == SystemCall::logfileReplay)
  {
    if(logAcknowledged)
    {
      if(logPlayer.currentFrameNumber == -1 && (logCycle || logOnce))
        logPlayer.play();
      logOnce = false;
      if(logPlayer.replay())
        logAcknowledged = false;
    }
  }
  else if(mode == SystemCall::physicalRobot)
  {
    int duration = SystemCall::getTimeSince(imageTimeStamp);
    if(duration >= 67)
    {
      imageTimeStamp = duration >= 133 ? SystemCall::getCurrentSystemTime() : imageTimeStamp + 67;
      image.timeStamp = SystemCall::getCurrentSystemTime();
    }
  }
  else if(mode == SystemCall::simulatedRobot)
  {
    bool isBioloid = spJoints[JointData::armLeft2] == -1;
    int imageDelay = isBioloid ? 67 : 33;
    if(moveOp != noMove)
    {
      if(moveOp == moveBoth)
        ctrl->moveObject(objRobot, movePos * 0.001, moveRot * (pi / 180));
      else
        ctrl->moveObject(objRobot, movePos * 0.001);
      moveOp = noMove;
    }
    int duration = SystemCall::getTimeSince(imageTimeStamp);
    if(duration >= imageDelay)
    {
      imageTimeStamp = duration >= 2 * imageDelay ? SystemCall::getCurrentSystemTime() : imageTimeStamp + imageDelay;
      if(spCamera != -1 && calculateImage) // If there is a PDA camera
      {
        unsigned char* pSrc;
        ctrl->getSensorValue(spCamera, pSrc);
        for(int y = 0; y < cameraResolutionHeight; ++y)
          for(int x = 0; x < cameraResolutionWidth; ++x)
          {
            int b = *pSrc++,
                g = *pSrc++,
                r = *pSrc++,
                yy =          306 * r + 601 * g + 117 * b,
                uu = 130048 + 512 * r - 429 * g -  83 * b,
                vv = 130048 - 173 * r - 339 * g + 512 * b;

            if(yy < 0) yy = 0; else if(yy > 261120) yy = 261120;
            if(uu < 0) uu = 0; else if(uu > 261120) uu = 261120;
            if(vv < 0) vv = 0; else if(vv > 261120) vv = 261120;

            image.image[y][x].y = (unsigned char)(yy >> 10);
            image.image[y][x].cr = (unsigned char)(uu >> 10);
            image.image[y][x].cb = (unsigned char)(vv >> 10);
          }
      }
      image.timeStamp = SystemCall::getCurrentSystemTime();
    }

    for(int i = 0; i < JointData::numOfJoints; ++i)
    {
      // Get angles
      double actualAngle;
      if(spJoints[i] == -1)  // if joint does not exist
        actualAngle = JointData::off;
      else
      {
        ctrl->getSensorValue(spJoints[i], actualAngle);
        jointData.angles[i] = fromDegrees(actualAngle) * jointCalibration.joints[i].sign - jointCalibration.joints[i].offset;
      }
      // Set angles
      double targetAngle = jointRequest.angles[i];
      if(targetAngle != JointData::off && 
         targetAngle != JointData::ignore &&
         spJoints[i] != -1) // if joint does exist
        ctrl->setActuatorport(apJoints[i], toDegrees(targetAngle + jointCalibration.joints[i].offset) * jointCalibration.joints[i].sign);
    }

    // Battery
    sensorData.data[SensorData::voltage] = isBioloid ? 12.0f : 1.0f;

    // LEDs
    const unsigned timer = SystemCall::getCurrentSystemTime();
    const bool slow = (timer / 50 & 8) != 0;
    const bool fast = (timer / 10 & 8) != 0;
    for(int i = 0; i < LEDRequest::numOfLEDs; ++i)
      if (apLeds[i] != -1)
        switch(ledRequest.ledStates[i])
        {
        case LEDRequest::on:
          ctrl->setActuatorport(apLeds[i], 1.);
          break;
        case LEDRequest::blinking:
          ctrl->setActuatorport(apLeds[i], slow ? 1. : 0.);
          break;
        case LEDRequest::fastBlinking:
          ctrl->setActuatorport(apLeds[i], fast ? 1. : 0.);
          break;
        default:
          ctrl->setActuatorport(apLeds[i], 0.);
        }

    //@todo: compare signs with real robot
    // Acceleration sensors
    double* acceleration;
    ctrl->getSensorValue(spAcc, acceleration);
    if(model == Settings::nao)
    {
      sensorData.data[SensorData::accX] = float(acceleration[0] / 9.81);
      sensorData.data[SensorData::accY] = float(acceleration[1] / 9.81);
    }
    else
    {
      sensorData.data[SensorData::accX] = float(-acceleration[0] / 9.81);
      sensorData.data[SensorData::accY] = float(-acceleration[1] / 9.81);
    }
    sensorData.data[SensorData::accZ] = float(acceleration[2] / 9.81);

    //@todo: compare signs and absolute values with real robot
    // Gyros
    double* rotationSpeed;
    ctrl->getSensorValue(spGyro, rotationSpeed);
    sensorData.data[SensorData::gyroX] = float(rotationSpeed[0]);
    sensorData.data[SensorData::gyroY] = float(rotationSpeed[1]);
    sensorData.data[SensorData::gyroZ] = float(rotationSpeed[2]);

    // Compass
    Matrix3d n = objBody->getRotation();                         //Hack, please remove if real implementation works
    n.rotateZ(pi); //Robot object is currently rotated in scene  //Hack, please remove if real implementation works
    Vector3d northVec(1.0,0.0,0.0); //north is at y=0, x=inf     //Hack, please remove if real implementation works
    northVec = n*northVec;                                       //Hack, please remove if real implementation works
    sensorData.data[SensorData::compassX] = float(northVec.v[0]) * 50.f + 650;      //Hack, please remove if real implementation works
    sensorData.data[SensorData::compassY] = float(northVec.v[1]) * 50.f + 650;      //Hack, please remove if real implementation works

    // Time stamps for data
    sensorData.timeStamp  = SystemCall::getCurrentSystemTime();
    jointData.timeStamp = SystemCall::getCurrentSystemTime();
  }

  std::string statusText;
  if(mode == SystemCall::logfileReplay)
  {
    statusText = std::string("replaying ") + logFile + " ";
    if(logPlayer.currentFrameNumber != -1)
    {
      char buf[10];
      statusText += itoa_s(logPlayer.currentFrameNumber + 1, buf, sizeof(buf), 10);
    }
    else
      statusText += "finished";
  }

  if(mode != SystemCall::logfileReplay && logPlayer.numberOfFrames != 0)
  {
    if(statusText != "")
      statusText += ", ";
    char buf[10];
    statusText += std::string("recorded ");
    statusText += itoa_s(logPlayer.numberOfFrames, buf, sizeof(buf), 10);
  }

  if(pollingFor)
  {
    statusText += statusText != "" ? ", polling for " : "polling for ";
    statusText += pollingFor;
  }

  if(statusText != "")
  {
    statusText = robotName.substr(robotName.find_last_of(".") + 1) + ": " + statusText;
    ctrl->printStatusText(statusText);
  }
}

MAKE_PROCESS(LocalRobot);
