/**
 * @file Platform/Win32Linux/TcpConnection.cpp
 *
 * Implementation of class TcpConnection.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
 */

#include "TcpConnection.h"
#include "Platform/GTAssert.h"

#include <errno.h>
#include <fcntl.h>

#ifdef _WIN32

#include <sys/types.h>

#define ERRNO WSAGetLastError()
#define RESET_ERRNO WSASetLastError(0)
#define EWOULDBLOCK WSAEWOULDBLOCK
#define NON_BLOCK(socket) ioctlsocket(socket, FIONBIO, (u_long*) "NONE")
#define CLOSE(socket) closesocket(socket)

class _WSAFramework
{
  public:
    _WSAFramework() 
    {
      WORD wVersionRequested = MAKEWORD(1, 0);
      WSADATA wsaData;
      WSAStartup(wVersionRequested, &wsaData);
    }
    ~_WSAFramework() {WSACleanup();}
} _wsaFramework;

#else

#include <sys/select.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>

#define ERRNO errno
#define RESET_ERRNO errno = 0
#define NON_BLOCK(socket) fcntl(socket,F_SETFL,O_NONBLOCK)
#define CLOSE(socket) close(socket)

#endif

TcpConnection::TcpConnection()
{
  transferSocket = 0;
  createSocket = 0;
  overallBytesSent = overallBytesReceived = 0;
  client = false;
}

TcpConnection::TcpConnection(const char* ip, int port, Handshake handshake, int maxPackageSendSize, int maxPackageReceiveSize)
{
  overallBytesSent = overallBytesReceived = 0;
  connect(ip, port, handshake, maxPackageSendSize, maxPackageReceiveSize);
}

TcpConnection::~TcpConnection()
{
  if(isConnected())
    closeTransferSocket();
  if(createSocket > 0)
    CLOSE(createSocket);
}

void TcpConnection::connect(const char* ip, int port, Handshake handshake, int maxPackageSendSize, int maxPackageReceiveSize)
{
  this->handshake = handshake;
  this->maxPackageSendSize = maxPackageSendSize;
  this->maxPackageReceiveSize = maxPackageReceiveSize;
  client = false;
  ack = false;
  createSocket = 0;
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  if(ip) // shall we try as client?
  {
    address.sin_addr.s_addr = inet_addr(ip);
    transferSocket = socket(AF_INET, SOCK_STREAM,0);
    ASSERT(isConnected());
    if(::connect(transferSocket, (sockaddr*) &address, sizeof(sockaddr_in)) == 0)
    {
      NON_BLOCK(transferSocket);
      if(maxPackageSendSize)
        VERIFY(!setsockopt(transferSocket, SOL_SOCKET, SO_SNDBUF, (char*)&maxPackageSendSize, sizeof(maxPackageSendSize)));
      if(maxPackageReceiveSize)
        VERIFY(!setsockopt(transferSocket, SOL_SOCKET, SO_RCVBUF, (char*)&maxPackageReceiveSize, sizeof(maxPackageReceiveSize)));
      client = true;
      return; // success, we are connected as client
    }
    else
    {
      CLOSE(transferSocket);
      transferSocket = 0;
    }
  }
  // connection failed, we try it as server
  createSocket = socket(AF_INET, SOCK_STREAM, 0);
  ASSERT(createSocket > 0);

  //allow reuse of port (after restart of application)
  const char optval = 1;
  setsockopt(createSocket,SOL_SOCKET,SO_REUSEADDR,&optval,1);

  address.sin_addr.s_addr = INADDR_ANY; 
  VERIFY(bind(createSocket, (sockaddr*) &address, sizeof(sockaddr_in)) == 0);
  VERIFY(listen(createSocket, SOMAXCONN) == 0);
  NON_BLOCK(createSocket);
}

bool TcpConnection::sendAndReceive(char* dataToSend, int sendSize,
                                   char*& dataRead, int& readSize)
{
  readSize = 0;

  // if transferSocket is invalid, try to create a new one
  if(!isConnected())
  {
#ifdef LINUX
    unsigned int addrlen = sizeof(sockaddr_in);
#else
    int addrlen = sizeof(sockaddr_in);
#endif

    if(createSocket) 
      transferSocket = accept(createSocket, (sockaddr*) &address, &addrlen);
    else
    {
      transferSocket = socket(AF_INET, SOCK_STREAM,0);
      ASSERT(isConnected());
      if(::connect(transferSocket, (sockaddr*) &address, sizeof(sockaddr_in)) != 0)
      {
        CLOSE(transferSocket);
        transferSocket = 0;
      }
    }

    if(isConnected())
    {
      NON_BLOCK(transferSocket); // switch socket to nonblocking
      if(maxPackageSendSize)
        VERIFY(!setsockopt(transferSocket, SOL_SOCKET, SO_SNDBUF, (char*)&maxPackageSendSize, sizeof(maxPackageSendSize)));
      if(maxPackageReceiveSize)
        VERIFY(!setsockopt(transferSocket, SOL_SOCKET, SO_RCVBUF, (char*)&maxPackageReceiveSize, sizeof(maxPackageReceiveSize)));
      if(handshake != noHandshake)
      {
        int heartbeat = 0; // connection establish, send heartbeat now
        if(!send((char*) &heartbeat, sizeof(heartbeat)))
          closeTransferSocket();
        else
          ack = false;
      }
    }
  }

  // Try to read data
  if(isConnected())
  {
    readSize = receive(dataRead);
    if(readSize < 0) // error, socket is invalid
      closeTransferSocket();
    else if(handshake == sender && readSize > 0 && !sendSize)
    {
      // we have received a package, but we don't want to send one now.
      // so, send heartbeat instead to acknowledge that we are still alive
      sendHeartbeat();
    }
  }

  // Try to send data
  if((handshake != receiver || ack) &&
     isConnected() && sendSize > 0)
  {
    if(!send((char*) &sendSize, sizeof(sendSize)) ||   // sends size of block
       !send(dataToSend, sendSize))                    // sends data
    {
      closeTransferSocket();                          // if this fails close socket
    }
    else
    {
      ack = false;
      return true;
    }
  }
  return false;
}

void TcpConnection::sendHeartbeat()
{
  int empty = 0;
  if(!send((char*) &empty, sizeof(empty)))
    closeTransferSocket();
}

int TcpConnection::receive(char*& buffer)
{
  RESET_ERRNO;
  int size,
      received = recv(transferSocket, (char*) &size, sizeof(size), 0);  
      // looks for data. (non blocking, tries to read size-field)
      // If the received is >0 data arrived
  if(received > 0)
  {
    if(received < (int)sizeof(size) &&   // not all data for size received
       !read(((char*) &size) + received, sizeof(size) - received))  // read size-field
      return -1; // cannot read the size

    if(size == 0)
    {
      ack = true;
      return 0; // nothing to read (maybe heartbeat)
    }
    else
    {
      // prevent from allocating to much buffer
      if ( size > MAX_PACKAGE_SIZE )
        return -1;
      
      buffer = new char[size];    // get buffer for complete package
      ASSERT(buffer);
      if(!read(buffer,size))    // read complete package (wait and read size bytes)
      {
        delete [] buffer;
        return -1; // error
      }
      else
      {
        ack = true;
        return size; // package received
      }
    }
  }
  else if(received < 0 && ERRNO == EWOULDBLOCK)
    return 0; // nothing read, but ok
  else
    return -1; // error
}

bool TcpConnection::read(char* buffer, int size)
{
  int received = 0;
  while(received < size)
  {
    RESET_ERRNO;

    int received2 = recv(transferSocket, buffer + received,
                         size - received, 0);

    if(received2 < 0 && ERRNO != EWOULDBLOCK)  // error during reading of package
    {
      return false;
    }
    else if(ERRNO == EWOULDBLOCK) // wait for the rest
    {
      received2 = 0;
      timeval timeout;
      timeout.tv_sec = 0;
      timeout.tv_usec = 100000;
      fd_set rset;
      FD_ZERO(&rset);
      FD_SET(transferSocket, &rset);
      if(select(transferSocket + 1, &rset, 0, 0, &timeout) == -1)
      {
        return false; // error while waiting
      }
    }
    received += received2;
    overallBytesReceived += received2;
  }
  return true; // ok, data received
}

bool TcpConnection::send(const char* buffer, int size)
{
  RESET_ERRNO;
  int sent = ::send(transferSocket, buffer, size, 0);
  if(sent > 0)
  {
    overallBytesSent += sent;
    while(sent < size && (ERRNO == EWOULDBLOCK || ERRNO == 0))
    {
      timeval timeout;
      timeout.tv_sec = 0;
      timeout.tv_usec = 100000;
      fd_set wset;
      FD_ZERO(&wset);
      FD_SET(transferSocket, &wset);
      RESET_ERRNO;
      if(select(transferSocket + 1, 0, &wset, 0, &timeout) == -1) 
        break;
      RESET_ERRNO;
      int sent2 = ::send(transferSocket, buffer + sent, size - sent, 0);
      if(sent2 >= 0)
      {
        sent += sent2;
        overallBytesSent += sent;
      }
    }
  }
  return ERRNO == 0 && sent == size;
}

void TcpConnection::closeTransferSocket()
{
  CLOSE(transferSocket);
  transferSocket = 0;
  ack = false;
}
