/**
 * @file Platform/Win32Linux/TcpConnection.h
 *
 * Declaration of class TcpConnection.
 *
 * @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
 */

#ifndef __TcpConnection_H__
#define __TcpConnection_H__

#ifdef _WIN32
#include <winsock.h>
#else
#include <netinet/in.h>
#endif

#define MAX_PACKAGE_SIZE    67108864      // max package size that can be received. prevent from allocating to much buffer (max ~64 MB)

/**
 * @class TcpConnection
 * The class implements a tcp connection.
 */
class TcpConnection
{
  public:
    enum Handshake
    {
      noHandshake,
      sender,
      receiver
    };
  private:
    int createSocket, /**< The handle of the basic socket. */
        transferSocket; /**< The handle of the actual transfer socket. */
    sockaddr_in address; /**< The socket address. */
    bool ack,
         client;
    int overallBytesSent, /**< The overall number of bytes sent so far. */
        overallBytesReceived, /**< The overall number of bytes received so far. */
        maxPackageSendSize, /**< The maximum size of an outgoing package. If 0, this setting is ignored. */
        maxPackageReceiveSize; /**< The maximum size of an incoming package. If 0, this setting is ignored. */
    Handshake handshake; /**< The handshake mode. */

    /**
     * The function tries to receive a package.
     * @param buffer A pointer that will be initialized with the address of
     *               the data received. It is valid and has to be freed only
     *               if the function returns a value larger than 0.
     * @return Success of the function: -1: failure, 0: nothing read, 
     *         > 0: success, size of data, and buffer points to data.
     */
    int receive(char*& buffer);

    /**
     * The function reads a number of bytes into a buffer.
     * @param buffer The buffer the data will be read into. 
     *               It must be large enough to contain "size" bytes.
     * @param size The number of bytes to be read.
     * @return Did the function succeed?
     */
    bool read(char* buffer, int size);

    /**
     * The function writes a number of bytes to a socket.
     * @param buffer The data that will be written. 
     * @param size The number of bytes to be written.
     * @return Did the function succeed?
     */
    bool send(const char* buffer, int size);

    /**
     * The function closes the transfer socket.
     */
    void closeTransferSocket();

  public:
    /**
     * Default constructor.
     */
    TcpConnection();

    /**
     * Constructor.
     * The constructor will first try to connect another process as
     * a client. If this does not work, it will open the port as a
     * server.
     * @param ip The ip address of the communication partner. If 0, the port
     *           will be opened as server.
     * @param port The port under which will be communicated. 
     * @param handshake The handshake mode.
     * @param maxPackageSendSize The maximum size of an outgoing package.
     *                           If 0, this setting is ignored. 
     * @param maxPackageReceiveSize The maximum size of an incouming package. 
     *                              If 0, this setting is ignored. 
     */
    TcpConnection(const char* ip, int port, Handshake handshake = noHandshake, 
                  int maxPackageSendSize = 0, int maxPackageReceiveSize = 0);

    /**
     * Destructor.
     */
    ~TcpConnection();

    /**
     * The function will first try to connect another process as
     * a client. If this does not work, it will open the port as a
     * server.
     * @param ip The ip address of the communication partner.
     * @param port The port under which will be communicated. 
     * @param handshake The handshake mode.
     * @param maxPackageSendSize The maximum size of packages to send.
     * @param maxPackageReceiveSize The maximum size of packages to receive.
     */
    void connect(const char* ip, int port, Handshake handshake = noHandshake,
                 int maxPackageSendSize = 0, int maxPackageReceiveSize = 0);

    /** 
     * The function sends and receives data.
     * @param dataToSend The data to be send. The function will not free the buffer.
     * @param sendSize The size of data to send. If 0, no data is sent.
     * @param dataRead If data has been read, the parameter is initialzed with
     *                 the address of a buffer pointing to that data. The 
     *                 buffer has to be freed manually.
     * @param readSize The size of the block read. "dataRead" is only valid 
     *                 (and has to be freed) if this parameter contains a 
     *                 positive number after the call to the function.
     * @return Returns true if the data has been sent.
     */
    bool sendAndReceive(char* dataToSend, int sendSize, char*& dataRead, int& readSize);

    /**
     * The function states whether the connection is still established.
     * @return Does the connection still exist?
     */
    bool isConnected() const {return transferSocket > 0;}

    /**
     * The function states whether this system is the client in the connection.
     * @return Is it the client?
     */
    bool isClient() const {return client;}

    /**
    * The function returns the overall number of bytes sent so far by this object.
    * @return The number of bytes sent since this object was created.
    */
    int getOverallBytesSent() const {return overallBytesSent;}

    /**
    * The function returns the overall number of bytes received so far by this object.
    * @return The number of bytes received since this object was created.
    */
    int getOverallBytesReceived() const {return overallBytesReceived;}

    /**
    * The functions sends a heartbeart.
    */
    void sendHeartbeat();
};

#endif
