/**
* \file Platform/linux/NaoCamera.cpp
* Interface to the Nao camera using linux-uvc.
* \author Colin Graf
* \author Thomas Röfer
*/

#undef __STRICT_ANSI__

#include <cstring>

#include "NaoCamera.h"
#include "Platform/GTAssert.h"
#include "Representations/Infrastructure/CameraInfo.h"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <linux/videodev.h>
#include <linux/version.h>

#ifndef V4L2_CID_POWER_LINE_FREQUENCY

#define V4L2_CID_POWER_LINE_FREQUENCY		(V4L2_CID_BASE+24) 
enum v4l2_power_line_frequency {
	V4L2_CID_POWER_LINE_FREQUENCY_DISABLED	= 0,
	V4L2_CID_POWER_LINE_FREQUENCY_50HZ	= 1,
	V4L2_CID_POWER_LINE_FREQUENCY_60HZ	= 2,
};

#define V4L2_CID_HUE_AUTO			(V4L2_CID_BASE+25) 
#define V4L2_CID_WHITE_BALANCE_TEMPERATURE	(V4L2_CID_BASE+26) 
#define V4L2_CID_SHARPNESS			(V4L2_CID_BASE+27) 
#define V4L2_CID_BACKLIGHT_COMPENSATION 	(V4L2_CID_BASE+28) 

#define V4L2_CID_CAMERA_CLASS_BASE 		(V4L2_CTRL_CLASS_CAMERA | 0x900)
#define V4L2_CID_CAMERA_CLASS 			(V4L2_CTRL_CLASS_CAMERA | 1)

#define V4L2_CID_EXPOSURE_AUTO			(V4L2_CID_CAMERA_CLASS_BASE+1)
enum  v4l2_exposure_auto_type {
	V4L2_EXPOSURE_MANUAL = 0,
	V4L2_EXPOSURE_AUTO = 1,
	V4L2_EXPOSURE_SHUTTER_PRIORITY = 2,
	V4L2_EXPOSURE_APERTURE_PRIORITY = 3
};
#define V4L2_CID_EXPOSURE_ABSOLUTE		(V4L2_CID_CAMERA_CLASS_BASE+2)
#define V4L2_CID_EXPOSURE_AUTO_PRIORITY		(V4L2_CID_CAMERA_CLASS_BASE+3)

#define V4L2_CID_PAN_RELATIVE			(V4L2_CID_CAMERA_CLASS_BASE+4)
#define V4L2_CID_TILT_RELATIVE			(V4L2_CID_CAMERA_CLASS_BASE+5)
#define V4L2_CID_PAN_RESET			(V4L2_CID_CAMERA_CLASS_BASE+6)
#define V4L2_CID_TILT_RESET			(V4L2_CID_CAMERA_CLASS_BASE+7)

#define V4L2_CID_PAN_ABSOLUTE			(V4L2_CID_CAMERA_CLASS_BASE+8)
#define V4L2_CID_TILT_ABSOLUTE			(V4L2_CID_CAMERA_CLASS_BASE+9)

#define V4L2_CID_FOCUS_ABSOLUTE			(V4L2_CID_CAMERA_CLASS_BASE+10)
#define V4L2_CID_FOCUS_RELATIVE			(V4L2_CID_CAMERA_CLASS_BASE+11)
#define V4L2_CID_FOCUS_AUTO			(V4L2_CID_CAMERA_CLASS_BASE+12)
 
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26) */

NaoCamera* NaoCamera::theInstance = 0;

NaoCamera::NaoCamera() :
timeStamp(0)
{
  ASSERT(theInstance == 0);
  theInstance = this;

  // open device
  fd = open("/dev/video0", O_RDWR | O_NONBLOCK);
  ASSERT(fd != -1);

  v4l2_std_id esid0 = WIDTH == 320 ? 0x04000000UL : 0x08000000UL;
  VERIFY(!ioctl(fd, VIDIOC_S_STD, &esid0));

  // set format
  struct v4l2_format fmt;
  memset(&fmt, 0, sizeof(struct v4l2_format));
  fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  fmt.fmt.pix.width = WIDTH;
  fmt.fmt.pix.height = HEIGHT;
  fmt.fmt.pix.pixelformat = V4L2_PIX_FMT_YUYV;
  fmt.fmt.pix.field = V4L2_FIELD_NONE;
  VERIFY(!ioctl(fd, VIDIOC_S_FMT, &fmt));

  ASSERT(fmt.fmt.pix.sizeimage == SIZE);

  // set frame rate
  struct v4l2_streamparm fps;  
  memset(&fps, 0, sizeof(struct v4l2_streamparm));
  fps.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  VERIFY(!ioctl(fd, VIDIOC_G_PARM, &fps));
  fps.parm.capture.timeperframe.numerator = 1;
  fps.parm.capture.timeperframe.denominator = 15;
  fps.parm.capture.capability = V4L2_CAP_TIMEPERFRAME;
  VERIFY(!ioctl(fd, VIDIOC_S_PARM, &fps));

  // request camera's default control settings
  settings.exposure = getControlSetting(V4L2_CID_EXPOSURE);
  settings.brightness = getControlSetting(V4L2_CID_BRIGHTNESS);
  settings.red = getControlSetting(V4L2_CID_RED_BALANCE); 
  settings.blue = getControlSetting(V4L2_CID_BLUE_BALANCE); 
  settings.gain = getControlSetting(V4L2_CID_GAIN);
  // make sure automatic stuff is off
  setControlSetting(V4L2_CID_AUTO_WHITE_BALANCE, 0);
  setControlSetting(V4L2_CID_AUTOGAIN, 0);
}

NaoCamera::~NaoCamera()
{
  theInstance = 0;
}

bool NaoCamera::capturedNew()
{
  if(read(fd, buffer, SIZE) == SIZE)
  {
    timeStamp = SystemCall::getCurrentSystemTime();
    return true;
  }
  else
    return false;
}

const unsigned char* NaoCamera::getImage() const
{
  ASSERT(timeStamp);
  return buffer;
}

unsigned int NaoCamera::getTimeStamp() const
{
  ASSERT(timeStamp);
  return timeStamp;
}

int NaoCamera::getControlSetting(unsigned int id)
{
  struct v4l2_queryctrl queryctrl;
  queryctrl.id = id;
  if(ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl) < 0)
    return -1;
  if(queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
    return -1; // not available
  if(queryctrl.type != V4L2_CTRL_TYPE_BOOLEAN && queryctrl.type != V4L2_CTRL_TYPE_INTEGER)
    return -1; // not supported

  struct v4l2_control control_s;
  control_s.id = id;
  if(ioctl(fd, VIDIOC_G_CTRL, &control_s) < 0) 
    return -1;
  if(control_s.value == queryctrl.default_value)
    return -1;
  return control_s.value;
}

bool NaoCamera::setControlSetting(unsigned int id, int value)
{
  struct v4l2_queryctrl queryctrl;
  queryctrl.id = id;
  if(ioctl(fd, VIDIOC_QUERYCTRL, &queryctrl) < 0)
    return false;
  if(queryctrl.flags & V4L2_CTRL_FLAG_DISABLED)
    return false; // not available
  if(queryctrl.type != V4L2_CTRL_TYPE_BOOLEAN && queryctrl.type != V4L2_CTRL_TYPE_INTEGER)
    return false; // not supported

  // clip value
  if(value < queryctrl.minimum)
    value = queryctrl.minimum;
  if(value > queryctrl.maximum)
    value = queryctrl.maximum;
  if(value < 0)
    value = queryctrl.default_value;

  struct v4l2_control control_s;
  control_s.id = id;
  control_s.value = value;
  if(ioctl(fd, VIDIOC_S_CTRL, &control_s) < 0) 
    return false;
  return true;
}

void NaoCamera::setSettings(const CameraSettings& newset)
{
  if(newset.exposure != settings.exposure)
    setControlSetting(V4L2_CID_EXPOSURE, (settings.exposure = newset.exposure));
  if(newset.brightness != settings.brightness)
    setControlSetting(V4L2_CID_BRIGHTNESS, (settings.brightness = newset.brightness));
  if(newset.blue != settings.blue)
    setControlSetting(V4L2_CID_BLUE_BALANCE, (settings.blue = newset.blue));
  if(newset.red != settings.red)
    setControlSetting(V4L2_CID_RED_BALANCE, (settings.red = newset.red));
  if(newset.gain != settings.gain)
    setControlSetting(V4L2_CID_GAIN, (settings.gain = newset.gain));
}
