/** 
* @file Blackboard.h
* Declaration of a class representing the blackboard.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef __Blackboard_h_
#define __Blackboard_h_

#include "Platform/SystemCall.h"

// Declare prototypes of all representations here:

// Infrastructure
class JointData;
class SafeJointData;
class JointRequest;
class SensorData;
class SafeSensorData;
class SensorSAPOutput;
class KeyStates;
class LEDRequest;
class Image;
class CameraInfo;
class FrameInfo;
class RobotInfo;
class OwnTeamInfo;
class OpponentTeamInfo;
class GameInfo;
class TeamStatus;
class SoundRequest;
class SoundOutput;
class TeamMateData;
class MotionRobotHealth;
class RobotHealth;

// Configuration
class RobotName;
class ColorTable64;
class CameraSettings;
class FieldDimensions;
class RobotDimensions;
class JointCalibration;
class SensorCalibration;
class CameraCalibration;
class BehaviorConfiguration;
class WalkingParameters;
class BalanceParameters;

// Perception
class CameraMatrix;
class ImageCoordinateSystem;
class BallSpot;
class BallPercept;
class PointsPercept;
class SegmentsPercept;

// Modeling
class FallDownState;
class BallModel;
class GroundTruthBallModel;
class GoalModel;
class FreeSpaceModel;
class RobotPose;
class GroundTruthRobotPose;
class DribbleModel;
class SelfLocatorParameter;

// BehaviorControl
class BehaviorControlOutput;

// MotionControl
class OdometryData;
class MotionRequest;
class HeadMotionRequest;
class MotionSelection;
class SpecialActionsOutput;
class WalkingEngineOutput;
class WalkingEngineStandOutput;
class MotionInfo;

// friends
class Process;
class Cognition;
class Motion;

/**
* @class Blackboard
* The class represents the blackboard that contains all representation.
* Note: The blackboard only contains references to the objects as attributes.
* The actual representations are constructed on the heap, because many copies of
* of the blackboard exist but only a single set of the representations shared
* by all instances.
*/
class Blackboard
{
protected:
  // Add all representations as constant references here:
  // Infrastructure
  const JointData& theJointData;
  const SafeJointData& theSafeJointData;
  const JointRequest& theJointRequest;
  const SensorData& theSensorData;
  const SafeSensorData& theSafeSensorData;
  const SensorSAPOutput& theSensorSAPOutput;
  const KeyStates& theKeyStates;
  const LEDRequest& theLEDRequest;
  const Image& theImage;
  const CameraInfo& theCameraInfo;
  const FrameInfo& theFrameInfo;
  const RobotInfo& theRobotInfo;
  const OwnTeamInfo& theOwnTeamInfo;
  const OpponentTeamInfo& theOpponentTeamInfo;
  const GameInfo& theGameInfo;
  const TeamStatus& theTeamStatus;
  const SoundRequest& theSoundRequest;
  const SoundOutput& theSoundOutput;
  const TeamMateData& theTeamMateData;
  const MotionRobotHealth& theMotionRobotHealth;
  const RobotHealth& theRobotHealth;

  // Configuration
  const RobotName& theRobotName;
  const ColorTable64& theColorTable64;
  const CameraSettings& theCameraSettings;
  const FieldDimensions& theFieldDimensions;
  const RobotDimensions& theRobotDimensions;
  const JointCalibration& theJointCalibration;
  const SensorCalibration& theSensorCalibration;
  const CameraCalibration& theCameraCalibration;
  const BehaviorConfiguration& theBehaviorConfiguration;
  const WalkingParameters& theWalkingParameters;
  const BalanceParameters& theBalanceParameters;

  // Perception
  const CameraMatrix& theCameraMatrix;
  const ImageCoordinateSystem& theImageCoordinateSystem;
  const BallSpot& theBallSpot;
  const BallPercept& theBallPercept;
  const PointsPercept& thePointsPercept;
  const SegmentsPercept& theSegmentsPercept;
  
  // Modeling
  const FallDownState& theFallDownState;
  const BallModel& theBallModel;
  const GroundTruthBallModel& theGroundTruthBallModel;
  const GoalModel& theGoalModel;
  const FreeSpaceModel& theFreeSpaceModel;
  const RobotPose& theRobotPose;
  const GroundTruthRobotPose& theGroundTruthRobotPose;
  const DribbleModel& theDribbleModel;
  const SelfLocatorParameter& theSelfLocatorParameter;

  // BehaviorControl
  const BehaviorControlOutput& theBehaviorControlOutput;

  // MotionControl
  const OdometryData& theOdometryData;
  const MotionRequest& theMotionRequest;
  const HeadMotionRequest& theHeadMotionRequest;
  const MotionSelection& theMotionSelection;
  const SpecialActionsOutput& theSpecialActionsOutput;
  const WalkingEngineOutput& theWalkingEngineOutput;
  const WalkingEngineStandOutput& theWalkingEngineStandOutput;
  const MotionInfo& theMotionInfo;

  PROCESS_WIDE_STORAGE_STATIC Blackboard* theInstance; /**< The only real instance in the current process. */

  /**
  * The method is a dummy that is called to prevent the compiler from certain
  * optimizations in a method generated in Module.h.
  * It is empty, but important, not defined inline.
  */
  static void distract();

private:
  /**
  * Default constructor.
  */
  Blackboard();

public:
  /**
  * Virtual destructor.
  * Required for derivations of this class.
  */
  virtual ~Blackboard() {}

  /**
  * Assignment operator.
  * Note: copies will share all representations.
  * @param other The instance that is cloned.
  */
  void operator=(const Blackboard& other);

  /**
  * The operator allocates a memory block that is zeroed.
  * Therefore, all members of this class are initialized with 0.
  * @attention This operator is only called if this class is instantiated by
  * a separate call to new, i.e. it cannot be created as a part of another class.
  * @param size The size of the block in bytes.
  * @return A pointer to the block.
  */
  static void* operator new(unsigned size);

  /**
  * The operator frees a memory block.
  * @param p The address of the block to free.
  */
  static void operator delete(void* p);

  friend class Process; /**< The class Process can set theInstance. */
  friend class Cognition; /**< The class Cognition can read theInstance. */
  friend class Motion; /**< The class Motion can read theInstance. */
};

#endif //__Blackboard_h_
