/**
* @file FieldDimensions.cpp
*
* Some useful functions regarding field dimensions.
*
* @author Max Risler
*/

#include "FieldDimensions.h"
#include "Tools/Settings.h"
#include "Tools/Streams/InStreams.h"

FieldDimensions::FieldDimensions()
: Boundary<double>(0,0)
{
}

void FieldDimensions::load()
{
  int* const valuesInt[]={
    &xPosOpponentFieldBorder, &xPosOpponentGoal, &xPosFrontFlags, &xPosOpponentGoalpost,
    &xPosOpponentGroundline, &xPosOpponentSideCorner, &xPosOpponentPenaltyArea, &xPosHalfWayLine,
    &xPosOwnPenaltyArea, &xPosOwnSideCorner, &xPosOwnGroundline, &xPosOwnGoalpost,
    &xPosBackFlags, &xPosOwnGoal, &xPosOwnFieldBorder,
    &yPosLeftFieldBorder, &yPosLeftFlags, &yPosLeftSideline, &yPosLeftGroundline,
    &yPosLeftPenaltyArea, &yPosLeftGoal, &yPosCenterGoal, &yPosRightGoal,
    &yPosRightPenaltyArea, &yPosRightGroundline, &yPosRightSideline, &yPosRightFlags, &yPosRightFieldBorder,
    &centerCircleRadius, &flagHeight, &flagColorHeight,
    &goalWidth, &goalHeight, &fieldLinesWidth,
    &xPosThrowInPointOpponentHalf, &xPosThrowInPointCenter, &yPosThrowInPointLeft,
    &xPosThrowInPointOwnHalf, &yPosThrowInPointRight, &numberOfRobotsPerTeam, &fieldHasBorder,
    &flagRadius, &ballRadius
  };
  const char* const namesInt[]={
    "xPosOpponentFieldBorder", "xPosOpponentGoal", "xPosFrontFlags", "xPosOpponentGoalpost",
    "xPosOpponentGroundline", "xPosOpponentSideCorner", "xPosOpponentPenaltyArea", "xPosHalfWayLine",
    "xPosOwnPenaltyArea", "xPosOwnSideCorner", "xPosOwnGroundline", "xPosOwnGoalpost",
    "xPosBackFlags", "xPosOwnGoal", "xPosOwnFieldBorder",
    "yPosLeftFieldBorder", "yPosLeftFlags", "yPosLeftSideline", "yPosLeftGroundline",
    "yPosLeftPenaltyArea", "yPosLeftGoal", "yPosCenterGoal", "yPosRightGoal",
    "yPosRightPenaltyArea", "yPosRightGroundline", "yPosRightSideline", "yPosRightFlags", "yPosRightFieldBorder",
    "centerCircleRadius", "flagHeight", "flagColorHeight",
    "goalWidth", "goalHeight", "fieldLinesWidth",
    "xPosThrowInPointOpponentHalf", "xPosThrowInPointCenter", "yPosThrowInPointLeft",
    "xPosThrowInPointOwnHalf", "yPosThrowInPointRight", "numberOfRobotsPerTeam", "fieldHasBorder",
    "flagRadius", "ballRadius"
  };

  const int numOfValuesInt = sizeof(valuesInt)/sizeof(int*);
  ASSERT(sizeof(namesInt)/sizeof(char*) == numOfValuesInt);

  bool initializedInt[numOfValuesInt];

  int i;
  for (i = 0; i < numOfValuesInt; initializedInt[i++] = false);

  readDimensions(valuesInt, namesInt, initializedInt, numOfValuesInt);
  readLines(valuesInt, namesInt, initializedInt, numOfValuesInt);
  readCorners(valuesInt, namesInt, initializedInt, numOfValuesInt);
  readLandmarks(valuesInt, namesInt, initializedInt, numOfValuesInt);
  readFlags(valuesInt, namesInt, initializedInt, numOfValuesInt);

  add(Vector2<double>(xPosOpponentFieldBorder,yPosLeftFieldBorder));
  add(Vector2<double>(xPosOwnFieldBorder,yPosRightFieldBorder));
}

void FieldDimensions::readDimensions(int* const* valuesInt, const char* const* namesInt, bool* initializedInt, int numOfValuesInt)
{
  InConfigFile file(Global::getSettings().expandLocationFilename("field.cfg"), "dimensions");
  std::string buf;
  while(file.exists() && !file.eof())
  {
    file >> buf;
    int i;
    for (i = 0; i < numOfValuesInt; i++)
      if(buf == namesInt[i])
      {
        if (initializedInt[i])
        {
          ASSERT(false);
        }
        file >> buf;
        *valuesInt[i] = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        initializedInt[i] = true;
        break;
      }
    ASSERT(i < numOfValuesInt);
  }
  int i;
  for (i = 0; i < numOfValuesInt; i++)
    if (!initializedInt[i])
    {
      ASSERT(false);
    }
}

void FieldDimensions::readLines(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt)
{
  LinesTable* linesTables[] = {
    &carpetBorder, &fieldBorder, 
    &table[ownBeacon], &table[opponentBeacon], 
    &table[ownGoal], &table[opponentGoal], 
    &table[fieldLines]
  };
  const char* linesTableNames[] = {
    "carpetBorder", "fieldBorder",
    "ownBeacon", "opponentBeacon",
    "ownGoal", "opponentGoal",
    "fieldLines"
  };
  const int numOfLinesTables = sizeof(linesTables)/sizeof(LinesTable*);
  ASSERT(sizeof(linesTableNames)/sizeof(char*) == numOfLinesTables);

  int i;
  for(i = 0; i < numOfLinesTables; i++)
  {
    InConfigFile file(Global::getSettings().expandLocationFilename("field.cfg"), linesTableNames[i]);

    while(file.exists() && !file.eof())
    {
      std::string buf;
      file >> buf;
      if (buf == "(")
      {
        Vector2<double> p1,p2;
        file >> buf;
        p1.x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        p1.y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        if (buf != ")")
        {
          ASSERT(false);
        }
        file >> buf;
        if (buf != "-")
        {
          ASSERT(false);
        }
        file >> buf;
        if (buf != "(")
        {
          ASSERT(false);
        }
        file >> buf;
        p2.x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        p2.y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        if (buf != ")")
        {
          ASSERT(false);
        }

        linesTables[i]->push(p1, p2);
      } 
      else if (buf == "circle")
      {
        Vector2<double> center;
        double radius;
        int numOfSegments;
        file >> buf;
        if (buf != "(")
        {
          ASSERT(false);
        }
        file >> buf;
        center.x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        center.y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        radius = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> numOfSegments;
        file >> buf;
        if (buf != ")")
        {
          ASSERT(false);
        }

        linesTables[i]->pushCircle(center, radius, numOfSegments);
      }
      else if (buf == "quartercircle")
      {
        Vector2<double> center;
        double radius;
        int numOfSegments;
        int angle;
        file >> buf;
        if (buf != "(")
        {
          ASSERT(false);
        }
        file >> buf;
        center.x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        center.y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        radius = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> numOfSegments;
        file >> angle;
        file >> buf;
        if (buf != ")")
        {
          ASSERT(false);
        }

        linesTables[i]->pushQuarterCircle(center, radius, numOfSegments, angle);
      }
      else
      {
        ASSERT(false);
      }
    }
  }
}

void FieldDimensions::readCorners(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt)
{
  std::string buf;
  InConfigFile file(Global::getSettings().expandLocationFilename("field.cfg"), "corners");
  while(file.exists() && !file.eof())
  {
    file >> buf;
    int i;
    for (i = 0; i < numOfCornerClasses; i++)
      if(buf == getCornerClassName((CornerClass)i))
      {
        int orientation;
        Vector2<double> corner;
        file >> orientation;
        file >> buf;
        corner.x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        file >> buf;
        corner.y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
        
        cornerTable[i].addCorner(orientation, corner);
        break;
      }
    if (i >= numOfCornerClasses)
    {
      ASSERT(false);
      file >> buf; file >> buf; file >> buf;
    }
  }
}

void FieldDimensions::readLandmarks(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt)
{
  for (int i=0; i < numberOfLandmarkIDs; i++)
  {
    std::string buf;
    InConfigFile file(Global::getSettings().expandLocationFilename("field.cfg"), getLandmarkName(static_cast<LandmarkIDs>(i)));
    while(file.exists() && !file.eof())
    {
      file >> buf;
      landmarkPosition[i].x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
      file >> buf;
      landmarkPosition[i].y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
    }
  }
}

void FieldDimensions::readFlags(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt)
{
  for (int i=0; i < numberOfFlagTypes; i++)
  {
    flagTypeTable[i].numberOfFlags = 0;
    std::string buf;
    InConfigFile file(Global::getSettings().expandLocationFilename("field.cfg"), getFlagTypeName(static_cast<FlagType>(i)));
    while(file.exists() && !file.eof())
    {
      file >> buf;
      flagTypeTable[i].flagPosition[flagTypeTable[i].numberOfFlags].x = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
      file >> buf;
      flagTypeTable[i].flagPosition[flagTypeTable[i].numberOfFlags++].y = readValue(buf, valuesInt, namesInt, initializedInt, numOfValuesInt);
    }
  }
}

int FieldDimensions::readValue(const std::string& buf, const int* const* values, const char* const* names, const bool* initialized, int numOfValues) const
{
  std::string n = buf;
  bool neg = false;
  if (buf[0] == '-')
  {
    neg=true;
    n = n.substr(1);
  }
  for (int j = 0; j < numOfValues; j++)
    if (n == names[j])
    {
      if (!initialized[j])
      {
        ASSERT(false);
      }
      return (neg) ? (-*values[j]) : (*values[j]);
    }
  return (int)strtol(buf.c_str(),(char**)NULL,0);
}

double FieldDimensions::distanceToOpponentPenaltyArea(const Vector2<double> &p) const
{
  double distanceX, distanceY;
  double minDistance = 0;
  
  if ((p.x <= xPosOpponentPenaltyArea) ||
    (p.y < yPosRightPenaltyArea) ||
    (p.y > yPosLeftPenaltyArea))
  {
    if (p.x > xPosOpponentPenaltyArea)
      distanceX = 0;
    else
      distanceX = xPosOpponentPenaltyArea - p.x;
    
    if (p.y < yPosRightPenaltyArea)
      distanceY = yPosRightPenaltyArea - p.y;
    else if (p.y > yPosLeftPenaltyArea)
      distanceY = p.y - yPosLeftPenaltyArea;
    else
      distanceY = 0;
    return sqrt(distanceX*distanceX+distanceY*distanceY);
  }
  else
  {
    if ((yPosRightPenaltyArea - p.y) > (p.y - yPosLeftPenaltyArea))
    {
      minDistance = yPosRightPenaltyArea - p.y;
    }
    else
    {
      minDistance = p.y - yPosLeftPenaltyArea;
    }
    if ((xPosOpponentPenaltyArea - p.x) > minDistance)
    {
      minDistance = xPosOpponentPenaltyArea - p.x;
    }
    return minDistance;
  }
}

double FieldDimensions::distanceToOwnPenaltyArea(const Vector2<double> &p) const
{
  double distanceX, distanceY;
  double minDistance = 0;
  
  if ((p.x >= xPosOwnPenaltyArea) ||
    (p.y < yPosRightPenaltyArea) ||
    (p.y > yPosLeftPenaltyArea))
  {
    if (p.x < xPosOwnPenaltyArea)
      distanceX = 0;
    else
      distanceX = p.x - xPosOwnPenaltyArea;
    
    if (p.y < yPosRightPenaltyArea)
      distanceY = yPosRightPenaltyArea - p.y;
    else if (p.y > yPosLeftPenaltyArea)
      distanceY = p.y - yPosLeftPenaltyArea;
    else
      distanceY = 0;
    
    return sqrt(distanceX*distanceX+distanceY*distanceY);
  }
  else
  {
    if ((yPosRightPenaltyArea - p.y) >
      (p.y - yPosLeftPenaltyArea))
    {
      minDistance = yPosRightPenaltyArea - p.y;
    }
    else
    {
      minDistance = p.y - yPosLeftPenaltyArea;
    }
    if ((p.x - xPosOwnPenaltyArea) > minDistance)
    {
      minDistance = p.x - xPosOwnPenaltyArea;
    }
    return minDistance;
  }
}

double FieldDimensions::distanceToOwnPenaltyArea(const Pose2D &p) const
{
  return distanceToOwnPenaltyArea(p.translation);
}


bool FieldDimensions::isInsideGoal(const Vector2<int> &p) const
{
  return (
    p.y > yPosRightGoal && p.y < yPosLeftGoal &&
    (
    (p.x > xPosOwnGoal && p.x < xPosOwnGroundline) 
    ||
    (p.x > xPosOpponentGoal && p.x < xPosOpponentGroundline)
    )
    );
}

bool FieldDimensions::isInsideOwnGoal(const Vector2<int> &p) const
{
  return (
    p.y > yPosRightGoal && p.y < yPosLeftGoal &&
    p.x > xPosOwnGoal && p.x < xPosOwnGroundline 
    );
}

bool FieldDimensions::isInsideOpponentGoal(const Vector2<int> &p) const
{
  return (
    p.y > yPosRightGoal && p.y < yPosLeftGoal &&
    p.x < xPosOpponentGoal && p.x > xPosOpponentGroundline 
    );
}

bool FieldDimensions::isOnOwnGoalGroundline(const Vector2<int> &p) const
{
  return (
    p.x == xPosOwnGroundline && 
    p.y < yPosLeftGoal && 
    p.y > yPosRightGoal 
    );
}

bool FieldDimensions::isOnOpponentGoalGroundline(const Vector2<int> &p) const
{
  return (
    p.x == xPosOpponentGroundline && 
    p.y < yPosLeftGoal && 
    p.y > yPosRightGoal 
    );
}

Pose2D FieldDimensions::randomPoseOnField() const
{ 
  Pose2D pose;
  do
    pose = Pose2D::random(x,y,Range<double>(-pi,pi));
  while(!isInsideField(pose.translation));
  return pose;
}

Pose2D FieldDimensions::randomPoseOnCarpet() const
{ 
  Pose2D pose;
  do
    pose = Pose2D::random(x,y,Range<double>(-pi,pi));
  while(!isInsideCarpet(pose.translation));
  return pose;
}

// needed for GTGoalie2006SelfLocator
Pose2D FieldDimensions::randomPoseNearOwnGoal() const
{ 
  Pose2D pose;
  do
    pose = Pose2D::random(x,y,Range<double>(-pi,pi));
  while(!isInsideCarpet(pose.translation) || ((pose.translation - Vector2<double>(-2700.0,0.0)).abs() > 1200));
  return pose;
}

void FieldDimensions::declareDrawings() const
{
  DECLARE_DEBUG_DRAWING("field lines", "drawingOnField", "displays the field lines");
  DECLARE_DEBUG_DRAWING("field polygons", "drawingOnField", "displays the field polygons");
}

void FieldDimensions::drawCrosses() const
{
  ColorRGBA lineColor(255,255,255);
  int lineWidth = fieldLinesWidth;
  LINE( "field lines", -lineWidth, 0,  lineWidth, 0, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", -lineWidth, yPosThrowInPointLeft,  lineWidth, yPosThrowInPointLeft, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", -lineWidth, yPosThrowInPointRight,  lineWidth, yPosThrowInPointRight, lineWidth, Drawings::ps_solid, lineColor);
  
  LINE( "field lines", -lineWidth + xPosThrowInPointOpponentHalf, 0,  lineWidth + xPosThrowInPointOpponentHalf, 0, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOpponentHalf, -lineWidth, xPosThrowInPointOpponentHalf,  lineWidth, lineWidth, Drawings::ps_solid, lineColor);
  
  LINE( "field lines", -lineWidth + xPosThrowInPointOpponentHalf, yPosThrowInPointLeft,  lineWidth + xPosThrowInPointOpponentHalf, yPosThrowInPointLeft, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOpponentHalf, -lineWidth + yPosThrowInPointLeft, xPosThrowInPointOpponentHalf,  lineWidth + yPosThrowInPointLeft, lineWidth, Drawings::ps_solid, lineColor);
  
  LINE( "field lines", -lineWidth + xPosThrowInPointOpponentHalf, yPosThrowInPointRight,  lineWidth + xPosThrowInPointOpponentHalf, yPosThrowInPointRight, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOpponentHalf, -lineWidth + yPosThrowInPointRight, xPosThrowInPointOpponentHalf,  lineWidth + yPosThrowInPointRight, lineWidth, Drawings::ps_solid, lineColor);

  LINE( "field lines", -lineWidth + xPosThrowInPointOwnHalf, 0,  lineWidth + xPosThrowInPointOwnHalf, 0, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOwnHalf, -lineWidth, xPosThrowInPointOwnHalf,  lineWidth, lineWidth, Drawings::ps_solid, lineColor);

  LINE( "field lines", -lineWidth + xPosThrowInPointOwnHalf, yPosThrowInPointLeft,  lineWidth + xPosThrowInPointOwnHalf, yPosThrowInPointLeft, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOwnHalf, -lineWidth + yPosThrowInPointLeft, xPosThrowInPointOwnHalf,  lineWidth + yPosThrowInPointLeft, lineWidth, Drawings::ps_solid, lineColor);

  LINE( "field lines", -lineWidth + xPosThrowInPointOwnHalf, yPosThrowInPointRight,  lineWidth + xPosThrowInPointOwnHalf, yPosThrowInPointRight, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosThrowInPointOwnHalf, -lineWidth + yPosThrowInPointRight, xPosThrowInPointOwnHalf,  lineWidth + yPosThrowInPointRight, lineWidth, Drawings::ps_solid, lineColor);

}

void FieldDimensions::drawLines() const
{
  int lineWidth = fieldLinesWidth;
  ColorRGBA lineColor(255,255,255);

  //side lines
  LINE( "field lines", xPosOwnGroundline, yPosLeftSideline, xPosOpponentGroundline, yPosLeftSideline,       lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOwnGroundline, yPosRightSideline, xPosOpponentGroundline, yPosRightSideline,     lineWidth, Drawings::ps_solid, lineColor);
  //ground lines
  LINE( "field lines", xPosOwnGroundline, yPosLeftSideline, xPosOwnGroundline, yPosRightSideline,           lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOpponentGroundline, yPosLeftSideline, xPosOpponentGroundline, yPosRightSideline, lineWidth, Drawings::ps_solid, lineColor);
  //center line
  LINE( "field lines", xPosHalfWayLine, yPosLeftSideline, xPosHalfWayLine, yPosRightSideline, lineWidth, Drawings::ps_solid, lineColor);
  //center circle
  CIRCLE( "field lines", xPosHalfWayLine, yPosCenterGoal, centerCircleRadius, lineWidth, Drawings::ps_solid, lineColor, Drawings::bs_null, lineColor);
  
  //own penalty area
  LINE( "field lines", xPosOwnGroundline, yPosLeftPenaltyArea, xPosOwnPenaltyArea, yPosLeftPenaltyArea,   lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOwnGroundline, yPosRightPenaltyArea, xPosOwnPenaltyArea, yPosRightPenaltyArea, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOwnPenaltyArea, yPosLeftPenaltyArea, xPosOwnPenaltyArea, yPosRightPenaltyArea, lineWidth, Drawings::ps_solid, lineColor);

  //opponent penalty area
  LINE( "field lines", xPosOpponentGroundline, yPosLeftPenaltyArea, xPosOpponentPenaltyArea, yPosLeftPenaltyArea,   lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOpponentGroundline, yPosRightPenaltyArea, xPosOpponentPenaltyArea, yPosRightPenaltyArea, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOpponentPenaltyArea, yPosLeftPenaltyArea, xPosOpponentPenaltyArea, yPosRightPenaltyArea, lineWidth, Drawings::ps_solid, lineColor);

  // own goal area
  LINE( "field lines", xPosOwnGroundline, yPosLeftGoal, xPosOwnGoal, yPosLeftGoal,   lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOwnGroundline, yPosRightGoal, xPosOwnGoal, yPosRightGoal, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOwnGoal, yPosLeftGoal, xPosOwnGoal, yPosRightGoal, lineWidth, Drawings::ps_solid, lineColor);

  // opponent goal area
  LINE( "field lines", xPosOpponentGroundline, yPosLeftGoal, xPosOpponentGoal, yPosLeftGoal,   lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOpponentGroundline, yPosRightGoal, xPosOpponentGoal, yPosRightGoal, lineWidth, Drawings::ps_solid, lineColor);
  LINE( "field lines", xPosOpponentGoal, yPosLeftGoal, xPosOpponentGoal, yPosRightGoal, lineWidth, Drawings::ps_solid, lineColor);
}

void FieldDimensions::drawAxes() const
{
  int minX = -3000;
  int maxX = +3000;
  int minY = -3000;
  int maxY = +3000;
  int rasterWidth = 200;
  int rasterWidthBold = 1000;
  for(int x = minX; x <= maxX; x+= rasterWidth)
  {
    LINE( "coordinate system", x, minY, x, maxY, 10, Drawings::ps_solid, ColorRGBA(100, 100, 100));
  }
  for(int x = minX; x <= maxX; x+= rasterWidthBold)
  {
    LINE( "coordinate system", x, minY, x, maxY, 20, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  }
  for(int y = minY; y <= maxY; y+= rasterWidth)
  {
    LINE( "coordinate system", minX, y, maxX, y, 10, Drawings::ps_solid, ColorRGBA(100, 100, 100));
  }
  for(int y = minY; y <= maxY; y+= rasterWidthBold)
  {
    LINE( "coordinate system", minX, y, maxX, y, 20, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  }
  LINE( "coordinate system", 0,        minY-300, 0,    minY, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", minX-300, 0,        minX, 0,    30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", minX,     minY-300, minX,     maxY+500, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", minX-300, minY,     maxX+500, minY,     30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", minX,     maxY+500, minX-100, maxY+200, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", minX,     maxY+500, minX+100, maxY+200, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", maxX+500, minY,     maxX+200, minY-100, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system", maxX+500, minY,     maxX+200, minY+100, 30, Drawings::ps_solid, ColorRGBA(0, 0, 0));
}

void FieldDimensions::drawAxesTimeLine() const
{
  int minX = -100;
  int maxX = +600;
  int minY =    0;
  int maxY = +215;
  int rasterWidth = 10;
  int rasterWidthBold = 100;

  for(int x = minX; x <= maxX; x+= 100)
  {
    LINE( "coordinate system time line", x, 1000 - minY * 30, x, 1000 - maxY * 30, 10, Drawings::ps_solid, ColorRGBA(100, 100, 100));
  }

  for(int y = minY; y <= maxY; y+= rasterWidth)
  {
    LINE( "coordinate system time line", minX, 1000 - y * 30, maxX, 1000 - y * 30, 10, Drawings::ps_solid, ColorRGBA(100, 100, 100));
  }
  for(int y = minY; y <= maxY; y+= rasterWidthBold)
  {
    LINE( "coordinate system time line", minX - 100, 1000 - y * 30, maxX, 1000 - y * 30, 20, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  }

  LINE( "coordinate system time line", minX, 1000 - minY * 30, minX, 1000 - maxY * 30, 20, Drawings::ps_solid, ColorRGBA(0, 0, 0));
  LINE( "coordinate system time line", maxX, 1000 - minY * 30, maxX, 1000 - maxY * 30, 20, Drawings::ps_solid, ColorRGBA(0, 0, 0));
}

void FieldDimensions::drawPolygons(RoboCup::uint32 ownColor) const
{
  Vector2<double>* points = new Vector2<double>[carpetBorder.numberOfEntries];
  for (int i = 0; i < carpetBorder.numberOfEntries; ++i)
  {  
    points[i] = carpetBorder.lines[i].corner.translation;
  }
  POLYGON("field polygons", carpetBorder.numberOfEntries, points, 
          1, Drawings::ps_solid, ColorRGBA(0, 180, 0), Drawings::bs_solid, ColorRGBA(0, 140, 0));
  delete [] points;

  ColorClasses::Color own = ownColor == TEAM_BLUE ? ColorClasses::blue : ColorClasses::yellow,
                      opp = ownColor != TEAM_BLUE ? ColorClasses::blue : ColorClasses::yellow;

  Vector2<double> goal[4];
  goal[0] = Vector2<double>(xPosOwnGoalpost, yPosLeftGoal);
  goal[1] = Vector2<double>(xPosOwnGoalpost, yPosRightGoal);
  goal[2] = Vector2<double>(xPosOwnGoal, yPosRightGoal);
  goal[3] = Vector2<double>(xPosOwnGoal, yPosLeftGoal);
  POLYGON("field polygons", 4, goal, 1, Drawings::ps_solid, own, Drawings::bs_solid, own);

  goal[0] = Vector2<double>(xPosOpponentGoalpost, yPosLeftGoal);
  goal[1] = Vector2<double>(xPosOpponentGoalpost, yPosRightGoal); 
  goal[2] = Vector2<double>(xPosOpponentGoal, yPosRightGoal); 
  goal[3] = Vector2<double>(xPosOpponentGoal, yPosLeftGoal);
  POLYGON("field polygons", 4, goal, 1, Drawings::ps_solid, opp, Drawings::bs_solid, opp);


  //Beacons:
  for (int i = 0; i < numberOfFlagTypes; i++)
  {
    ColorClasses::Color beacon = i == 0 ? opp : own; 
    for(int j = 0; j < flagTypeTable[i].numberOfFlags; j++)
    {
      CIRCLE( "field lines", flagTypeTable[i].flagPosition[j].x, flagTypeTable[i].flagPosition[j].y, flagRadius, 1, Drawings::ps_solid, beacon, Drawings::bs_solid, beacon );
    }
  }
}

void FieldDimensions::LinesTable::push(const Pose2D& p, double l)
{
  ASSERT(numberOfEntries<maxNumberOfEntries);
  lines[numberOfEntries].corner = p;
  lines[numberOfEntries++].length = l;
}

void FieldDimensions::LinesTable::push(const Vector2<double>& s, const Vector2<double>& e)
{
  Vector2<double> d = e - s;
  push(Pose2D(d.angle(), s), d.abs());
}

void FieldDimensions::LinesTable::pushCircle(const Vector2<double>& center, double radius, int numOfSegments)
{
  Vector2<double> p1, p2;
  for (double a = 0; a <= pi_4; a += pi2/numOfSegments)
  {
    p1 = Vector2<double>(sin(a), cos(a)) * radius;
    if (a > 0)
    {
      push(center + p1, center + p2);
      push(center + Vector2<double>(p1.x,-p1.y), center + Vector2<double>(p2.x,-p2.y));
      push(center + Vector2<double>(-p1.x,p1.y), center + Vector2<double>(-p2.x,p2.y));
      push(center - p1, center - p2);
      push(center + Vector2<double>(p1.y,p1.x), center + Vector2<double>(p2.y,p2.x));
      push(center + Vector2<double>(p1.y,-p1.x), center + Vector2<double>(p2.y,-p2.x));
      push(center + Vector2<double>(-p1.y,p1.x), center + Vector2<double>(-p2.y,p2.x));
      push(center + Vector2<double>(-p1.y,-p1.x), center + Vector2<double>(-p2.y,-p2.x));
    }
    p2 = p1;
  }
}

void FieldDimensions::LinesTable::pushQuarterCircle(const Vector2<double>& center, double radius, int numOfSegments, int angle)
{
  Vector2<double> p1, p2;
  for (double a = 0; a <= pi_4; a += pi2/numOfSegments)
  {
    p1 = Vector2<double>(sin(a), cos(a)) * radius;
    if (a > 0)
    {
      switch(angle)
      {
        case 0:
          push(center + p1, center + p2);
          push(center + Vector2<double>(p1.y,p1.x), center + Vector2<double>(p2.y,p2.x));
          break;
        case 1:
          push(center + Vector2<double>(-p1.x,p1.y), center + Vector2<double>(-p2.x,p2.y));
          push(center + Vector2<double>(-p1.y,p1.x), center + Vector2<double>(-p2.y,p2.x));
          break;
        case 2:
          push(center - p1, center - p2);
          push(center + Vector2<double>(-p1.y,-p1.x), center + Vector2<double>(-p2.y,-p2.x));
          break;
        case 3:
          push(center + Vector2<double>(p1.x,-p1.y), center + Vector2<double>(p2.x,-p2.y));
          push(center + Vector2<double>(p1.y,-p1.x), center + Vector2<double>(p2.y,-p2.x));
          break;
      }
    }
    p2 = p1;
  }
}

void FieldDimensions::LinesTable::doubleSided(double width, const FieldDimensions::LinesTable& single)
{
  numberOfEntries = 0;

  for (int i = 0; i < single.numberOfEntries; i++)
  {
    push(single.lines[i].corner + 
          Pose2D(Vector2<double>(-width/2,width/2)), 
          single.lines[i].length + width);
    push(single.lines[i].corner + 
          Pose2D(pi, Vector2<double>(single.lines[i].length,0)) + 
          Pose2D(Vector2<double>(-width/2,width/2)), 
          single.lines[i].length + width);
  }
}

bool FieldDimensions::LinesTable::isInside(const Vector2<double>& v) const
{
  //note:
  //This function assumes that the point (0,0) is inside and
  //that for any point inside the area the line to (0,0) belongs to the area too.

  Geometry::Line testLine(v, -v);
  for (int i = 0; i < numberOfEntries; ++i)
  {
    double factor;
    Geometry::Line border(lines[i].corner, lines[i].length);
    if (Geometry::getIntersectionOfRaysFactor(border, testLine, factor))
      return false;
  }
  return true;
}

double FieldDimensions::LinesTable::clip(Vector2<double>& v) const
{
  if(isInside(v))
    return 0;
  else
  {
    Vector2<double> old = v,
                          v2;
    double minDist = 100000;
    for(int i = 0; i < numberOfEntries; ++i)
    {
      Vector2<double> diff = (Pose2D(old) - lines[i].corner).translation;
      if(diff.x < 0)
        v2 = lines[i].corner.translation;

      else if(diff.x > lines[i].length)
        v2 = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length,0))).translation;
      else
        v2 = (lines[i].corner + Pose2D(Vector2<double>(diff.x,0))).translation;
      double dist = (old - v2).abs();
      if(minDist > dist)
      {
        minDist = dist;
        v = v2;
      }
    }
    return (v - old).abs();
  }
}

bool FieldDimensions::LinesTable::getClosestPoint(Vector2<double>& vMin, const Pose2D& p, int numberOfRotations, double minLength) const
{
  int trueNumberOfRotations = numberOfRotations;
  if(numberOfRotations == 2)
    numberOfRotations = 4;

  // target angle -> target index
  double r = p.rotation / pi2 * numberOfRotations + 0.5;
  if(r < 0)
    r += numberOfRotations;
  int targetRot = int(r);
  ASSERT(targetRot >= 0 && targetRot < numberOfRotations);
  targetRot %= trueNumberOfRotations;
  Vector2<double> v2;
  double minDist = 100000;
  for(int i = 0; i < numberOfEntries; ++i)
    if(lines[i].length >= minLength)
    {
      // angle -> index
      double r = (lines[i].corner.rotation + pi_2) / pi2 * numberOfRotations + 0.5;
      if(r < 0)
        r += numberOfRotations;
      else if(r >= numberOfRotations)
        r -= numberOfRotations;
      int rot = int(r);
      ASSERT(rot >= 0 && rot < numberOfRotations);
      rot %= trueNumberOfRotations;

      // index must be target index
      if(rot == targetRot)
      {
        Vector2<double> diff = (p - lines[i].corner).translation;
        if(diff.x < 0)
          v2 = lines[i].corner.translation;
        else if(diff.x > lines[i].length)
          v2 = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length,0))).translation;
        else
          v2 = (lines[i].corner + Pose2D(Vector2<double>(diff.x,0))).translation;
        Vector2<double> vDiff = v2 - p.translation;
        double dist = vDiff.abs();
        if(minDist > dist)
        {
          minDist = dist;
          vMin = v2;
        }
      }
    }
  return (minDist<100000);
}

double FieldDimensions::LinesTable::getDistance(const Pose2D& pose) const
{
  double minDist = 100000;
  for(int i = 0; i < numberOfEntries; ++i)
  {
    Vector2<double> v1 = (lines[i].corner - pose).translation,
                    v2 = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length,0)) 
                          - pose).translation;
    if(v1.y < 0 && v2.y > 0)
    {
      double dist = v1.x + (v2.x - v1.x) * -v1.y / (v2.y - v1.y);
      if(dist >= 0 && dist < minDist)
        minDist = dist;
    }
  }
  return minDist == 100000 ? -1 : minDist;
}

void FieldDimensions::LinesTable::draw(const ColorRGBA& color, bool drawNormals) const
{
  for(int i = 0; i < numberOfEntries; ++i)
  {
    Vector2<double> s = lines[i].corner.translation;
    Vector2<double> p = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length,0))).translation;
    LINE("field lines",
       (int) s.x,
       (int) s.y,
       (int) p.x,
       (int) p.y,
       1, Drawings::ps_solid, color);
    if(drawNormals)
    {
      p = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length/2,0))).translation;
      Vector2<double> p2 = (lines[i].corner + Pose2D(Vector2<double>(lines[i].length/2,100))).translation;
      LINE("field lines",
         (int) p2.x,
         (int) p2.y,
         (int) p.x,
         (int) p.y,
         1, Drawings::ps_solid, color);    
    }
  }
}

const char* FieldDimensions::getCornerClassName(FieldDimensions::CornerClass cornerClass)
{
  switch(cornerClass)
  {
    case lCorner:
      return "l";
    case tCorner:
      return "t";
    case xCorner:
      return "x";
    case iCorner:
      return "i";
    case virtualCorner:
      return "virtual";
    case outerLCorner:
      return "outerL";
    case outerTCorner:
      return "outerT";
    case outerVirtualCorner:
      return "outerVirtual";
    case falseCorner:
      return "false";
    default:
      return "";
  }
}

void FieldDimensions::serialize(In *in, Out *out)
{
  STREAM_REGISTER_BEGIN();
  STREAM(xPosOpponentFieldBorder);
  STREAM(xPosOpponentGoal);
  STREAM(xPosFrontFlags);
  STREAM(xPosOpponentGoalpost);
  STREAM(xPosOpponentGroundline);
  STREAM(xPosOpponentSideCorner);
  STREAM(xPosOpponentPenaltyArea);
  STREAM(xPosHalfWayLine);
  STREAM(xPosOwnPenaltyArea);
  STREAM(xPosOwnSideCorner);
  STREAM(xPosOwnGroundline);
  STREAM(xPosOwnGoalpost);
  STREAM(xPosBackFlags);
  STREAM(xPosOwnGoal);
  STREAM(xPosOwnFieldBorder);
  STREAM(yPosLeftFieldBorder);
  STREAM(yPosLeftFlags);
  STREAM(yPosLeftSideline);
  STREAM(yPosLeftGroundline);
  STREAM(yPosLeftPenaltyArea);
  STREAM(yPosLeftGoal);
  STREAM(yPosCenterGoal);
  STREAM(yPosRightGoal);
  STREAM(yPosRightPenaltyArea);
  STREAM(yPosRightGroundline);
  STREAM(yPosRightSideline);
  STREAM(yPosRightFlags);
  STREAM(yPosRightFieldBorder);
  STREAM(centerCircleRadius);
  STREAM(flagRadius);
  STREAM(flagHeight);
  STREAM(flagColorHeight);
  STREAM(goalHeight);
  STREAM(goalWidth);
  STREAM(ballRadius);
  STREAM(fieldLinesWidth);
  STREAM(xPosThrowInPointOpponentHalf);
  STREAM(xPosThrowInPointCenter);
  STREAM(xPosThrowInPointOwnHalf);
  STREAM(yPosThrowInPointLeft);
  STREAM(yPosThrowInPointRight);
  STREAM(numberOfRobotsPerTeam);
  
  STREAM_ARRAY(table);
  STREAM(carpetBorder);
  STREAM(fieldBorder);

  STREAM_ARRAY(cornerTable);

  STREAM_ARRAY(flagTypeTable);

  STREAM_REGISTER_FINISH();
}


