/**
* @file FieldDimensions.h
*
* Description of the dimensions of the field.
*
* @author Matthias Jngel
* @author Thomas Rfer
*/

#ifndef __FieldDimensions_h_ 
#define __FieldDimensions_h_ 

#include "Tools/Math/Geometry.h"
#include "Platform/GTAssert.h"
#include "Representations/Infrastructure/RoboCupGameControlData.h"
#include "Tools/Boundary.h"
#include "Tools/Debugging/DebugDrawings.h"

/**
* Class containing definitions and functions
* regarding field dimensions.
*
* @author Max Risler
*/
class FieldDimensions : public Boundary<double>, public Streamable
{
public:
  int xPosOpponentFieldBorder;
  int xPosOpponentGoal;
  int xPosFrontFlags;
  int xPosOpponentGoalpost;
  int xPosOpponentGroundline;
  int xPosOpponentSideCorner;
  int xPosOpponentPenaltyArea;
  int xPosHalfWayLine;

  int xPosOwnPenaltyArea;
  int xPosOwnSideCorner;
  int xPosOwnGroundline;
  int xPosOwnGoalpost;
  int xPosBackFlags;
  int xPosOwnGoal;
  int xPosOwnFieldBorder;

  int yPosLeftFieldBorder;
  int yPosLeftFlags;
  int yPosLeftSideline;
  int yPosLeftGroundline;
  int yPosLeftPenaltyArea;
  int yPosLeftGoal;
  int yPosCenterGoal;
  int yPosRightGoal;
  int yPosRightPenaltyArea;
  int yPosRightGroundline;
  int yPosRightSideline;
  int yPosRightFlags;
  int yPosRightFieldBorder;

  //other dimensions
  int centerCircleRadius;
  int flagRadius;
  /** Height of flag from ground to bottom of lower colored part */
  int flagHeight;
  /** Height of one of the two colored parts of the flag */
  int flagColorHeight;
  int goalHeight;
  int goalWidth;
  int ballRadius;
  int fieldLinesWidth;

  //throw-in points
  int xPosThrowInPointOpponentHalf;
  int xPosThrowInPointCenter;
  int xPosThrowInPointOwnHalf;
  int yPosThrowInPointLeft;
  int yPosThrowInPointRight;

  // number of robots in one team
  int numberOfRobotsPerTeam;

  // flag, if field has a "border" or not
  int fieldHasBorder;

  /**
  * This is a collection of line- or boundary segments with start-Pose2D and length.
  */
  class LinesTable : public Streamable
  {
  private:
    enum { 
      maxNumberOfEntries = 106 /**< The maximum number of lines in the table. */
    };

  public:
    class Line : public Streamable
    {
    public:
      Pose2D corner; /**< The field corners. */
      double length; /**< The lengths of the border segments starting at a corresponding corner. */
      virtual void serialize(In *in, Out *out)
      {
        STREAM_REGISTER_BEGIN();
        STREAM(corner);
        STREAM(length);
        STREAM_REGISTER_FINISH();
      }
    };
    Line lines[maxNumberOfEntries];

    int numberOfEntries; /**< The number of corners in the table. */

    LinesTable() :
      numberOfEntries(0)
    {}

    ~LinesTable() {}

    void push(const Pose2D& p, double l);

    void push(const Vector2<double>& s, const Vector2<double>& e);

    void pushCircle(const Vector2<double>& center, double radius, int numOfSegments);

    void pushQuarterCircle(const Vector2<double>& center, double radius, int numOfSegments, int angle);

    /**
    * Doubles the line segments creating double sided segments that have edges on both sides.
    * @param width The distance between both edges of each line segment.
    * @param single A lines table with single-sided lines.
    */
    void doubleSided(double width, const LinesTable& single);

    /*
    * Returns whether a given point is inside the polygon described by the line segments.
    * Only valid if the line segment table describes a closed polygon. 
    */

    bool isInside(const Vector2<double>& v) const;

    /**
    * The function clips a point to the polygon described by the line segments.
    * Only valid if the line segment table describes a closed polygon. 
    * @param v The point.
    * @return How far was the point moved?
    */
    double clip(Vector2<double>& v) const;
    
    /**
    * The function returns the point on a line of a certain type closest to given a point.
    * @param point The point on a line.
    * @param p The reference point and the rotation of the line.
    * @param numberOfRotations The number of discretizations of line rotations.
    * @param minLength The minimum length of the line segments that are considered.
    * @return whether there is a matching point in that direction
    */
    bool getClosestPoint(Vector2<double>& point, const Pose2D& p, int numberOfRotations, double minLength) const;

    /**
    * The function returns the distance between a point and the closest point on a line of a certain type in a certain direction.
    * @param pose The reference point and direction.
    * @return The distance. It is -1 if no line of that type exists in the certain direction.
    */
    double getDistance(const Pose2D& pose) const;

    /**
    * Draws a debug field drawing that displays the set of lines.
    */
    void draw(const ColorRGBA& color, bool drawNormals = true) const;

    virtual void serialize(In *in, Out *out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM_DYN_ARRAY(lines, numberOfEntries);
      STREAM_REGISTER_FINISH();
    }
  };

  enum TableType
  {
    ownBeacon,        //*< lower own beacon borders
    opponentBeacon,   //*< lower opponent beacon borders
    ownGoal,          //*< lower own goal borders
    opponentGoal,     //*< lower opponent goal borders
    fieldLines,       //*< the field lines, two line segments at the borders of each field line
    numberOfTableTypes
  };

  /**
  * Tables of line segments
  */
  LinesTable table[numberOfTableTypes];

  /** 
  * Describes a polygon around the border of the field carpet.
  * All legal robot positions are inside this polygon.
  */
  LinesTable carpetBorder;

  /** 
  * Describes a polygon around the border of the playing field.
  * All legal ball positions are inside this polygon.
  */
  LinesTable fieldBorder;

  /**
  * A table containing field line corners
  */
  class CornerTable : public Streamable
  {
  public:
    class DirectedCornerTable : public Streamable
    {
    public:
      enum{
        maxNumOfCorners = 7
      };

      Vector2<double> corners[maxNumOfCorners];
      int numOfCorners;

      DirectedCornerTable() :
        numOfCorners(0)
      {}

      void addCorner(const Vector2<double> &corner)
      {
        ASSERT(numOfCorners < maxNumOfCorners);
        corners[numOfCorners++] = corner;
      }
      virtual void serialize(In *in, Out *out)
      {
        STREAM_REGISTER_BEGIN();
        STREAM_DYN_ARRAY(corners, numOfCorners);
        STREAM_REGISTER_FINISH();
      }
    };

    DirectedCornerTable table[4];

    void addCorner(int orientation, const Vector2<double> &corner)
    {
      table[orientation].addCorner(corner);
    }

    virtual void serialize(In *in, Out *out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM_ARRAY(table);
      STREAM_REGISTER_FINISH();
    }
  }; 
    
  enum CornerClass{
    virtualCorner,
    lCorner,
    tCorner,
    xCorner,
    outerLCorner,
    outerTCorner,
    outerVirtualCorner,
    iCorner,
    falseCorner,
      numOfCornerClasses
  };
  
  static const char* getCornerClassName(CornerClass cornerClass);
    
  CornerTable cornerTable[numOfCornerClasses];

  /** A description of landmarks on the field 
  * TODO: Currently there are two different descriptions of landmark positions. Merge this with FlagType */
  enum LandmarkIDs
  {
    // 8 Ls
    opponentLeftL,
    opponentRightL,
    ownRightL,
    ownLeftL,
    opponentLeftPenaltyL,
    opponentRightPenaltyL,
    ownRightPenaltyL,
    ownLeftPenaltyL,
    // 12 Ls inside Ts
    ownLeftPenaltyTInnerL,
    ownLeftPenaltyTOuterL,
    ownRightPenaltyTInnerL,
    ownRightPenaltyTOuterL,
    
    centerLeftTOwnL,
    centerLeftTOpponentL,
    centerRightTOwnL,
    centerRightTOpponentL,
    
    opponentLeftPenaltyTInnerL,
    opponentLeftPenaltyTOuterL,
    opponentRightPenaltyTInnerL,
    opponentRightPenaltyTOuterL,
    // 6 Ts
    ownLeftPenaltyT,
    ownRightPenaltyT,
    centerLeftT,
    centerRightT,
    opponentLeftPenaltyT,
    opponentRightPenaltyT,
    // 4 goal posts
    opponentLeftGoalPost,
    opponentRightGoalPost,
    ownRightGoalPost,
    ownLeftGoalPost,
    // 2 beacons
    centerLeftFlag,
    centerRightFlag,
    // 4 old beacons
    opponentLeftFlag,
    opponentRightFlag,
    ownRightFlag,
    ownLeftFlag,
    // ...
    undefinedLandmark,
    numberOfLandmarkIDs
  };

  static const char* getLandmarkName(LandmarkIDs landmark)
  {
    switch (landmark)
    {
      // 8 Ls
      case opponentLeftL: return "opponentLeftL";
      case opponentRightL: return "opponentRightL";
      case ownRightL: return "ownRightL";
      case ownLeftL: return "ownLeftL";
      case opponentLeftPenaltyL: return "opponentLeftPenaltyL";
      case opponentRightPenaltyL: return "opponentRightPenaltyL";
      case ownRightPenaltyL: return "ownRightPenaltyL";
      case ownLeftPenaltyL: return "ownLeftPenaltyL";
      // 6 Ts
      case ownLeftPenaltyT: return "ownLeftPenaltyT";
      case ownRightPenaltyT: return "ownRightPenaltyT";
      case centerLeftT: return "centerLeftT";
      case centerRightT: return "centerRightT";
      case opponentLeftPenaltyT: return "opponentLeftPenaltyT";
      case opponentRightPenaltyT: return "opponentRightPenaltyT";
      // 12 Ls inside Ts
      case ownLeftPenaltyTInnerL: return "ownLeftPenaltyTInnerL";
      case ownLeftPenaltyTOuterL: return "ownLeftPenaltyTOuterL";
      case ownRightPenaltyTInnerL: return "ownRightPenaltyTInnerL";
      case ownRightPenaltyTOuterL: return "ownRightPenaltyTOuterL";
      
      case centerLeftTOwnL: return "centerLeftTOwnL";
      case centerLeftTOpponentL: return "centerLeftTOpponentL";
      case centerRightTOwnL: return "centerRightTOwnL";
      case centerRightTOpponentL: return "centerRightTOpponentL";
      
      case opponentLeftPenaltyTInnerL: return "opponentLeftPenaltyTInnerL";
      case opponentLeftPenaltyTOuterL: return "opponentLeftPenaltyTOuterL";
      case opponentRightPenaltyTInnerL: return "opponentRightPenaltyTInnerL";
      case opponentRightPenaltyTOuterL: return "opponentRightPenaltyTOuterL";
      // 4 goal posts
      case opponentLeftGoalPost: return "opponentLeftGoalPost";
      case opponentRightGoalPost: return "opponentRightGoalPost";
      case ownRightGoalPost: return "ownRightGoalPost";
      case ownLeftGoalPost: return "ownLeftGoalPost";
      // 2 beacons
      case centerLeftFlag: return "centerLeftFlag";
      case centerRightFlag: return "centerRightFlag";
      // 4 old beacons
      case opponentLeftFlag: return "opponentLeftFlag";
      case opponentRightFlag: return "opponentRightFlag";
      case ownRightFlag: return "ownRightFlag";
      case ownLeftFlag: return "ownLeftFlag";
      // ...
      case undefinedLandmark: return "undefinedLandmark";
      default: return "";
    }
  }

  /** The positions of flags of given type on the field */
  Vector2<double> landmarkPosition[numberOfLandmarkIDs];

  enum FlagType
  {
    blueYellowBlue,
    yellowBlueYellow,
    numberOfFlagTypes
  }; /**< The type is used to identify the different flags on the field. */

  static const char* getFlagTypeName(FlagType flagType)
  {
    switch (flagType)
    {
      case blueYellowBlue: return "blueYellowBlue";
      case yellowBlueYellow: return "yellowBlueYellow";
      default: return "";
    }
  }

  enum
  {
    maxNumberOfFlagsPerType = 2
  };

  /** Table of flag positions for one type */
  class FlagTypeTable : public Streamable
  {
  public:
    /** The number of times a given flag appears on the field */
    int numberOfFlags;

    /** The positions of flags of given type on the field */
    Vector2<double> flagPosition[maxNumberOfFlagsPerType];
    
    virtual void serialize(In *in, Out *out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM_DYN_ARRAY(flagPosition, numberOfFlags);
      STREAM_REGISTER_FINISH();
    }
  };

private:
  /** Flag positions */
  FlagTypeTable flagTypeTable[numberOfFlagTypes];

public:
  /**
   * Default constructor.
   */
  FieldDimensions();

  /**
  * Read field dimensions from configuration file.
  */
  void load();

  /**
  * The function returns the number of times a given flag appears on the field.
  * NOTE: this is only useful for ambiguous flags, eg demos on MSL field, for standard flags return value is always 1
  * @param flagType the type of the flag
  */
  int getNumberOfFlags(FlagType flagType) const {return flagTypeTable[flagType].numberOfFlags;}

  /**
  * The function returns a position of a flag of a given type on the field.
  * @param flagType the type of the flag
  * @param ownColor The color of the own team.
  * @param index the index of the flag, for multiple flag positions of the same type
  * NOTE: index is only useful for ambiguous flags, eg demos on MSL field
  */
  Vector2<double> getFlagPosition(FlagType flagType, RoboCup::uint32 ownColor, int index = 0) const 
  {
    int flip = ownColor == TEAM_BLUE ? -1 : 1;
    return flagTypeTable[flagType].flagPosition[index] * flip;
  }

  /**
  * Returns true when p is inside the carpet.
  */
  bool isInsideCarpet(const Vector2<double> &p) const
  {
    return carpetBorder.isInside(p);
  }

  /**
  * The function clips a point to the carpet.
  * @param v The point.
  * @return How far was the point moved?
  */
  double clipToCarpet(Vector2<double>& v) const
  {
    return carpetBorder.clip(v);
  }

  /**
  * Returns true when p is inside the playing field.
  */
  bool isInsideField(const Vector2<double> &p) const
  {
    return fieldBorder.isInside(p);
  }

  /**
  * The function clips a point to the field.
  * @param v The point.
  * @return How far was the point moved?
  */
  double clipToField(Vector2<double>& v) const
  {
    return fieldBorder.clip(v);
  }

  /**
  * Returns true when p is inside one of the goals.
  */
  bool isInsideGoal(const Vector2<int> &p) const;

  /**
  * Returns true when p is inside the own goal.
  */
  bool isInsideOwnGoal(const Vector2<int> &p) const;

  /**
  * Returns true when p is inside the opponent goal.
  */
  bool isInsideOpponentGoal(const Vector2<int> &p) const;

  /**
  * Returns true when p is on the groundline of the own goal.
  */
  bool isOnOwnGoalGroundline(const Vector2<int> &p) const;

  /**
  * Returns true when p is on the groundline of the opponent goal.
  */
  bool isOnOpponentGoalGroundline(const Vector2<int> &p) const;

  /**
  * Returns the distance of p to opponents penalty area.
  */
  double distanceToOpponentPenaltyArea(const Vector2<double> &p) const;

  /**
  * Returns the distance of p to own penalty area.
  */
  double distanceToOwnPenaltyArea(const Vector2<double> &p) const;
  double distanceToOwnPenaltyArea(const Pose2D &p) const;

  /**
  * The function returns a random pose inside the field.
  * @return The random pose.
  */
  Pose2D randomPoseOnField() const;

  /**
  * The function returns a random pose on the carpet.
  * @return The random pose.
  */
  Pose2D randomPoseOnCarpet() const;

  /**
  * The function returns a random pose near own goal (wirh maximal distance <= 1200mm)
  * @return The random pose near own goal.
  */
  Pose2D randomPoseNearOwnGoal() const;

  /**
  * The method declares the drawings provided by this class.
  */
  void declareDrawings() const;

  /**
  * The method draws the field crosses 
  */
  void drawCrosses() const;
  
  /**
  * The method draws the field lines.
  */
  void drawLines() const;

  /**
  * The method draws the axes of the field coordinate system.
  */
  void drawAxes() const;

  /**
  * The method draws the axes of the time line diagram.
  */
  void drawAxesTimeLine() const;

  /**
  * The method draws the field polygons.
  * @param ownColor The color of the own team.
  */
  void drawPolygons(RoboCup::uint32 ownColor) const;

  virtual void serialize(In *in, Out *out);

private:
  void readDimensions(int* const* valuesInt, const char* const* namesInt, bool* initializedInt, int numOfValuesInt);
  void readLines(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt);
  void readCorners(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt);
  void readLandmarks(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt);
  void readFlags(const int* const* valuesInt, const char* const* namesInt, const bool* initializedInt, int numOfValuesInt);

  int readValue(const std::string& buf, const int* const* values, const char* const* names, const bool* initialized, int numOfValues) const;
};

#endif //FieldDimensions
