/**
* @file WalkingParameters.h
* Declaration of a class for representing the walking parameters of a Robot.
* @author <a href="mailto:sancho@informatik.uni-bremen.de">Dario Sancho-Pradel</a>
*/

#ifndef __WalkingParameters_H__
#define __WalkingParameters_H__

#include "Tools/Streams/StreamHandler.h"
#include "Tools/Math/Trajectories.h"
#include "Tools/Math/Vector3.h"
#include "Tools/Math/Pose2D.h"

struct GaitOptParameters
{
  unsigned int stepDuration;
  double bodyTilt;
  double stepPhase[3];
  unsigned int stepHeight;
  double stepTiltForward, stepTiltBackward, stepTiltSideways;
  double stepOrigin[3];
  double bodyShiftPhases[3];
  double bodyShiftPhaseShift;
  double armOrigin;
  double armScale;
};

class WalkingParameters : public Streamable
{
  public:
    unsigned stepDuration; /**< Duration of a step in ms. */
    Pose2D maxSpeedChange; /**< Max speed change in mm/s and radians/s. */
    Vector2<double> odometryScale; /**< Scales the current walk request so that it matches the expected motion for translational motion. */
    double odometryRotationScaleCCW, /**< Scales the current walk request so that it matches the expected motion for turning counter clockwise (positive). */
           odometryRotationScaleCW, /**< Scales the current walk request so that it matches the expected motion for turning clockwise (negative). */
           maxForwardSpeed, /**< The maximum forward speed in mm/s. */
           maxBackwardSpeed, /**< The maximum backward speed in mm/s. */
           maxSidewaysSpeed, /**< The maximum sideways speed in mm/s. */
           maxRotationSpeed, /**< The maximum rotation speed in radians/s. */
           elbowAngle, /**< The origin of the arm's angular tilt motion. */
           armDeactivationRatio, /**< A ratio for deactivating the arms. */
           bodyTiltOrigin, /**< The origin of the body's tilt motion. */
           bodyTiltStandOrigin, /**< The origin of the body's tilt while standing. */
           bodyTiltSpeedFactor, /**<A factor the speed is multiplied with and added to the bodyTilt */
           bodyTiltCorrection, /**< A correction ratio for forward/backward direction. */
           bodyShiftCorrection, /**< A correction ratio for sideways direction. */
           bodyShiftToAccel, /**< Ratio between body shift and accelerator value. */
           footPanCorrection, /**< Correction of the foot pan (actually the whole legs). */
           footRollCorrection, /**< Correction of the foot roll. */
           bodyShiftSidewaysOriginFactorLeft,
           bodyShiftSidewaysOriginFactorRight,
           stepHeight, /**< The height of a step.**/
           bodyShiftScale; /**< The ratio between of the body's sideward motion and the stepSize. */ 

  
    Trajectory3D::Shape stepShape; /**< The shape of the feet's trajectory. */
    Vector3<double> stepOrigin; /**< Origin of the position of the left foot. */
    double stepTiltForward, /**< The tilt of the forward step. */
           stepTiltBackward, /**< The tilt of the backward step. */
           stepTiltSideways, /**< The tilt of the sideways step in step direction. */
           stepPhases[3]; /**< Phase ratios for ground time, up time, and air time for each step. */

    Trajectory1D::Shape bodyShiftShape; /**< The shape of the body's sideward motion. */
    double bodyShiftOrigin; /**< The origin of the body's sideward motion. */
    Trajectory1D::ScaleMode bodyShiftScaleMode; /**< Scale the resulting values based on stepSize. */
    double bodyShiftPhaseShift, /**< The phase shift. */
           bodyShiftPhases[3]; /**< Phase ratios for forward time, stop front time, and backward time. */

    Trajectory1D::Shape rotationShape; /**< The shape of the rotation motion. */
    double rotationOrigin; /**< The origin of the rotation motion. */
    Trajectory1D::ScaleMode rotationScaleMode; /**< Scale the resulting values based on rotation speed. */
    double rotationScale, /**< The ratio between of the rotation and the desired rotation speed. */
           rotationPhaseShift, /**< The phase shift. */
           rotationPhases[3]; /**< Phase ratios for forward time, stop front time, and backward time. */

    Trajectory1D::Shape armShape; /**< The shape of the arm motion. */
    double armOrigin; /**< The origin of the arm motion. */
    Trajectory1D::ScaleMode armScaleMode; /**< Scale the resulting values based on forward speed. */
    double armScale, /**< The ratio between of the forward speed and the desired arm speed. */
           armPhaseShift, /**< The phase shift. */
           armPhases[3]; /**< Phase ratios for forward time, stop front time, and backward time. */

    /**
    * The method serializes this object.
    * @param in Points to an In-stream, when currently reading.
    * @param out Points to an Out-stream, when currently writing.
    */
    virtual void serialize(In* in, Out* out);

    /**
    * The method initializes this object after some private attributes were changed.
    */
    virtual void init();

    /**
    * The method frees the dynamic space allocated by this object.
    */
    virtual void free();

  
    // these must be the last attributes!
    Trajectory3D* footPositions[2]; /**< The trajectories of the left and the right feet. */
    Trajectory1D* bodyShift;
    Trajectory1D* rotation[2];
    Trajectory1D* arm[2];

    /**
    * Constructor.
    * For parameter describtions, see above.
    */
    WalkingParameters()
    {
      footPositions[0] = 0;

      //this is need because otherwise the trajectorys can't be initialized
      this->armScaleMode = Trajectory1D::linear;
      this->armShape = Trajectory1D::constant;
      this->bodyShiftShape = Trajectory1D::constant;
      this->bodyShiftScaleMode = Trajectory1D::linear;
      this->rotationScaleMode = Trajectory1D::linear;
      this->rotationShape = Trajectory1D::constant;
      this->stepShape = Trajectory3D::ellipse;
      
    }

    /**
    * Copy constructor.
    * @param other The other object this one is copied from.
    */
    WalkingParameters(const WalkingParameters& other)
    {
      footPositions[0] = 0;
      *this = other;
    }

    /**
    * Destructor.
    */
    virtual ~WalkingParameters()
    {
      free();
    }

    /**
    * Assignment operator.
    * @param other The other object that is assigned to this one.
    * @return This object.
    */
    WalkingParameters& operator=(const WalkingParameters& other);

    bool operator==(const WalkingParameters& other) const;

    /**
    * Set the optimized parameters
    * @param params set of parameters
    */
    void setOptParameters(const struct GaitOptParameters& params);
  };

#endif // __WalkingParameters_H__
