/**
* @file CameraInfo.h
* 
* Declaration of class CameraInfo
*
* @author <a href="mailto:juengel@informatik.hu-berlin.de">Matthias Juengel</a>
* @author <a href="mailto:walter.nistico@uni-dortmund.de">Walter Nistico</a>
*/ 

#ifndef __CameraInfo_h_
#define __CameraInfo_h_

#include "Tools/Math/Vector2.h"
#include "Tools/Settings.h"


//!@name camera resolution (currently used mode rather then all modes)
//!@{
const int cameraResolutionWidth_LOOX720  = 320;
const int cameraResolutionHeight_LOOX720 = 240;
const int cameraResolutionWidth_UVC  = 320;
const int cameraResolutionHeight_UVC = 240;
const int cameraResolutionWidth_Nao  = 320;
const int cameraResolutionHeight_Nao = 240;
const int cameraResolutionWidth = 320;
const int cameraResolutionHeight = 240;
//!@}

//!@name opening angles of the camera
//!@{
const double openingAngleWidth_LOOX720 = 0.78690; // 45.1
const double openingAngleHeight_LOOX720  = 0.60714; // 34.8

const double openingAngleWidth_UVC = 0.70592; // 40.45
const double openingAngleHeight_UVC  = 0.54991; // 31.51

const double openingAngleWidth_Nao = 0.834267; // 47.8 (unchecked)
const double openingAngleHeight_Nao  = 0.642281; // 36.8 (unchecked)
//!@}

//!@name camera intrinsic parameters
//!@{
const double focalLength_LOOX720 = 384.25;
const double opticalCenterX_LOOX720 = 165.2; 
const double opticalCenterY_LOOX720 = 118.5; 

const double focalLength_UVC = 434.33;
const double opticalCenterX_UVC = 150.1;
const double opticalCenterY_UVC = 126.0;

const double focalLength_Nao = 361.06; // unchecked
const double opticalCenterX_Nao = 160.0; // unchecked
const double opticalCenterY_Nao = 120.0; // unchecked
//!@}

//FIXME
#ifdef LINUX

const int cameraResolutionWidth_Bioloid  = cameraResolutionWidth_UVC;
const int cameraResolutionHeight_Bioloid = cameraResolutionHeight_UVC;

const double openingAngleWidth_Bioloid = openingAngleWidth_UVC;
const double openingAngleHeight_Bioloid  = openingAngleHeight_UVC;

const double focalLength_Bioloid = focalLength_UVC;
const double opticalCenterX_Bioloid = opticalCenterX_UVC;
const double opticalCenterY_Bioloid = opticalCenterY_UVC;

#else

const int cameraResolutionWidth_Bioloid  = cameraResolutionWidth_LOOX720;
const int cameraResolutionHeight_Bioloid = cameraResolutionHeight_LOOX720;

const double openingAngleWidth_Bioloid = openingAngleWidth_LOOX720;
const double openingAngleHeight_Bioloid  = openingAngleHeight_LOOX720;

const double focalLength_Bioloid = focalLength_LOOX720;
const double opticalCenterX_Bioloid = opticalCenterX_LOOX720;
const double opticalCenterY_Bioloid = opticalCenterY_LOOX720;

#endif



/**
* Information about the camera which provides the images for the robot
*/
class CameraInfo : public Streamable
{
public:
  /**
  * Default constructor.
  */

  CameraInfo() :
    resolutionWidth(Global::getSettings().model == Settings::bioloid ? cameraResolutionWidth_Bioloid : cameraResolutionWidth_Nao),
    resolutionHeight(Global::getSettings().model == Settings::bioloid ? cameraResolutionHeight_Bioloid : cameraResolutionHeight_Nao),
    openingAngleWidth(Global::getSettings().model == Settings::bioloid ? openingAngleWidth_Bioloid : openingAngleWidth_Nao),
    openingAngleHeight(Global::getSettings().model == Settings::bioloid ? openingAngleHeight_Bioloid : openingAngleHeight_Nao),
    focalLength(Global::getSettings().model == Settings::bioloid ? focalLength_Bioloid : focalLength_Nao),
    opticalCenter(Global::getSettings().model == Settings::bioloid 
                  ? Vector2<double>(opticalCenterX_Bioloid, opticalCenterY_Bioloid)
                  : Vector2<double>(opticalCenterX_Nao, opticalCenterY_Nao))
  {
    calcAdditionalConstants();
  }

  CameraInfo(Settings::Model model) :
    resolutionWidth(model == Settings::bioloid ? cameraResolutionWidth_Bioloid : cameraResolutionWidth_Nao),
    resolutionHeight(model == Settings::bioloid ? cameraResolutionHeight_Bioloid : cameraResolutionHeight_Nao),
    openingAngleWidth(model == Settings::bioloid ? openingAngleWidth_Bioloid : openingAngleWidth_Nao),
    openingAngleHeight(model == Settings::bioloid ? openingAngleHeight_Bioloid : openingAngleHeight_Nao),
    focalLength(model == Settings::bioloid ? focalLength_Bioloid : focalLength_Nao),
    opticalCenter(model == Settings::bioloid 
                  ? Vector2<double>(opticalCenterX_Bioloid, opticalCenterY_Bioloid)
                  : Vector2<double>(opticalCenterX_Nao, opticalCenterY_Nao))
  {
    calcAdditionalConstants();
  }

  virtual void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(resolutionWidth);
    STREAM(resolutionHeight);
    STREAM(openingAngleWidth);
    STREAM(openingAngleHeight);
    STREAM(focalLength);
    STREAM(opticalCenter);
    STREAM_REGISTER_FINISH();
    if(in)
      calcAdditionalConstants();
  }

  int resolutionWidth;
  int resolutionHeight;
  double openingAngleWidth;
  double openingAngleHeight;

  /** Intrinsic camera parameters: axis skew is modelled as 0 (90 perfectly orthogonal XY) 
  * and the same has been modeled for focal axis aspect ratio; distortion is considering
  * only 2nd and 4th order coefficients of radial model, which account for about 95% of total.
  */
  double focalLength;
  double focalLengthInv; // (1/focalLength) used to speed up certain calculations 
  Vector2<double> opticalCenter;
  double focalLenPow2;  
  double focalLenPow4;

private:

  /**
  * Calculates some additional constants based on focal length for faster calculations.
  */
  void calcAdditionalConstants()
  {
    focalLenPow2 = focalLength * focalLength;
    focalLenPow4 = focalLenPow2 * focalLenPow2;
    focalLengthInv = 1.0 / focalLength;
  }

};

#endif //__CameraInfo_h_
