/**
* @file Representations/Infrastructure/JointData.h
*
* This file declares a class to represent the joint angles sent to the robot.
*
* @author <A href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</A>
*/

#ifndef __JointData_H__
#define __JointData_H__

#include "Tools/Streams/Streamable.h"
#include "Tools/Math/Common.h"
#include "Platform/GTAssert.h"

/**
* @class JointData
* A class to represent the joint angles sent to the robot.
*/
class JointData : public Streamable
{
protected:
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    STREAM_ARRAY(angles);
    STREAM(timeStamp);
    STREAM_REGISTER_FINISH();
  }

public:
  enum Joint
  {
    headPan,
    headTilt,
    armLeft0,
    armLeft1,
    armLeft2,
    armLeft3,
    armRight0,
    armRight1,
    armRight2,
    armRight3,
    legLeft0,
    legLeft1,
    legLeft2,
    legLeft3,
    legLeft4,
    legLeft5,
    legRight0,
    legRight1,
    legRight2,
    legRight3,
    legRight4,
    legRight5,
    numOfJoints
  };


  // If you change those values be sure to change them in MofCompiler.cpp too. (Line ~280)
  enum 
  {
    off    = 1000, /**< Special angle for switching off a joint. */
    ignore = 2000  /**< Special angle for not overwriting the previous setting. */
  };
  double angles[numOfJoints]; /**< The angles of all joints. */
  unsigned timeStamp; /**< The time when these angles were received. */

  /**
  * Default constructor.
  * Switches off all joints.
  */
  JointData() : timeStamp(0)
  {
    for(int i = 0; i < numOfJoints; ++i)
      angles[i] = off;
  }

  /**
  * The method returns the angle of the mirror (left/right) of the given joint.
  * @param joint The joint the mirror of which is returned.
  * @return The angle of the mirrored joint.
  */
  double mirror(Joint joint) const
  {
    switch (joint)
    {
      // don't mirror an invalid join value (!)
      case headPan: return angles[headPan] == off || angles[headPan] == ignore ? angles[headPan]: -angles[headPan];
      case armLeft0: return angles[armRight0]; 
      case armLeft1: return angles[armRight1]; 
      case armLeft2: return angles[armRight2]; 
      case armLeft3: return angles[armRight3]; 
      case armRight0: return angles[armLeft0]; 
      case armRight1: return angles[armLeft1]; 
      case armRight2: return angles[armLeft2]; 
      case armRight3: return angles[armLeft3]; 
      case legLeft0: return angles[legRight0]; 
      case legLeft1: return angles[legRight1]; 
      case legLeft2: return angles[legRight2]; 
      case legLeft3: return angles[legRight3]; 
      case legLeft4: return angles[legRight4]; 
      case legLeft5: return angles[legRight5]; 
      case legRight0: return angles[legLeft0]; 
      case legRight1: return angles[legLeft1]; 
      case legRight2: return angles[legLeft2]; 
      case legRight3: return angles[legLeft3]; 
      case legRight4: return angles[legLeft4]; 
      case legRight5: return angles[legLeft5];
      default: return angles[joint];
    }
  }

  /**
  * The method initializes the joint angles as a mirror of a set of other joint angles.
  * @param other The set of joint angles that are mirrored.
  */
  void mirror(const JointData& other)
  {
    for(int i = 0; i < numOfJoints; ++i)
      angles[i] = other.mirror((Joint) i);
    timeStamp = other.timeStamp;
  }

  /**
  * The function returns the name of a joint.
  * @param joint The joint the name of which will be returned.
  * @return The name.
  */
  static const char* getJointName(Joint joint)
  {
    switch(joint)
    {
    case headPan: return "headPan";
    case headTilt: return "headTilt";
    case armLeft0: return "armLeft0";
    case armLeft1: return "armLeft1";
    case armLeft2: return "armLeft2";
    case armLeft3: return "armLeft3";
    case armRight0: return "armRight0";
    case armRight1: return "armRight1";
    case armRight2: return "armRight2";
    case armRight3: return "armRight3";
    case legLeft0: return "legLeft0";
    case legLeft1: return "legLeft1";
    case legLeft2: return "legLeft2";
    case legLeft3: return "legLeft3";
    case legLeft4: return "legLeft4";
    case legLeft5: return "legLeft5";
    case legRight0: return "legRight0";
    case legRight1: return "legRight1";
    case legRight2: return "legRight2";
    case legRight3: return "legRight3";
    case legRight4: return "legRight4";
    case legRight5: return "legRight5";
    default:
      return "unknown";
    }
  }
};

/**
* @class JointDataDeg
* A class that wraps joint data to be transmitted in degrees.
*/
class JointDataDeg : public JointData
{
private:
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    if(jointData)
    {
      ASSERT(out);
      for(int i = 0; i < JointData::numOfJoints; ++i)
        angles[i] = jointData->angles[i] == JointData::off ? JointData::off
                    : floor(toDegrees(jointData->angles[i]) * 10.0 + 0.5) / 10.0;
      timeStamp = jointData->timeStamp;

      STREAM_ARRAY(angles);
      STREAM(timeStamp);
    }
    else
    {
      STREAM_BASE(JointData);
    }
    STREAM_REGISTER_FINISH();
  }

  JointData* jointData; /**< The joint data that is wrapped. */

public:
  /**
  * Default constructor.
  */
  JointDataDeg() : jointData(0) {}

  /**
  * Constructor.
  * @param jointData The joint data that is wrapped.
  */
  JointDataDeg(JointData& jointData) : jointData(&jointData) {}

  /**
  * Assignment operator.
  */
  JointDataDeg& operator=(const JointDataDeg& other)
  {
    if(jointData)
      for(int i = 0; i < JointData::numOfJoints; ++i)
        jointData->angles[i] = other.angles[i] == JointData::off ? JointData::off
                              : fromDegrees(other.angles[i]);
    else
      *((JointData*) this) = other;
    return *this;
  }
};

class JointRequest : public JointData {};
class SafeJointData : public JointData {};
class StandOutput : public JointRequest {};

#endif // __JointData_H__
