/**
* @file RobotHealth.h
* The file declares two classes to transport information about the current robot health
* @author <a href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</a>
*/ 

#ifndef __RobotHealth_h_
#define __RobotHealth_h_

#include "Tools/Streams/Streamable.h"


/**
* @class MotionRobotHealth
* All information collected within motion process.
*/
class MotionRobotHealth : public Streamable
{
public:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read (if in != 0).
  * @param out The stream to which the object is written (if out != 0).
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
      STREAM(motionFrameRate);
      STREAM(voltage);
      STREAM(maxTemperature);
      STREAM(robotFramesPerSecond);
    STREAM_REGISTER_FINISH();
  }

  /** Constructor */
  MotionRobotHealth(): motionFrameRate(0.0), voltage(0.0), 
    maxTemperature(0.0), robotFramesPerSecond(0.0) {}

  float motionFrameRate;       /*< Frames per second within process "Motion" */
  float voltage;               /*< Current voltage of robot battery */
  float maxTemperature;        /*< Highest temperature of a robot actuator */
  float robotFramesPerSecond;  /*< Average number of received frames from robot per second */
};


/**
* @class RobotHealth
* Full information about the robot.
*/
class RobotHealth : public MotionRobotHealth
{
private:
  /**
  * The method makes the object streamable.
  * @param in The stream from which the object is read (if in != 0).
  * @param out The stream to which the object is written (if out != 0).
  */
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
      STREAM_BASE(MotionRobotHealth);
      STREAM(cognitionFrameRate);
      STREAM(memoryLoad);
      STREAM(robotName);
    STREAM_REGISTER_FINISH();
  }

public:
  /** Constructor */
  RobotHealth():cognitionFrameRate(0.0), memoryLoad(0) {}

  /** Assigning MotionRobotHealth 
  * @param motionRobotHealth Information from the motion process
  */
  void operator=(const MotionRobotHealth& motionRobotHealth)
  {
    motionFrameRate = motionRobotHealth.motionFrameRate;
    voltage = motionRobotHealth.voltage;
    maxTemperature = motionRobotHealth.maxTemperature;
    robotFramesPerSecond = motionRobotHealth.robotFramesPerSecond;
  }

  float cognitionFrameRate;    /*< Frames per second within process "Cognition" */
  unsigned char memoryLoad;    /*< Percentage of used memory 0==empty, 100==full*/
  std::string robotName;       /*< For fancier drawing :-) */
};

#endif //__RobotHealth_h_
