/**
* @file Representations/Infrastructure/SensorData.h
*
* This file declares a class to represent the sensor data received from the robot.
*
* @author <A href="mailto:Harlekin@tzi.de">Philippe Schober</A>
*/

#ifndef __SensorData_H__
#define __SensorData_H__

#include "Tools/Streams/Streamable.h"
#include "Representations/Infrastructure/JointData.h"

/**
* @class SensorData
* A class to represent the sensor datas received from the robot.
*/
class SensorData : public Streamable
{
protected:
  virtual void serialize(In* in, Out* out)
  {  
    STREAM_REGISTER_BEGIN();
    STREAM_ARRAY(data);
    STREAM_ARRAY(loads);
    STREAM_ARRAY(temperatures);
    STREAM(timeStamp);
    STREAM_REGISTER_FINISH();
  }

public:
  enum Sensor
  {
    gyroX,
    gyroY,
    gyroZ,
    accX,
    accY,
    accZ,
    compassX,
    compassY,
    voltage,
    fsrLFL,     //the feetsensors of the Nao-Robot
    fsrLFR,
    fsrLBL,
    fsrLBR,
    fsrRFL,
    fsrRFR,
    fsrRBL,
    fsrRBR,
    us,
    numOfSensors
  };

  //special value to indicate that the sensor does not exist
  enum
  {
    doesNotExist=-9999999
  };

  float data[numOfSensors]; /**< The data of all sensors. */
  short loads[JointData::numOfJoints]; /**< The load of all sensors. */  
  unsigned char temperatures[JointData::numOfJoints]; /**< The temperature of all sensors. */  
  unsigned timeStamp; /**< The time when the sensor data was received. */

  /**
  * Default constructor.
  */
  SensorData() : timeStamp(0)
  {
    for(int i = 0; i < numOfSensors; ++i)
      if( i < fsrLFL || i > fsrRBR)
        data[i] = 0;
      else
        data[i] = doesNotExist;
    for(int i = 0; i < JointData::numOfJoints; ++i)
      loads[i] = temperatures[i] = 0;
  }

  /**
  * The function returns the name of a sensor.
  * @param sensor The sensor the name of which will be returned.
  * @return The name.
  */
  static const char* getSensorName(Sensor sensor)
  {
    switch(sensor)
    {
    case gyroX: return "gyroX";
    case gyroY: return "gyroY";
    case gyroZ: return "gyroZ";
    case accY: return "accY";
    case accZ: return "accZ";
    case accX: return "accX";
    case compassX: return "compassX";
    case compassY: return "compassY";
    case voltage: return "voltage";
    case fsrLFL: return "fsrLFL";
    case fsrLFR: return "fsrLFR";
    case fsrLBL: return "fsrLBL";
    case fsrLBR: return "fsrLBR";
    case fsrRFL: return "fsrRFL";
    case fsrRFR: return "fsrRFR";
    case fsrRBL: return "fsrRBL";
    case fsrRBR: return "fsrRBR";
    case us: return "us";
    default: return "unknown";
    }
  }
};

class SafeSensorData : public SensorData {};

#endif // __SensorData_H__
