/**
* @file BallModel.h
*
* Declaration of class BallModel
*
* @author <A href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</A>
*/

#ifndef __BallModel_h_
#define __BallModel_h_

#include "Tools/Streams/Streamable.h"
#include "Tools/Math/Geometry.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Representations/Modeling/RobotPose.h"


/**
 * @class BallState
 *
 * Base class for a model of a ball
 */
class BallState : public Streamable
{
protected: 
  /** The position of the ball relative to the robot (in mm)*/
  Vector2<double> position;
  
  /** The velocity of the ball relative to the robot (in mm/s)*/
  Vector2<double> velocity;

public:

  Vector2<double> getPositionInFieldCoordinates(const RobotPose& rp) const
  {
    return Geometry::relative2FieldCoord(rp, position.x, position.y);
  }

  Vector2<double> getVelocityInFieldCoordinates(const RobotPose& rp) const
  {
    double c(rp.getCos());
    double s(rp.getSin());
    return Vector2<double>(velocity.x*c - velocity.y*s, velocity.x*s + velocity.y*c);
  }

  Vector2<double> getPositionInRelativeCoordinates() const
  {
    return position;
  }

  Vector2<double> getVelocityInRelativeCoordinates() const
  {
    return velocity;
  }

  double getAngle() const
  {
    return atan2(position.y, position.x);
  }

  double getDistance() const
  {
    return position.abs();
  }

  void setPositionAndVelocityInRelativeCoordinates(const Vector2<double>& position,
                                                   const Vector2<double>& velocity = Vector2<double>(0.0,0.0))
  {
    this->position = position;
    this->velocity = velocity;
  }

  void setPositionAndVelocityInFieldCoordinates(const Vector2<double>& positionOnField,
                                                const Vector2<double>& velocityOnField,
                                                const RobotPose& rp)
  {
    position = Geometry::fieldCoord2Relative(rp, positionOnField);
    double c(rp.getCos());
    double s(rp.getSin());
    velocity = Vector2<double>(velocityOnField.x*c + velocityOnField.y*s,
                              -velocityOnField.x*s + velocityOnField.y*c);
  }

  /** Streaming (with specifications) */
  virtual void serialize(In *in, Out *out)
  {
    STREAM_REGISTER_BEGIN();
      STREAM(position);
      STREAM(velocity);
    STREAM_REGISTER_FINISH();
  }
};


/**
 * @class BallModel
 *
 * Contains all current knowledge about the ball.
 */
class BallModel : public Streamable
{
public:
  /** Constructor */
  BallModel():timeWhenLastSeen(0)
  {}

  /** The last seen position of the ball */
  BallState lastPerception;
  
  /** The state of the ball estimated from own observations;
      it is propagated even if the ball is not seen */
  BallState estimate;

  /** The last seen estimate */
  BallState lastSeenEstimate;

  /** Time stamp, indicating what its name says*/
  unsigned timeWhenLastSeen;

  /** Streaming (with specifications) */
  virtual void serialize(In *in, Out *out)
  {
    STREAM_REGISTER_BEGIN();
      STREAM(lastPerception);
      STREAM(estimate);
      STREAM(lastSeenEstimate);
      STREAM(timeWhenLastSeen);
    STREAM_REGISTER_FINISH();
  }

  /** Draws something*/
  void draw() 
  {
    DECLARE_DEBUG_DRAWING("representation:BallModel", "drawingOnField", "drawing of the ball model");
    Vector2<double> estBallPos = estimate.getPositionInRelativeCoordinates();
    Vector2<double> v = estimate.getVelocityInRelativeCoordinates();
    CIRCLE("representation:BallModel", 
      estBallPos.x, 
      estBallPos.y, 
      45, 
      1, // pen width 
      Drawings::ps_solid, 
      ColorClasses::orange,
      Drawings::bs_solid, 
      ColorClasses::orange);
    ARROW("representation:BallModel", estBallPos.x, estBallPos.y, 
      estBallPos.x+(v.x), estBallPos.y+(v.y),5,1,ColorClasses::orange);

    Vector2<double> seenBallPos = lastSeenEstimate.getPositionInRelativeCoordinates();
    v = lastSeenEstimate.getVelocityInRelativeCoordinates();
    CIRCLE("representation:BallModel", 
      seenBallPos.x, 
      seenBallPos.y, 
      45, 
      1, // pen width 
      Drawings::ps_solid, 
      ColorRGBA(255, 128, 128, 220),
      Drawings::bs_solid, 
      ColorRGBA(255, 128, 128, 220));
    ARROW("representation:BallModel", seenBallPos.x, seenBallPos.y, 
      seenBallPos.x+(v.x), seenBallPos.y+(v.y),5,1,ColorRGBA(255, 128, 0, 220));
  }
};

class GroundTruthBallModel : public BallModel 
{
public:
    /** Draws something*/
  void draw() 
  {
    DECLARE_DEBUG_DRAWING("representation:GroundTruthBallModel", "drawingOnField", "drawing of the ground truth ball model");
    Vector2<double> seenBallPos = lastPerception.getPositionInRelativeCoordinates();
    Vector2<double> v = estimate.getVelocityInRelativeCoordinates();
    CIRCLE("representation:GroundTruthBallModel", 
      seenBallPos.x, 
      seenBallPos.y, 
      45, 
      1, // pen width 
      Drawings::ps_solid, 
      ColorRGBA(255, 128, 0, 192),
      Drawings::bs_solid, 
      ColorRGBA(255, 128, 0, 192));
    ARROW("representation:GroundTruthBallModel", seenBallPos.x, seenBallPos.y, 
      seenBallPos.x+(v.x), seenBallPos.y+(v.y),5,1,ColorRGBA(255, 128, 0, 192));
  }
};
#if 0
class TeamMateBallModel : public BallModel {};

class TeamBallModels : public Streamable
{
public:

  /** Constructor */
  TeamBallModels(): 
      lastSeenRobotNumber(0)
  {}

  BallModel ballModels[12]; /**< The last received ball model of each team mate. (0 = currently unused) */
  int lastSeenRobotNumber; /**< The number of the robot who has seen the ball at last. (not including self) */

  /** Streaming (with specifications) */
  virtual void serialize(In *in, Out *out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(ballModels[0]);
    STREAM(ballModels[1]);
    STREAM(ballModels[2]);
    STREAM(ballModels[3]);
    STREAM(ballModels[4]);
    STREAM(ballModels[5]);
    STREAM(ballModels[6]);
    STREAM(ballModels[7]);
    STREAM(ballModels[8]);
    STREAM(ballModels[9]);
    STREAM(ballModels[10]);
    STREAM(ballModels[11]);
    STREAM(lastSeenRobotNumber);;
    STREAM_REGISTER_FINISH();
  }
};
#endif // 0


#endif //__BallModel_h_
