/**
 * @file GoalModel.h
 * 
 * Declaration of class GoalModel
 *
 * * @author <A href="mailto:timlaue@informatik.uni-bremen.de">Tim Laue</A>
 */ 

#ifndef __GoalModel_h_
#define __GoalModel_h_

#include "Tools/Range.h"
#include "Tools/Streams/Streamable.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Geometry.h"
#include "Representations/Modeling/RobotPose.h"


/**
* @class GoalModel
* Simple information currently available about the opponent's goal
*/
class GoalModel : public Streamable
{
private:
  /** Streaming function 
  * @param in Object for streaming in the one direction
  * @param out Object for streaming in the other direction
  */
  virtual void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
      STREAM(timeWhenLastSeen);
      STREAM(goalColor);
      STREAM(largestFreePart.min);
      STREAM(largestFreePart.max);
      STREAM(distancesToFreePart.min);
      STREAM(distancesToFreePart.max);
    STREAM_REGISTER_FINISH();
  }

public:
  unsigned timeWhenLastSeen;       /**< The last time, a goal has been seen*/
  ColorRGBA goalColor;             /**< Sorry, but the color is needed for the drawing*/
  Range<double> largestFreePart;   /**< Angles to the corners of the free part*/
  Range<int> distancesToFreePart;  /**< Closest and farthest distance to parts of the free part*/

  /** Constructor*/
  GoalModel():timeWhenLastSeen(0), largestFreePart(0.0,0.0)
  {}

  /** Draws the angles to the free part*/
  void draw() 
  {
    DECLARE_DEBUG_DRAWING("representation:GoalModel:largestFreePart", "drawingOnField", "drawing of the angles to the free part of the goal");
    if(SystemCall::getTimeSince(timeWhenLastSeen) < 9000)
    {
      Vector2<double> straightLine(distancesToFreePart.max,0);
      Vector2<double> vecL = Geometry::rotate(straightLine, largestFreePart.max);
      Vector2<double> vecR = Geometry::rotate(straightLine, largestFreePart.min);
      Vector2<double> shortLine(distancesToFreePart.min,0);
      Vector2<double> vecLShort = Geometry::rotate(shortLine, largestFreePart.max);
      Vector2<double> vecRShort = Geometry::rotate(shortLine, largestFreePart.min);
      ColorRGBA freePartLineColor(goalColor);
      freePartLineColor.a = 180;
      LINE("representation:GoalModel:largestFreePart",0,0,vecR.x,vecR.y,50,Drawings::ps_solid,freePartLineColor);
      LINE("representation:GoalModel:largestFreePart",0,0,vecL.x,vecL.y,50,Drawings::ps_solid,freePartLineColor);
      LINE("representation:GoalModel:largestFreePart",vecR.x,vecR.y,vecL.x,vecL.y,50,Drawings::ps_dash, ColorClasses::black);
      LINE("representation:GoalModel:largestFreePart",vecRShort.x,vecRShort.y,vecLShort.x,vecLShort.y,50,Drawings::ps_dash, ColorClasses::black);
    }
  }
};


#endif //__GoalModel_h_
