/**
* @file PointsPercept.h
* Declaration of a class that represents the points found in an image.
* @author <a href="mailto:Thomas.Roefer@dfki.de">Thomas Rfer</a>
*/

#ifndef __PointsPercept_h_
#define __PointsPercept_h_

#include "Tools/Streams/Streamable.h"
#include "Representations/Infrastructure/RoboCupGameControlData.h"
#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Pose2D.h"
#include <vector>

/**
* @class PointsPercept
* A class that represents the points found in an image.
*/
class PointsPercept : public Streamable
{
private:
  void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(origin);
    STREAM_VECTOR(points);
    STREAM_REGISTER_FINISH();
  }

public:
  enum PointType
  {
    blueBeacon = TEAM_BLUE,
    yellowBeacon = TEAM_YELLOW,
    numberOfBeaconPointTypes, // until here, only a single table per type is required
    blueGoal = numberOfBeaconPointTypes,
    yellowGoal,
    numberOfTeamDependentPointTypes, // until here, team color is important
    fieldLine = numberOfTeamDependentPointTypes,
    numberOfFieldPointTypes, // until here, points are relevant for localization
    robot = numberOfFieldPointTypes,
    cyanBeacon,
    magentaBeacon,
    numberOfPointTypes
  };

  class Point : public Streamable
  {
  private:
    void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM(inImage);
      STREAM(onField);
      STREAM_ENUMASINT(type);
      STREAM(isCloser);
      STREAM(link);
      STREAM(length2);
      STREAM(segmentId);
      STREAM_REGISTER_FINISH();
    }

  public:
    Vector2<short> inImage; /**< The point in image coordinates. */
    Vector2<int> onField; /**< The point in field coordinates. */
    PointType type; /**< The type of a line point (fieldLine, blueGoal or yellowGoal). */
    bool isCloser; /**< Set if the actual object is closer, because it was seen at the bottom of the image. */
    short link; /**< Link to another point of the same line segment. */
    int length2; /**< The squared length of the connected segment. Note: The connected segment is not necessarily a straight line. */
    int segmentId;
  };

  Pose2D origin; /**< All points are specified relative to this origin. */
  std::vector<Point> points; /**< The points in relative field coordinates. */
  
  /**
  * Default constructor.
  */
  PointsPercept() {points.reserve(1000);}

  /**
  * The method draws the percept.
  */
  void draw() const
  {
    DECLARE_DEBUG_DRAWING("representation:PointsPercept:Field", "drawingOnField", "Draws the points to the field");
    COMPLEX_DRAWING("representation:PointsPercept:Field",
      for(std::vector<Point>::const_iterator i = points.begin(); i != points.end(); ++i)
      {
        Vector2<double> v = (origin + Pose2D(i->onField.x, i->onField.y)).translation;
        ColorRGBA c = ColorRGBA(ColorClasses::none);
        switch(i->type)
        {
        case numberOfPointTypes: break;
        case blueGoal: c = ColorClasses::blue; break;
        case yellowGoal: c = ColorClasses::yellow; break;
        case fieldLine: c = ColorClasses::white; break;
        case yellowBeacon: c = ColorClasses::white; break;
        case blueBeacon: c = ColorClasses::white; break;
        case robot: c = ColorRGBA(200,0,0); break;
        case cyanBeacon: c = ColorRGBA(0,255,255); break;
        case magentaBeacon: c = ColorRGBA(255,0,255); break;
        }
        if(i->isCloser)
        {
          LINE("representation:PointsPercept:Field", 0, 0, v.x, v.y, 1, Drawings::ps_solid, c);
        }
        LARGE_DOT("representation:PointsPercept:Field", v.x, v.y, ColorClasses::black, c);
      }
    );
    DECLARE_DEBUG_DRAWING("representation:PointsPercept:Image", "drawingOnImage", "Draws the edge points found in the image");
    COMPLEX_DRAWING("representation:PointsPercept:Image",
      for(std::vector<Point>::const_reverse_iterator i = points.rbegin(); i != points.rend(); ++i)
      {
        ColorRGBA c = ColorRGBA(ColorClasses::none);
        switch(i->type)
        {
        case numberOfPointTypes: break;
        case blueGoal: c = ColorRGBA(255, 255, 0); break;
        case yellowGoal: c = ColorRGBA(0, 0, 255); break;
        case fieldLine: c = ColorRGBA(237, 11, 242); break;
        case yellowBeacon: c = ColorClasses::white; break;
        case blueBeacon: c = ColorClasses::white; break;
        case robot: c = ColorRGBA(200,0,0); break;
        case cyanBeacon: c = ColorRGBA(0,255,255); break;
        case magentaBeacon: c = ColorRGBA(255, 0, 255); break;
        }
        const Vector2<short>& v(points[i->link].inImage);
        LINE("representation:PointsPercept:Image", i->inImage.x, i->inImage.y, v.x, v.y, 1, Drawings::ps_solid, c);
        DOT("representation:PointsPercept:Image", i->inImage.x, i->inImage.y, ColorClasses::black, c);
      }
    );
  }
};

#endif //__PointsPercept_h_
