/**
* @file Representations/Perception/SegmentsPercept.h
*
* Representention of the seen segments
*
* @author <a href="mailto:rieskamp@informatik.uni-bremen.de">Andrik Rieskamp</a>
* @author <a href="mailto:felipe@informatik.uni-bremen.de">Felipe Lopez</a>
*/

#ifndef __SegmentsPercept_h_
#define __SegmentsPercept_h_

#include "Tools/Debugging/DebugDrawings.h"
#include "Tools/Math/Pose2D.h"
#include <vector>

class SegmentsPercept: public Streamable
{
private:
  /** Streaming function
  * @param in  streaming in ...
  * @param out ... streaming out.
  */
  void serialize(In* in, Out* out)
  {
    STREAM_REGISTER_BEGIN();
    STREAM(origin);
    STREAM_VECTOR(segmentsHorizontal);
    STREAM_VECTOR(segmentsVertical);
    STREAM_VECTOR(fieldBorders);
    STREAM_REGISTER_FINISH();
  }

public:

  /** Constructor */
  SegmentsPercept() 
  {
    segmentsHorizontal.reserve(1000); // TODO: Check default segment count
    segmentsVertical.reserve(1000);   // TODO: Check default segment count
  }

  class Segment : public Streamable
  {
  private:
    void serialize(In* in, Out* out)
    {
      STREAM_REGISTER_BEGIN();
      STREAM(x);
      STREAM(y);
      STREAM(length);
      STREAM(onHalfScanLine);
      STREAM_ENUMASINT(color);
      STREAM_REGISTER_FINISH();
    }

  public:
    int x; /**< x coordinate of start point in image */
    int y; /**< y coordinate of start point in image */
    int length; /**< The length of the segment line in the image. */
    bool onHalfScanLine;
    ColorClasses::Color color; /**< The primary color of the segment. */
  };

  Pose2D origin; /**< All segments are specified relative to this origin. */
  std::vector<Segment> segmentsHorizontal; /**< The segments detected in horizontal scan */
  std::vector<Segment> segmentsVertical; /**< The segments detected in vertical scan  */

  std::vector<Vector2<int> > fieldBorders;

  /**
  * The method draws the percept.
  */
  void draw() const
  {
    DECLARE_DEBUG_DRAWING("representation:SegmentsPercept:Image", "drawingOnImage", "Draws the segment to the image");
    COMPLEX_DRAWING("representation:SegmentsPercept:Image",
      for(std::vector<Segment>::const_iterator i = segmentsVertical.begin(); i != segmentsVertical.end(); ++i)
      {
        ColorClasses::Color lineColor = i->color;
        if (lineColor == ColorClasses::black) lineColor = ColorClasses::white;
        else if (lineColor == ColorClasses::white){ lineColor = ColorClasses::black;}
        LINE("representation:SegmentsPercept:Image", i->x, i->y, i->x, i->y - i->length, 1, Drawings::ps_solid, lineColor);
      }			
 
      for(std::vector<Segment>::const_iterator i = segmentsHorizontal.begin(); i != segmentsHorizontal.end(); ++i)
      {
        ColorClasses::Color lineColor = i->color;
        if (lineColor == ColorClasses::black) lineColor = ColorClasses::white;
        else if (lineColor == ColorClasses::white) lineColor = ColorClasses::black;
        LINE("representation:SegmentsPercept:Image", i->x, i->y, i->x + i->length, i->y, 1, Drawings::ps_solid, lineColor);
      }

      for(std::vector<Vector2<int> >::const_iterator i = fieldBorders.begin(); i != fieldBorders.end(); ++i)
      {
       DOT("representation:SegmentsPercept:Image", i->x, i->y, ColorRGBA::ColorRGBA(255,0,0,255), ColorRGBA::ColorRGBA(255,0,0,255));
      } 

      for(unsigned int i=0; (i+1)<fieldBorders.size(); i++) {
        LINE("representation:SegmentsPercept:Image", fieldBorders.at(i).x, fieldBorders.at(i).y, fieldBorders.at(i+1).x, fieldBorders.at(i+1).y, 1, Drawings::ps_solid, ColorRGBA::ColorRGBA(255,0,0,255));
      }
     ); 
  }

};

#endif// __SegmentsPercept_h_
