/** 
* @file SimRobotGUI/Document.cpp
* Implementation of class Document.
* @author Colin Graf
*/

#include <QFile>
#include <QMessageBox>
#include <QTextStream>
#include <QFileInfo>
#include <QFileDialog>
#include <QApplication>
#include <QMainWindow>
#include <QTimer>

#include "Document.h"
#include "StatusBar.h"
#include "Main.h"
#include "MenuBar.h"
#include "../SimRobotCore/Controller/Controller.h"
#include "../SimRobotCore/Simulation/Simulation.h"

Connection* Connection::start = 0;
Simulation* Connection::simulation;
SimulatorGui* Connection::simulatorGui;

Connection::Connection(const char* sceneName)
: sceneName(sceneName)
{
  next = start;
  start = this;
}

void SimulatorGui::print(const std::string& text)
{
  Document::document->print(text.c_str());
}

void SimulatorGui::clear()
{
  Document::document->clearConsole();
}

void SimulatorGui::setUpdateDelay(unsigned ms)
{
  Document::document->setUpdateViewsDelay(ms);
}

void SimulatorGui::setStatusText(const std::string& text)
{
  Document::document->setStatusText(text.c_str());
}

Document* Document::document = 0;

Document::Document(QMainWindow* parent) : QObject(parent), 
  opened(false), changed(false),
  simulation(0), controller(0), running(false), performStep(false), updateViews(false),
  lastUpdateViewsTime(0), updateViewsDelay(100), layoutSettings(0),
  settings("B-Human", "SimRobot"), recentFiles(settings.value("RecentFiles").toStringList())
{
  connect(MenuBar::menuBar->fileNewAction, SIGNAL(triggered()), SLOT(newSlot()));
  connect(MenuBar::menuBar->fileOpenAction, SIGNAL(triggered()), SLOT(openSlot()));
  connect(MenuBar::menuBar->fileSaveAction, SIGNAL(triggered()), SLOT(saveSlot()));
  connect(MenuBar::menuBar->fileSaveAsAction, SIGNAL(triggered()), SLOT(saveAsSlot()));
  connect(MenuBar::menuBar->fileCloseAction, SIGNAL(triggered()), SLOT(closeSlot()));
  connect(MenuBar::menuBar->simStartAction, SIGNAL(triggered()), SLOT(simStartSlot()));
  connect(MenuBar::menuBar->simResetAction, SIGNAL(triggered()), SLOT(simResetSlot()));
  connect(MenuBar::menuBar->simStepAction, SIGNAL(triggered()), SLOT(simStepSlot()));
  MenuBar::menuBar->fileNewAction->setEnabled(true);
  MenuBar::menuBar->fileOpenAction->setEnabled(true);

  Q_ASSERT(document == 0);
  document = this;
}

Document::~Document()
{
  Q_ASSERT(!opened);
  Q_ASSERT(!simulation);
  Q_ASSERT(!controller);
  Q_ASSERT(!running);
  Q_ASSERT(!layoutSettings);

  Q_ASSERT(document == this);
  document = 0;
}

void Document::newSlot()
{
  closeSlot();

  file.clear();
  fileContent.clear();  
  statusMessage = tr("File created");
  emit updateStatusBarSignal();
  initDocument();
}

void Document::openSlot()
{
  QString fileName = QFileDialog::getOpenFileName(qobject_cast<QMainWindow *>(parent()), 
    tr("Open File"), settings.value("Opendir", "").toString(), tr("Robot Simulation Files (*.ros)"));

  if(fileName.isEmpty())
    return;
  settings.setValue("Opendir", QFileInfo(fileName).dir().path());

  openFileSlot(fileName);
}

void Document::openFileSlot(const QString& fileName)
{
  qobject_cast<QMainWindow *>(parent())->setUpdatesEnabled(false);

  closeSlot();

  QFile xmlFile(fileName);
  if(!xmlFile.open(QFile::ReadOnly | QFile::Text))
  {
     QMessageBox::warning(qobject_cast<QMainWindow *>(parent()), tr("SimRobot"),
                          tr("Cannot read file %1:\n%2.").arg(fileName).arg(xmlFile.errorString()));
     qobject_cast<QMainWindow *>(parent())->setUpdatesEnabled(true);
     return;
  }

  QFileInfo fileInfo(fileName);
  QTextStream in(&xmlFile);
  file = fileInfo.absoluteDir().canonicalPath() + '/' + fileInfo.fileName();
  fileContent = in.readAll(); 

  // add clean filepath to recent file list
  recentFiles.removeAll(file);
  recentFiles.prepend(file);
  while(recentFiles.count() > 8)
    recentFiles.removeLast();
  settings.setValue("RecentFiles", recentFiles);

  statusMessage = tr("File loaded");
  emit updateStatusBarSignal();

  initDocument();

  qobject_cast<QMainWindow *>(parent())->setUpdatesEnabled(true);
}

bool Document::isOpened() const
{
  return opened;
}

void Document::initDocument()
{
  opened = true;
  changed = false;
  running = false;
  performStep = false;
  updateViews = false;
  openedObjectsRestored = false;
  MenuBar::menuBar->fileSaveAction->setEnabled(false);
  MenuBar::menuBar->fileSaveAsAction->setEnabled(true);
  MenuBar::menuBar->fileCloseAction->setEnabled(true);
  MenuBar::menuBar->simStartAction->setEnabled(true);
  MenuBar::menuBar->simResetAction->setEnabled(true);
  MenuBar::menuBar->simStepAction->setEnabled(true);
  Q_ASSERT(!layoutSettings);

  layoutSettings = new QSettings("B-Human", "SimRobot");
  layoutSettings->beginGroup("Layouts");
  layoutSettings->beginGroup(file.isEmpty() ? "default" : getUserFriendlyFile());

  emit openSignal();

  restoreLayout();
}

void Document::saveSlot()
{
  if(file.isEmpty())
  {
    saveAsSlot();
    return;
  }

  QFile xmlFile(file);
  if(!xmlFile.open(QFile::WriteOnly | QFile::Text))
  {
     QMessageBox::warning(qobject_cast<QMainWindow *>(parent()), tr("SimRobot"),
                          tr("Cannot write file %1:\n%2.")
                          .arg(file)
                          .arg(xmlFile.errorString()));
     return;
  }

  QTextStream out(&xmlFile);
  emit aboutToSaveSignal();
  out << fileContent;
  changed = false;
  MenuBar::menuBar->fileSaveAction->setEnabled(false);
  emit saveSignal();
  statusMessage = tr("File saved");
  emit updateStatusBarSignal();
}

void Document::saveAsSlot()
{
  QString fileName = QFileDialog::getSaveFileName(qobject_cast<QMainWindow *>(parent()), tr("Save File As"), file, tr("Robot Simulation Files (*.ros)"));
  if (fileName.isEmpty())
     return;

  QFileInfo fileInfo(fileName);
  fileName = fileInfo.absoluteDir().canonicalPath() + '/' + fileInfo.fileName();
  if(file == fileName)
  {
    saveSlot();
    return;
  }

  file = fileName;
  saveSlot();

  Q_ASSERT(layoutSettings);
  delete layoutSettings;
  layoutSettings = new QSettings("B-Human", "SimRobot");
  layoutSettings->beginGroup("Layouts");
  layoutSettings->beginGroup(file.isEmpty() ? "default" : getUserFriendlyFile());

  recentFiles.removeAll(fileName);
  recentFiles.prepend(fileName);
  while(recentFiles.count() > 8)
    recentFiles.removeLast();
  settings.setValue("RecentFiles", recentFiles);
}

void Document::closeSlot()
{
  if(opened)
  {
    emit aboutToCloseSignal();
    writeLayout();
    for(QLinkedList<OpenedObject>::Iterator i = openedObjects.begin(), end  = openedObjects.end(); i != end; i = openedObjects.begin())
    {
      QString name(i->name);
      DocumentObjectType type = i->type;
      openedObjects.erase(i);
      emit closeObjectSignal(name, type);
    }
    file.clear();
    fileContent.clear();
    opened = false;
    changed = false;
    running = false;
    performStep = false;
    updateViews = false;
    openedObjectsRestored = false;
    MenuBar::menuBar->fileSaveAction->setEnabled(false);
    MenuBar::menuBar->fileSaveAsAction->setEnabled(false);
    MenuBar::menuBar->fileCloseAction->setEnabled(false);
    MenuBar::menuBar->simStartAction->setEnabled(false);
    MenuBar::menuBar->simStartAction->setChecked(false);
    MenuBar::menuBar->simResetAction->setEnabled(false);
    MenuBar::menuBar->simStepAction->setEnabled(false);
    Q_ASSERT(layoutSettings);
    delete layoutSettings;
    layoutSettings = 0;
    
    if(controller)
    {
      controller->destroy();
      delete controller;
      controller = 0;
    }
    if(simulation)
    {
      delete simulation;
      simulation = 0;
      objectDescriptionTree.clear();
      emit updateObjectTreeSignal();
    }
    clearConsole();
    emit closeSignal();
    statusMessage = tr("File closed");
    emit updateStatusBarSignal();
  }
}

QString Document::getUserFriendlyFile() const
{
  return QFileInfo(file).fileName();
}

const QString& Document::getFileContent() const
{
  return fileContent;
}

void Document::setFileContent(const QString& content)
{
  fileContent = content;
}

void Document::setFileChanged()
{
  if(!changed)
  {
    changed = true;
    MenuBar::menuBar->fileSaveAction->setEnabled(true);
    emit changeSignal();
  }
}

bool Document::getFileChanged() const
{
  return changed;
}

const QString& Document::getStatusMessage() const
{
  if(controllerStatusMessage.isEmpty())
    return statusMessage;
  return controllerStatusMessage;
}

void Document::simResetSlot() 
{
  bool wasRunning = running;
  running = false;
  MenuBar::menuBar->simStartAction->setChecked(running);
  if(controller)
  {
    controller->destroy();
    delete controller;
    controller = 0;
  }
  if(simulation)
  {
    delete simulation;
    simulation = 0;
    objectDescriptionTree.clear();
    emit updateObjectTreeSignal();
    updateObjects();
  }
  compile();
  if(simulation)
  {
    running = wasRunning;
    MenuBar::menuBar->simStartAction->setChecked(running);
    if(running)
      startTimer(0);
  }
}

void Document::simStartSlot() 
{
  if(running)
  {
    running = false;
    MenuBar::menuBar->simStartAction->setChecked(running);
  }
  else
  {
    if(changed || !simulation)
      simResetSlot();
    if(simulation)
    {
      running = true;
      MenuBar::menuBar->simStartAction->setChecked(running);
      startTimer(0);
    }
  }
}

void Document::simStepSlot() 
{
  if(running)
  {
    running = false;
    MenuBar::menuBar->simStartAction->setChecked(running);
    performStep = true;
    return;
  }
  if(changed || !simulation)
    simResetSlot();
  if(simulation)
  {
    performStep = true;
    startTimer(0);
  }
}

void Document::compile()
{
  unsigned int startTime = Main::getSystemTime();

  // check if document has a name
  if(changed || file.isEmpty())
  {
    saveSlot();
    if(changed || file.isEmpty())
      return;
  }
  
  // set directory for relative paths in texture load commands
  QDir::setCurrent(QFileInfo(file).dir().path());
  QString roSiFilePath = QFileInfo(Main::main->appPath).dir().path() + "/RoSi.xsd";

  Q_ASSERT(simulation == 0);
  Q_ASSERT(controller == 0);

  // compile
  simulation = new Simulation();
  // check for an error
  std::string stdFile(file.toAscii().constData());
  std::string stdRoSiFile(roSiFilePath.toAscii().constData());
  if(!simulation->loadFile(stdFile, stdRoSiFile))
  {

    std::vector<ErrorDescription> errors;
    simulation->getFirstError(errorDescription);
    simulation->getAllErrors(errors);
    statusMessage = tr(errorDescription.text.c_str());
    emit updateStatusBarSignal();
    delete simulation;
    simulation = 0;
    QApplication::beep();
    std::vector<ErrorDescription>::const_iterator it;
    print("----------------------\n");
    print("Scene Compiler Errors:\n");
    print(statusMessage + "\n");
    for(it = errors.begin(); it < errors.end(); it++)
      print(((*it).text+"\n").c_str());
    print("----------------------\n");
  }
  else
  {
    // search for controller
    SimObject* objectTree = simulation->getObjectReference("");
    const std::string& sceneName(objectTree->getName());
    Connection* pThis = Connection::start;
    while (pThis && !(sceneName == pThis->sceneName))
      pThis = pThis->next;
    if(pThis)
    {
      controller = pThis->createController(*simulation, simulatorGui);
      controller->init();
      simulation->getObjectDescriptionTree(objectDescriptionTree);
      simulation->resetSceneGraphChanged();
      emit updateObjectTreeSignal();
      updateObjects();
      statusMessage = tr("Compilation successful (%1ms)").arg(Main::getSystemTime() - startTime);
      emit updateStatusBarSignal();
    }
    else
    {
      simulation->getObjectDescriptionTree(objectDescriptionTree);
      simulation->resetSceneGraphChanged();
      emit updateObjectTreeSignal();
      updateObjects();
      statusMessage = tr("There's no controller for this scene");
      emit updateStatusBarSignal();
    }    
  }
}

DocumentObjectType Document::getObjectType(const QString& fullName) const
{
  if(simulation)
  {
    std::string compName(fullName.toAscii().constData());
    for(std::vector<ObjectDescription>::const_iterator i = objectDescriptionTree.begin(), end = objectDescriptionTree.end(); i != end; ++i)
      if(i->fullName == compName)
        return DocumentObjectType(i->type);
  }
  if(fullName == ".Editor")
    return ObjectTypeEditorWindow;
  if(fullName == ".Actuator")
    return ObjectTypeActuatorWindow;
  return ObjectTypeNone;
}

Simulation* Document::getSimulation() const
{
  return simulation;
}

bool Document::openObject(const QString& object)
{
  if(!opened || !openedObjectsRestored)
    return false;

  for(QLinkedList<OpenedObject>::iterator i = openedObjects.begin(), end = openedObjects.end(); i != end; ++i)
    if(i->name == object)
    {
      emit raiseObjectSignal(object, i->type);
      return true;
    }

  DocumentObjectType type = getObjectType(object);
  switch(type)
  {
  case ObjectTypeObject:
  case ObjectTypeActuator:
  case ObjectTypeSensor:
  case ObjectTypeEditorWindow:
  case ObjectTypeActuatorWindow:
    {
      openedObjects.append(OpenedObject(object, type));
      emit openObjectSignal(object, type);
      return true;
    }
    break;
  default:
    break;
  }

  return false;
}

void Document::raiseObject(const QString& object)
{
  for(QLinkedList<OpenedObject>::iterator i = openedObjects.begin(), end = openedObjects.end(); i != end; ++i)
    if(i->name == object)
    {
      DocumentObjectType type = i->type;
      openedObjects.erase(i);
      openedObjects.append(OpenedObject(object, type));
      return;
    }
}

bool Document::closeObject(const QString& object)
{
  for(QLinkedList<OpenedObject>::iterator i = openedObjects.begin(), end = openedObjects.end(); i != end; ++i)
    if(i->name == object)
    {
      DocumentObjectType type = i->type;
      openedObjects.erase(i);
      emit closeObjectSignal(object, type);
      return true;
    }
  return false;
}

void Document::timerEvent(QTimerEvent* event) 
{
  if(performStep || running)
  {
    performStep = false;
    if(controller)
    {
      controller->execute();
      if(simulation->getResetFlag())
        simulation->resetSimulation();
      if(simulation->hasSceneGraphChanged())
      {
        simulation->getObjectDescriptionTree(objectDescriptionTree);
        simulation->resetSceneGraphChanged();
        updateObjects();
        emit updateObjectTreeSignal();
      }
    }
    emit updateActuatorsSignal();
    simulation->doSimulationStep();
    updateViews = true;
  }

  if(updateViews)
  {
    unsigned int t = Main::getSystemTime();
    if(t - lastUpdateViewsTime >= updateViewsDelay)
    {      
      lastUpdateViewsTime = t;
      updateViews = false;
      emit updateViewsSignal();
      emit updateStatusBarSignal();
    }
  }
  else
    killTimer(event->timerId());
}

void Document::setUpdateViewsDelay(unsigned int ms)
{
  updateViewsDelay = ms;
}

const std::vector<ObjectDescription>& Document::getObjectTree() const
{
  return objectDescriptionTree;
}

QDataStream& operator>>(QDataStream& s, Document::OpenedObject& obj)
{
  int type;
  s >> obj.name >> type;
  obj.type = DocumentObjectType(type);
  return s;
}

QDataStream& operator<<(QDataStream& s, const Document::OpenedObject& obj)
{
  s << obj.name << int(obj.type);
  return s;
}

void Document::restoreLayout()
{
  Q_ASSERT(layoutSettings);
  bool hasFile = !file.isEmpty();

  emit restoreLayoutSignal();

  if(hasFile)
  {
    if(layoutSettings->value("Compile", false).toBool())
      compile();

    if(simulation)
    {
      if(layoutSettings->value("Run", false).toBool())
        simStartSlot();
    }
    
    QTimer::singleShot(1000, this, SLOT(restoreOpenedObjectsSlot()));
  }
}

void Document::restoreOpenedObjectsSlot()
{
  if(!opened || !layoutSettings)
    return;
  
  
  QLinkedList<OpenedObject> newOpenedObjects;
  const QByteArray& buffer(layoutSettings->value("Views").toByteArray());
  QDataStream in(buffer);
  in.setVersion(QDataStream::Qt_4_3);
  in >> newOpenedObjects;

  qobject_cast<QMainWindow *>(parent())->setUpdatesEnabled(false);
  openedObjects.clear();
  for(QLinkedList<OpenedObject>::Iterator i = newOpenedObjects.begin(), end = newOpenedObjects.end(); i != end; ++i)
    if(i->type != ObjectTypeNone)
    {
      emit openObjectSignal(i->name, i->type);
      openedObjects.append(OpenedObject(i->name, i->type));
    }
  qobject_cast<QMainWindow *>(parent())->setUpdatesEnabled(true);
  
  openedObjectsRestored = true;
}

void Document::writeLayout()
{
  Q_ASSERT(layoutSettings);
  bool hasFile = !file.isEmpty();

  if(hasFile)
  {
    layoutSettings->setValue("Compile", simulation ? true : false);
    layoutSettings->setValue("Run", running);

    QByteArray buffer;
    QDataStream out(&buffer, QIODevice::WriteOnly);
    out.setVersion(QDataStream::Qt_4_3);
    out << openedObjects;
    layoutSettings->setValue("Views", buffer);
  }
  emit writeLayoutSignal();
}

QSettings* Document::getLayoutSettings()
{
  return layoutSettings;
}

void Document::print(const QString& text)
{    
  emit printSignal(text);
}

void Document::clearConsole()
{
  emit clearConsoleSignal();
}

void Document::setStatusText(const QString& text)
{
  controllerStatusMessage = text;
}

void Document::handleConsoleCommand(const QString& text)
{
  if(controller)
    controller->onConsoleCommand(text.toAscii().constData());
}

void Document::completeConsoleCommand(QString& text, bool forward, bool nextSection)
{
  if(controller)
  {
    std::string cmd(text.toAscii().constData());
    controller->onConsoleCompletion(cmd, forward, nextSection);
    text = cmd.c_str();
  }
}

void Document::updateObjects()
{
  QHash<QString, OpenedObject*> openedObjectsHash;
  openedObjectsHash.reserve(openedObjects.size());
  for(QLinkedList<OpenedObject>::iterator i = openedObjects.begin(), end = openedObjects.end(); i != end; ++i)
    openedObjectsHash[i->name] = &(*i);

  for(std::vector<ObjectDescription>::const_iterator i = objectDescriptionTree.begin(), end = objectDescriptionTree.end(); i != end; ++i)
  {
    QString name(i->fullName.c_str());
    if(openedObjectsHash.contains(name))
      openedObjectsHash[name]->type = DocumentObjectType(i->type);
  }
  for(QLinkedList<OpenedObject>::Iterator i = openedObjects.begin(), end  = openedObjects.end(); i != end; ++i)
  {
    emit updateObjectSignal(i->name, i->type);
  }
}

const QStringList& Document::getRecentFiles() const
{
  return recentFiles;
}
