/** 
* @file SimRobotGUI/Document.h
* Declaration of class Document.
* @author Colin Graf
*/

#ifndef Document_H
#define Document_H

#include <QObject>
#include <QSettings>
#include <QHash>
#include <QLinkedList>
#include <QStringList>
#include <QSettings>
#include <vector>

#include "../SimRobotCore/Simulation/APIDatatypes.h"
#include "../SimRobotCore/Controller/Controller.h"

class QMainWindow;
class Simulation;
class Controller;
class QSettings;


enum DocumentObjectType // using Document::ObjectType didn't work with qt slot/signals
                        // (but probably i just did something wrong)
{
  ObjectTypeNone = OBJECT_TYPE_NONE,
  ObjectTypeObject = OBJECT_TYPE_OBJECT,
  ObjectTypeSensor = OBJECT_TYPE_SENSORPORT, 
  ObjectTypeActuator = OBJECT_TYPE_ACTUATORPORT,
  ObjectTypeEditorWindow,
  ObjectTypeActuatorWindow,
};

/** 
* @class Document
* A global accessible class for managing an opened document.
*/
class Document : public QObject
{
  Q_OBJECT

public:

  class OpenedObject
  {
  public:
    QString name;
    DocumentObjectType type;

    OpenedObject() : type(ObjectTypeNone) {}
    OpenedObject(const QString& name, DocumentObjectType type) : name(name), type(type) {}
  };

  static Document* document;

  Document(QMainWindow* parent);
  ~Document();
  QString getUserFriendlyFile() const;
  const QString& getFileContent() const;
  void setFileContent(const QString& content);
  void setFileChanged();
  bool getFileChanged() const;
  const QString& getStatusMessage() const;
  bool openObject(const QString& object);
  void raiseObject(const QString& object);
  bool closeObject(const QString& object);
  Simulation* getSimulation() const;
  void setUpdateViewsDelay(unsigned int ms);
  const std::vector<ObjectDescription>& getObjectTree() const;
  QSettings* getLayoutSettings();
  void setStatusText(const QString& text);
  void print(const QString& text);
  void clearConsole();
  void handleConsoleCommand(const QString& text);
  void completeConsoleCommand(QString& text, bool forward, bool nextSection);
  bool isOpened() const;
  const QStringList& getRecentFiles() const;

public slots:
  void newSlot();
  void closeSlot();
  void openFileSlot(const QString& filePath);

signals:
  void openSignal();
  void aboutToCloseSignal();
  void closeSignal();
  void changeSignal();
  void aboutToSaveSignal();
  void saveSignal();
  void updateObjectTreeSignal();
  void openObjectSignal(const QString& object, DocumentObjectType type);
  void raiseObjectSignal(const QString& object, DocumentObjectType type);
  void updateObjectSignal(const QString& object, DocumentObjectType type);
  void closeObjectSignal(const QString& object, DocumentObjectType type);
  void updateStatusBarSignal();
  void updateViewsSignal();
  void updateActuatorsSignal();
  void restoreLayoutSignal();
  void writeLayoutSignal();
  void clearConsoleSignal();
  void printSignal(const QString& text);

private slots:  
  void saveSlot();
  void saveAsSlot();
  void openSlot();
  void simResetSlot();
  void simStartSlot();
  void simStepSlot();
  void restoreOpenedObjectsSlot(); // delayed restoring (via QTimer::singleShot)

private:
  QString file;
  QString fileContent;
  bool opened;
  bool changed;
  Simulation* simulation;
  Controller* controller;
  bool running;
  bool performStep;
  bool updateViews;
  unsigned int lastUpdateViewsTime;
  unsigned int updateViewsDelay;
  ErrorDescription errorDescription;
  QString statusMessage;
  QString controllerStatusMessage;
  SimulatorGui simulatorGui;
  std::vector<ObjectDescription> objectDescriptionTree;
  QSettings* layoutSettings;
  QLinkedList<OpenedObject> openedObjects; /**< Opened objects in window z-order. */
  bool openedObjectsRestored;
  QSettings settings;
  QStringList recentFiles;

  void compile();
  void timerEvent(QTimerEvent* event);
  void initDocument();
  void restoreLayout();
  void writeLayout();
  DocumentObjectType getObjectType(const QString& fullName) const;
  void updateObjects();
};

#endif // Document_H
