
#include <QFont>
#include <QCloseEvent>
#include <QMenu>
#include <QTextEdit>
#include <QMdiArea>

#include "MenuBar.h"
#include "Editor.h"
#include "Document.h"

Editor::Editor(QMdiArea *mdiArea) : SubWindow(mdiArea, ".Editor"), 
  canCopy(false), canUndo(false), canRedo(false)
{
  setWindowTitle(tr("Editor"));
  setWindowIcon(QIcon(":/icons/Editor.png"));

  textEdit = new QTextEdit(this);
  setWidget(textEdit);
  textEdit->setPlainText(Document::document->getFileContent());

#ifdef _WIN32
  QFont font("Courier New", 10);
#else
  QFont font("Bitstream Vera Sans Mono", 9);
#endif
  textEdit->setFont(font);
  textEdit->setLineWrapMode(QTextEdit::NoWrap);

  connect(textEdit, SIGNAL(copyAvailable(bool)), SLOT(copyAvailableSlot(bool)));
  connect(textEdit, SIGNAL(redoAvailable(bool)), SLOT(redoAvailableSlot(bool)));
  connect(textEdit, SIGNAL(undoAvailable(bool)), SLOT(undoAvailableSlot(bool)));
  connect(textEdit, SIGNAL(textChanged()), this, SLOT(textChangedSlot()));

  connect(Document::document, SIGNAL(aboutToSaveSignal()), SLOT(aboutToSaveDocumentSlot()));

  MenuBar::menuBar->viewEditorAction->setChecked(true);
}

Editor::~Editor()
{
  MenuBar::menuBar->viewEditorAction->setChecked(false);
}

void Editor::copyAvailableSlot(bool yes)
{
  canCopy = yes;
  MenuBar::menuBar->editCopyAction->setEnabled(canCopy);
  MenuBar::menuBar->editCutAction->setEnabled(canCopy);
}

void Editor::redoAvailableSlot(bool available)
{
  canRedo = available;
  MenuBar::menuBar->editRedoAction->setEnabled(canRedo);
}

void Editor::undoAvailableSlot(bool available)
{
  canUndo = available;
  MenuBar::menuBar->editUndoAction->setEnabled(canUndo);
}

void Editor::textChangedSlot()
{
  Document::document->setFileChanged();
}

void Editor::connectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->editCopyAction->setEnabled(canCopy);
  menuBar->editCutAction->setEnabled(canCopy);
  menuBar->editPasteAction->setEnabled(textEdit->canPaste());
  menuBar->editRedoAction->setEnabled(canRedo);
  menuBar->editUndoAction->setEnabled(canUndo);

  connect(menuBar->editCopyAction, SIGNAL(triggered()), textEdit, SLOT(copy()));
  connect(menuBar->editCutAction, SIGNAL(triggered()), textEdit, SLOT(cut()));
  connect(menuBar->editPasteAction, SIGNAL(triggered()), textEdit, SLOT(paste()));
  connect(menuBar->editRedoAction, SIGNAL(triggered()), textEdit, SLOT(redo()));
  connect(menuBar->editUndoAction, SIGNAL(triggered()), textEdit, SLOT(undo()));

  QMenu* editMenu = menuBar->editMenu;
  editMenu->addAction(menuBar->editUndoAction);
  editMenu->addAction(menuBar->editRedoAction);
  editMenu->addSeparator();
  editMenu->addAction(menuBar->editCutAction);
  editMenu->addAction(menuBar->editCopyAction);
  editMenu->addAction(menuBar->editPasteAction);
}

void Editor::disconnectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->editCopyAction->setEnabled(false);
  menuBar->editCutAction->setEnabled(false);
  menuBar->editPasteAction->setEnabled(false);
  menuBar->editUndoAction->setEnabled(false);
  menuBar->editRedoAction->setEnabled(false);

  disconnect(menuBar->editCopyAction, SIGNAL(triggered()), textEdit, SLOT(copy()));
  disconnect(menuBar->editCutAction, SIGNAL(triggered()), textEdit, SLOT(cut()));
  disconnect(menuBar->editPasteAction, SIGNAL(triggered()), textEdit, SLOT(paste()));
  disconnect(menuBar->editRedoAction, SIGNAL(triggered()), textEdit, SLOT(redo()));
  disconnect(menuBar->editUndoAction, SIGNAL(triggered()), textEdit, SLOT(undo()));

  menuBar->resetEditMenu();
}

void Editor::aboutToSaveDocumentSlot()
{
  if(Document::document->getFileChanged())
    Document::document->setFileContent(textEdit->toPlainText());
}

void Editor::restoreLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup(object);
  restoreGeometry(*settings);
  settings->endGroup();
}

void Editor::writeLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup(object);
  writeGeometry(*settings);
  settings->endGroup();
}
