
#include <QMdiArea>
#include <QApplication>
#include <QMessageBox>
#include <QDragEnterEvent>
#include <QUrl>
#include <QFileInfo>

#include "MainWindow.h"
#include "ActuatorWindow.h"
#include "ObjectWindow.h"
#include "SensorWindow.h"
#include "Editor.h"
#include "MenuBar.h"
#include "Document.h"
#include "ToolBar.h"
#include "StatusBar.h"
#include "Console.h"
#include "Tree.h"

MainWindow::MainWindow(int argc, char *argv[]) : QMainWindow(0),
  windowActionGroup(0), actuatorWindow(0), activeSubWindow(0)
{
  setWindowTitle(tr("SimRobot"));
  setWindowIcon(QIcon(":/icons/Simulator.png"));
  setAcceptDrops(true);

  // create actions and menu
  menuBar = new MenuBar(this);
  setMenuBar(menuBar);
  resizeMapper.setMapping(menuBar->resize160x120, 1);
  resizeMapper.setMapping(menuBar->resize320x240, 2);
  resizeMapper.setMapping(menuBar->resize480x360, 3);
  resizeMapper.setMapping(menuBar->resize640x480, 4);

  // create document handler
  document = new Document(this);

  // create toolbar
  toolBar = new ToolBar(this);
  addToolBar(toolBar);

  // create Status Bar
  statusBar = new StatusBar(this);
  setStatusBar(statusBar);

  // create mdi workspace
  mdiArea = new QMdiArea(this);
  mdiArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
  mdiArea->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
  setCentralWidget(mdiArea);  
  
  // create Console
  console = new Console(this);
  addDockWidget(Qt::BottomDockWidgetArea, console);

  // create Tree
  tree = new Tree(this);
  addDockWidget(Qt::LeftDockWidgetArea, tree);

  //
  createConnections();

  // handle command line args
  for(int i = 1; i < argc; i++)
    if(*argv[i] != '-')
    {
      document->openFileSlot(argv[i]);
      break;
    }
  if(!document->isOpened())
    document->newSlot();
}

void MainWindow::createConnections()
{
  connect(menuBar->fileExitAction, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));
  menuBar->fileExitAction->setEnabled(true);

  connect(menuBar->winCloseAction, SIGNAL(triggered()), mdiArea, SLOT(closeActiveSubWindow()));
  connect(menuBar->winCloseAllAction, SIGNAL(triggered()), mdiArea, SLOT(closeAllSubWindows()));
  connect(menuBar->winTileAction, SIGNAL(triggered()), mdiArea, SLOT(tileSubWindows()));
  connect(menuBar->winCascadeAction, SIGNAL(triggered()), mdiArea, SLOT(cascadeSubWindows()));
  connect(menuBar->winNextAction, SIGNAL(triggered()), mdiArea, SLOT(activateNextSubWindow()));
  connect(menuBar->winPreviousAction, SIGNAL(triggered()), mdiArea, SLOT(activatePreviousSubWindow()));

  connect(menuBar->helpAboutAction, SIGNAL(triggered()), this, SLOT(aboutSlot()));
  menuBar->helpAboutAction->setEnabled(true);
  connect(menuBar->helpAboutQtAction, SIGNAL(triggered()), qApp, SLOT(aboutQt()));  
  menuBar->helpAboutQtAction->setEnabled(true);


  connect(menuBar->resize160x120, SIGNAL(triggered()), &resizeMapper, SLOT(map()));
  connect(menuBar->resize320x240, SIGNAL(triggered()), &resizeMapper, SLOT(map()));
  connect(menuBar->resize480x360, SIGNAL(triggered()), &resizeMapper, SLOT(map()));
  connect(menuBar->resize640x480, SIGNAL(triggered()), &resizeMapper, SLOT(map()));
  connect(&resizeMapper, SIGNAL(mapped(int)), SLOT(resizeSubWindow(int)));
  

  connect(menuBar->viewEditorAction, SIGNAL(triggered(bool)), this, SLOT(showEditorSlot(bool)));
  menuBar->viewEditorAction->setCheckable(true);
  menuBar->viewEditorAction->setEnabled(true);

  connect(menuBar->fileMenu, SIGNAL(aboutToShow()), this, SLOT(updateFileMenuSlot()));
  connect(&recentFileMapper, SIGNAL(mapped(const QString&)), document, SLOT(openFileSlot(const QString&)));

  connect(menuBar->winMenu, SIGNAL(aboutToShow()), this, SLOT(updateWindowMenuSlot()));


  connect(mdiArea, SIGNAL(subWindowActivated(QMdiSubWindow*)), this, SLOT(subWindowActivatedSlot(QMdiSubWindow*)));
  connect(&windowMapper, SIGNAL(mapped(QWidget*)), this, SLOT(setActiveSubWindowSlot(QWidget*)));


  connect(document, SIGNAL(openObjectSignal(const QString&, DocumentObjectType)), this, SLOT(openObjectSlot(const QString&, DocumentObjectType)));
  connect(document, SIGNAL(updateObjectSignal(const QString&, DocumentObjectType)), this, SLOT(updateObjectSlot(const QString&, DocumentObjectType)));
  connect(document, SIGNAL(closeObjectSignal(const QString&, DocumentObjectType)), this, SLOT(closeObjectSlot(const QString&, DocumentObjectType)));
  connect(document, SIGNAL(raiseObjectSignal(const QString&, DocumentObjectType)), this, SLOT(raiseObjectSlot(const QString&, DocumentObjectType)));
  connect(document, SIGNAL(openSignal()), this, SLOT(openDocumentSlot()));
  connect(document, SIGNAL(closeSignal()), this, SLOT(closeDocumentSlot()));
  connect(document, SIGNAL(changeSignal()), this, SLOT(changeDocumentSlot()));
  connect(document, SIGNAL(saveSignal()), this, SLOT(saveDocumentSlot()));
  connect(document, SIGNAL(restoreLayoutSignal()), this, SLOT(restoreLayoutSlot()));
  connect(document, SIGNAL(writeLayoutSignal()), this, SLOT(writeLayoutSlot()));
}

void MainWindow::updateWindowMenuSlot()
{
  menuBar->resetWinMenu();
  QMenu* menu = menuBar->winMenu;
  QList<QMdiSubWindow *> windows = mdiArea->subWindowList();

  if(windowActionGroup)
  {
    delete windowActionGroup;
    windowActionGroup = 0;
  }

  bool hasWindows = false;
  if(!windows.isEmpty())
  {
    menu->addSeparator();

    windowActionGroup = new QActionGroup(this);

    int pos = 1;
    for(QList<QMdiSubWindow*>::iterator i = windows.begin(), end = windows.end(); i != end; ++i, ++pos)
    {
      QMdiSubWindow *subWindow = *i;

      if(!subWindow->isVisible())
        continue;
      hasWindows = true;

      QString text = tr(pos < 10 ? "&%1 %2" : "%1 %2").arg(pos).arg(subWindow->windowTitle());

      QAction *action  = menu->addAction(text);
      action->setCheckable(true);      
      connect(action, SIGNAL(triggered()), &windowMapper, SLOT(map()));
      windowMapper.setMapping(action, subWindow);
      windowActionGroup->addAction(action);
      if(subWindow == mdiArea->activeSubWindow())
        action->setChecked(true);
    }
  }

  menuBar->winCloseAction->setEnabled(hasWindows);
  menuBar->winCloseAllAction->setEnabled(hasWindows);
  menuBar->winTileAction->setEnabled(hasWindows);
  menuBar->winCascadeAction->setEnabled(hasWindows);
  menuBar->winNextAction->setEnabled(hasWindows);
  menuBar->winPreviousAction->setEnabled(hasWindows);
  menuBar->resizeMenu->menuAction()->setEnabled(hasWindows);
}

void MainWindow::updateFileMenuSlot()
{
  menuBar->resetFileMenu();
  QMenu* menu = menuBar->fileMenu;

  const QStringList& recentFiles(Document::document->getRecentFiles());
  if(!recentFiles.isEmpty())
  {
    menu->addSeparator();
    char shortcut = '1';
    for(QStringList::const_iterator i = recentFiles.begin(), end = recentFiles.end(); i != end; ++i)
    {
      const QString& file(*i);
      QAction* action = menu->addAction("&" + QString(shortcut++) + " " + QFileInfo(file).fileName());
      connect(action, SIGNAL(triggered()), &recentFileMapper, SLOT(map()));
      recentFileMapper.setMapping(action, file);
    }
  }

  menu->addSeparator();
  menu->addAction(menuBar->fileExitAction);
}

void MainWindow::aboutSlot()
{
  QMessageBox::about(this, tr("About SimRobot"),
    tr("<b>SimRobot</b> 2008 for Qt4<br><table><tr><td>Authors:</td><td>Tim Laue</td></tr><tr><td></td><td>Thomas Rfer</td></tr><tr><td></td><td>Kai Spiess</td></tr><tr><td></td><td>Dennis Pachur</td></tr><tr><td></td><td>Armin Burchardt</td></tr><tr><td></td><td>Christian Bttner</td></tr><tr><td></td><td>Markus Kch</td></tr><tr><td></td><td>Tobias Schade</td></tr><tr><td></td><td>Colin Graf</td></tr><tr><td></td><td>Thijs Jeffry de Haas</td></tr></table><br><br>German Research Center for Artificial Intelligence (DFKI)<br>University of Bremen\n\n"));
}

void MainWindow::openDocumentSlot()
{
  QString fileTitle(document->getUserFriendlyFile());
  if(fileTitle.isEmpty())
    setWindowTitle(tr("SimRobot - New"));
  else
    setWindowTitle(tr("SimRobot - %1").arg(fileTitle));
}

void MainWindow::saveDocumentSlot()
{
  QString fileTitle(document->getUserFriendlyFile());
  if(fileTitle.isEmpty())
    setWindowTitle(tr("SimRobot - New"));
  else
    setWindowTitle(tr("SimRobot - %1").arg(fileTitle));
}

void MainWindow::closeDocumentSlot()
{
  setWindowTitle(tr("SimRobot"));
}

void MainWindow::changeDocumentSlot()
{
  QString fileTitle(document->getUserFriendlyFile());
  if(fileTitle.isEmpty())
    setWindowTitle(tr("SimRobot - New*"));
  else
    setWindowTitle(tr("SimRobot - %1*").arg(fileTitle));
}

void MainWindow::setActiveSubWindowSlot(QWidget *window)
{
  if(!window)
    return;
  mdiArea->setActiveSubWindow(qobject_cast<QMdiSubWindow*>(window));
}

void MainWindow::subWindowActivatedSlot(QMdiSubWindow* widget)
{
  SubWindow* subWindow = qobject_cast<SubWindow*>(widget); 
  if(activeSubWindow)
    activeSubWindow->disconnectMenuBar();
  activeSubWindow = subWindow;
  if(activeSubWindow)
  {
    activeSubWindow->connectMenuBar();
    Document::document->raiseObject(activeSubWindow->getObject());
  }
}

void MainWindow::writeLayoutSlot()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup("MainWindow");

  settings->setValue("State", saveState()); // tool and dockwidgets stuff
  writeGeometry(*settings);

  settings->endGroup();
}

void MainWindow::restoreLayoutSlot()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup("MainWindow");

  restoreGeometry(*settings);
  const QVariant& var(settings->value("State")); // tool and dockwidgets stuff
  if(var.isValid())
    restoreState(var.toByteArray());

  settings->endGroup();
}

void MainWindow::writeGeometry(QSettings& settings)
{
  settings.setValue("Geometry", normalGeometry);
  settings.setValue("WindowState", int(windowState()));
}

void MainWindow::restoreGeometry(QSettings& settings)
{
  const QVariant& var(settings.value("Geometry"));
  if(!var.isValid())
    return;
  const QRect& geometry(var.toRect());
  if(geometry.isEmpty())
    return;
  Qt::WindowStates newWindowState = Qt::WindowStates(settings.value("WindowState").toInt());

  Qt::WindowStates oldWindowState = windowState();
  if(oldWindowState & (Qt::WindowMinimized | Qt::WindowMaximized | Qt::WindowFullScreen))
  { // Make sure the window is not maximized or minimized. Otherwise we can't restore the window geometry.
#ifdef _WIN32
      WINDOWPLACEMENT wp;
      GetWindowPlacement(effectiveWinId(), &wp);
      wp.showCmd = SW_SHOWNORMAL;
      SetWindowPlacement(effectiveWinId(), &wp);
#endif
      setWindowState(oldWindowState & ~(Qt::WindowMinimized | Qt::WindowMaximized | Qt::WindowFullScreen));
  }

  Q_ASSERT(!(windowState() & (Qt::WindowMinimized | Qt::WindowMaximized | Qt::WindowFullScreen)));
  setGeometry(geometry);
#ifdef _WIN32
  if(newWindowState & Qt::WindowMaximized)
  {
    WINDOWPLACEMENT wp;
    GetWindowPlacement(effectiveWinId(), &wp);
    wp.showCmd = SW_SHOWMAXIMIZED;
    SetWindowPlacement(effectiveWinId(), &wp);
  }
#endif
  setWindowState(newWindowState);
  normalGeometry = geometry;
}

void MainWindow::closeEvent(QCloseEvent* event)
{
  setVisible(false); // optical close speedup
  document->closeSlot();

  QMainWindow::closeEvent(event);
}

void MainWindow::moveEvent(QMoveEvent* event)
{
  if(!(windowState() & (Qt::WindowMinimized | Qt::WindowMaximized | Qt::WindowFullScreen)))
    normalGeometry = geometry();
  
  QMainWindow::moveEvent(event);
}

void MainWindow::resizeEvent(QResizeEvent* event)
{
  if(!(windowState() & (Qt::WindowMinimized | Qt::WindowMaximized | Qt::WindowFullScreen)))
    normalGeometry = geometry();

  QMainWindow::resizeEvent(event);
}

void MainWindow::openObjectSlot(const QString& object, DocumentObjectType type)
{
  if(type == ObjectTypeActuator) // sepcial handling for the gathered actuator window
  {
    document->openObject(".Actuator"); // open or raise window, this may fail while layout restoring
    if(!actuatorWindow) // in cause it failed, we force opening the .Actuator window
      openObjectSlot(object, ObjectTypeActuatorWindow);
    actuatorWindow->addObject(object);
    return;
  }

  SubWindow* subWindow = 0;
  switch(type)
  {
  case ObjectTypeObject:
    subWindow = new ObjectWindow(mdiArea, object);    
    subWindow->setWindowIcon(tree->objectIcon);
    break;
  case ObjectTypeSensor:
    subWindow = new SensorWindow(mdiArea, object);
    subWindow->setWindowIcon(tree->sensorIcon);
    break;
  case ObjectTypeEditorWindow:
    subWindow = new Editor(mdiArea);
    break;
  case ObjectTypeActuatorWindow:
    if(!actuatorWindow)
      subWindow = actuatorWindow = new ActuatorWindow(mdiArea);
    break;
  default:
    Q_ASSERT(false);
    break;
  }

  if(subWindow)
  {
    subWindow->setAttribute(Qt::WA_DeleteOnClose);
    mdiArea->addSubWindow(subWindow);
    subWindow->show();
  }
}

void MainWindow::raiseObjectSlot(const QString& object, DocumentObjectType type)
{
  if(type == ObjectTypeActuator) // sepcial handling for the gathered actuator window
  {
    Q_ASSERT(actuatorWindow);
    document->openObject(".Actuator"); // raise the actuator window
    return;
  }

  QList<QMdiSubWindow*> subWindows = mdiArea->subWindowList();
  for(QList<QMdiSubWindow*>::iterator i = subWindows.begin(), end = subWindows.end(); i != end; ++i)
    if(qobject_cast<SubWindow*>(*i)->getObject() == object)
    {
      setActiveSubWindowSlot(*i);
      return;
    }
}

void MainWindow::updateObjectSlot(const QString& object, DocumentObjectType type)
{
  QList<QMdiSubWindow*> subWindows = mdiArea->subWindowList();
  for(QList<QMdiSubWindow*>::iterator i = subWindows.begin(), end = subWindows.end(); i != end; ++i)
    if(qobject_cast<SubWindow*>(*i)->getObject() == object)
    {
      SubWindow* subWindow = qobject_cast<SubWindow*>(*i);
      subWindow->updateObject();
      return;
    }
}

void MainWindow::closeObjectSlot(const QString& object, DocumentObjectType type)
{
  if(type == ObjectTypeActuator) // sepcial handling for the gathered actuator window
  {
    actuatorWindow->removeObject(object);
    if(actuatorWindow->getObjectCount() == 0)
      document->closeObject(".Actuator");
    return;
  }

  QList<QMdiSubWindow*> subWindows = mdiArea->subWindowList();
  for(QList<QMdiSubWindow*>::iterator i = subWindows.begin(), end = subWindows.end(); i != end; ++i)
    if(qobject_cast<SubWindow*>(*i)->getObject() == object)
    {
      SubWindow* subWindow = qobject_cast<SubWindow *>(*i);      
      mdiArea->removeSubWindow(subWindow);
      if(subWindow == activeSubWindow)
        subWindowActivatedSlot(0);
      //if(subWindow->close())
      //  delete subWindow;
      subWindow->close();
      if(subWindow == actuatorWindow)
        actuatorWindow = 0;
      return;
    }
}

void MainWindow::showEditorSlot(bool show)
{
  if(show)
    document->openObject(".Editor");
  else
    document->closeObject(".Editor");
}

void MainWindow::dragEnterEvent(QDragEnterEvent *event)
{  
  if(event->mimeData()->hasUrls())
    event->acceptProposedAction();
}

void MainWindow::dropEvent(QDropEvent *event)
{
  QList<QUrl> urls = event->mimeData()->urls();
  QString chdir;
  foreach(QUrl url, urls)
  {
    QString file(url.toLocalFile());
    if(!file.isEmpty())
    {
      Document::document->openFileSlot(file);
      break;
    }
  }
  event->acceptProposedAction();
}

void MainWindow::resizeSubWindow(int size)
{
  QMdiSubWindow* activeSubWindow = mdiArea->activeSubWindow();
  if(!activeSubWindow)
    return;
  QRect rect(activeSubWindow->contentsRect());
  int innerWidth = rect.width();
  int innerHeight = rect.height();
  QSize newSize;
  switch(size)
  {
  case 1:
    newSize = QSize(160, 120);
    break;
  case 2:
    newSize = QSize(320, 240);
    break;
  case 3:
    newSize = QSize(480, 360);
    break;
  case 4:
    newSize = QSize(640, 480);
    break;
  }
  activeSubWindow->resize(newSize.width() + activeSubWindow->width() - innerWidth,
    newSize.height() + activeSubWindow->height() - innerHeight);
}
