
#include <QKeyEvent>
#include <QApplication>
#include <QClipboard>

#include "ObjectWindow.h"
#include "Document.h"
#include "MenuBar.h"

ObjectWindow::ObjectWindow(QMdiArea* mdiArea, const QString& object) : SubWindow(mdiArea, object),
  objectRenderer(Document::document->getSimulation(), object.toAscii().constData()),
  objectGlWidget(0)
{
  setWindowTitle(tr("Object - ") + object);
  setMinimumSize(QSize(200, 150));

  MenuBar* menuBar = MenuBar::menuBar;
  surfaceMapper.setMapping(menuBar->surfaceWire, VisualizationParameterSet::WIREFRAME);
  surfaceMapper.setMapping(menuBar->surfaceFlat, VisualizationParameterSet::FLAT_SHADING);
  surfaceMapper.setMapping(menuBar->surfaceSmooth, VisualizationParameterSet::SMOOTH_SHADING);
  surfaceMapper.setMapping(menuBar->surfaceTexture, VisualizationParameterSet::TEXTURE_SHADING);
  surfaceMapper.setMapping(menuBar->surfaceCameraSim, VisualizationParameterSet::CAMERA_SIMULATION);
  surfaceMapper.setMapping(menuBar->surfacePhysicsWire, VisualizationParameterSet::PHYSICS_WIREFRAME);
  surfaceMapper.setMapping(menuBar->surfacePhysicsFlat, VisualizationParameterSet::PHYSICS_FLAT);
  surfaceMapper.setMapping(menuBar->surfacePhysicsSmooth, VisualizationParameterSet::PHYSICS_SMOOTH);
  surfaceMapper.setMapping(menuBar->surfaceMixed1, VisualizationParameterSet::MIXED_PHYSICS);
  surfaceMapper.setMapping(menuBar->surfaceMixed2, VisualizationParameterSet::MIXED_APPEARANCE);
  connect(&surfaceMapper, SIGNAL(mapped(int)), SLOT(setSurfaceStyleSlot(int)));

  cameraMapper.setMapping(menuBar->camTarget, VisualizationParameterSet::TARGETCAM);
  cameraMapper.setMapping(menuBar->camFree, VisualizationParameterSet::FREECAM);
  cameraMapper.setMapping(menuBar->camPresentation, VisualizationParameterSet::PRESENTATION);
  connect(&cameraMapper, SIGNAL(mapped(int)), SLOT(setCameraModeSlot(int)));

  fovMapper.setMapping(menuBar->fov20, 20);
  fovMapper.setMapping(menuBar->fov40, 40);
  fovMapper.setMapping(menuBar->fov60, 60);
  fovMapper.setMapping(menuBar->fov80, 80);
  fovMapper.setMapping(menuBar->fov100, 100);
  fovMapper.setMapping(menuBar->fov120, 120);
  connect(&fovMapper, SIGNAL(mapped(int)), SLOT(setFovYSlot(int)));

  dragPlaneMapper.setMapping(menuBar->dragXY, XY_PLANE);
  dragPlaneMapper.setMapping(menuBar->dragXZ, XZ_PLANE);
  dragPlaneMapper.setMapping(menuBar->dragYZ, YZ_PLANE);
  connect(&dragPlaneMapper, SIGNAL(mapped(int)), SLOT(setDragPlaneSlot(int)));

  dragModeMapper.setMapping(menuBar->dragKeepDynamics, KEEP_DYNAMICS);
  dragModeMapper.setMapping(menuBar->dragResetDynamics, RESET_DYNAMICS);
  dragModeMapper.setMapping(menuBar->dragApplyDynamics, APPLY_DYNAMICS);
  connect(&dragModeMapper, SIGNAL(mapped(int)), SLOT(setDragModeSlot(int)));

  updateObject();
}

void ObjectWindow::updateObject()
{
  Simulation* sim = Document::document->getSimulation();
  if(objectRenderer.getSimulation() != sim || (!objectGlWidget && sim))
  {
    objectRenderer.setSimulation(sim);
    QWidget* oldWidget = objectGlWidget;
    objectGlWidget = sim ? new ObjectGLWidget(objectRenderer, this) : 0;
    setWidget(objectGlWidget);
    if(oldWidget)
      delete oldWidget;
  }
  else
    objectRenderer.setObject(object.toAscii().constData());
}

void ObjectWindow::connectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->viewSurfaceAction->setEnabled(true);
  menuBar->viewCameraAction->setEnabled(true);
  menuBar->viewFovAction->setEnabled(true);
  menuBar->editCopyAction->setEnabled(true);
  menuBar->editPlaneAction->setEnabled(true);
  menuBar->viewDetailSensorsAction->setEnabled(true);
  menuBar->viewControllerDrawingsAction->setEnabled(true);


  connect(menuBar->editCopyAction, SIGNAL(triggered()), this, SLOT(copySlot()));
  
  connect(menuBar->surfaceWire, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceFlat, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceSmooth, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceTexture, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceCameraSim, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfacePhysicsWire, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfacePhysicsFlat, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfacePhysicsSmooth, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceMixed1, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  connect(menuBar->surfaceMixed2, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));

  connect(menuBar->camTarget, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  connect(menuBar->camFree, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  connect(menuBar->camPresentation, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  connect(menuBar->camReset, SIGNAL(triggered()), this, SLOT(resetCameraSlot()));
  connect(menuBar->camToggle, SIGNAL(triggered()), this, SLOT(toggleCameraModeSlot()));
  connect(menuBar->camFit, SIGNAL(triggered()), this, SLOT(fitCameraSlot()));
  connect(menuBar->camSnapToRoot, SIGNAL(triggered()), this, SLOT(snapToRootSlot()));

  connect(menuBar->fov20, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  connect(menuBar->fov40, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  connect(menuBar->fov60, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  connect(menuBar->fov80, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  connect(menuBar->fov100, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  connect(menuBar->fov120, SIGNAL(triggered()), &fovMapper, SLOT(map()));

  connect(menuBar->dragXY, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));
  connect(menuBar->dragXZ, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));
  connect(menuBar->dragYZ, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));

  connect(menuBar->dragKeepDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));
  connect(menuBar->dragResetDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));
  connect(menuBar->dragApplyDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));

  connect(menuBar->viewDetailSensorsAction, SIGNAL(triggered()), this, SLOT(showDetailSensorsSlot()));
  connect(menuBar->viewControllerDrawingsAction, SIGNAL(triggered()), this, SLOT(showControllerDrawingsSlot()));


  qobject_cast<QAction*>(surfaceMapper.mapping(objectRenderer.getSurfaceStyle()))->setChecked(true);
  qobject_cast<QAction*>(cameraMapper.mapping(objectRenderer.getCameraMode()))->setChecked(true);
  qobject_cast<QAction*>(fovMapper.mapping(objectRenderer.getFovY()))->setChecked(true);
  qobject_cast<QAction*>(dragPlaneMapper.mapping(objectRenderer.getDragPlane()))->setChecked(true);
  qobject_cast<QAction*>(dragModeMapper.mapping(objectRenderer.getDragMode()))->setChecked(true);
  menuBar->camSnapToRoot->setChecked(objectRenderer.hasSnapToRoot());


  QMenu* viewMenu = menuBar->viewMenu;
  viewMenu->addSeparator();
  viewMenu->addAction(menuBar->viewCameraAction);
  viewMenu->addAction(menuBar->viewFovAction);
  viewMenu->addAction(menuBar->viewSurfaceAction);
  viewMenu->addSeparator();
  viewMenu->addAction(menuBar->viewGridAction);
  viewMenu->addAction(menuBar->viewPlaneAction);
  viewMenu->addAction(menuBar->viewDetailSensorsAction);
  viewMenu->addAction(menuBar->viewControllerDrawingsAction);

  QMenu* editMenu = menuBar->editMenu;
  editMenu->addAction(menuBar->editCopyAction);
  editMenu->addSeparator();
  editMenu->addAction(menuBar->editPlaneAction);

}

void ObjectWindow::disconnectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->viewSurfaceAction->setEnabled(false);
  menuBar->viewCameraAction->setEnabled(false);
  menuBar->viewFovAction->setEnabled(false);
  menuBar->editCopyAction->setEnabled(false);
  menuBar->editPlaneAction->setEnabled(false);
  menuBar->viewDetailSensorsAction->setEnabled(false);
  menuBar->viewControllerDrawingsAction->setEnabled(false);
  
  disconnect(menuBar->editCopyAction, SIGNAL(triggered()), this, SLOT(copySlot()));

  disconnect(menuBar->surfaceWire, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceFlat, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceSmooth, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceTexture, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceCameraSim, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfacePhysicsWire, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfacePhysicsFlat, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfacePhysicsSmooth, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceMixed1, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));
  disconnect(menuBar->surfaceMixed2, SIGNAL(triggered()), &surfaceMapper, SLOT(map()));

  disconnect(menuBar->camTarget, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  disconnect(menuBar->camFree, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  disconnect(menuBar->camPresentation, SIGNAL(triggered()), &cameraMapper, SLOT(map()));
  disconnect(menuBar->camReset, SIGNAL(triggered()), this, SLOT(resetCameraSlot()));
  disconnect(menuBar->camToggle, SIGNAL(triggered()), this, SLOT(toggleCameraModeSlot()));
  disconnect(menuBar->camFit, SIGNAL(triggered()), this, SLOT(fitCameraSlot()));
  disconnect(menuBar->camSnapToRoot, SIGNAL(triggered()), this, SLOT(snapToRootSlot()));

  disconnect(menuBar->fov20, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  disconnect(menuBar->fov40, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  disconnect(menuBar->fov60, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  disconnect(menuBar->fov80, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  disconnect(menuBar->fov100, SIGNAL(triggered()), &fovMapper, SLOT(map()));
  disconnect(menuBar->fov120, SIGNAL(triggered()), &fovMapper, SLOT(map()));

  disconnect(menuBar->dragXY, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));
  disconnect(menuBar->dragXZ, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));
  disconnect(menuBar->dragYZ, SIGNAL(triggered()), &dragPlaneMapper, SLOT(map()));

  disconnect(menuBar->dragKeepDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));
  disconnect(menuBar->dragResetDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));
  disconnect(menuBar->dragApplyDynamics, SIGNAL(triggered()), &dragModeMapper, SLOT(map()));

  disconnect(menuBar->viewDetailSensorsAction, SIGNAL(triggered()), this, SLOT(showDetailSensorsSlot()));
  disconnect(menuBar->viewControllerDrawingsAction, SIGNAL(triggered()), this, SLOT(showControllerDrawingsSlot()));

  menuBar->resetViewMenu();
  menuBar->resetEditMenu();
}

void ObjectWindow::setSurfaceStyleSlot(int style)
{
  objectRenderer.setSurfaceStyle(VisualizationParameterSet::SurfaceStyle(style));
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(surfaceMapper.mapping(style))->setChecked(true);
}

void ObjectWindow::setCameraModeSlot(int mode)
{
  objectRenderer.setCameraMode(VisualizationParameterSet::CameraMode(mode));
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(cameraMapper.mapping(mode))->setChecked(true);
}

void ObjectWindow::setFovYSlot(int fovy)
{
  objectRenderer.setFovY(fovy);
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(fovMapper.mapping(fovy))->setChecked(true);
}

void ObjectWindow::setDragPlaneSlot(int plane)
{
  objectRenderer.setDragPlane(DragAndDropPlane(plane));
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(dragPlaneMapper.mapping(plane))->setChecked(true);
}

void ObjectWindow::setDragModeSlot(int mode)
{
  objectRenderer.setDragMode(DragAndDropMode(mode));
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(dragModeMapper.mapping(mode))->setChecked(true);
}

void ObjectWindow::resetCameraSlot()
{
  objectRenderer.resetCamera();
  if(objectGlWidget)
    objectGlWidget->updateGL();
}

void ObjectWindow::toggleCameraModeSlot()
{
  objectRenderer.toggleCameraMode();
  if(objectGlWidget)
    objectGlWidget->updateGL();
  qobject_cast<QAction*>(cameraMapper.mapping(objectRenderer.getCameraMode()))->setChecked(true);
}

void ObjectWindow::fitCameraSlot()
{
  objectRenderer.fitCamera();
  if(objectGlWidget)
    objectGlWidget->updateGL();
}

void ObjectWindow::snapToRootSlot()
{
  objectRenderer.setSnapToRoot(!objectRenderer.hasSnapToRoot());
  if(objectGlWidget)
    objectGlWidget->updateGL();
  MenuBar::menuBar->camSnapToRoot->setChecked(objectRenderer.hasSnapToRoot());
}

void ObjectWindow::showDetailSensorsSlot()
{
  objectRenderer.setShowDetailSensors(!objectRenderer.hasShowDetailSensors());
  if(objectGlWidget)
    objectGlWidget->updateGL();
  MenuBar::menuBar->viewDetailSensorsAction->setChecked(objectRenderer.hasShowDetailSensors());
}

void ObjectWindow::showControllerDrawingsSlot()
{
  objectRenderer.setShowControllerDrawings(!objectRenderer.hasShowControllerDrawings());
  if(objectGlWidget)
    objectGlWidget->updateGL();
  MenuBar::menuBar->viewControllerDrawingsAction->setChecked(objectRenderer.hasShowControllerDrawings());
}

void ObjectWindow::keyPressEvent(QKeyEvent* event)
{
  SubWindow::keyPressEvent(event);

  switch(event->key())
  {
  case Qt::Key_PageUp:
    event->accept();
    objectRenderer.zoom(0.05);
    if(objectGlWidget)
      objectGlWidget->updateGL();
    break;

  case Qt::Key_PageDown:
    event->accept();
    objectRenderer.zoom(-0.05);
    if(objectGlWidget)
      objectGlWidget->updateGL();
    break;
  }
}

void ObjectWindow::wheelEvent(QWheelEvent* event)
{
  SubWindow::wheelEvent(event);

  if(event->orientation() == Qt::Vertical)
  {
    objectRenderer.zoom(-0.05 * event->delta() / 160);
    if(objectGlWidget)
      objectGlWidget->updateGL();
    event->accept();
  }
}

void ObjectWindow::focusInEvent(QFocusEvent* event)
{
  SubWindow::focusInEvent(event);

  objectRenderer.setFocus(true);
}

void ObjectWindow::focusOutEvent(QFocusEvent* event)
{
  SubWindow::focusOutEvent(event);

  objectRenderer.setFocus(objectGlWidget && objectGlWidget->hasFocus());
}

void ObjectWindow::restoreLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup(object);
  restoreGeometry(*settings);

  objectRenderer.setSurfaceStyle(VisualizationParameterSet::SurfaceStyle(settings->value("SurfaceStyle", int(objectRenderer.getSurfaceStyle())).toInt()));
  objectRenderer.setCameraMode(VisualizationParameterSet::CameraMode(settings->value("CameraMode", int(objectRenderer.getCameraMode())).toInt()));
  objectRenderer.setSnapToRoot(settings->value("SnapToRoot", objectRenderer.hasSnapToRoot()).toBool());
  objectRenderer.setFovY(settings->value("FovY", objectRenderer.getFovY()).toInt());
  objectRenderer.setDragPlane(DragAndDropPlane(settings->value("DragPlane", int(objectRenderer.getDragPlane())).toInt()));
  objectRenderer.setDragMode(DragAndDropMode(settings->value("DragMode", int(objectRenderer.getDragMode())).toInt()));
  objectRenderer.setShowDetailSensors(settings->value("ShowDetailSensors", objectRenderer.hasShowDetailSensors()).toBool());
  objectRenderer.setShowControllerDrawings(settings->value("ShowControllerDrawings", objectRenderer.hasShowControllerDrawings()).toBool());

  Vector3d pos, target;
  objectRenderer.getCamera(pos, target);

  pos.v[0] = settings->value("cameraPosX", pos.v[0]).toDouble();
  pos.v[1] = settings->value("cameraPosY", pos.v[1]).toDouble();
  pos.v[2] = settings->value("cameraPosZ", pos.v[2]).toDouble();
  target.v[0] = settings->value("cameraTargetX", target.v[0]).toDouble();
  target.v[1] = settings->value("cameraTargetY", target.v[1]).toDouble();
  target.v[2] = settings->value("cameraTargetZ", target.v[2]).toDouble();

  objectRenderer.setCamera(pos, target);

  settings->endGroup();
}

void ObjectWindow::writeLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup(object);
  writeGeometry(*settings);

  settings->setValue("SurfaceStyle", int(objectRenderer.getSurfaceStyle()));
  settings->setValue("CameraMode", int(objectRenderer.getCameraMode()));
  settings->setValue("SnapToRoot", objectRenderer.hasSnapToRoot());
  settings->setValue("FovY", objectRenderer.getFovY());
  settings->setValue("DragPlane", int(objectRenderer.getDragPlane()));
  settings->setValue("DragMode", int(objectRenderer.getDragMode()));
  settings->setValue("ShowDetailSensors", objectRenderer.hasShowDetailSensors());
  settings->setValue("ShowControllerDrawings", objectRenderer.hasShowControllerDrawings());

  Vector3d pos, target;
  objectRenderer.getCamera(pos, target);

  settings->setValue("cameraPosX", pos.v[0]);
  settings->setValue("cameraPosY", pos.v[1]);
  settings->setValue("cameraPosZ", pos.v[2]);
  settings->setValue("cameraTargetX", target.v[0]);
  settings->setValue("cameraTargetY", target.v[1]);
  settings->setValue("cameraTargetZ", target.v[2]);

  settings->endGroup();
}

void ObjectWindow::copySlot()
{
  QApplication::clipboard()->clear();
  if(objectGlWidget)
  {
    /*
    unsigned int width, height;
    objectRenderer.getSize(width, height);
    char* buffer = new char[width * height * 4];
    char tmp;
    objectRenderer.renderOffscreen(buffer);
    for(char* pos  = buffer, * end = buffer + width * height * 4; pos < end; pos += 4)
    {
      tmp = pos[0];
      pos[0] = pos[2];
      pos[2] = tmp;
    }
    QImage image((uchar*)buffer, width, height, QImage::Format_RGB32);
    QApplication::clipboard()->setPixmap(QPixmap::fromImage(image));
    delete[] buffer;
    */
    QApplication::clipboard()->setPixmap(QPixmap::fromImage(objectGlWidget->grabFrameBuffer()));
  }
}

QList<ObjectGLWidget*> ObjectGLWidget::objectGLWidgetList;

ObjectGLWidget::ObjectGLWidget(ObjectRenderer& renderer, ObjectWindow* parent) : QGLWidget(parent, objectGLWidgetList.isEmpty() ? 0 : objectGLWidgetList.first()),
  objectRenderer(renderer), parent(parent)
{
  objectGLWidgetList.append(this);
  objectRenderer.setFocus(hasFocus() || parent->hasFocus());

  connect(Document::document, SIGNAL(updateViewsSignal()), SLOT(updateGL()));
}

ObjectGLWidget::~ObjectGLWidget()
{
  objectGLWidgetList.removeAll(this);
}

void ObjectGLWidget::initializeGL()
{
  objectRenderer.init(isSharing());
}

void ObjectGLWidget::paintGL()
{
  objectRenderer.draw();
}

void ObjectGLWidget::resizeGL(int width, int height)
{
  objectRenderer.setSize(width, height);
}

void ObjectGLWidget::focusInEvent(QFocusEvent * event)
{
  QGLWidget::focusInEvent(event);

  objectRenderer.setFocus(true);
}

void ObjectGLWidget::focusOutEvent(QFocusEvent * event)
{
  QGLWidget::focusOutEvent(event);

  objectRenderer.setFocus(parent->hasFocus());
}

void ObjectGLWidget::mouseMoveEvent(QMouseEvent* event)
{
  QGLWidget::mouseMoveEvent(event);

  if(objectRenderer.moveDrag(event->x(), event->y()))
  {
    event->accept();
    updateGL();
  }
}

void ObjectGLWidget::mousePressEvent(QMouseEvent* event)
{
  QGLWidget::mousePressEvent(event);

  if(!objectRenderer.hasFocus())
    return;
  updateGL();

  if(event->button() == Qt::LeftButton || event->button() == Qt::MidButton)
    if(objectRenderer.startDrag(event->x(), event->y(), QApplication::keyboardModifiers() & Qt::ShiftModifier ? ObjectRenderer::DRAG_ROTATE : ObjectRenderer::DRAG_NORMAL))
    {
      event->accept();
      updateGL();
    }
}

void ObjectGLWidget::mouseReleaseEvent(QMouseEvent* event)
{
  QGLWidget::mouseReleaseEvent(event);

  if(objectRenderer.releaseDrag(event->x(), event->y()))
  {
    event->accept();
    updateGL();
  }
  /*
  else if(event->button() == Qt::RightButton)
  {

  } */
}

void ObjectGLWidget::contextMenuEvent(QContextMenuEvent* event)
{
  event->accept();
  MenuBar* menuBar = MenuBar::menuBar;
  QMenu menu;
  menu.addAction(menuBar->editCopyAction);
  menu.addSeparator();
  menu.addAction(menuBar->editPlaneAction);
  menu.addSeparator();
  menu.addAction(menuBar->viewCameraAction);
  menu.addAction(menuBar->viewFovAction);
  menu.addAction(menuBar->viewSurfaceAction);
  menu.addSeparator();
  menu.addAction(menuBar->viewGridAction);
  menu.addAction(menuBar->viewPlaneAction);
  menu.addAction(menuBar->viewDetailSensorsAction);
  menu.addAction(menuBar->viewControllerDrawingsAction);
  menu.addSeparator();
  menu.addAction(menuBar->simResetAction);
  menu.addAction(menuBar->simStartAction);
  menu.addAction(menuBar->simStepAction);
  menu.addSeparator();
  menu.addAction(menuBar->simGenerateCubeMapAction);
  menu.addAction(menuBar->simShowFpsAction);
  menu.addAction(menuBar->simShowStatistics);
  menu.exec(mapToGlobal(QPoint(event->x(), event->y())));
}
