
#include <GL/glew.h>
#include <QLabel>
#include <QPainter>
#include <QContextMenuEvent>
#include <QApplication>
#include <QClipboard>
#include <QLocale>
#include <QGLWidget>

#include "SensorWindow.h"
#include "Document.h"
#include "MenuBar.h"
#include "../SimRobotCore/Controller/DirectView.h"
#include "../SimRobotCore/Simulation/Simulation.h"

class SensorViewWidget : public QWidget
{
public:
  QPainter painter;
  QString sensorName;

  SensorViewWidget(QString &sensorName, QWidget* parent=0) : QWidget( parent) 
  {
    this->sensorName = sensorName;
  }

  void paintEvent(QPaintEvent *event)
  {
    Simulation *sim = Document::document->getSimulation();
    int sensorPortId = sim->getSensorportId(sensorName.toAscii().constData());
    int sensorType = sim->getSensorportType(sensorPortId);
    QPen pen(QColor::fromRgb(255, 0, 0));
    
    painter.begin(this);

    switch(sensorType)
    {
      case doubleSensor:
        {
          double d_val, d_min, d_max;
          char str_val[32];
          sim->getSensorportValue(sensorPortId, d_val);
          sprintf(str_val, "%.03f", d_val);
          d_min = sim->getSensorportMinValue(sensorPortId);
          d_max = sim->getSensorportMaxValue(sensorPortId);
          d_val -= d_min;
          d_val /= (d_max-d_min);
          QBrush brush(QColor::fromRgb(d_val*255, d_val*255, d_val*255));
          painter.setBrush(brush);
          painter.drawRect(0,0, this->width(), this->height());
          painter.setPen(pen);
          painter.drawText(this->width()/2, this->height()/2, str_val);
          break;
        }
      case doubleArraySensor:
        {
          std::vector<int> dims = sim->getSensorDimensions(sensorPortId);
          if(dims.size() == 1)
          {
            double* vals;
            sim->getSensorportValue(sensorPortId, vals);
            double d_min = sim->getSensorportMinValue(sensorPortId);
            double d_max = sim->getSensorportMaxValue(sensorPortId);
            double d_val;
            float widthPerVal = this->width()/(float)dims[0];
            for(int i = 0; i < dims[0]; i++)
            {
              d_val = vals[i] - d_min;
              d_val /= (d_max-d_min);

              QBrush brush(QColor::fromRgb(d_val*255, d_val*255, d_val*255));
              painter.setBrush(brush);
              pen.setColor(QColor::fromRgb(0, 0, 0));
              painter.setPen(pen);

              painter.drawRect(i*widthPerVal, 0, (i+1)*widthPerVal, this->height());

              pen.setColor(QColor::fromRgb(255, 0, 0));
              painter.setPen(pen);
              char str[32];
              sprintf(str, "%.03f", vals[i]);
              painter.drawText(i*widthPerVal, 20, str);
            }
          }
          else
            painter.drawText(this->width()/2, this->height()/2, "More than 1 dimension\n->not implemented");
          break;
        }
      case boolSensor:
        {
          bool b_val;
          char c_val[32];

          sim->getSensorportValue(sensorPortId, b_val);

          QBrush brush(QColor::fromRgb(0,0,0));
          if(b_val)
          {
            brush.setColor(QColor::fromRgb(255, 255, 255));
            sprintf(c_val, "true");
          }
          else
            sprintf(c_val, "false");

          pen.setColor(QColor::fromRgb(255, 0, 0));
          painter.setPen(pen);
          painter.drawRect(0, 0, this->width(), this->height());
          painter.drawText(this->width()/2, this->height()/2, c_val);
          break;
        }
      case cameraSensor:
        {
          std::vector<int> dims = sim->getSensorDimensions(sensorPortId);
          int xsize = dims[0], ysize = dims[1];
          
          unsigned char *vals;
          sim->getSensorportValue(sensorPortId, vals);
          unsigned char* end = vals + xsize * ysize * 3,
                       * buffer = new unsigned char[xsize * ysize * 4],
                       * pDest = buffer;
          for(unsigned char* pSrc = vals; pSrc < end; pSrc += 3)
          {
            *pDest++ = pSrc[0];
            *pDest++ = pSrc[1];
            *pDest++ = pSrc[2];
            *pDest++ = 0xff;
          }
          QImage img(buffer, xsize, ysize, QImage::Format_RGB32);
          
          painter.drawImage(0,0, img.scaled(this->width(), this->height()));
          delete [] buffer;
          break;
        }
    }
    painter.end();
  }

};

SensorWindow::SensorWindow(QMdiArea* mdiArea, const QString& object) : SubWindow(mdiArea, object),
  directView(0), widget(0)
{
  setWindowTitle(tr("Sensor - ") + object);
  setMinimumSize(QSize(200, 150));

  connect(Document::document, SIGNAL(updateViewsSignal()), SLOT(updateSlot()));

  updateObject();
}

void SensorWindow::updateObject()
{
  QWidget* newWidget = widget;
  DirectView* newDirectView = 0;
  Simulation* simulation = Document::document->getSimulation();
  int sensorPortId = -1;
  if(simulation)
  {
    sensorPortId = simulation->getSensorportId(object.toAscii().constData());    
    if(sensorPortId >= 0)
    {
      int sensorType = simulation->getSensorportType(sensorPortId);
      if(sensorType == viewSensor)
        newDirectView = (DirectView*)simulation->getView(sensorPortId);
      else if( sensorType != noSensor)
        newWidget = new SensorViewWidget(object, this);
      else
        newWidget = new QLabel("noSensor");
    }    
  }

  if(newDirectView != directView)
  {
    directView = newDirectView;
    newWidget = directView ? directView->createWidget() : 0;
  }

  if(!newWidget && sensorPortId >= 0)
    newWidget = new QLabel(tr("TODO: implement me"), this);

  if(newWidget != widget)
  {
    QWidget* oldWidget = widget;
    widget = newWidget;
    setWidget(newWidget);
    if(oldWidget)
      delete oldWidget;
  }
}

void SensorWindow::updateSlot()
{
  if(directView)
    directView->updateWidget(*widget);
  else
  {
    Simulation* simulation = Document::document->getSimulation();
    int sensorPortId = simulation->getSensorportId(object.toAscii().constData());
    if(sensorPortId != -1)
    {
      int sensorType = simulation->getSensorportType(sensorPortId);
      if(sensorType != noSensor) 
        widget->update();
    }
  }
}


void SensorWindow::restoreLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();

  settings->beginGroup(object);
  restoreGeometry(*settings);

  if(directView)
  {
    const QVariant& var(settings->value("Configuration"));
    if(var.isValid())
    {
      std::string conf(var.toByteArray().constData());
      directView->setConfiguration(*widget, conf);
    }
  }
  
  settings->endGroup();
}

void SensorWindow::writeLayout()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup(object);
  writeGeometry(*settings);

  if(directView)
    settings->setValue("Configuration", QByteArray(directView->getConfiguration(*widget).c_str()));

  settings->endGroup();
}

void SensorWindow::connectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->editCopyAction->setEnabled(true);

  connect(menuBar->editCopyAction, SIGNAL(triggered()), this, SLOT(copySlot()));

  /*
  QMenu* viewMenu = menuBar->viewMenu;
  viewMenu->addSeparator();
  viewMenu->addAction(menuBar->viewLineAction);
  viewMenu->addAction(menuBar->viewColumnAction);
  viewMenu->addAction(menuBar->viewMonoAction);
  viewMenu->addAction(menuBar->viewColorAction);
  viewMenu->addAction(menuBar->viewStereoAction);
  viewMenu->addSeparator();
  viewMenu->addAction(menuBar->viewGridAction);
  viewMenu->addMenu(menuBar->brighteningMenu);
  */

  QMenu* editMenu = menuBar->editMenu;
  editMenu->addAction(menuBar->editCopyAction);
}

void SensorWindow::disconnectMenuBar()
{
  MenuBar* menuBar = MenuBar::menuBar;

  menuBar->editCopyAction->setEnabled(false);

  disconnect(menuBar->editCopyAction, SIGNAL(triggered()), this, SLOT(copySlot()));

  menuBar->resetViewMenu();
  menuBar->resetEditMenu();
}

void SensorWindow::contextMenuEvent(QContextMenuEvent* event)
{
  QRect content(contentsRect());
  if(!content.contains(event->x(), event->y()))
  { // klick on window frame
    SubWindow::contextMenuEvent(event);
    return;
  };

  event->accept();
  MenuBar* menuBar = MenuBar::menuBar;
  QMenu menu;  
  menu.addAction(menuBar->editCopyAction);
  menu.addSeparator();
  menu.addMenu(menuBar->resizeMenu);
  menu.addSeparator();
  menu.addAction(menuBar->simResetAction);
  menu.addAction(menuBar->simStartAction);
  menu.addAction(menuBar->simStepAction);
  menu.addSeparator();
  menu.addAction(menuBar->simGenerateCubeMapAction);
  menu.addAction(menuBar->simShowFpsAction);
  menu.addAction(menuBar->simShowStatistics);
  menu.exec(mapToGlobal(QPoint(event->x(), event->y())));
}

void SensorWindow::copySlot()
{
  QApplication::clipboard()->clear();
  Simulation* sim = Document::document->getSimulation();
  int sensorPortId = sim->getSensorportId(object.toAscii().constData());
  std::vector<int> dims = sim->getSensorDimensions(sensorPortId);
  QMimeData* mineData = new QMimeData;
  setClipboardGraphics(*mineData);
  int count = 1;
  for(int i = 0; i < int(dims.size()); i++)
    count *= dims[i];
  if(count < 10000 && !directView)
    setClipboardText(*mineData);
  QApplication::clipboard()->setMimeData(mineData);  
}

void SensorWindow::setClipboardGraphics(QMimeData& mimeData)
{
  if(widget)
  {
    // just take a screenshot of opengl, normal or mixed widgets
    if(widget->inherits("QGLWidget"))
    {
      mimeData.setImageData(QVariant(qobject_cast<QGLWidget*>(widget)->grabFrameBuffer()));
      return;
    }
    QPixmap pixmap(QPixmap::grabWidget(widget));    
    QWidget* subWidget = widget->focusProxy();
    if(subWidget && subWidget->inherits("QGLWidget"))
    {
      QGLWidget* glWidget = qobject_cast<QGLWidget*>(subWidget);
      QPainter painter(&pixmap);
      painter.drawImage(glWidget->mapTo(widget, QPoint(0, 0)), glWidget->grabFrameBuffer());
    }
    mimeData.setImageData(QVariant(pixmap.toImage()));
  }
}

void SensorWindow::setClipboardText(QMimeData& mimeData)
{
  double doubleValue,
    *pDouble;
  bool deletePDouble = false;
  Simulation* pSim = Document::document->getSimulation();
  int sensorportNumber = pSim->getSensorportId(object.toAscii().constData());
  int dimSize[3] = {1, 1, 1};
  {
    std::vector<int> dims = pSim->getSensorDimensions(sensorportNumber);
    for(int i = 0; i < 3 && i < int(dims.size()); i++)
      if(dims[i] > 0)
        dimSize[i] = dims[i];
  }
  switch(pSim->getSensorportType(sensorportNumber))
  {
  case boolSensor:
    {
      bool boolValue;
      pSim->getSensorportValue(sensorportNumber,boolValue);
      doubleValue = boolValue ? 1 : 0;
      pDouble = &doubleValue;
      break;
    }
  case doubleSensor:
    pSim->getSensorportValue(sensorportNumber,doubleValue);
    pDouble = &doubleValue;
    break;
  case cameraSensor:
    {
      pDouble = new double[dimSize[0] * dimSize[1] * dimSize[2]];
      deletePDouble = true;
      unsigned char *vals;
      pSim->getSensorportValue(sensorportNumber, vals);
      for(int i = dimSize[0] * dimSize[1] * dimSize[2] - 1; i >= 0; --i)
        pDouble[i] = vals[i];
      break;
    }
  case doubleArraySensor:
    {
      pSim->getSensorportValue(sensorportNumber,pDouble);
      break;
    }
  case viewSensor:
  case noSensor:
    {
      break;
    }
  }

  QString text;
  const QLocale& locale(QLocale::system());
  double* pData = pDouble;
  char sBuffer[30];
  for(int y = 0; y < dimSize[1] * dimSize[2]; y++)
    for(int x = 0; x < dimSize[0]; x++)
    {      
      double d = *pData;
      sprintf(sBuffer, "%lg", d);
      for(char* str = sBuffer; *str; ++str)
        if(*str == '.')
          text += locale.decimalPoint();
        else
          text += *str;
      text += x + 1 == dimSize[0] ? '\n' : '\t';
      pData++;
    }
  mimeData.setText(text);

  if(deletePDouble)
    delete[] pDouble;
}

