
#include <QHeaderView>
#include <QHideEvent>
#include <QTreeView>

#include "Document.h"
#include "Tree.h"
#include "MenuBar.h"

class ObjectTreeModel : public QAbstractItemModel
{
public:
  ObjectTreeModel(Tree* tree) : QAbstractItemModel(tree),
    tree(tree) {}
  ~ObjectTreeModel() {}

  void clear()
  {
    if(root.childs.count() > 0)
    {
      beginRemoveRows(QModelIndex(), 0, root.childs.count() - 1);
      root.remove(0, root.childs.count());
      endRemoveRows();
    }
  }

  void update(const std::vector<ObjectDescription> &objs, const QSet<QString>& expandedItems, QList<QModelIndex>& itemsToExpand)
  {
    if(objs.size() == 0)
    {
      clear();
      return;
    }

    int currentDepth = 0;
    Item* currentParent = &root;
    currentParent->currentIndex = 0;
    QModelIndex currentModelIndex;

    bool initialBuild = false;
    if(root.count() == 0)
    {
      initialBuild = true;
      beginInsertRows(currentModelIndex, 0, 0);
    }

    for(std::vector<ObjectDescription>::const_iterator i = objs.begin(), end = objs.end(); i != end; ++i)
    {
      if(i->depth < currentDepth) //if we need to go upwards
      {
        do //go upwards till we reach the level we want to
        {
          // remove all not passed items
          int childCount = currentParent->count();
          if(currentParent->currentIndex < childCount)
          {
            Q_ASSERT(!initialBuild);
            beginRemoveRows(currentModelIndex, currentParent->currentIndex, childCount - 1);
            currentParent->remove(currentParent->currentIndex, childCount - currentParent->currentIndex);
            endRemoveRows();
            Q_ASSERT(currentParent->currentIndex >= currentParent->childs.count());
          }

          // now really go upwards
          currentParent = currentParent->parent;
          Q_ASSERT(currentParent);
          currentDepth--;
          currentModelIndex = parent(currentModelIndex);
        } while(i->depth < currentDepth); 
      }
      if(i->depth > currentDepth) //if we need to go downwards
      {
        Q_ASSERT(currentParent->currentIndex > 0);
        Q_ASSERT(currentParent->currentIndex - 1 < currentParent->childs.count());
        int row = currentParent->currentIndex - 1;
        currentParent = currentParent->childs[row];
        currentDepth++;
        Q_ASSERT(i->depth == currentDepth);
        currentParent->currentIndex = 0;
        currentModelIndex = index(row, 0, currentModelIndex);
      }

      int childCount = currentParent->count();
      if(currentParent->currentIndex < childCount &&
        currentParent->childs[currentParent->currentIndex]->name == i->name)
      { // item already exists, pass it
        currentParent->currentIndex++;
      }
      else
      { // try to find the item in "yonger" childrens
        int j;
        for(j = currentParent->currentIndex + 1; j < childCount; j++)
          if(currentParent->childs[j]->name == i->name)
            break;

        if(j < childCount)
        { // if found remove all items above
          Q_ASSERT(!initialBuild);
          beginRemoveRows(currentModelIndex, currentParent->currentIndex, j - 1);
          currentParent->remove(currentParent->currentIndex, j - currentParent->currentIndex);
          endRemoveRows();

          // and item already exists, pass it
          currentParent->currentIndex++;
        }
        else
        {
          // insert new item
          if(!initialBuild)
            beginInsertRows(currentModelIndex, currentParent->currentIndex, currentParent->currentIndex);
          currentParent->insert(currentParent->currentIndex, i->name, i->type);
          if(!initialBuild)
            endInsertRows();
          if(expandedItems.contains(QString(i->fullName.c_str())))
            itemsToExpand.append(index(currentParent->currentIndex, 0, currentModelIndex));
          currentParent->currentIndex++; // and pass it
        }
      }
    }

    if(initialBuild)
    {
      Q_ASSERT(root.count() == 1);
      endInsertRows();
    }
  }

  class Item
  {
  public:

    inline const Item* at(int index) const    
    {
      return childs.at(index);
    }

    inline int count() const
    {
      return childs.count();
    }

    const std::string& getName() const
    {
      return name;
    }

    const Item* getParent() const
    {
      return parent;
    }

    ObjectType getType() const
    {
      return type;
    }

  private:
    std::string name;
    ObjectType type;
    QVector<Item*> childs;
    Item* parent;
    int currentIndex; /**< A walkthrough index used in update() */ 

    Item() : parent(0) {}
    Item(const std::string& name, ObjectType type, Item* parent) : 
      name(name), type(type), parent(parent) {}

    ~Item()
    {
      for(int i = childs.count() - 1; i >= 0; --i)
        delete childs[i];
    }

    inline void insert(int index, const std::string& name, ObjectType type)
    {
      childs.insert(index, new Item(name, type, this));
    }

    inline void remove(int index, int count)
    {
      for(int i = index, end = index + count; i < end; ++i)
        delete childs[i];
      childs.remove(index, count);
    }

    inline int row() const
    {
      return parent->childs.indexOf((Item*)this);
    }

    friend class ObjectTreeModel;
  };

  const Item* getRoot() const
  {
    return &root;
  }

private:
  Tree* tree;
  Item root;

  QModelIndex index(int row, int column, const QModelIndex &parent = QModelIndex()) const
  {
    if(!hasIndex(row, column, parent))
      return QModelIndex();

    const Item *parentItem;
    if(parent.isValid())
      parentItem = (const Item*)parent.internalPointer();
    else
      parentItem = &root;
    
    return createIndex(row, column, (void*)parentItem->at(row));
  }

  QModelIndex parent(const QModelIndex& index) const
  {
    if(!index.isValid())
      return QModelIndex();
    
    const Item* childItem = (const Item*)index.internalPointer();
    const Item* parentItem = childItem->parent;
    if(parentItem == &root)
       return QModelIndex();

    return createIndex(parentItem->row(), 0, (void*)parentItem);
  }

  int rowCount(const QModelIndex& index) const
  {
    if(index.column() > 0)
       return 0;

    const Item* item;
    if(index.isValid())
      item = (const Item*)index.internalPointer();
    else
      item = &root;

    return item->count();
  }

  int columnCount(const QModelIndex& index) const
  {
    return 1;
  }

  QVariant data(const QModelIndex& index, int role) const
  {
    if(!index.isValid())
       return QVariant();
    if(role == Qt::DisplayRole)
    {
      const Item *item = (const Item*)index.internalPointer();
      return QVariant(QString(item->name.c_str()));
    }
    else if(role == Qt::DecorationRole)
    {
      
      const Item *item = (const Item*)index.internalPointer();
      switch(item->type)
      {
      case OBJECT_TYPE_OBJECT:
        return QVariant(tree->objectIcon);
        break;
      case OBJECT_TYPE_ACTUATORPORT:
        return QVariant(tree->actuatorIcon);
        break;
      case OBJECT_TYPE_SENSORPORT:
        return QVariant(tree->sensorIcon);
        break;
      default:
        return QVariant(tree->categoryIcon);
        break;
      }
      
    }
    
    return QVariant();
  }

  Qt::ItemFlags flags(const QModelIndex& index) const
  {
    if(!index.isValid())
      return 0;
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
  }
};

class TreeView : public QTreeView
{
public:
  TreeView(QWidget* parent) : QTreeView(parent) {}

private:
  void contextMenuEvent(QContextMenuEvent* event)
  {
    event->accept();
    MenuBar* menuBar = MenuBar::menuBar;
    QMenu menu;
    menu.addAction(menuBar->simResetAction);
    menu.addAction(menuBar->simStartAction);
    menu.addAction(menuBar->simStepAction);
    menu.addSeparator();
    menu.addAction(menuBar->simGenerateCubeMapAction);
    menu.addAction(menuBar->simShowFpsAction);
    menu.addAction(menuBar->simShowStatistics);
    menu.exec(mapToGlobal(QPoint(event->x(), event->y())));
  }
};

Tree::Tree(QWidget* parent) : QDockWidget(parent),
  objectIcon(":/icons/ObjectSilk.png"), sensorIcon(":/icons/SensorSilk.png"), 
  actuatorIcon(":/icons/ActuatorSilk.png"), categoryIcon(":/icons/CategorySilk.png")
{
  setObjectName(tr("Tree"));
  setWindowTitle(tr("Tree"));

  objectTreeModel = new ObjectTreeModel(this);
  treeView = new TreeView(this);
  treeView->setModel(objectTreeModel);
#if (QT_VERSION >= QT_VERSION_CHECK(4, 4, 0))
  treeView->setExpandsOnDoubleClick(false);
#endif
  setWidget(treeView);
  treeView->header()->hide();

  connect(MenuBar::menuBar->viewTreeAction, SIGNAL(triggered(bool)), SLOT(setVisible(bool)));
  MenuBar::menuBar->viewTreeAction->setEnabled(true);

  connect(treeView, SIGNAL(activated(const QModelIndex&)), this, SLOT(itemActivatedSlot(const QModelIndex&)));
  connect(treeView, SIGNAL(collapsed(const QModelIndex&)), this, SLOT(itemCollapsedSlot(const QModelIndex&)));
  connect(treeView, SIGNAL(expanded(const QModelIndex&)), this, SLOT(itemExpandedSlot(const QModelIndex&)));

  connect(Document::document, SIGNAL(updateObjectTreeSignal()), SLOT(updateTreeSlot()));
  connect(Document::document, SIGNAL(closeSignal()), SLOT(closeTreeSlot()));
  connect(Document::document, SIGNAL(restoreLayoutSignal()), SLOT(restoreLayoutSlot()));
  connect(Document::document, SIGNAL(writeLayoutSignal()), SLOT(writeLayoutSlot()));
}

void Tree::updateTreeSlot()
{
  const std::vector<ObjectDescription> &objs(Document::document->getObjectTree());
  QList<QModelIndex> itemsToExpand;
  objectTreeModel->update(objs, expandedItems, itemsToExpand);

  for(QList<QModelIndex>::iterator i = itemsToExpand.begin(), end = itemsToExpand.end(); i != end; ++i)
    treeView->setExpanded(*i, true);
}

void Tree::closeTreeSlot()
{
  objectTreeModel->clear();
  expandedItems.clear();
}

/**
 * This function updates the items of the tree from a given 
 * ObjectDescription vector. This should be called from outside
 * everytime the content of the tree changes.
 **/
/*
void Tree::updateTreeSlot()
{
  const std::vector<ObjectDescription> &objs(Document::document->getObjectTree());

  //clear the tree
  treeWidget->removeItemWidget(treeWidget->takeTopLevelItem(0), 0);

  std::vector<ObjectDescription>::const_iterator i;
  QTreeWidgetItem *root, *parent = 0, *grandParent = 0, *lastChild = 0;
  int actualDepth = 0;

  for(i = objs.begin(); i != objs.end(); i++)
  {
    if(i->depth == 0) //root node
    {
      root = new QTreeWidgetItem(treeWidget, QStringList(QString(i->name.c_str())));
      parent = root;
      grandParent = root;
      lastChild = root;
      treeWidget->setCurrentItem(root);
    }
    else
    {
      if(i->depth < actualDepth) //if we need to go upwards
      {
        while(i->depth < actualDepth) //go upwards till we reach the level we want to
        {
          parent = grandParent;
          if(grandParent->parent() != 0)
            grandParent = grandParent->parent();
          actualDepth--;
        }
      }
      if(i->depth > actualDepth) //if we need to go downwards
      {
        grandParent = parent;
        parent = lastChild;
      }
      //create Node
      lastChild = new QTreeWidgetItem(parent, QStringList(QString(i->name.c_str())));
    }
    actualDepth = i->depth;
  }
}
*/

void Tree::hideEvent(QHideEvent * event)
{
  QDockWidget::hideEvent(event);

  MenuBar::menuBar->viewTreeAction->setChecked(false);
  event->accept();
}

void Tree::showEvent(QShowEvent * event)
{
  QDockWidget::showEvent(event);

  MenuBar::menuBar->viewTreeAction->setChecked(true);
  event->accept();
}


void Tree::itemActivatedSlot(const QModelIndex& itemIndex)
{
  const ObjectTreeModel::Item* item = (const ObjectTreeModel::Item*)itemIndex.internalPointer();
  if(item->getType() != OBJECT_TYPE_NONE)
    Document::document->openObject(getFullName(itemIndex));
  else
    treeView->setExpanded(itemIndex, !treeView->isExpanded(itemIndex));
}

void Tree::itemCollapsedSlot(const QModelIndex &index)
{
  expandedItems.remove(getFullName(index));
}

void Tree::itemExpandedSlot(const QModelIndex& index)
{
  expandedItems.insert(getFullName(index));
}

QString Tree::getFullName(const QModelIndex& index) const
{
  const ObjectTreeModel::Item* root = objectTreeModel->getRoot();
  const ObjectTreeModel::Item* item = (const ObjectTreeModel::Item*)index.internalPointer();
  std::string fullName(item->getName());
  for(;;)
  {
    item = item->getParent();
    if(item == root)
      break;
    fullName = item->getName() + '.' + fullName;
  }
  return QString(fullName.c_str());
}

void Tree::restoreLayoutSlot()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup("Tree");

  const QByteArray& buffer(settings->value("ExpandedItems").toByteArray());
  QDataStream in(buffer);
  in.setVersion(QDataStream::Qt_4_3);
  in >> expandedItems;

  settings->endGroup();
}

void Tree::writeLayoutSlot()
{
  QSettings* settings = Document::document->getLayoutSettings();
  settings->beginGroup("Tree");

  QByteArray buffer;
  QDataStream out(&buffer, QIODevice::WriteOnly);
  out.setVersion(QDataStream::Qt_4_3);
  out << expandedItems;
  settings->setValue("ExpandedItems", buffer);

  settings->endGroup();
}
