/**
* @file Tools/Debugging/Debugging.h
*
* Macros and functions for debugging
* 
* @author Martin Ltzsch
*/ 
#ifndef __Debugging_h_
#define __Debugging_h_

#include "Tools/MessageQueue/OutMessage.h"
#include "Tools/Debugging/DebugRequest.h"
#include "Tools/Global.h"
#include <iostream>

#ifdef RELEASE

#define EXECUTE_ONLY_IN_DEBUG(expression) /**/
#define INFO(key,type,format,expression) /**/
#define OUTPUT(type,format,expression) /**/
#define DEBUG_RESPONSE(id, expression) /**/
#define DEBUG_RESPONSE_OR_RELEASE(id, expression) { expression }
#define DEBUG_RESPONSE_NOT(id, expression) { expression }
#define NOT_POLLABLE_DEBUG_RESPONSE(id, expression) /**/
#else // RELEASE

/**
* A macro which executes an expression only if in debug mode.
*/
#define EXECUTE_ONLY_IN_DEBUG(expression) \
  expression

/**
* A macro for sending debug messages depending on a debug key.
*
* @param key A key.
* @param type The type of the message from the MessageID enum in MessageIDs.h
* @param format The message format of the message (bin or text)
* @param expression A streamable expression
*
* Examples:
* <pre>
* INFO(sendImages,idImage,bin,myImage);
* INFO(sendInterestingValue,idInterestingValue,text,"value: " << value);
* INFO(sendWorldState,idWorldState,bin,worldState);
* </pre>
*/
#define INFO(key,type,format,expression) \
  DEBUG_RESPONSE( "info:"#key , \
    {\
      Global::getDebugOut().format << expression;\
      Global::getDebugOut().finishMessage(type);\
    }\
  );

/**
* A macro for sending debug messages.
*
* @param type The type of the message from the MessageID enum in MessageIDs.h
* @param format The message format of the message (bin or text)
* @param expression A streamable expression
*
* Examples:
* <pre>
* OUTPUT(idImage, bin, *pMyImage);
* OUTPUT(idText, text, "MyObject:myFunction() invoked");
* OUTPUT(idText, text, "i: " << i << ", j:" << j);
* </pre>
*/

#ifdef CONSOLE_OUTPUT
#define OUTPUT(type,format,expression) \
  { Global::getDebugOut().format << expression;\
    Global::getDebugOut().finishMessage(type); \
    copyToCout__##format (expression) \
    }
#define copyToCout__text(expression) std::cout << expression << std::endl;
#define copyToCout__bin(expression)
#else
#define OUTPUT(type,format,expression) \
  { Global::getDebugOut().format << expression;\
    Global::getDebugOut().finishMessage(type); }
#endif

/**
* A debugging switch, allowing the enabling or disabling of expressions.
* @param id The id of the debugging switch
* @param expression The expression to be executed if id is enabled
*/
#define DEBUG_RESPONSE(id, expression) \
{\
  if(Global::getDebugRequestTable().poll && Global::getDebugRequestTable().notYetPolled(id)) \
  { \
    OUTPUT(idDebugResponse, text, id << \
           int(Global::getDebugRequestTable().isActive(id) && !Global::getDebugRequestTable().once(id))); \
  } \
  if(Global::getDebugRequestTable().isActive(id)) \
  { \
    if(Global::getDebugRequestTable().once(id)) \
      Global::getDebugRequestTable().disable(id); \
    expression \
  } \
}

/**
* A debugging switch, allowing the enabling or disabling of expressions.
* @param id The id of the debugging switch
* @param expression The expression to be executed if id is enabled
*/
#define DEBUG_RESPONSE_OR_RELEASE(id, expression) \
{\
  if(Global::getDebugRequestTable().poll && Global::getDebugRequestTable().notYetPolled(id)) \
  { \
    OUTPUT(idDebugResponse, text, id << \
           int(Global::getDebugRequestTable().isActive(id) && !Global::getDebugRequestTable().once(id))); \
  } \
  if(Global::getDebugRequestTable().isActive(id)) \
  { \
    if(Global::getDebugRequestTable().once(id)) \
      Global::getDebugRequestTable().disable(id); \
    expression \
  } \
}

/**
* A debugging switch, allowing the enabling or disabling of expressions.
* @param id The id of the debugging switch
* @param expression The expression to be executed if id is disabled or software is compiled in release mode
*/
#define DEBUG_RESPONSE_NOT( id, expression) \
{\
  if(Global::getDebugRequestTable().poll && Global::getDebugRequestTable().notYetPolled(id)) \
  { \
    OUTPUT(idDebugResponse, text, id << \
           int(Global::getDebugRequestTable().isActive(id) && !Global::getDebugRequestTable().once(id))); \
  } \
  if(! (Global::getDebugRequestTable().isActive(id))) \
  { \
    if(Global::getDebugRequestTable().once(id)) \
      Global::getDebugRequestTable().disable(id); \
    expression \
  } \
}

/** A debugging switch not to be polled */
#define NOT_POLLABLE_DEBUG_RESPONSE( id, expression) \
{\
  if(Global::getDebugRequestTable().isActive(id)) \
  { \
    if(Global::getDebugRequestTable().once(id)) \
      Global::getDebugRequestTable().disable(id); \
    expression \
  } \
}
#endif // RELEASE

#endif //__Debugging_h_
