/*
* @file Tools/Debugging/Modify.h
* Macros make data visible/editable through RobotControl.
*
* @author Michael Spranger
* @author Tobias Oberlies
*/


#include "DebugDataTable.h"

/**
 * == Definition of the idDebugDataResponse message format ==
 *
 * The overall syntax is
 *   [identifier string] [flags] [data]
 * where flags is a 32 bit integer specifying the traits of the message (see also 
 * constants defined below):
 *  - bit 0 set means the debug data can't be modified
 *  - bit 1 set means the message contains a compound/list of objects
 * 
 * If the message contains just a single object (bit 1 must be unset), [data] consists of
 *   [class name] [object]
 * If the message contains several objects (bit 1 must be set), [data] consists of
 *   ( [part identifyer string] [class name string] [object] )+
 * 
 * The [part identifyer string] may be arbitrarily chosen, for example to distinguish two 
 * objects of the same type. If the name is left empty (char* array of lenght 0), it 
 * should default to the class name with an initial lowercase letter.
 */

#define DEBUG_DATA_MODIFYABLE 0
#define DEBUG_DATA_READ_ONLY 1
#define DEBUG_DATA_SINGLE_OBJECT 0
#define DEBUG_DATA_COMPOUND 2


#ifdef RELEASE
#define MODIFY( name, object) {}
#define MODIFY_ENUM( name, object, numberOfEnumElements, getNameFunctionPtr) {}
#define SEND( name, object) {}
#define SEND_COMPOUND( id, parts) {}
#define SEND_PART( object) ;
#define SEND_NAMED_PART( name, object) ;

#else

/**
 * Allows for the modification of an object implementing a complete streaming operator 
 * with macros from Tools/Streams/StreamHandler.h (or scalar).
 * @param id An identifier string of the object. Separate hirarchy levels by a single 
 *     colon ':', i.e. "cognition:ball percept". The value must be a string constant. 
 * @param object The object to be modified.
 */
#define MODIFY( id, object) \
  { \
    DebugDataTable::testForStreamable(object); \
    Global::getDebugDataTable().updateObject( id, object); \
    DEBUG_RESPONSE( "debug data:"id, \
    OUTPUT( idDebugDataResponse, bin, id << (long)(DEBUG_DATA_MODIFYABLE+DEBUG_DATA_SINGLE_OBJECT) << StreamHandler::demangle(typeid(object).name()) << object); \
    ); \
  }

/**
 * Allows for the modification of an enumerated value.
 * @param id An identifier string of the variable.
 * @param object The variable to be modified.
 * @param numberOfEnumElements The number of elements of the enumeration type.
 * @param getNameFunctionPtr A function that returns a string for each enumeration value.
 */
#define MODIFY_ENUM( id, object, numberOfEnumElements, getNameFunctionPtr) \
  { \
    DebugDataTable::testForStreamable(object); \
    Global::getStreamHandler().registerEnum( object, numberOfEnumElements, getNameFunctionPtr); \
    Global::getDebugDataTable().updateEnum( id, object); \
    DEBUG_RESPONSE( "debug data:"id, \
      OUTPUT( idDebugDataResponse, bin, id << (long)(DEBUG_DATA_MODIFYABLE+DEBUG_DATA_SINGLE_OBJECT) << StreamHandler::demangle(typeid(object).name()) << (int)object); \
    ); \
  }


/**
 * Sends an object that uses the STREAM macros for streaming to RobotControl if requested.
 * @param id An identifier string for the object. Separate hirarchy levels by a single 
 *     colon, i.e. "cognition:ball percept". The value must be a string constant. 
 * @param object The object to be sent. The object must extend Streamable or 
 *     ImplicitlyStreamable and use the STREAM or STREAM_EXT macros for streaming.
 */
#define SEND( id, object) \
  { \
    DebugDataTable::testForStreamable(object); \
    DEBUG_RESPONSE( "debug data:"id, \
      OUTPUT( idDebugDataResponse, bin, id << (long)(DEBUG_DATA_READ_ONLY+DEBUG_DATA_SINGLE_OBJECT) << StreamHandler::demangle(typeid(object).name()) << object); \
    ); \
  }

/**
 * Sends a bundle of objects to RobotControl if requested. This is particularly useful 
 * for representations which consist of several objects which are not contained in a 
 * streamable wrapper object.
 * @param id An identifying string constant.
 * @param parts A sequence of SEND_PART and/or SEND_NAMED_PART macros with the objects to 
 *     be sent separated by whitespaces.
 */
#define SEND_COMPOUND( id, parts) \
  { \
  DEBUG_RESPONSE( "debug data:"id, \
      Global::getDebugOut().bin << id << (long)(DEBUG_DATA_READ_ONLY+DEBUG_DATA_COMPOUND); \
      int __inside_send_compound__ = 0; \
      parts; \
      Global::getDebugOut().finishMessage(idDebugDataResponse); \
    ); \
  }

/**
 * Sends an object as part of a compound. The label of the part will be the class name 
 * with a lowercase initial letter. Only use inside a SEND_COMPOUND macro!
 */
#define SEND_PART( object) \
  __inside_send_compound__ = 0; \
  DebugDataTable::testForStreamable(object); \
  Global::getDebugOut().bin << #object << StreamHandler::demangle(typeid(object).name()) << object;

/**
 * Sends an object as part of a compound, labeling it with the given string. Only use 
 * inside of a SEND_COMPOUND macro!
 */
#define SEND_NAMED_PART( label, object) \
  __inside_send_compound__ = 0; \
  DebugDataTable::testForStreamable(object); \
  Global::getDebugOut().bin << label << StreamHandler::demangle(typeid(object).name()) << object;

#endif

